// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_channel.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: Thierry Gautier
//
//
//
// ==========================================================================

#include <CORNET_network.h>
#include <string>
#include <fstream>

namespace CORNET {


// --------------------------------------------------------------------------
/** send  to a node (AM,AWS,WS protocols)
    @param   req  request (IN)
    @param   ch  channel  (IN)
*/
void CORNET_net::post( ACOM::Network::Request* req , ACOM::Network::OutChannel* ch)
      throw (ACOM::ComFailure) 
{
  if (req) 
  {
    if (req->get_net()!=this)
      ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );
    else {
      ch->post(req);
      return;
    }
  }
  ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );
}


// --------------------------------------------------------------------------
CORNET_net::OutChannel::OutChannel( Network*	net, int peer, INUKCORBA::Server_ptr s )
 : ACOM::Network::OutChannel()
{
  init(net, peer);
  _server = s;
}


// --------------------------------------------------------------------------
CORNET_net::OutChannel::~OutChannel()
{
}


// --------------------------------------------------------------------------
//
// InukCorba::OutChannel::post
//
void CORNET_net::OutChannel::post( ACOM::Network::Request* req ) 
  throw(ACOM::ComFailure)
{
  ACOM_ASSERT_DEBUG( req !=0, "[CORNET_net::OutChannel::post] nul request");
  ACOM_ASSERT_DEBUG( get_net() !=0, "[CORNET_net::OutChannel::post] nul network");

  /* close the sequences */
  if (req->get_state() != ACOM::Network::S_POSTED)
  {
    req->set_state( ACOM::Network::S_POSTED);
    req->_datastream.packEos();
    req->_iovstream.putEos();
  }
  
  /* complete the fixed part of header */
  Header* h = &req->_header;
  h->src    = get_net()->rank();
  h->dst    = _peer;
  h->size_eh= req->_size_eh;
  h->size_h = req->_datastream.getFillSize();
  h->size_d = req->_iovstream.getByteSize();
  
//std::cout << "[Post]: size_h=" << h->size_h << std::endl;
//std::cout << "[Post]: size_d=" << h->size_d << std::endl;

  /* number of blocks of defered data */
  h->iovsz_d = req->_iovstream.getBlockSize();
  h->no=0;

  /* we need to pack the vector containing the deferred block sizes
  */
  INUKCORBA::CIOVector iiov;
  INUKCORBA::CIOVector diov;
  int nbBlocHeader = 1 + (h->size_eh >0 ? 1 : 0);
  iiov.length( req->_datastream.getBlockSize()+nbBlocHeader );
  diov.length( h->iovsz_d );
  
  // - for the header part:
  iiov[0].replace( sizeof(ACOM::Network::Header), 
                   sizeof(ACOM::Network::Header), 
                   (CORBA::Octet*)h,
                   false );

  if (h->size_eh >0) {
    iiov[1].replace( h->size_eh,
                     h->size_eh,
                     (CORBA::Octet*)req->_ext_header,
                     false );
  }
  
  // - for the immediat data part:
  int i =nbBlocHeader;
  for( ACOM::IOVectSeq::IteratorEntry ie= req->_datastream.beginEntry();
		   ie!=req->_datastream.endEntry();
       ++i, ++ie)
  {
    iiov[i].replace( ie->len, ie->len, (CORBA::Octet*)ie->data, false );
  }

  // - for the defered data part:
  i =0;
	for( ACOM::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
		   ie!=req->_iovstream.endEntry();
       ++i, ++ie)
  {
    diov[i].replace( ie->len, ie->len, (CORBA::Octet*)ie->data, false );
  }

  /* send this request */
  try {
    _server->doIt( iiov, diov );
    req->notify();
  } 
  catch (CORBA::COMM_FAILURE& ex) 
  {
#if 0
		std::cerr << "Caught system exception COMM_FAILURE -- unable to contact the "
				      << "object." << std::endl;
    std::cerr << /*getrank() <<*/ "::in [OutChannel::post] header\n"
              << "size_h= " << h->size_h << std::endl
              << "size_d= " << h->size_d << std::endl
              << "node  = " << h->dst << std::endl
              << "proto = " << h->protocol << std::endl;
#endif
    ACOM::ComFailure::Code e = ACOM::ComFailure::FAIL;
    req->set_error(e);
    get_net()->deallocate_request( req );
		ACOM::throwit( ACOM::ComFailure(e) );
  } 
  catch (CORBA::SystemException& ex) {
    ACOM::ComFailure::Code e = ACOM::ComFailure::FAIL;
    req->set_error(e);
    get_net()->deallocate_request( req );
    ACOM::throwit( ACOM::ComFailure(e) );
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    ACOM::ComFailure::Code e = ACOM::ComFailure::FAIL;
    req->set_error(e);
    get_net()->deallocate_request( req );
    ACOM::throwit(ACOM::ComFailure(e) );
  }
}



} // namespace InukCorba

