// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_dns_impl.cpp,v $
// Copyright(c)'2001 by Apache Team
// acom_corba.idl : 
// Author: Thierry Gautier
// $Id: CORNET_dns_impl.cpp,v 1.4 2003/07/28 20:21:19 thierry Exp $
// ==========================================================================

#include <CORBA.h>
#include <CORNET_dns_impl.h>
#include <string>
#include <fstream>

namespace CORNET {

//#define CORNET_TRACE_CALL

// --------------------------------------------------------------------------
// Constructor
//
DNS_impl::DNS_impl(
  PortableServer::POA_ptr poa, CORBA::ORB_ptr orb
)
  : _poa (PortableServer::POA::_duplicate(poa) ),
    _orb (CORBA::ORB::_duplicate(orb) )
{
  /* not need with current impl of PTH: ctsor call init */
  _access.init();
  _is_defined.clear();
  _all_servers.clear();
  _all_servers.reserve(128);
}

DNS_impl::~DNS_impl()
{
  //_access.destroy();
}

PortableServer::POA_ptr DNS_impl::_default_POA()
{
  return PortableServer::POA::_duplicate(_poa);
}

CORBA::ORB_ptr DNS_impl::_the_ORB()
{
  return CORBA::ORB::_duplicate(_orb);
}



// --------------------------------------------------------------------------
//
// IDL:InukCorba/Server/get_server:1.0
//
INUKCORBA::Server_ptr DNS_impl::get_server(CORBA::Long rank)
      throw(CORBA::SystemException)
{
#ifdef CORNET_TRACE_CALL
  std::cout << "In : " << __PRETTY_FUNCTION__ << ":" << rank << std::endl;
#endif
  PTH::CriticalGuard critical(_access);
  if ((rank <0) || ((size_t)rank >= _all_servers.size()))
    return INUKCORBA::Server::_nil();
  if ( !_is_defined[rank] )
    return INUKCORBA::Server::_nil();
#ifdef CORNET_TRACE_CALL
  std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return INUKCORBA::Server::_duplicate(_all_servers[rank]);
}



// --------------------------------------------------------------------------
//
// IDL:InukCorba/Server/get_servers:1.0
//
INUKCORBA::ListServers* DNS_impl::get_servers()
      throw(CORBA::SystemException)
{
  INUKCORBA::ListServers* ls = new INUKCORBA::ListServers;
  PTH::CriticalGuard critical(_access);
  ls->length( _all_servers.size() );
  for (size_t i=0; i<_all_servers.size(); ++i)
    (*ls)[i] = INUKCORBA::Server::_duplicate(_all_servers[i]);
  return ls;
}


// --------------------------------------------------------------------------
//
// IDL:InukCorba/Server/add_server:1.0
//
CORBA::Long DNS_impl::add_server( INUKCORBA::Server_ptr s )
      throw(CORBA::SystemException)
{
#ifdef CORNET_TRACE_CALL
  std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl << std::flush;
  std::cout << "Server IOR : " << _orb->object_to_string(s) << std::endl << std::flush;
#endif

  PTH::CriticalGuard critical(_access);
  long newrank = _all_servers.size();

  _all_servers.resize( newrank + 1);
  _is_defined.resize( newrank + 1);
  _is_defined[newrank] = true;
  _all_servers[newrank] = INUKCORBA::Server::_duplicate(s);
  
  /* Signal to all other nodes that a new process has been added */
  for (int i=0; i<= newrank; ++i)
  {
    //if (i == rank) continue;
    if (_is_defined[i]) {
      _all_servers[i]->signal_server( newrank, _all_servers[newrank] );
    }
  }
#ifdef CORNET_TRACE_CALL
  std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl << std::flush;
#endif
  return newrank;
}


} // -namespace

