// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_network.cpp,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#include <CORNET_network.h>

namespace CORNET {

// --------------------------------------------------------------------------
//
// Constructor
//
CORNET_net::CORNET_net()
 : acom::Network()
{
  initAllocators();
  _primary_net = NULL;
  _is_init = false;
  _nb_init = 0;
  _is_initcommit = false;
  _is_terminate = false;
  _access.init();
  _cond.init();
 

  set_property("name","CORNET");
  set_property("primary","true");
  set_property("secondary","true");
  set_name("CORNET");
}


// --------------------------------------------------------------------------
CORNET_net::~CORNET_net()
{
  _is_init = false;
  _nb_init = 0;
  _is_initcommit = false;
  _is_terminate = false;
  _is_defined.clear();
  _access.destroy();
  _cond.destroy();
  _local_servant = 0;
  _dns_servant = 0;
  _all_servers.clear();
  _all_channels.clear();
}


// ---------------------------------------------------------------------------
/** Network specific barrier
    \warning  parameters count, time out etc ...( voir avec Cyrille)
*/
void CORNET_net::barrier()
  throw (acom::ComFailure) 
{
  static int tag = 1;
  _barrier->barrier(++tag);
}


// ---------------------------------------------------------------------------
void CORNET_net::initAllocators() 
{
  _comClock=0;
  _reqA.init(_request_threshold);
  _bA.init(_block_sz, _block_threshold  );
  _iovA.init( sizeof(acom::IOVectCell)+(_nb_entries*sizeof(acom::IOVectEntry)),
              _iovect_threshold );
}


// ---------------------------------------------------------------------------
/** plug the current node to a  network address
      @param   nodelabel(IN) network label of remote node
    \warning ajouter les aliasing d'adresse de noeud struct  ( A voir  partir
    IP/port, url corba ..)
*/
void CORNET_net::bind_node( std::string nodelabel )
  throw(ACOM::NetworkError)
{
}


// ---------------------------------------------------------------------------
/** add a new node to the network
  @param  nodename(IN) network identification of remote node
  @return local node number of the new node
  \warning MUST fail when multiple add of same addreses
  \warning ajouter les aliasing d'adresse de noeud   ( A voir  partir fd,
  IP/port, url corba ..)
*/
void CORNET_net::add_node( std::string nodelabel, LocalNodeNo  peer )
      throw(ACOM::NetworkError)
{
}

// ---------------------------------------------------------------------------
/** del a previously added node to the network
    @param  peer  local node Number of remote node to disconnect
    \warning ajouter les aliasing d'adresse de noeud ( A voir  partir
    IP/port, url corba ..)
*/
void CORNET_net::del_node( ACOM::Network::LocalNodeNo  peer  )
  throw(ACOM::NetworkError) 
{
  ACOM::throwit( ACOM::NetworkError("[[CORNET_net::del_node::]]  not yet implemented") );
}


// ---------------------------------------------------------------------------
/** Return the route (channel) to send data to LocalNodeNo peer
    @param   peer(IN) remote node id
*/
ACOM::Network::OutChannel* CORNET_net::get_default_route( LocalNodeNo peer )
    throw(ACOM::NetworkError)
{
  {
    PTH::CriticalGuard critical(_access);
    
    if ((peer < _all_channels.size()) && (_is_defined[peer])) 
    {
      /* route to peer */
      return _all_channels[peer];
    }
    
    if (peer >= _all_servers.size())
    {
      size_t sz = _all_servers.size();
      _all_channels.resize( peer + 1);
      _all_servers.resize( peer + 1);
      _is_defined.resize( peer + 1);
      for (size_t i= sz; i<=peer; ++i)
      {
        _is_defined[i] = false;
        _all_servers[i] = INUKCORBA::Server::_nil();
        _all_channels[i] =0;
      }
    }
  }
  
  /* get the server: */
  _all_servers[peer] = _dns_servant->get_server( peer );
  if ( CORBA::is_nil( _all_servers[peer] ) )
    ACOM::throwit( ACOM::NetworkError() );
  _is_defined[peer] = true;
  return _all_channels[peer] = new OutChannel( this, peer, _all_servers[peer] );
}  


// ---------------------------------------------------------------------------
/** Set a channel as a default channel to a destination node
    @param   ch default channel
*/
void CORNET_net::set_default_route(  ACOM::Network::OutChannel* ch  )
    throw(ACOM::NetworkError)
{
}  


// ---------------------------------------------------------------------------
/** Open a channel between local node  to remote peer
    @param   peer(IN) remote node id
    \warning ajouter les aliasing d'adresse de noeud
*/
ACOM::Network::OutChannel* CORNET_net::open_channel( LocalNodeNo peer )
      throw(ACOM::ComFailure)
{
  return 0;
}

// ---------------------------------------------------------------------------
/** Close a channel between local node  to remote peer
    @param   ch (IN)channel to close
    \warning ajouter les aliasing d'adresse de noeud
*/
void CORNET_net::close_channel( ACOM::Network::OutChannel* ch )
    throw(ACOM::ComFailure)
{
}  


// ---------------------------------------------------------------------------
ACOM::Network::Request* CORNET_net::allocate_request () 
    throw (ACOM::BadAlloc) 
{
  CORNET_net::Request* req = _reqA.alloc() ;
  if (!req) ACOM::throwit( ACOM::BadAlloc("[[CORNET_net::allocate_request::]] fail") );
  if (!req->init( _block_sz, &_bA, _nb_entries, &_iovA,(ACOM::Network*)this ))
  {
    _reqA.free(req);
    ACOM::throwit( ACOM::BadAlloc("[[CORNET_net::allocate_request::]] fail") );
  }
  return req;
}

/** free request descriptor when unused
*/
// ---------------------------------------------------------------------------
void CORNET_net::deallocate_request( ACOM::Network::Request* rq ) 
    throw (ACOM::InvalidArgument) 
{
  CORNET_net::Request* req = (CORNET_net::Request*)rq;
  if (req) 
  {
    if (req->get_net()!= this )
      ACOM::throwit( ACOM::InvalidArgument("[[CORNET_net::deallocate_request::]] not aloopback request") );
    else
      _reqA.free(req);
      //delete req; 
    return;
  }
  else 
    ACOM::throwit( ACOM::InvalidArgument("[[CORNET_net::deallocate_request::]] null request") );
}


// --------------------------------------------------------------------------
const std::string& CORNET_net::get_urlconnect( ) 
      throw (ACOM::NetworkError)
{
  static std::string urlserver = _orb->object_to_string( _local_servant );
  return urlserver;
}


// --------------------------------------------------------------------------
void CORNET_net::CallContext::reset()
{
  /* free all user immediate data blocs which are a CORBA allocated blocs */
  for (size_t i=0; i< _iiov_sz; ++i)
  {
    if (_iiov[i].data !=0) ::free(_iiov[i].data);
    _iiov[i].data = 0;
    _iiov[i].len  = 0;
  }
  _iiov_sz = 0;

  /* free all deferred data blocs which are a CORBA allocated blocs for AM protocol */
  if ((header !=0) && (header->protocol == ACOM::Network::P_AM))
  {
    for (size_t i=0; i< iov_sz; ++i)
    {
      if (/*(iov[i].len >0) && */(iov[i].data !=0)) ::free(iov[i].data);
      iov[i].data = 0;
      iov[i].len  = 0;
    }
  }
  iov_sz = 0;

  /* free extended header bloc which is a CORBA allocated bloc */
  if (ext_header !=0) ::free( ext_header );
  ext_header = 0;

  /* free header bloc which is a CORBA allocated bloc */
  if (header !=0) ::free( header );
  header = 0;
  ACOM::Network::CallContext::reset();
  //destroy();
}


// --------------------------------------------------------------------------
ACOM::Network::CallContext* CORNET_net::allocate_callcontext(size_t sz) throw (ACOM::BadAlloc)
{
  //std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
  if (sz == (size_t)-1) sz = 2048;
  CORNET_net::CallContext* cc = new CORNET_net::CallContext;
  cc->init( this, sz );
  //std::cout << "Init OK : " << std::endl;
  cc->_iiov_sz = 0;
  cc->_iiov = 0;

  //std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
  return cc;
}


// --------------------------------------------------------------------------
void CORNET_net::deallocate_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument)
{
  cc->destroy( );
  delete cc;
}


// --------------------------------------------------------------------------
ACOM::Network::CallContext* CORNET_net::resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) 
  throw (ACOM::BadAlloc)
{
  if (cc->_stack_size > sz) return cc;
  size_t offset = cc->size();
  ACOM::Network::CallContext*  new_cc = allocate_callcontext(sz);
  memcpy( new_cc->_base, cc->_base, offset );
  deallocate_callcontext( cc );
  return new_cc;
}


// --------------------------------------------------------------------------
ACOM::Network::CallContext* CORNET_net::acquire_callcontext() throw (ACOM::BadAlloc)
{
  //std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
  PTH::CriticalGuard critical( _access);
  return ACOM::Network::acquire_callcontext();
}
  

// --------------------------------------------------------------------------
void CORNET_net::release_callcontext(ACOM::Network::CallContext* c) throw (ACOM::InvalidArgument)
{
  PTH::CriticalGuard critical( _access);
  ACOM::Network::release_callcontext(c);
}

} //End of namespace CORNET_NET

extern "C" {

	static CORNET::CORNET_net* corbanet =0;

	// - dynamic library initialization:
	static void factory_initialize() 
	{
		corbanet = new CORNET::CORNET_net();
		acom::Acom::bind_network("CORNET", corbanet );
	}
	extern void CORNET_factory_initialize() 
	{
		//dbg.log("","factory init\n");
		factory_initialize();
		//dbg.log("","factory init OK\n");
	}
	
	// - dynamic library terminaison:
	static void factory_destroy() 
	{
		delete corbanet;
	}
	extern void CORNET_factory_destroy() 
	{
		return factory_destroy();
	}
}

