// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_network.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#ifndef __CORNET_NET_H
#define __CORNET_NET_H

#include <acom>
#include <CORNET_servers.h>
#include <ics>

namespace CORNET {
class Server_impl;
class DNS_impl;

/** CORBA Network implementation.
  The network implementation use CORBA as transport layer for the data 
of the Inuktitut active messages (protocol AM,WS,AWS).
Each Inuktitut's message was transformed into a sequence of iovectors 
that will be send to the destination.
  CORBA Inuktitut network was called 'CORNET' (in the properties/name).
It is a primary kind of network and a secondary kind of network that 
could be initialized over an other existing Inuktitut network.
  The initialisation process is quite simple : a master should be spawned
before others processes. The master generates its IOR that should be
passed on the command line for the other slaves. At initialisation, the
master reach an initial group that could be of any size : when a corresponding
number of slaves were connected to the master, the master continues its
execution and the slaves returns.
  'CORNET' network builts at runtime the routing table bewteen nodes: after
the initialisation, the slave know only the master node (and vis et versa).
A slave node that want to send a message an other node not in its routing table,
call a kind of DNS (the master node) that returns its the IOR of the destination
node or 'unknown'. Then the message is send to the destination node.
  With 'CORNET' network it is possible to add at runtime new node for the 
computation, each new node contact any of the running node and ask to reach
the current network (in sens of Inuktitut), this is a join.
*/
class CORNET_net: public ACOM::Network {
  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------

public:
  // --------------------------------------------------------------------------
  //
  // Constructor
  //
  CORNET_net();
  ~CORNET_net();

  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
  int initialize( int* argc, char** argv[] ) 
    throw (ACOM::NetworkError);

  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
  int initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
    throw (ACOM::NetworkError);


  /** Second stage of the initialisation process.
      End of the first stage of initialisation: all local initialisations
      are done, communication could occurs after the second stage (after
      the call to initcommit) or during the call of initcommit in an other
      module.
  */
  int commit () throw (ACOM::NetworkError);

  /** Terminate the net...
  */
  int terminate () throw (ACOM::NetworkError);


public:
  /** Allocate a new request.
  */
  ACOM::Network::Request* allocate_request () throw (ACOM::BadAlloc);

  /** free request descriptor when unused
  */
  void deallocate_request( ACOM::Network::Request* req ) throw (ACOM::InvalidArgument);

  /** Cancel a request allocated by allocate_request ( true == cancel done)
      Cancelation of request depend on state of progress
      @param req (IN)
  */
  bool cancel_request( ACOM::Network::Request* req )
    throw (ACOM::InvalidArgument)
  { return false; }


  // ---------------------------------------------------------------------------
  /** Allocate  an immediate buffer of sz bytes in network memory
      @param  sz (IN) in bytes
  */
  void* allocate_data( size_t sz )
    throw (ACOM::BadAlloc) 
  { return std::malloc(sz); }


  /** Deallocate a previously allocated immediate buffer  in network memory
      @param  data (IN) byte block to deallocate
  */
  void deallocate_data(  void* data )
    throw (ACOM::InvalidArgument) 
  { std::free(data); }


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** plug the current node to a  network address
       @param   nodelabel(IN) network label of remote node
      \warning ajouter les aliasing d'adresse de noeud struct  ( A voir  partir
      IP/port, url corba ..)
  */
  void bind_node( std::string nodelabel )
    throw (ACOM::NetworkError);

  /** add a new node to the network
    @param  nodename(IN) network identification of remote node
    @return local node number of the new node
    \warning MUST fail when multiple add of same addreses
    \warning ajouter les aliasing d'adresse de noeud   ( A voir  partir fd,
    IP/port, url corba ..)
  */
  void add_node( std::string nodelabel, LocalNodeNo  peer )
    throw (ACOM::NetworkError);

  /** del a previously added node to the network
      @param  peer  local node Number of remote node to disconnect
      \warning ajouter les aliasing d'adresse de noeud ( A voir  partir
      IP/port, url corba ..)
  */
  void del_node( LocalNodeNo  peer  )
    throw (ACOM::NetworkError);


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Return the route (channel) to send data to LocalNodeNo peer
      @param   peer(IN) remote node id
  */
  ACOM::Network::OutChannel* get_default_route( LocalNodeNo peer )
    throw(ACOM::NetworkError);

  /** Set a channel as a default channel to a destination node
      @param   ch default channel
  */
  void set_default_route(  ACOM::Network::OutChannel* ch  )
    throw(ACOM::NetworkError);


  /** Open a channel between local node  to remote peer
      @param   peer(IN) remote node id
      \warning ajouter les aliasing d'adresse de noeud
  */
	ACOM::Network::OutChannel* open_channel( LocalNodeNo peer )
       throw(ACOM::ComFailure);

  /** Close a channel between local node  to remote peer
      @param   ch (IN)channel to close
      \warning ajouter les aliasing d'adresse de noeud
  */
  void close_channel( ACOM::Network::OutChannel* ch )
      throw(ACOM::ComFailure);

  /** send  to a node (AM,AWS,WS protocols)
      @param   req  request (IN)
      @param   ch  channel  (IN)
  */
  void post( ACOM::Network::Request* req, ACOM::Network::OutChannel* ch)
    throw (ACOM::ComFailure);

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */
  void barrier()
    throw (ACOM::ComFailure);


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  class CallContext : public ACOM::Network::CallContext {
  public:
    void reset();
  protected:
    ACOM::IOVectEntry* _iiov; /* to store allocated bloc from CORBA */
    size_t             _iiov_sz;   /* size of _user_iiov */
    friend class Server_impl;
    friend class CORNET_net;
  };

  /** Allocate a new call context to treats incomming message
  */
  ACOM::Network::CallContext* allocate_callcontext(size_t sz) throw (ACOM::BadAlloc);

  /** Deallocate a call context
  */
  void deallocate_callcontext(ACOM::Network::CallContext*) throw (ACOM::InvalidArgument);

  /** Reallocate of the stack of the call context
  */
  ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* ctxt, size_t sz ) throw (ACOM::BadAlloc);

  /** Acquire a call context for a new incomming communication.
  */
  ACOM::Network::CallContext* acquire_callcontext() throw (ACOM::BadAlloc);
  
  /** Release a call context.
  */
  void release_callcontext(ACOM::Network::CallContext* ) throw (ACOM::InvalidArgument);

public:
	/** the only channel class
  */
  class OutChannel: public ACOM::Network::OutChannel {
  public:
    void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);
    INUKCORBA::Server_ptr& get_server()
    { return _server; }
  protected:
    /** Constructor
    */
    OutChannel( Network*	net, int peer, INUKCORBA::Server_ptr s );

    /** destructor
    */
    ~OutChannel();

    friend class CORNET_net;

  protected:
    INUKCORBA::Server_ptr _server;
  };


  /** The request class 
  */
  class Request: public ACOM::Network::Request, public PTH::Linkable<Request> {
  protected:
    friend class CORNET_Channel;
    friend class CORNET_net;
    friend class CORNET_router;
  };

protected:

  /** Receipt class
  */
  class Receipt: public ACOM::Network::Receipt {
  protected:
    friend class CORNET_Channel;
    friend class CORNET_net;
  };


  /** iovect entries in a cell
   */
  static const int _nb_entries=64;

  /** packed block size
   */
  static const int _block_sz =16384;

  /** Logical clock for timestamping request
   */
  PTH::AtomicCounter _comClock;

  /** size of cached free blocks
   */
  static const int _block_threshold=32; 

  /** size of cached free blocks
   */
  static const int _iovect_threshold=32;

  /** size of cached free blocks
   */
  static const int _request_threshold=32;


  /** byte cell allocator
   */
  ACOM::ByteCellAllocator _bA;

  /** iovec allocator
   */
  ACOM::IOVectCellAllocator _iovA;

  void initAllocators();

  /** request allocator
   */
  typedef PTH::SyncCellAllocator<Request> RequestAllocator ;
  RequestAllocator _reqA;

  // ------------------------------------------------------------------------
  //
  // Methods to initialize the CORBA object
  //
  //@{
  /*@name init_corba
  Initialize the CORBA ORB and accept incomming message */
  void init_corba(int* argc, char*** argv);

  /*@name init_master
  Initialize the master object if this network,
  return the ior of the the servant. After this operation
  the master accept all incomming message.
  A call of this method should be done on the master object. */
  void init_master( ACOM::Network* net, const std::string& file_ior );

  /*@name init_slave
  Initialize a slave object. After this call the CORBA object
  could accept incomming message. */
  void init_slave( ACOM::Network* net, const std::string& ior_master);


  /*@name join_group
    Call by a new node to join the current group.
    The rank of the new added node is returned.
  */
  void join_group( long rank, INUKCORBA::Server_ptr& s );

  //@}
  // ------------------------------------------------------------------------
  const std::string& get_urlconnect( ) 
       throw (ACOM::NetworkError);

protected:
  bool   _is_init;         // to avoid multiple calls to the associated
  int    _nb_init;         // number of calls
  bool   _is_initcommit;   // methods
  bool   _is_terminate;

  ACOM::Network*          _primary_net;
  PTH::Mutex              _access; // to lock access to internal data
  PTH::Condition          _cond;   // to wait initial group
  PortableServer::POA_var _poa;
  CORBA::ORB_var          _orb;

  ICS::BarrierService*    _barrier;            // - service barrier
  int                     _size_initial_group; // - initial slave to wait
  INUKCORBA::Server_var   _local_servant;      // - local servant
  INUKCORBA::Server_var   _master_servant;     // - master servant
  INUKCORBA::DNS_var      _dns_servant;        // - DNS servant to get node
  std::vector<bool>       _is_defined;         // - [i] true iff defined
  std::vector<INUKCORBA::Server_var>  _all_servers;  // - all known servers
  std::vector<OutChannel*> _all_channels; // - routes to other 

  friend class Server_impl;
}; // -- end class Network

} // End of namespace

#endif //end of define __CORNET_NET_H
