// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_network_init.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: Thierry Gautier
//
//
//
// ==========================================================================
#include <CORBA.h>
#include <CORNET_network.h>
#include <CORNET_dns_impl.h>
#include <CORNET_servers_impl.h>

#include <string>
#include <fstream>
#include <sstream>

//#define CORNET_TRACE_CALL

namespace CORNET {

#define MAXRETRY 1 

// --------------------------------------------------------------------------
//
// Static object that will register to the list of modules
//
// --------------------------------------------------------------------------//


// --------------------------------------------------------------------------//
// Worker thread that run the ORB
//
class Worker : public PTH::ThreadModel<Worker> {
public:
  ~Worker( ) {}
  Worker( CORBA::ORB_ptr theorb )
   : _theorb(theorb)
  {}

  void* run()
  {
    _theorb->run();
  }
public:
  CORBA::ORB_ptr _theorb;
};
#if 0 /*DEPRECATED */
void* handler_orb(void* arg)
{
  Worker* theworker = (Worker*)arg;
  theworker->run(); 
  return 0;
}
#endif


// --------------------------------------------------------------------------//
// Init the master node : 
//
void CORNET_net::init_corba(int* argc, char*** argv)
{
#ifdef CORBA_USE_OMNIORB
#ifndef CORBA_USE_OMNIORB4
  omniORB::maxTcpConnectionPerServer = 64;
  omniORB::scanGranularity ( 10 );
#endif
#endif
  // - Connect to the ORB
    _orb = CORBA::ORB_init (*argc, *argv);

  CORBA::Object_var poaObj;
  poaObj = _orb->resolve_initial_references ("RootPOA");
  _poa = PortableServer::POA::_narrow (poaObj);
}


// --------------------------------------------------------------------------//
// Init the master node : 
//
void CORNET_net::init_master( ACOM::Network* primary_net, const std::string& file_ior )
{
#ifdef CORNET_TRACE_CALL
  std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
#endif  
  ACOM_ASSERT_DEBUG( _node_rank ==0, "[bad node rank for master]");

  try {
    DNS_impl* dns_server = new DNS_impl ( _poa, _orb );
    _poa->activate_object( dns_server );
    _dns_servant= dns_server->_this ();
    dns_server->_remove_ref();

#ifdef CORNET_TRACE_CALL
    std::cout << "New DNS object, IOR=  " << _orb->object_to_string( _dns_servant ) << std::endl;
#endif    
    
    int myrank  = _dns_servant->add_server( _local_servant );
    ACOM_ASSERT_DEBUG( myrank ==0, "[bad node rank for master]");
    
    /* register master into the DNS */
    //_dns_servant->add_server( myrank, _local_servant );
    std::string master_ior_name = _orb->object_to_string( _master_servant );
    std::string ior_name = _orb->object_to_string( _local_servant );

    if (primary_net ==0) {
      std::ofstream fout(file_ior.c_str());
#ifdef CORNET_TRACE_CALL
      std::cout << " ior master= : " << master_ior_name  << std::endl;
      std::cout << " ior connect= : " << ior_name  << std::endl;
#endif
      fout << ior_name;
    }
#ifdef CORNET_TRACE_CALL
    std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
#endif
    return;
  } 
  catch (...) {
    std::cerr << "ERROR: unknown exception during master node initialization -stop."
              << std::endl << std::flush;
    throw;
  }
  return;
}


// --------------------------------------------------------------------------//
// Init the slave node : 
//
void CORNET_net::init_slave( ACOM::Network* primary_net, const std::string& in_ior_master )
{
  ACOM_ASSERT_DEBUG( _node_rank !=0, "[bad node rank for slave]");
  try
  {
#ifdef CORNET_TRACE_CALL
    std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
#endif    
    std::string ior_master;
    if (primary_net !=0) 
    {
      //ior_master = init_service->get_iormaster();
    }
    else {
      ior_master = in_ior_master;
    }
    
    CORBA::Object_var objRef = 
        _orb->string_to_object( ior_master.c_str() );

#ifdef CORNET_TRACE_CALL
    std::cout << "get master object : " <<  ior_master << std::endl;
#endif
    PTH::Thread::sleep( 2*(ACOM::Acom::get_pid() % 100), 0);
    // - get the master and dns
    try {
      _master_servant = INUKCORBA::Server::_narrow (objRef);
      if (CORBA::is_nil( _master_servant )) 
      {
        std::cout << "nul master object" << std::endl;
        throw;
      }
    } catch (const CORBA::Exception& ex ) {
      std::cerr << __LINE__ << "::ERROR: CORBA exception, when binding to master"
                << std::endl;
      throw;
    } catch (...) {
      std::cerr << __LINE__ << "::ERROR: Cannot get Master Node, unknown exception -stop."
                << std::endl;
      throw;
    }

#ifdef CORNET_TRACE_CALL
    std::cout << "get Master object IOR=" << _orb->object_to_string(_master_servant) << std::endl;
#endif

    bool dns_ok = false;
    for (int i=0; i<MAXRETRY; ++i)
    {
      try {
        _dns_servant = _master_servant->get_dns();
        if (CORBA::is_nil( _dns_servant )) 
        {
          std::cout << "nul dns object" << std::endl;
          continue;
        }
        dns_ok = true;
      } catch (const CORBA::Exception& ex ) {
        std::cerr << __LINE__ << "::ERROR: CORBA exception, when get dns from master"
                  << std::endl;
      } catch (...) {
        std::cerr << __LINE__ << "::ERROR: Cannot get DNS object, unknown exception -stop."
                  << std::endl;
      }
      if (!dns_ok)
        PTH::Thread::sleep( 2*(ACOM::Acom::get_pid() % 100), 0);      
    }
    if (!dns_ok) 
      throw;
#ifdef CORNET_TRACE_CALL
    std::cout << "get DNS object, IOR=" << _orb->object_to_string( _dns_servant ) << std::endl;
#endif


    /* register to the master */
    bool register_ok = false;
    for (int i=0; i<MAXRETRY; ++i)
    {
      try {
        _node_rank = _dns_servant->add_server( _local_servant );
        register_ok = true;
      } catch (const CORBA::Exception& ex ) {
        std::cerr << __LINE__ << "::ERROR: CORBA exception, when add my self"
                  << std::endl;
      } catch (...) {
        std::cerr << __LINE__ << "::ERROR: get add my self, unknown exception -stop."
                  << std::endl;
      }
      if (!register_ok)
        PTH::Thread::sleep( 2*(ACOM::Acom::get_pid() % 100), 0);      
    }
    if (!register_ok) 
      throw;


    /* get all server from 0 to _node_rank */
    for (int i=0; i < _node_rank; ++i)
    {
      bool get_ok = false;
      for (int k =0; k<MAXRETRY; ++k)
      {
        try {
          INUKCORBA::Server_ptr s = _dns_servant->get_server( i );
          join_group( i, s );
          get_ok = true;
        } catch (const CORBA::Exception& ex ) {
          std::cerr << __LINE__ << "::ERROR: CORBA exception, when get other servers"
                    << std::endl;
        } catch (...) {
          std::cerr << __LINE__ << "::ERROR: get other servers, unknown exception -stop."
                    << std::endl;
        }
        if (!get_ok)
          PTH::Thread::sleep( 2*(ACOM::Acom::get_pid() % 100), 0);      
      }
    }
    
#ifdef CORNET_TRACE_CALL
    std::cout << "New rank: " << _node_rank << std::endl;    
#endif
    ACOM_ASSERT_DEBUG( _node_rank !=0, "[bad node rank for slave]");
#ifdef CORNET_TRACE_CALL
    std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
#endif
  }
  catch (const CORBA::Exception& ex ) {
    std::cerr << "ERROR: CORBA exception, when binding to master"
              << std::endl;
  } catch (...) {
    std::cerr << "ERROR: Cannot get Master Node, unknown exception -stop."
              << std::endl;
    throw;
  }
}


// ---------------------------------------------------------------------------
/*@name reach_group
  Signal the current node that a new server has reach the group and
  with given rank.
*/
void CORNET_net::join_group( long rank, INUKCORBA::Server_ptr& s )
{
#ifdef CORNET_TRACE_CALL
  std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
#endif
  PTH::CriticalGuard critical(_access);
  size_t sz = _all_servers.size();
  if ((size_t)rank >= sz) 
  {
    _all_servers.resize( rank + 1);
    _all_channels.resize( rank + 1);
    _is_defined.resize( rank + 1);
    for (int i=sz; i<=rank; ++i)
    {
      _is_defined[i] = false;
      _all_servers[i] = INUKCORBA::Server::_nil();
      _all_channels[i] = 0;
    }
  }
  _is_defined[rank] = true;
  _all_servers[rank] = INUKCORBA::Server::_duplicate(s);
  _all_channels[rank] = new OutChannel( this, rank, _all_servers[rank] );

  ++_node_count;  
  if (--_size_initial_group ==0)
    _cond.signal();
#ifdef CORNET_TRACE_CALL
  std::cout << "Size initial group=" << _size_initial_group << std::endl;
  std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
#endif
}



// --------------------------------------------------------------------------
//
// CORBA object
//

// Initialization of the corba network.
// Deployment of the application occurs in a two steps, using
// a two phase protocol to built the network:
// 1/ first the master is spawned with only command line arguments:
// -a0corba-ior-file <filename> 
// -a0corba-count <#nodes>
// and the master writes into filename its ior and it is waiting for
// the connection of <#nodes>-1 slaves. 
// 2/ each slave is spawned with only command line arguments:
// -a0corba-ior-master <ior> 
// -a0corba-fileior-master <file> 
// First, it declares its ior to the master. In a second step, 
// the master callback to each slave the set of all IOR when it receives
// ior for <#nodes>-1 slaves.
//
int CORNET_net::initialize( int* argc, char** argv[] ) 
    throw (ACOM::NetworkError)
{
  return initialize( argc, argv, 0 );
}

int CORNET_net::initialize( int* argc, char** argv[], ACOM::Network* primary_net )
    throw (ACOM::NetworkError)
{
  // to avoid multiple call of this method
  _nb_init++;
  if (_is_init) return 0;
  _is_init = true;
  _primary_net = primary_net;

  //PTH::Pth::initialize(argc, argv);
  //PTH::Pth::commit();

  // not need by the current implementation of PTH, cstor initial them
  _all_servers.clear();
  _all_servers.reserve(128);
  _all_channels.clear();

#ifdef CORNET_TRACE_CALL
  std::cout << "In " << __PRETTY_FUNCTION__ << std::endl;
#endif

  _barrier = new ICS::FlatBarrier;
  ACOM::Service::Manager.bind( ACOM::ServiceManager::CORBA_LNS, 2, _barrier );
  _barrier->initialize( this );


  std::string ior_master = "";
  std::string ior_file = "";

  // - get the number of node and my node id in the command line
  int i;
  bool iam_master = true;
  _node_rank  = 0;   // - I assume that I am the master ...
  _node_count = 0;   // ... and I was alone, incr by join_group
  _size_initial_group = 1; // I was alone
  int curr2set =0;
#ifdef CORBANET_DEBUG_ARGV
  std::cout << "[CORBA::CORNET_net::initialize], input argv" <<  std::endl;
  for (i =0; i < *argc; ++i) 
    std::cout << "CORBA, argv[" << i << "]=" << (*argv)[i] << std::endl;
#endif

  //
  // parse command line argument, delete from it the reconnized arguments
  //
  for (i =0; i < *argc; ++i) 
  {
    if (strcmp((*argv)[i], "-cornet-usage") ==0)
    {
      std::cout << "[Usage Module CORBA NET]: options are \n" 
          << "[-cornet-count <#nodes>] for all nodes\n"
          << "[-cornet-ior-master <IOR>] for all nodes except for the master\n" 
          << "[-cornet-ior-file <filename>] on the master node\n"
          << std::endl;
      ACOM::Acom::abort(0);
    }
    else if (strcmp((*argv)[i], "-cornet-ior-file") ==0)
    {
      if (i+1 < *argc) {
        ior_file   = (*argv)[i+1]; 
      }
      else {
        std::cerr << "[usage] -cornet-ior-file <filename>" << std::endl;
        ACOM::Acom::abort(0);
      }
      ++i;
      iam_master = true;
    }
    else if (strcmp((*argv)[i], "-cornet-ior-master") ==0)
    {
      if (i+1 < *argc) {
        ior_master = (*argv)[i+1]; 
      }
      else {
        std::cerr << "[usage] -cornet-ior-master <iior>" << std::endl;
        ACOM::Acom::abort(0);
      }
      ++i;
      iam_master = false;
    }
    else if (strcmp((*argv)[i], "-cornet-count") ==0)
    {
      if (i+1 < *argc) 
        _size_initial_group = atoi((*argv)[i+1]);
      else {
        std::cerr << "[usage] -cornet-count <#nodes>" << std::endl;
        ACOM::Acom::abort(0);
      }
      ++i;
    }
    else {
      // - compact the array of args to the beginning
      if (i > curr2set) (*argv)[curr2set] = (*argv)[i];
        ++curr2set;
      }
    }
    // - set up the new argc
    *argc = curr2set;

#ifdef CORBANET_DEBUG_ARGV
  std::cout << "CORBA, output argv" <<  std::endl;
  for (i =0; i < *argc; ++i) 
    std::cout << "CORBA, argv[" << i << "]=" << (*argv)[i] << std::endl;
  std::cout << std::endl;
#endif

//std::cout << "[CORBA]: self = " << _node_rank  << std::endl;
//std::cout << "[CORBA]: count= " << _size_initial_group << std::endl;


  // - test if all arguments are OK:
  if (primary_net ==0) 
    {
    if ( iam_master && (ior_file == "") && (_size_initial_group >1)) 
    {
      std::cerr << "[usage] for MASTER corba implementation: the two \n";
      std::cerr << "following command line options should be set up.\n";
      std::cerr << "   -cornet-ior-file <filename>\n";
      std::cerr << "   -cornet-count <#nodes>\n" << std::endl;
      ACOM::Acom::abort(0);
    }
    if ( !iam_master && (ior_master == "") ) 
    {
      std::cerr << "[usage] for SLAVE corba implementation: the \n";
      std::cerr << "following command line option should be set up.\n";
      std::cerr << "   -cornet-ior-master <ior>\n";
      ACOM::Acom::abort(0);
    }

  }
  else {
    _size_initial_group = primary_net->node_count();
    _node_rank = primary_net->rank();
    if (_node_rank ==0) 
      iam_master = true;
    else 
      iam_master = false;
  }

  // - set up the _node_rank 
  if (!iam_master) 
    _node_rank = (LocalNodeNo)-1;


  // CORBA initialization
  init_corba(argc, argv);


  // POA Manager
	PortableServer::POAManager_var pman = _poa->the_POAManager();
  // - activate the POA
  pman->activate();


  // - Create the local servant object 
  //std::cout << "[CORBA::CORNET_net::initialize], create server " <<  std::endl;
  Server_impl* local_server = new Server_impl ( this, _poa, _orb );
	_poa->activate_object( local_server );
  _local_servant= local_server->_this ();
  local_server->_remove_ref();
#ifdef CORNET_TRACE_CALL
  std::cout << "New object, IOR=  " << _orb->object_to_string( _local_servant ) << std::endl;
#endif

  //std::cout << "[CORBA::CORNET_net::initialize], create server ok " <<  std::endl;

#ifdef CORNET_TRACE_CALL
  std::cout << "[CORBA] rank=" << _node_rank << ", ior=" 
            << _orb->object_to_string( _local_servant )  << std::endl;
#endif

  // - Start the worker
  Worker* worker = new Worker(_orb );
  //worker->start();
#if 0 /* DEPRECATED */
#ifndef CORBA_USE_OMNIORB
  /* SIGSEGV if using Pthread start method */
  //worker->start(&_sv_wait_end_worker);
  pthread_t tid;
  pthread_create(&tid, 0, &handler_orb, worker);
#else
  pthread_t tid;
  pthread_create(&tid, 0, &handler_orb, worker);
#endif
#endif


  // - First phase of the protocol
  if (_node_rank ==0) 
  {
    // - I'm the master pman is activate after creation of DNS, write ior into a file
#ifdef CORNET_TRACE_CALL
    std::cout << "[CORBA::CORNET_net::initialize], init master fileior=" 
              << ior_file <<  std::endl;
    std::cout << "[CORBA::CORNET_net::initialize], init master " <<  std::endl;
#endif
    init_master( primary_net, ior_file );

#ifdef CORNET_TRACE_CALL
    std::cout << "[CORBA::CORNET_net::initialize], init master ok" <<  std::endl;
#endif
  }
  else
  {
#ifdef CORNET_TRACE_CALL
    std::cout << "[CORBA::CORNET_net::initialize], init slave " <<  std::endl;
#endif
    init_slave( primary_net, ior_master );
#ifdef CORNET_TRACE_CALL
    std::cout << "[CORBA::CORNET_net::initialize], init slave ok" <<  std::endl;
#endif
  }

#ifdef CORNET_TRACE_CALL
  std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 1;
}


// --------------------------------------------------------------------------
//
// InitCommit
//
int CORNET_net::commit( )
    throw (ACOM::NetworkError)
{
	// to avoid multiple call of this method
	if (_is_initcommit) return 0;
	_is_initcommit = true;

#ifdef CORNET_TRACE_CALL
	std::cout << "In " << __PRETTY_FUNCTION__ << std::endl;
#endif

	// - second phase of the protocol : wait all servers... 
  if ( _node_rank ==0) {
    PTH::CriticalGuard critical(_access);
    while (_size_initial_group >0) {
      _cond.wait(_access);
    }
  }
  
  //_barrier->barrier(1);
  
#ifdef CORNET_TRACE_CALL
	std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 1;
}



// --------------------------------------------------------------------------
//
// Terminate
//
int CORNET_net::terminate( )
    throw (ACOM::NetworkError)
{
#ifdef CORNET_TRACE_CALL
  std::cout << "In " << __PRETTY_FUNCTION__ << std::endl;
#endif
  _nb_init--;
  if ((_nb_init > 0) || (_is_terminate)) return 0;
  _is_terminate = true;

  try {
	  PortableServer::POAManager_var pman = _poa->the_POAManager();
    // - deactivate the POA
    pman->deactivate(false,false);
    //_local_servant->release();
    //_master_servant->release();
    //_dns_servant->release();
    //_orb->shutdown(false);
    //_orb->destroy();
    PTH::Pth::terminate();
  }
  catch(const CORBA::Exception& ex) {
    //WARN TG: shutdown always returns or may return a CORBA::Exception
    //std::cerr << "CORBA exception" << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
#ifdef CORNET_TRACE_CALL
  std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 1;
}


} // -namespace INUKCORBA

