// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_servers_impl_call.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: Thierry Gautier
// $Id: CORNET_servers_impl_call.cpp,v 1.5 2003/07/30 12:56:46 thierry Exp $
// ==========================================================================

#include <CORBA.h>
#include <CORNET_network.h>
#include <CORNET_servers_impl.h>
#include <string>
#include <fstream>

namespace CORNET {

//#define ACOM_TRACE_DEBUG


// --------------------------------------------------------------------------
//
// IDL:InukCorba/Server/doWS:1.0
//
void Server_impl::doIt
 ( const INUKCORBA::CIOVector& iiov, const INUKCORBA::CIOVector& odiov )
  throw(CORBA::SystemException)
{
//std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;

  /* Get a new Call Context */
  ACOM::Network::Header* header =0;
  CORNET_net::CallContext* cc = 0;
  cc = (CORNET_net::CallContext*)_net->acquire_callcontext();
  
//std::cout << "CC acquired : " << cc << std::endl;
  /* to avoid warning to to non const usage in get_buffer(true) */
  INUKCORBA::CIOVector& diov = (INUKCORBA::CIOVector&)odiov;
  
  try {  
    /* initialize the cc :
       - get the owner ship of data in iiov and store pointers into
         in _iiov and _iiov_sz the data of the active
    */
#if 0
std::cout << "CC reset : free size=" << cc->free_size() << std::endl;
std::cout << "CORBA iiov sz:" << iiov.length() << std::endl;
for (int i=0; i<iiov.length(); ++i)
  std::cout << "CORBA iiov[" << i << "] sz:" << iiov[i].length() << std::endl;
std::cout << "CORBA diov sz:" << diov.length() << std::endl;
for (int i=0; i<diov.length(); ++i)
  std::cout << "CORBA diov[" << i << "] sz:" << diov[i].length() << std::endl;
#endif
    int nbBlocHeader = 1;
    cc->error_no  = ACOM::ComFailure::OK;
    header = cc->header 
      = (ACOM::Network::Header*)((INUKCORBA::CIOVectEntry&)iiov[0]).get_buffer(true);
    if (cc->header->size_eh >0) {
      cc->ext_header = ((INUKCORBA::CIOVectEntry&)iiov[1]).get_buffer(true);
      ++nbBlocHeader;
    }

    cc->_iiov_sz  = iiov.length()-nbBlocHeader;
    cc->iov_sz    = diov.length();
    cc->_im_data_seq.init( cc->_iiov_sz, &_net->_iovA);
//std::cout << "CC _im_data_seq init : " << std::endl;
    /* store iiov in cc as a iovector in order to free bloc after */
    cc->_iiov     = (ACOM::IOVectEntry*)cc->allocate(sizeof(ACOM::IOVectEntry)*cc->_iiov_sz); 
    cc->iov       = (ACOM::IOVectEntry*)cc->allocate(sizeof(ACOM::IOVectEntry)*cc->iov_sz); 
  
    /* resolve the service object */
    ACOM::Service* stub_server = ACOM::Service::Manager.resolv(cc->header->id);
//std::cout << "Service resolved : " << std::endl;
  
    /* build sequence of immediate data : always get owner ship of CORBA buffers */
    for (int i=0; i < cc->_iiov_sz; ++i)
    {
      cc->_iiov[i].len  = iiov[nbBlocHeader+i].length();
      /* don't get owner ship */
      cc->_iiov[i].data = (void*)((INUKCORBA::CIOVectEntry&)iiov[nbBlocHeader+i]).get_buffer(true); 
      cc->_im_data_seq.put(&cc->_iiov[i]);
    }
//std::cout << "Sequence IM ok : " << std::endl;

    /* fill in deferred IOV description */    
    switch (cc->header->protocol) {
      case ACOM::Network::P_AM:
      case ACOM::Network::P_AWS:
        /* AM + AWS */
        for (int i=0; i < cc->iov_sz; ++i)
        {
          cc->iov[i].data = 0;
          cc->iov[i].len  = diov[i].length();
          //if (cc->iov[i].len >0) 
            cc->iov[i].data = (void*)diov[i].get_buffer(true);
        }
      break;
      case ACOM::Network::P_WS:      
        /* WS */
        for (int i=0; i < cc->iov_sz; ++i)
        {
          cc->iov[i].len = diov[i].length();
          cc->iov[i].data = 0;
        }
      break;
      default:
        throw CORBA::COMM_FAILURE();
    }
//std::cout << "Sequence DM ok : " << std::endl;

    /* initialize the receipt data structure : */
    cc->receipt.init(&cc->_im_data_seq, cc->iov, cc->iov_sz, _net);
//std::cout << "Receipt INIT : " << std::endl;

    /* If WS was make a call to user service + recopy data */
    if (cc->header->protocol == ACOM::Network::P_WS) 
    {
      /* decode address */
      stub_server->server_get_iovect( cc );
//std::cout << "Call get_iovect OK : " << std::endl;
      
      /* copy diov into cc->iov */
      for (int i=0; i< cc->iov_sz; ++i)
      {
        ACOM_ASSERT_DEBUG(cc->iov[i].len == diov[i].length(), "bad assertion");
        ACOM_ASSERT_DEBUG(cc->iov[i].data != 0, "bad assertion");
        if (true) //cc->iov[i].len >0)
          memcpy( cc->iov[i].data, diov[i].get_buffer(), cc->iov[i].len);
        else 
          cc->iov[i].data = 0;
      }
    }
    
    /* Make a call to the user service entry point */
//std::cout << "Call user stub" << std::endl;
    stub_server->server_call(cc);
//std::cout << "Call user stub OK : " << std::endl;
    _net->release_callcontext( cc );
    
  } catch (ACOM::InvalidArgument &ex) {
    std::cerr << "[CORBA::impl_call] unknow service id" << std::endl;
//std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
    throw CORBA::COMM_FAILURE();
  } catch (ACOM::ComFailure& e) {
//std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
    throw CORBA::COMM_FAILURE();
  } catch (...) {
    std::cerr << "[CORBA::impl_call] unknow exception" << std::endl;
//std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
    throw CORBA::COMM_FAILURE();
  }

#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << "Out [CORBA::impl_call]" << endl;
#endif
//std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
}


} // -namespace
