// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_servers_impl_doAM.cpp,v $
// Copyright(c)'2001 by Apache Team
// acom_corba.idl : 
// Author: Thierry Gautier
// $Id: CORNET_servers_impl_doAM.cpp,v 1.2 2003/07/15 17:52:36 thierry Exp $
// ==========================================================================

#include <CORBA.h>
#include <CORNET_network.h>
#include <CORNET_servers_impl.h>
#include <string>
#include <fstream>

namespace CORNET {

//#define ACOM_TRACE_DEBUG


// --------------------------------------------------------------------------
//
// IDL:INUKCORBA/Server/doAM:1.0
//
void Server_impl::doAM
  ( CORBA::Long dest, const INUKCORBA::CIOVector& iiov, const INUKCORBA::CIOVector& diov )
      throw(CORBA::SystemException)
{
#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << ":: In [Server_impl::doAM]" << endl;
#endif
  ACOM::ComFailure::Code error_no = ACOM::ComFailure::OK;

#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << ":: In [Server_impl::doAM], get buffer" << endl;
  std::cout << "In [Server_impl::doAM] header\n"
            << "size_h= " << header->size_h << std::endl
            << "size_d= " << header->size_d << std::endl
            << "node  = " << header->node << std::endl
            << "proto = " << header->protocol << std::endl
            << "id    = " << header->id << std::endl
            << "#iov  = " << ciov.length() << std::endl
            << "iov[0]= <header>" << std::endl;
#endif

  try {

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() << ":: In [Server_impl::doAM]: service resolv" 
              << std::endl;
#endif

    // - get the header:
    ACOM::Network::Header* header = (ACOM::Network::Header*)iiov[0].get_buffer();
  
    ACOM::IOVectSize iiosz = iiov.length()-1;
    ACOM::IOVectSize diosz = diov.length();
    ACOM::IOVectSeq  im_data_seq;
    im_data_seq.init(iiosz, &_net->_iovA);

    // - get pointer to the service id:
    ACOM::Service* stub_server = 0;
    try {
      stub_server = ACOM::Service::Manager.resolv(header->id);
    } catch (ACOM::InvalidArgument &ex) {
      std::cerr << _net->rank() << "::[doAM] unknow service id: " << header->id << std::endl;
      throw CORBA::COMM_FAILURE();
    }

    // - destination iov allocate a Inuktitut::IOV, recopy entries from ciov 
    ACOM::IOVectEntry*  user_iiov = (ACOM::IOVectEntry*)::alloca(sizeof(ACOM::IOVectEntry)*iiosz); 
    ACOM::IOVectEntry*  user_diov = (ACOM::IOVectEntry*)::alloca(sizeof(ACOM::IOVectEntry)*diosz); 
    for (int i=0; i < iiosz; ++i)
    {
      user_iiov[i].len  = iiov[1+i].length();
      user_iiov[i].data = (void*)iiov[1+i].get_buffer(); /* don't get owner ship */
			im_data_seq.put(&user_iiov[i]);
    }
    for (int i=0; i < diosz; ++i)
    {
      user_diov[i].len  = diov[i].length();
      user_diov[i].data = (void*)diov[i].get_buffer(); /* don't get owner ship */
    }

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() 
              << ":: In [Server_impl::doAM]: begin call" << std::endl;
#endif
    ACOM::Network::Receipt rcpt;
		rcpt.init((ACOM::IOVectSeq*)&im_data_seq, user_diov, diosz, _net);

    stub_server->call(error_no, &rcpt, header, user_diov, diosz, 0 );
    im_data_seq.destroy();
    rcpt.finish();

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() 
              << ":: In [Server_impl::doAM]: end call " << std::endl;
#endif

  } catch (ACOM::InvalidArgument &ex) {
    std::cerr << "::[doAM] unknow service id " << std::endl;
    throw CORBA::COMM_FAILURE();
  } catch (ACOM::ComFailure& e) {
    throw CORBA::COMM_FAILURE();
  } catch (...) {
    std::cerr << "[doAM] unknow exception" << std::endl;
    throw CORBA::COMM_FAILURE();
  }

#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << "Out [Server_impl::doAM]" << endl;
#endif
}


} // -namespace
