// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/COR_net/src/CORNET_servers_impl_doWS.cpp,v $
// Copyright(c)'2001 by Apache Team
// acom_corba.idl : 
// Author: Thierry Gautier
// $Id: CORNET_servers_impl_doWS.cpp,v 1.2 2003/07/15 17:52:27 thierry Exp $
// ==========================================================================

#include <CORBA.h>
#include <CORNET_network.h>
#include <CORNET_servers_impl.h>
#include <string>
#include <fstream>

namespace CORNET {

//#define ACOM_TRACE_DEBUG


// --------------------------------------------------------------------------
//
// IDL:InukCorba/Server/doWS:1.0
//
void Server_impl::doWS
 ( CORBA::Long dest, const INUKCORBA::CIOVector& iiov, const INUKCORBA::CIOVector& diov )
      throw(CORBA::SystemException)
{
#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << ":: In [Server_impl::doAWS]" << endl;
#endif

  ACOM::ComFailure::Code error_no = ACOM::ComFailure::OK;

  try {

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() << ":: In [Server_impl::doAWS]: service resolv" 
              << std::endl;
#endif

    // - get the header:
    ACOM::Network::Header* header = (ACOM::Network::Header*)iiov[0].get_buffer();

#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << ":: In [Server_impl::doAWS], get buffer" << std::endl;
  std::cout << "In [Server_impl::doAWS] header\n"
            << "size_h= " << header->size_h << std::endl
            << "size_d= " << header->size_d << std::endl
            << "dest  = " << header->dst << std::endl
            << "src   = " << header->src << std::endl
            << "proto = " << header->protocol << std::endl
            << "iovsz_d= " << header->iovsz_d << std::endl
            << "no    = "  << header->no << std::endl
            << std::endl;
#endif
  
    ACOM::IOVectSize iiosz = iiov.length()-1;
    ACOM::IOVectSize diosz = diov.length();
    ACOM::IOVectSize dest_iovsz = diosz;

    ACOM::IOVectSeq  im_data_seq;
    im_data_seq.init( iiosz, &_net->_iovA);

    // - get pointer to the service id:
    ACOM::Service* stub_server = ACOM::Service::Manager.resolv(header->id);

    // - destination iov allocate a Inuktitut::IOV, recopy entries from ciov 
    ACOM::IOVectEntry*  user_iiov = (iiosz ==0 ? 0 : (ACOM::IOVectEntry*)::alloca(sizeof(ACOM::IOVectEntry)*iiosz)); 
    ACOM::IOVectEntry*  user_diov = (diosz ==0 ? 0 : (ACOM::IOVectEntry*)::alloca(sizeof(ACOM::IOVectEntry)*diosz)); 
    ACOM::IOVectEntry*  dest_iov  = (diosz ==0 ? 0 : (ACOM::IOVectEntry*)::alloca(sizeof(ACOM::IOVectEntry)*diosz));

    for (int i=0; i < iiosz; ++i)
    {
      user_iiov[i].len  = iiov[1+i].length();
      user_iiov[i].data = (void*)iiov[1+i].get_buffer(); /* don't get owner ship */
			im_data_seq.put(&user_iiov[i]);
    }
    for (int i=0; i < diosz; ++i)
    {
      dest_iov[i].len = user_diov[i].len  = diov[i].length();
      dest_iov[i].data =0;
      user_diov[i].data = (void*)diov[i].get_buffer(); /* don't get owner ship */
    }

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() 
              << ":: In [Server_impl::doAWS]: begin call" << std::endl;
#endif
    ACOM::Network::Receipt rcpt;
		rcpt.init(&im_data_seq, 0, 0, _net);

    // - allocate the destination iovector
    void* cont = stub_server->get_iovect( &rcpt, header, dest_iov, dest_iovsz );
    size_t sz_copy =0;

    if ((dest_iovsz >0) && (diosz >0) && (header->size_d >0))
    {
      //[TG] bug in dma move ??? dma.set_dest(dest_iov, dest_iovsz);
      int curr_entry = 0;
      char* curr_pos = (char*)dest_iov[curr_entry].data;
      size_t curr_len = dest_iov[curr_entry].len;
      ACOM_ASSERT( curr_pos != 0, "[CORBA::doWS] bad iov entry");
      ACOM_ASSERT( curr_len > 0,  "[CORBA::doWS] bad size");

      for (int i=0; i < diosz; ++i)
      {
        size_t sz_2cpy, len = user_diov[i].len;
        const char*  pos = (const char*)user_diov[i].data;
        while (len >0) {
          if (curr_entry == diosz) {
            error_no = ACOM::ComFailure::TRUNCATED;
            goto continue_call;
          }
          sz_2cpy = (len <= curr_len ? len : curr_len);
          memcpy( curr_pos, pos, sz_2cpy );
          len -= sz_2cpy; curr_len -= sz_2cpy;
          pos += sz_2cpy; curr_pos += sz_2cpy;
          sz_copy += sz_2cpy;
          ACOM_ASSERT( curr_len >= 0, "[CORBA::doWS] internal error");
          if (curr_len ==0) 
          {
            ++curr_entry;
            curr_pos = (char*)dest_iov[curr_entry].data;
            curr_len = dest_iov[curr_entry].len;
          }
        } 
      }
    }
    else if (header->size_d >0) { 
      dest_iov = 0;
      dest_iovsz = 0;
      sz_copy = header->size_d -1; // - assume that is was not copied

#ifdef ACOM_TRACE_DEBUG
      std::cout << _node_rank 
                << ":: In [Server_impl::doIt]: null iovect" << std::endl;
      std::cout << _node_rank 
                << ":: In [Server_impl::doIt]: copy data" << std::endl;
#endif
    }

continue_call:
    // - error ?
    if (sz_copy != header->size_d)
      error_no = ACOM::ComFailure::TRUNCATED;

    stub_server->call(error_no, &rcpt, header, dest_iov, dest_iovsz, cont );
    im_data_seq.destroy();
    rcpt.finish();

#ifdef ACOM_TRACE_DEBUG
    std::cout << _net->rank() 
              << ":: In [Server_impl::doAWS]: end call " << std::endl;
#endif

  } catch (ACOM::InvalidArgument &ex) {
    std::cerr << "[doAWS] unknow service id" << std::endl;
    throw CORBA::COMM_FAILURE();
  } catch (ACOM::ComFailure& e) {
    throw CORBA::COMM_FAILURE();
  } catch (...) {
    std::cerr << "[doAWS] unknow exception" << std::endl;
    throw CORBA::COMM_FAILURE();
  }

#ifdef ACOM_TRACE_DEBUG
  std::cout << _net->rank() << "Out [Server_impl::doAWS]" << endl;
#endif
}


} // -namespace
