// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_ack.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// Implementation of ack service
// $Id: ics_ack.cpp,v 1.6 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#include <ics_ack.h>
#include <iostream>

namespace ICS {


// -----------------------------------------------------------------------------
  /** */
AckService::~AckService()
{ }


// -----------------------------------------------------------------------------
/* */
void AckService::initialize( ACOM::Network* net )
{ 
  _net = net;
  _access.init(); _cond.init(); 
}



// -----------------------------------------------------------------------------
/** */
void AckService::wait( long key, long count)
{
  /* I will receive data */
  PTH::CriticalGuard cg(_access); // - begin critical section

  /* search if receive already occurs */
  ack_com_context declared;
  ack_com_context* used =0;

  Trigger_set::iterator curr = _postrecv.find(key);
  //if (curr == _postrecv.end()) 
  if (curr == 0) 
  {
    used = &declared;
    curr = _postrecv.insert(key, used);
    //_postrecv[key] = used;
  }
  else {
    used = curr->second;
  }

  used->count = count;
  while (used->recv != used->count) 
  {
#if 0
std::cout << "In wait: key=" << key 
          << ", recv:" << used->recv << ", waiting:" << used->count << std::endl;
#endif
    used->is_wait = true;
    _cond.wait(_access); 
  }
#if 0
std::cout << "Out wait: key=" << key 
          << ", recv:" << used->recv << ", waiting:" << used->count << std::endl;
#endif
  _postrecv.erase(curr);
  //_postrecv.erase(key);
  if (used != &declared) delete used;
} 




// -----------------------------------------------------------------------------
void AckService::signal( int node, long key )
{
  /* get a channel */
#if 0
std::cout << "In signal to node=" << node << ", key=" << key 
          << std::endl;
#endif
  //ACOM::OutChannel* channel = _net->get_local_route( node );
  ACOM::Network::OutChannel* channel = _net->get_default_route( node );

  /* Form a message */
  ACOM::ActiveMessage<AckService> am (_net, this, 0);
  am->pack(&key, sizeof(key)); 
  am.post(channel);
#if 0
std::cout << "Out signal to node=" << node << ", key=" << key 
          << std::endl;
#endif
}





// -----------------------------------------------------------------------------
/* appele apres que le message soit arrive 
 */
void AckService::call ( 
     ACOM::ComFailure::Code e,      // if any 
     ACOM::Network::Receipt* r,
     ACOM::Network::Header* header,
     ACOM::IOVectEntry* iov,
     ACOM::IOVectSize   iov_sz,
     void* cc
) throw (ACOM::ComFailure)
{
#if 0
std::cout << "In call" << std::endl;
#endif
  if (e != ACOM::ComFailure::OK) {
    std::cerr << "Error in msg='" //<< e.what(); // ACOM::ComFailure::what(e) 
              << "', #iov" << iov_sz  << ", len[0]" << (iov_sz > 0 ? iov[0].len : 0)
              << std::endl;
      ACOM::throwit( ACOM::ComFailure(e) ) ;
  }

  long key;
  r->unpack(&key, sizeof(long));

  /* find tag in _postrecv */
  PTH::CriticalGuard cg(_access); // - begin critical section
#if 0
std::cout << "In signal/service key=" << key 
          << std::endl;
#endif
  ack_com_context* ctxt =0;
  Trigger_set::iterator curr = _postrecv.find(key);
  //if (curr == _postrecv.end())
  if (curr == 0)
  {
    ctxt  = new ack_com_context;
    curr = _postrecv.insert(key, ctxt);
   // _postrecv[key] = ctxt;
  } else {
    ctxt = curr->second;
  }
  ++ctxt->recv;
#if 0
std::cout << "In signal/service recv:" << ctxt->recv 
          << ", waiting:" << ctxt->count 
          << std::endl;
#endif
  if (ctxt->is_wait)
    _cond.broadcast();
}

}
