// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_alphabarrier.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// Implementation of barrier as a service
// $Id: ics_alphabarrier.cpp,v 1.7 2003/07/18 11:26:00 olobry Exp $
//
// ==========================================================================
#include <ics_alphabarrier.h>
#include <iostream>

namespace ICS {


// -----------------------------------------------------------------------------
AlphaBarrier::AlphaBarrier()
 : _alpha(0.5)
{
}

// -----------------------------------------------------------------------------
/** Default implementation: use a binomial tree, if n > 3, else use a flat tree.
 */
void AlphaBarrier::barrier( int key )
{
  barrier(key, _alpha);
}

void AlphaBarrier::barrier( int key, double alpha )
{
#if 0
std::cout << "IN: " << std::endl;
#endif
  /* I will receive data */
  PTH::CriticalGuard cg(_access); // - begin critical section

  /* search if receive already occurs */
  bar_com_context declared;
  bar_com_context* used =0;

  int rank = _net->rank();
  int nodecount = _net->node_count();
  if (nodecount <=1) return;
  
  /* compute alpha parameters */
  int myrank = rank;
  int n = nodecount;
  int parent = 0;
  int nOthers;
  int threshold;

  while(myrank != 0) {
    nOthers = (int)round(alpha*n);
    threshold = n-nOthers;
    if(myrank == threshold) {
      n = nOthers;
      myrank -= threshold;
    } else {
      if(myrank < threshold)
        n = threshold;
      else {
        n = nOthers;
        parent += threshold;
        myrank -= threshold;
      }
    }
  }

  /* count the number of recv/send to do */
  int nn = n;
  int nbsendrecv = 0;
  while(nn > 1) {
    nn -= (int)round(alpha*nn);
    ++nbsendrecv;
  }

  /* if myself== 0, then wait for all other node in network */
  Trigger_set::iterator curr = _postrecv.find(key);
  if (curr == 0) 
  {
#if 0
    std::cout << "Post to recv ack" << std::endl;
#endif
    used = &declared;
    curr = _postrecv.insert(key, used);
  }
  else {
    used = curr->second;
  }


#if 0
std::cout << rank << "::is my rank over #node= " << nodecount << ", #childs= " << nbsendrecv << std::endl;
#endif

  /*
   * Phase 1: reduction 
   */

  /* set count with the number of recv to wait */
  used->count = nbsendrecv;

  if ((nbsendrecv == 0) && (rank !=0)) 
  { /* I'm a leaf: send to parent */
#if 0
    std::cout << "Leaf node=" << rank << ", send to :" << parent << std::endl;
#endif
    ACOM::Network::OutChannel* channel = _net->get_default_route( parent );
    ACOM::ActiveMessage<BarrierService> am (_net, this, 0);
    am->pack(&key, sizeof(key)); 
    am.post(channel);
  }

  /* else wait for reduction or of broadcast */
  if (nbsendrecv !=0 ) {
#if 0
    std::cout << "Node=" << rank << ", wait reduction :" << nbsendrecv << std::endl;
#endif
    while (used->recv != used->count) 
    {
      used->is_wait = true;
      _cond.wait(_access); 
    }
#if 0
    std::cout << "Node=" << rank << ", end wait reduction :" << nbsendrecv << std::endl;
#endif
  }

  if ((nbsendrecv >0) && (rank !=0)) 
  { /* I'm an intermediate node, re-send to parent */
#if 0
    std::cout << "Node=" << rank << ", re-send to :" << parent << std::endl;
#endif
    ACOM::Network::OutChannel* channel = _net->get_default_route( parent );
    ACOM::ActiveMessage<BarrierService> am (_net, this, 0);
    am->pack(&key, sizeof(key)); 
    am.post(channel);

  }

  if (rank !=0) {
    /* Wait respond from my parent */
    used->count =1;

#if 0
    std::cout << "Node=" << rank << ", wait broadcast :" << parent << std::endl;
#endif
    while (used->recv != used->count) 
    {
      used->is_wait = true;
      _cond.wait(_access); 
    }
#if 0
    std::cout << "Node=" << rank << ", end wait broadcast :" << parent << std::endl;
#endif
  }

  
  /*
   * Phase 2: broadcast 
   */
  if (nbsendrecv >0) {
    nn = n;
    while(n > 1) {
      n -= (int)round(alpha*n);
      /* send message to (myrank + n) % nodecount */
      int dest = (myrank + n) % nodecount;
#if 0
      std::cout << "Node=" << rank << ", send end to:" << dest << std::endl;
#endif
      ACOM::Network::OutChannel* channel = _net->get_default_route( dest );
      ACOM::ActiveMessage<BarrierService> am (_net, this, 0);
      am->pack(&key, sizeof(key)); 
      /* Send it */
      am.post(channel);
    }
  }
  
  _postrecv.erase(curr);
  if (used != &declared) delete used;
} 



} // ICS namespace


