// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_base.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithm
// $Id: ics_base.cpp,v 1.2 2003/06/05 16:43:23 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_base.h>
#include <iostream>

namespace ICS {
// -----------------------------------------------------------------------------
BaseService::com_context* BaseService::_freelist =0;     // - free list of cached com_context

// -----------------------------------------------------------------------------
PTH::Mutex   BaseService::_freelock;     // - lock free list

// -----------------------------------------------------------------------------
size_t BaseService::_ccalloc =0;     // number of internal allocated cc

// -----------------------------------------------------------------------------
size_t BaseService::_ccfree =0;      // number of internal freed cc

// -----------------------------------------------------------------------------
PTH::AtomicCounter  BaseService::_cached;       // - number of items in _freelist


// -----------------------------------------------------------------------------
/** */
BaseService::BaseService()
  : _net(0), _postrecv(256, &_ccalloc, &_ccfree)
{
}


// -----------------------------------------------------------------------------
BaseService::~BaseService()
{
}


// -----------------------------------------------------------------------------
/* */
void BaseService::initialize( ACOM::Network* net )
{ 
  _net         = net;
  _nbcalled    = 0;
  _nbmatchpost = 0;
  _nbmatchrecv = 0;
  _sizealloc   = 0;
  _sizefree    = 0;
  _access.init(); 
  _cond.init(); 
}


// -----------------------------------------------------------------------------
/* */
BaseService::com_context* BaseService::allocate_cc(  )
{
  ++_ccalloc;
  com_context* cc = 0;
  if (_freelist !=0) {
    PTH::CriticalGuard critical(BaseService::_freelock);
    cc = _freelist;
    _freelist = (com_context*)cc->data_recv;
  } else {
    cc = new com_context;
  }

  cc->is_wait   = false;
  cc->is_join   = true;  /* default */
  cc->wait_recv = 0;
  cc->wait_post = 0;
  cc->fmt       = 0;
  cc->op        = 0;
  cc->count_post= 0;
  cc->data_post = 0;
  cc->count_recv= 0;
  cc->data_recv = 0;
  --_cached;
  return cc;
}


// -----------------------------------------------------------------------------
/* */
void BaseService::deallocate_cc( com_context* cc)
{
  ++_ccfree;
  if (_cached > 128) {
    delete cc;
    return;
  }
  PTH::CriticalGuard critical(BaseService::_freelock);
  cc->data_recv = (void*)_freelist;
  _freelist = cc;
  ++_cached;
}


// -----------------------------------------------------------------------------
std::ostream& BaseService::print( std::ostream& o ) const
{
  double s = _nbmatchpost+_nbmatchrecv;
  o << "Service statistics:" << std::endl;
  o << "  #call=" << _nbcalled << std::endl;
  o << "  #call/match post=" << _nbmatchpost << ", %=" << double(_nbmatchpost)/s << std::endl;
  o << "  #call/match recv=" << _nbmatchrecv << ", %=" << double(_nbmatchrecv)/s << std::endl;
  o << "  #cc alloc=" << _ccalloc << std::endl;
  o << "  #cc free =" << _ccfree << std::endl;
  o << "  #size alloc=" << _sizealloc << std::endl;
  o << "  #size free =" << _sizefree << std::endl;
  return o;
}

}
