// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_base.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Base service for ICS library
// $Id: ics_base.h,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// ==========================================================================
#ifndef _INUK_BASE_H_
#define _INUK_BASE_H_
#include <acom>
#include <pth>
#include <ics_lighthashmap.h>
#include <ics_redops.h>


namespace ICS {

/* 
 * Bcast base service for inuktitut
 * implementation provides algorithm.
 */
class BaseService : public ACOM::Service {
public:
  /** */
  virtual void initialize( ACOM::Network* net );

  /** */
  virtual std::ostream& print( std::ostream& o ) const;  

protected:
  /** */
  BaseService();

  /** */
  ~BaseService();

  /** Context for one communication
      The role of this object is to implement the rendez-vous between
      post recv (always a post occurs) and received data.
      As long as the post is not executed, data_post points to a 
      received temporary data. When the post occurs, wait_post >0
      and data_post is a pointer to the pointer of the post (a void**),
      and count_post is a pointer to the count of the post.
      These former pointers permit to reallocated data if the post was
      done with no data (allocation by ICS to the user).
      Please, read notes about bcast algorithm in ics_bcast.cpp and ics_reduce.cpp
      for reduce algorithm.
   */
  struct com_context {
    com_context( )
     : serial(0), is_wait(false), is_join(true), 
       wait_post(0), wait_recv(0), 
       fmt(0), op(0), 
       count_post(0), data_post(0), count_recv(0), data_recv(0), 
       cb_arg1(0), cb_arg2(0), cb_func(0)
    {}
    void print() 
    { std::cout << "wait post=" << wait_post << ", wait recv=" << wait_recv 
                << ", serial=" << serial << std::endl; 
    }
    int                        serial;
    bool                       is_wait;
    bool                       is_join;    /* true iff join on the context is requiered */
    size_t                     wait_post;
    size_t                     wait_recv;
    const ACOM::Format*        fmt; 
    const CumulOperatorFormat* op;         /* used only in reduce */
    size_t                     count_post;
    void*                      data_post; 
    size_t                     count_recv;
    void*                      data_recv;  /* also link in freelist */
    void*                      cb_arg1;
    void*                      cb_arg2;
    void                       (*cb_func)(void*, void*);
  };


protected:
  /** The network on which communications occur
   */
  ACOM::Network* _net;

  /* Mutex  to lock access to internal data structure
   */
  PTH::Mutex     _access;
  PTH::Condition _cond;

  /** hash_map[key] -> context de communication
   */
  struct Key {
    Key( ) 
     : tag(-1), serial(0) {}
    Key( long t ) 
     : tag(t), serial(0) {}
    Key( long t, int s ) 
     : tag(t), serial(s) {}
    bool operator==( const Key& k ) 
    { return (serial == k.serial) && (tag == k.tag); }
    int operator%( int sz ) 
    { return (serial + tag) % sz; }
    long tag;
    int  serial;
  };
  typedef LHashMap<Key,com_context*> Trigger_set;
  Trigger_set         _postrecv;     // - queue of message post or received
  static PTH::AtomicCounter  _cached;       // - number of items in _freelist
  static com_context* _freelist;     // - free list of cached com_context
  static PTH::Mutex   _freelock;     // - lock free list
  friend class Ics;
  
  //@{
  /** Allocate a new com_context
   */
  static com_context* allocate_cc( );

  /** Deallocate a com_context
   */
  static void deallocate_cc( com_context* ); 
  //@}


protected:
  /* Some statistic about the service */
  size_t _nbcalled;    // number it was called
  size_t _nbmatchpost; // number of calls where rendez-vous occurs in post
  size_t _nbmatchrecv; // number of calls where rendez-vous occurs in recv
  size_t _sizealloc;   // number of internal allocated data structure
  size_t _sizefree;    // number of internal freed data structure
  static size_t _ccalloc;     // number of internal allocated cc
  static size_t _ccfree;      // number of internal freed cc
};



/*
 * inline definitions
 */
 
}

#endif
