// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithm
// $Id: ics_bcast.cpp,v 1.4 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast.h>
#include <iostream>

namespace ICS {
/* Implementation note about the management of rendez-vous.
   For Bcast (One2All and Alpha tree), one node send data to
   other node : each receiver will receive a message only one time.
   - The maching occurs iff post are executed and message has been recveived.
   - Data where to received is stored into data_post and count_post field
     that are pointers to pointer (data_post) and to size (count_post).
   - Data received are stored in (data_recv, count_recv).
   - If data_recv differs from data_post (i.e. data_recv != *(void**)data_post)
     then that means that data has been received before the post occurs.
   - For the former case, notify_cc will recopy data received in the right
     location into *data_post. 
     
   - Two modes exist for using this API: either in asynchronous fashion
     (a call back will be called when the rendez-vous occurs), either in 
     non-blocking call (post_bcast/wait_bcast). Due to the possibility of
     waiting for the rendez-vous, internal data structure is not deleted
     until the wait is called.
*/

// -----------------------------------------------------------------------------
/* */
void BcastService::notify_cc( Trigger_set::iterator curr  )
{
  com_context* ctxt = curr->second;
  if (ctxt->wait_post != ctxt->wait_recv)
    return;

  void**  data  = (void**) ctxt->data_post;
  size_t* count = (size_t*)ctxt->count_post;

  if (*data ==0) {
    /* (data,count) get the temporary, no reduction ! */
    *data  = ctxt->data_recv;
    *count = ctxt->count_recv;
  } 
  else 
  {
    if (*count != ctxt->count_recv) {
      std::cout << "Bad count " << std::endl;
      return;
      //throw ACOM::Length();
    }

    /* recopy received data into requested data and destroy received data */
    if (*data != ctxt->data_recv) {
//std::cout << "Recopy data recv to posted" << std::endl;
      ctxt->fmt->copy( *data, ctxt->data_recv, ctxt->count_recv );
      ctxt->fmt->destroy( ctxt->data_recv, ctxt->count_recv );
      _sizefree += ctxt->fmt->get_size() * ctxt->count_recv;
    }

    /*  */
    ctxt->data_recv  = 0;
    ctxt->count_recv = 0;
  }

  /* call the call back function */
  if (ctxt->cb_func !=0) 
    (*ctxt->cb_func)(ctxt->cb_arg1, ctxt->cb_arg2);

  /* erase data if needed */
  if (!ctxt->is_join) 
  {
    deallocate_cc( ctxt );
  }
}



// -----------------------------------------------------------------------------
/** */
void BcastService::postsend_data
  ( ACOM::Network::Request* req, int tag, 
    const ACOM::Format* fmt, void* data, size_t count )
{
  int fid = fmt->get_id();
  req->pack(&tag, sizeof(int));
  req->pack(&fid, sizeof(int));
  req->pack(&count, sizeof(size_t));

  if (count >0) {
    ACOM::OStream msg ( req->get_datastream(), req->get_iovstream() );
    fmt->write(msg, data, count);
  }

#if 0
  std::cout << "Send serial = " << serial << std::endl;
  std::cout << "Send tag    = " << tag << std::endl;
  std::cout << "Send fid    = " << fid << std::endl;
  std::cout << "Send count  = " << count << std::endl;
  size_t size_h  = req->get_datastream()->getFillSize();
  size_t iovsz_h = req->get_datastream()->getBlockSize();
  size_t size_d  = req->get_iovstream()->getByteSize();
  size_t iovsz_d = req->get_iovstream()->getBlockSize();
  std::cout << "Send size_i      = " << size_h << std::endl;
  std::cout << "Send size_d      = " << size_d << std::endl;
  std::cout << "Send iovsz_i     = " << iovsz_h << std::endl;
  std::cout << "Send iovsz_d     = " << iovsz_d << std::endl;
#endif
}


// -----------------------------------------------------------------------------
/* Post recv of a iov splited by bloc of size at most szbloc bytes
*/
void BcastService::postrecv_data
  ( int tag, bool isjoinable, 
    const ACOM::Format* fmt, void** data, size_t* count,
    void (*cbf)(void*, void*), void* cba1, void* cba2
  )
{
  if ((count ==0) || (data ==0))
    ACOM::throwit ( ACOM::InvalidArgument() );
  
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

#if 0
  std::cout << "----------\nbcast post tag  =" << tag << std::endl;
  std::cout << "bcast post fid  =" << fmt->get_id() << std::endl;
  std::cout << "bcast post count=" << *count << std::endl;
#endif

  /* search if receive already occurs */
  Key key (tag);
  curr = _postrecv.find( key );
  if (curr == 0) 
  {
#if 0
  std::cout << "bcast post newtctxt  =" << tag << std::endl;
#endif
    ctxt             = allocate_cc();
    curr             = _postrecv.insert( key, ctxt);
    ctxt->fmt        = fmt;
  }
  else 
  {
    ctxt             = curr->second;
#if 0
  std::cout << "bcast post ctxt already exist  =" << tag 
            << ", wait_recv=" << ctxt->wait_recv 
            << ", data_post=" << ctxt->data_post 
            << ", data_recv=" << ctxt->data_recv 
            << std::endl;
#endif
    if (ctxt->fmt != fmt) {
      std::cout << "Formats differs between required and received: f req=" << fmt->get_id()
                << ", f recv=" << ctxt->fmt->get_id()
                << std::endl;
//      throw ACOM::InvalidArgument("Formats differs between required and received");
      return;
    }
    ++_nbmatchpost;
  }
  ctxt->is_join    = isjoinable;
  ctxt->count_post = (size_t)count;
  ctxt->data_post  = (void*)data;
  ctxt->cb_arg1    = cba1;
  ctxt->cb_arg2    = cba2;
  ctxt->cb_func    = cbf;
  ctxt->wait_post  = 1;
  if ((ctxt->wait_recv == ctxt->wait_post) && (!ctxt->is_join)) {
    notify_cc( curr );
    //std::cout << "Erase: " << ctxt << std::endl;  
    _postrecv.erase( curr );
  }
#if 0
  std::cout << "bcast end post =" << tag 
            << ", wait_post=" << ctxt->wait_post 
            << ", wait_recv=" << ctxt->wait_recv 
            << ", count_post=" << ctxt->count_post 
            << ", data_post=" << ctxt->data_post 
            << ", *count_post=" << *(size_t*)ctxt->count_post 
            << ", *data_post=" << *(void**)ctxt->data_post 
            << ", count_recv=" << ctxt->count_recv 
            << ", data_recv=" << ctxt->data_recv 
            << "\n-----" << std::endl;
#endif
}




// -----------------------------------------------------------------------------
BcastService::com_context* BcastService::wait_data( int tag )
{
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

  /* find tag in _postrecv */
  Key key (tag);
  curr = _postrecv.find( key );
  if (curr == 0) 
  {
    std::cerr << "Cannot find ctxt with key(" << tag << ")" << std::endl;
    ACOM::throwit ( ACOM::InvalidArgument() );
  }

  /* infinite loop */
  ctxt = curr->second;
  ctxt->is_wait = true;
  while (ctxt->wait_recv != ctxt->wait_post) {
    _cond.wait(_access);
  }

  /* notify the com_context */
  if (ctxt->is_join) {
    notify_cc( curr );
  }


  /* supress entry in hash table */
  _postrecv.erase( curr );

  return ctxt;
}


// -----------------------------------------------------------------------------
void BcastService::call ( 
     ACOM::ComFailure::Code e,      
     ACOM::Network::Receipt* r,
     ACOM::Network::Header* header,
     ACOM::IOVectEntry* iov,
     ACOM::IOVectSize   iov_sz,
     void* cc
) throw (ACOM::ComFailure)
{
  if (e != ACOM::ComFailure::OK) {
    std::cerr << "Error in msg='" //<< ACOM::ComFailure::what(e) 
              << "', #iov" << iov_sz  << ", len[0]" << (iov_sz > 0 ? iov[0].len : 0)
              << std::endl;
    ACOM::throwit (ACOM::ComFailure(e)) ;
  }

  Trigger_set::iterator curr = (Trigger_set::iterator)cc;
  com_context* ctxt = curr->second;
//std::cout << "in call: ctxt=" << ctxt << std::endl;

  { PTH::CriticalGuard cg(_access);
    ++ctxt->wait_recv;
    if (ctxt->wait_post == ctxt->wait_recv) 
    {
//std::cout << "::call signal or notify msg ctxt=" << ctxt << std::endl;
      if (!ctxt->is_join) {
        notify_cc( curr );
        _postrecv.erase( curr );
       //std::cout << "Erase: " << ctxt << std::endl;  
      }
      if (ctxt->is_wait)
        _cond.signal();
    }
#if 0
  std::cout << "bcast recv call =" << curr->first.tag 
            << "\n-----" << std::endl;
#endif
  }
  PTH::Thread::yield();
}


// -----------------------------------------------------------------------------
/* decodage de l'address pour stocker le message
 */
void* BcastService::get_iovect(
  ACOM::Network::Receipt* r,
  ACOM::Network::Header* header, 
  ACOM::IOVectEntry* iov, 
  ACOM::IOVectSize iovsz
)
{
  /* extract immediate data */
  int tag;
  int fid;
  size_t count;
  r->unpack(&tag, sizeof(int));
  r->unpack(&fid,sizeof(int));
  r->unpack(&count,sizeof(size_t));

  Key key(tag);

  /* some data */
  com_context* ctxt = 0;
  const ACOM::Format* fmt =0;
  void* data =0;
 
  PTH::CriticalGuard cg(_access);
  //_access.lock();

#if 0
  std::cout << "----------\nbcast recv tag  =" << tag << std::endl;
  std::cout << "bcast recv fid  =" << fid << std::endl;
  std::cout << "bcast recv count=" << count << std::endl;
#endif

  /* find tag in _postrecv */
  Trigger_set::iterator curr = _postrecv.find( key );

  /* common set up */
  fmt = ACOM::Format::get_format( fid );

  if (curr == 0) 
  {
    /* Here the message arrived before the post to was called
     */
    ctxt      = (com_context*) allocate_cc();
    /* allocate temporary for receiving message */
    if (count >0) {
      data = ctxt->data_recv = fmt->create(count);
      _sizealloc += fmt->get_size() * count;
    }
    curr = _postrecv.insert(key, ctxt);
#if 0
  std::cout << "::getiovect: recv newtctxt=" << ctxt << ", tag=" << tag << std::endl;
#endif
  } 
  else 
  {
    /* Here the message arrived after the post (the only way a key was
       in the hash table for broadcast, not true for reduce where multiple
       receives occur).
       HERE: wait_post >0 (invariant for Bcast)
    */
    ++_nbmatchrecv;
    ctxt = curr->second;

#if 0
  std::cout << "::getiovect: recv ctxt already exist  ctxt=" << ctxt << ", tag=" << tag 
            << ", wait_post=" << ctxt->wait_post 
            << ", isjoin=" << ctxt->is_join << std::endl;
#endif

    /* if buffer has no enough space, 
       TG: WARN reallocated it ? 
    */
    if (*(void**)ctxt->data_post ==0) 
    {
      /* note that data will be set to requested data 
         in the notify function 
      */
      data = ctxt->data_recv = fmt->create(count);
      _sizealloc += fmt->get_size() * count;
    } 
    else 
    {
      if (*(size_t*)ctxt->count_post != count) {
        std::cout << "Bad count recv=" << count << ", requested=" << *(size_t*)ctxt->count_post << std::endl;
        return 0;
        ACOM::throwit( ACOM::Length() );
      }
      data  = ctxt->data_recv  = *(void**)ctxt->data_post;
    }
  }
  
  /* common setup */
  ctxt->count_recv = count;
  ctxt->fmt = fmt;

  /* No need to lock until call method was called: because only one message
     arrived for one post (with differents tag)....
     TG: WARN management of FIFO communication ???
  */
  //_access.unlock();

  if (count >0) 
  {
    /* Unpack the data and get iovect description */
    ACOM::IStream fin ( r->get_datastream(), iov, iovsz );
    fmt->read(fin, data, count);
  }

  //PTH::Thread::yield();
  return curr; 
}



// -----------------------------------------------------------------------------
/** Shortcut for post_bcast + wait_bcast */
void BcastService::bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
  handler wc = post_bcast( 
    buffer, 
    count,
    fmt,
    sz_ranks, 
    ranks, 
    index_self, 
    index_src, 
    tag
  );
  wait_bcast( wc );
}



}
