// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_bcast.h,v 1.2 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_BCAST_H_
#define _INUK_BCAST_H_
#include <acom>
#include <ics_base.h>


namespace ICS {

/* 
 * Bcast base service for inuktitut
 * implementation provides algorithm.
 */
class BcastService: public BaseService {
public:
  /** */
  typedef void* handler;

  /** Bcast interface for algorithm.
   This method broadcast a data (buffer,count,fmt) to all sites in the array 'ranks'.
   All sites that participate in the broadcast should make a call to this method.
   On all sites except the site of the initiator, buffer will contains the data.
   If buffer is null and data non null, then buffer is reallocated to store the data.
   This function is non-blocking, the method wait_bcast SHOULD be called to ensure
   completness of the local receive. This does not imply that all sites have received
   the data.
   @param buffer IN/OUT
   @param count IN/OUTnumber of data of format fmt to broadcast
   @param fmt the IN number of element to send
   @param sz_ranks IN number of sites that participate to the collective communication
   @param ranks IN the arrays of sites 
   @param index_self IN ranks[index_self] should be the site of the caller 
   @param index_src IN  ranks[index_src] is the site of the initiator of the communication 
   @param tag IN ranks[index_src] is the site of the initiator of the communication 
  */
  virtual handler post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  ) = 0;

  /** */
  virtual void wait_bcast(
    handler wc 
  ) = 0;

  /** 
   This method broadcast a data (buffer,count,fmt) to all sites in the array 'ranks'.
   All sites that participate in the broadcast should make a call to this method.
   On all sites except the site of the initiator, buffer will contains the data.
   If buffer is null and data non null, then buffer is reallocated to store the data.
   This function is non-blocking and asynchronous: the function cb_func is called 
   when the data is completely received. 
   @param buffer IN/OUT
   @param count IN/OUTnumber of data of format fmt to broadcast
   @param fmt the IN number of element to send
   @param sz_ranks IN number of sites that participate to the collective communication
   @param ranks IN the arrays of sites 
   @param index_self IN ranks[index_self] should be the site of the caller 
   @param index_src IN  ranks[index_src] is the site of the initiator of the communication 
   @param cb_func IN a function that will be called on receive of the data
   @param cb_arg1 IN a parameter for cb_func
   @param cb_arg2 IN a parameter for cb_func
  */
  virtual void async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  ) = 0;

  /** Shortcut for post_bcast + wait_bcast 
  */
  void bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  );

protected:
  /** notify a com_context
      If the request was posted as a non joinable com_context,
      then entry in _postrecv and context will be destroy in notify.
      Else entry and context will be keep until call to wait.
   */
  void notify_cc( Trigger_set::iterator curr );

  /* Fill in a request
  */
  void postsend_data ( ACOM::Network::Request* req, int tag, 
                      const ACOM::Format* fmt, void* data, size_t count );

  /** Create a new com_context
  */
  void postrecv_data( int tag, bool isjoinable, 
                      const ACOM::Format* fmt, void** data, size_t* count,
                      void (*cb_func)(void*, void*), void* cb_arg1, void* cb_arg2
                    );

  /* wait for a packet for a particular (tag,serial)
  */
  com_context* wait_data( int tag );

  /* appele apres que le message soit arrive 
   */
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,      
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void* cc
  ) throw (ACOM::ComFailure);

  /* decodage de l'address pour stocker le message
   */
  void* get_iovect(
    ACOM::Network::Receipt* r,
    ACOM::Network::Header* header, 
    ACOM::IOVectEntry* vector, ACOM::IOVectSize count
  );

};



/*
 * inline definitions
 */
 
}

#endif
