// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_alphachain.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of flat bcast
// $Id: ics_bcast_alphachain.cpp,v 1.3 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast_alphachain.h>
#include <iostream>

#define ICS_DEBUG_INST(x) 

namespace ICS {

// -----------------------------------------------------------------------------
BcastAlphaChain::BcastAlphaChain()
  : _salpha(0), _schain(0), _length(2)
{

}


// -----------------------------------------------------------------------------
struct SAlphaChainContinuation {
  BcastAlphaChain*    serv;
  void**              buffer;
  size_t*             count;
  const ACOM::Format* fmt;
  int                 sz_roots; 
  int*                roots; 
  int                 sz_subranks; 
  int                 index_beg; 
  int                 sz_ranks; 
  int*                ranks; 
  int                 index_self;
  int                 tag;
  void                (*cb_func)(void*, void*);
  void*               cb_arg1;
  void*               cb_arg2;

  BcastAlphaTree::handler ctxt;
};


// -----------------------------------------------------------------------------
/** */
BcastAlphaChain::handler BcastAlphaChain::post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
ICS_DEBUG_INST(std::cout << index_self << "::IN " << __PRETTY_FUNCTION__ << std::endl;)
  ++_nbcalled;
  if (sz_ranks <=1) return 0;

  /* node where to send */
  int indextosend = (index_self+1)%sz_ranks;

  /* no broadcast : one node ! */
  if (indextosend == index_self)
    return 0;

  /* swap index_src at position 0 */
  if (index_src !=0) {
    int tmp = ranks[index_src]; 
    ranks[index_src] = ranks[0]; 
    ranks[0] = tmp; 
    if (index_self == index_src)
      index_self =0;
    else if (index_self ==0)
      index_self = index_src;
    index_src = 0;
  }

ICS_DEBUG_INST(std::cout << index_self << "::Sz chain =" << _length << std::endl;)

  SAlphaChainContinuation* cc = new SAlphaChainContinuation;
  cc->serv        = this;
  cc->buffer      = &buffer;
  cc->count       = &count;
  cc->fmt         = fmt;
  cc->sz_ranks    = sz_ranks;
  cc->ranks       = ranks;
  cc->index_self  = index_self;
  cc->sz_roots    = sz_ranks / _length;
  if (sz_ranks - cc->sz_roots * _length >0)
    ++cc->sz_roots;
  int index_group = index_self / _length; 
ICS_DEBUG_INST(std::cout << index_self << "::Index in group of " <<  index_self << " is " << index_group << std::endl;)
  cc->index_beg   = index_group * _length;
  if (index_group < cc->sz_roots -1) cc->sz_subranks = _length;
  else {
    cc->sz_subranks = sz_ranks - (cc->sz_roots-1) * _length;
  }
ICS_DEBUG_INST(std::cout << index_self << "::Sz of the group is " <<  cc->sz_subranks << std::endl;)

  cc->roots = new int[cc->sz_roots];
  for (int i=0; i<cc->sz_roots; ++i)
    cc->roots[i] = ranks[i*_length];

ICS_DEBUG_INST(std::cout << index_self << "::Roots are;";
  for (int i=0; i<cc->sz_roots; ++i)
    std::cout << ranks[cc->roots[i]] << " ";
std::cout << std::endl;)

  cc->tag       = tag;
  cc->cb_func   = 0;
  cc->cb_arg1   = 0;
  cc->cb_arg2   = 0;

  if ((index_self % _length ==0) && (cc->sz_roots >1) ) 
  { 
    /* I'am into the header set and I do first alpha algorithm 
     */
ICS_DEBUG_INST(std::cout << index_self << "::Index=" << index_self << " is a root ! of group =" << index_group << std::endl;)
    cc->ctxt = _salpha->post_bcast( buffer, count, fmt, cc->sz_roots, cc->roots, index_group, 0, tag );
    return cc;
  } 
  cc->ctxt = (handler)-1;
  return cc;
}



// -----------------------------------------------------------------------------
/** */
void BcastAlphaChain::wait_bcast(
  handler wc
)
{
  if (wc ==0) return;
  SAlphaChainContinuation* cc = (SAlphaChainContinuation*)wc;
ICS_DEBUG_INST(std::cout << cc->index_self << "::IN " << __PRETTY_FUNCTION__ << std::endl;)
  if ((cc->index_self % _length ==0) && (cc->sz_roots >1) )  {
    /* I'am a root */
    _salpha->wait_bcast( cc->ctxt );
    delete [] cc->roots;
  }

  /* resend to rank[roots] rank[roots+sz_firsts] */
  if (cc->sz_subranks >1) {
ICS_DEBUG_INST(
    std::cout << cc->index_self << "::Do sub bcast on ranks=" ;
    for (int i=0; i < cc->sz_subranks; ++i)
      std::cout << cc->ranks[cc->index_beg+i] << ' ';
    std::cout << std::endl;
    std::cout << cc->index_self << "::self index=" << cc->index_self-cc->index_beg << std::endl;
    std::cout << cc->index_self << "::src index is 0=" << std::endl;
)
    /* resend to rank[roots] rank[roots+sz_firsts] */
    handler ctxt2 = _schain->post_bcast( *cc->buffer, *cc->count, cc->fmt,
                                         cc->sz_subranks, &(cc->ranks[cc->index_beg]), 
                                         cc->index_self-cc->index_beg, 0, cc->tag
                                       );
    _schain->wait_bcast(ctxt2);
  };
  delete cc;
}



// -----------------------------------------------------------------------------
void BcastAlphaChain::continuation(void* a1, void* a2)
{
  SAlphaChainContinuation* cc = (SAlphaChainContinuation*)a1;
  if (cc->index_self % cc->serv->_length ==0)  {
    /* I'am a root and I am receive the message */
    delete [] cc->roots;
  }

  /* resend to rank[roots] rank[roots+sz_firsts] */
  if (cc->sz_subranks >1) {
ICS_DEBUG_INST(
    std::cout << cc->index_self << "::Do sub bcast on ranks=" ;
    for (int i=0; i < cc->sz_subranks; ++i)
      std::cout << cc->ranks[cc->index_beg+i] << ' ';
    std::cout << std::endl;
    std::cout << cc->index_self << "::self index=" << cc->index_self-cc->index_beg << std::endl;
    std::cout << cc->index_self << "::src index is 0=" << std::endl;
)
    /* resend to rank[roots] rank[roots+sz_firsts] */
    cc->serv->_schain->async_bcast( *cc->buffer, *cc->count, cc->fmt,
                                    cc->sz_subranks, &(cc->ranks[cc->index_beg]),
                                    cc->index_self-cc->index_beg, 0, cc->tag,
                                    cc->cb_func, cc->cb_arg1, cc->cb_arg2
                                  );
  }
  delete cc;
}


// -----------------------------------------------------------------------------
/** */
void BcastAlphaChain::async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return;

  /* node where to send */
  int indextosend = (index_self+1)%sz_ranks;

  /* no broadcast : one node ! */
  if (indextosend == index_self)
    return;

  /* swap index_src at position 0 */
  if (index_src !=0) {
    int tmp = ranks[index_src];
    ranks[index_src] = ranks[0];
    ranks[0] = tmp;
    if (index_self == index_src)
      index_self =0;
    else if (index_self ==0)
      index_self = index_src;
    index_src = 0;
  }

  SAlphaChainContinuation* cc = new SAlphaChainContinuation;
  cc->serv        = this;
  cc->buffer      = &buffer;
  cc->count       = &count;
  cc->fmt         = fmt;
  cc->sz_ranks    = sz_ranks;
  cc->ranks       = ranks;
  cc->index_self  = index_self;
  cc->sz_roots    = sz_ranks / _length;
  if (sz_ranks - cc->sz_roots * _length >0)
    ++cc->sz_roots;
  int index_group = index_self / _length;
  cc->index_beg   = index_group * _length;
  if (index_group < cc->sz_roots -1) cc->sz_subranks = _length;
  else {
    cc->sz_subranks = sz_ranks - (cc->sz_roots-1) * _length;
  }

  cc->roots = new int[cc->sz_roots];
  for (int i=0; i<cc->sz_roots; ++i)
    cc->roots[i] = ranks[i*_length];

  cc->tag       = tag;
  cc->cb_func   = cb_func;
  cc->cb_arg1   = cb_arg1;
  cc->cb_arg2   = cb_arg1;
  cc->ctxt = (handler)-1;

  if (index_self % _length ==0 )
  {
    /* I'am into the header set and I do first alpha algorithm 
     */
    _salpha->async_bcast( buffer, count, fmt, cc->sz_roots, cc->roots, 
                          index_group, 0, tag, BcastAlphaChain::continuation, cc, 0 );
    return;
  }

  BcastAlphaChain::continuation(cc,0);
}


} // - namespace
