// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_chain.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of flat bcast
// $Id: ics_bcast_chain.cpp,v 1.1 2003/05/27 20:43:27 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast_chain.h>
#include <iostream>

namespace ICS {

// -----------------------------------------------------------------------------
BcastChain::BcastChain()
{
}


// -----------------------------------------------------------------------------
struct SChainContinuation {
  BcastChain*         serv;
  void**              buffer;
  size_t*             count;
  const ACOM::Format* fmt;
  int                 ranknext;
  int                 rankself;
  int                 ranksrc;
  int                 tag;
  void                (*cb_func)(void*, void*);
  void*               cb_arg1;
  void*               cb_arg2;
};


/** */
void BcastChain::continuation( void* a1, void* a2 )
{
  SChainContinuation* cc = (SChainContinuation*)a1;

//std::cout << "Src data is  : " << cc->ranksrc << std::endl;
//std::cout << "Self rank is : " << cc->rankself << std::endl;
//std::cout << "Send data to : " << cc->ranknext << std::endl;
  if (cc->ranknext != cc->ranksrc) 
  {
    /* Form a message */
    ACOM::WriteAndSignal<BcastChain> ws (cc->serv->_net, cc->serv, 0, 1);
    cc->serv->postsend_data( ws.get_request(), cc->tag, cc->fmt, *cc->buffer, *cc->count );

    /* I initiate the send */
    /* get a channel */
    ACOM::Network::OutChannel* channel = cc->serv->_net->get_default_route( cc->ranknext );

    /* Send it */
    ws.post(channel);
  }

  if (cc->cb_func !=0)
    (*cc->cb_func)( cc->cb_arg1, cc->cb_arg2 );

  delete cc;
}



// -----------------------------------------------------------------------------
/** */
BcastChain::handler BcastChain::post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return 0;

  /* node where to send */
  int indextosend = (index_self+1)%sz_ranks;

  /* no broadcast : one node ! */
  if (indextosend == index_self)
    return 0;

  SChainContinuation* cc = new SChainContinuation;
  cc->serv     = this;
  cc->buffer   = &buffer;
  cc->count    = &count;
  cc->fmt      = fmt;
  cc->ranknext = ranks[indextosend];
  cc->rankself = ranks[index_self];
  cc->ranksrc  = ranks[index_src];
  cc->tag      = tag;
  cc->cb_func  = 0;
  cc->cb_arg1  = 0;
  cc->cb_arg2  = 0;

  if (index_self != index_src) 
  { 
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section
    postrecv_data( tag, true, fmt, &buffer, &count, &BcastChain::continuation, cc, 0);
    return cc;
  } 

  BcastChain::continuation(cc,0);
  return 0;
}



// -----------------------------------------------------------------------------
/** */
void BcastChain::wait_bcast(
  handler wc
)
{
  if (wc ==0) return;
  SChainContinuation* cc = (SChainContinuation*)wc;
  PTH::CriticalGuard cg(_access); // - begin critical section
  int tag = cc->tag;
  com_context* ctxt = wait_data( tag);
  deallocate_cc( ctxt );
}



// -----------------------------------------------------------------------------
/** */
void BcastChain::async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return;

  /* node where to send */
  int indextosend = (index_self+1)%sz_ranks;

  /* no broadcast : one node ! */
  if (indextosend == index_self)
    return;

  SChainContinuation* cc = new SChainContinuation;
  cc->buffer   = &buffer;
  cc->count    = &count;
  cc->fmt      = fmt;
  cc->ranknext = ranks[indextosend];
  cc->rankself = ranks[index_self];
  cc->ranksrc  = ranks[index_src];
  cc->tag      = tag;
  cc->cb_func  = cb_func;
  cc->cb_arg1  = cb_arg1;
  cc->cb_arg2  = cb_arg2;

  if (index_self != index_src)
  {
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section
    postrecv_data( tag, false, fmt, &buffer, &count, &BcastChain::continuation, cc, 0);
    return;
  }

  BcastChain::continuation(cc,0);
}


} // - namespace
