// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_network.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_bcast_network.h,v 1.5 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_BCAST_NETWORK_H_
#define _INUK_BCAST_NETWORK_H_
#include <ics_network.h>


namespace ICS {


// -----------------------------------------------------------------------------
/** BcastNetwork with bcast (a la MPI) capabilities.
 * Such network relies on an existing network to handle point 2 point communication.
 * Allocation of request and channel is specific to such network in order
 * to add extra information as immediate data. 
 In order to implement and algorithm for the broadcast collective operation, the
 derived class should defined a OutChannel that inherit from OutChannel. 
 This derived network class must implement the 'process' method.
 The new derived network class should also implement :
  - virtual OutChannel* open_channel( int tag, int count_peer, LocalNodeNo* peers, 
                                      int index_self )
       throw(ACOM::ComFailure);
  In order to allocate the right OutChannel.
 */
class BcastNetwork: public Network {
public:
  // --------------------------------------------------------------------------
  /** Constructor
  */
  BcastNetwork();

  /** Destructor
  */
  ~BcastNetwork();

  // ---------------------------------------------------------------------------
  /** class for extended header
  */
  struct Header : public Network::Header {
    Header() 
     : Network::Header(), index_src(0)
    {}
    uint32_t index_src;      /* index in _peers of OutChannel of the sender */
  };


  // ---------------------------------------------------------------------------
  /** extended classes for broadcast purpose.
  */
  class OutChannel: public Network::OutChannel {
  public:
    /** Constructor
    */
    OutChannel( );

    /** destructor
    */
    ~OutChannel();

    /** Initialize a Bcast based OutChannel.
        The memory zone pointed by peers should be valid until the channel is destroyed.
        @param net the network that do allocation
        @param tag the name of the OutChannel
        @param count_peer the number of nodes involved in the collective operation
        @param peers a pointer to the array of rank of the nodes
        @param index in peers of the local node
    */    
    virtual void init( BcastNetwork* net, int tag, int count_peer, LocalNodeNo* peers, int index_self );

  public: /* to all derived service of forwarding */
    uint32_t   _count_peers;   /* number of nodes involved in the communication */
    uint16_t*  _peers;         /* list of their ranks, relative to the sub network */
    uint32_t   _index_self;    /* index of current node in _peers */
  };


  // ---------------------------------------------------------------------------
  /* extended methods for broadcast purpose.
  */
  /** Open a bcast channel between a set of node including local node.
      @param tag(IN) tag to identify this Bcast channel among the system
      @param count_peer(IN) number of nodes for the bcast
      @param peers(IN) node's numbers (relative to the underlaying network)
      @param index_self(IN) index of self node in peers
      \warning ajouter les aliasing d'adresse de noeud
  */
  virtual OutChannel* open_channel( int tag, int count_peer, LocalNodeNo* peers, int index_self )
       throw(ACOM::ComFailure) =0;


  // ---------------------------------------------------------------------------
  /* inherited and redefined methods:
  */  
  
private:
  friend class OutChannel;
};



/*
 * inline definitions
 */
  
}

#endif
