// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_one_2_all.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of flat bcast
// $Id: ics_bcast_one_2_all.cpp,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast_one_2_all.h>
#include <iostream>

namespace ICS {

// -----------------------------------------------------------------------------
BcastOne2All::BcastOne2All()
{
}


// -----------------------------------------------------------------------------
/** */
BcastOne2All::handler BcastOne2All::post_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return 0;

  if (index_self != index_src) 
  { 
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section
    postrecv_data( tag, true, fmt, &buffer, &count, 0, 0, 0);
    return (handler*)tag;
  } 

  /* Form a message */
  ACOM::WriteAndSignal<BcastOne2All> ws (_net, this, 0, sz_ranks-1);
  postsend_data( ws.get_request(), tag, fmt, buffer, count );

  /* I initiate the send */
  for (int i=0; i<sz_ranks; ++i) 
  {
    if (i == index_src) continue;
    /* get a channel */
    ACOM::Network::OutChannel* channel = _net->get_default_route( ranks[i] );
    /* Send it */
    ws.post(channel);
  }
  return 0;
}



// -----------------------------------------------------------------------------
/** */
void BcastOne2All::wait_bcast(
  handler wc
)
{
  if (wc ==0) return;
  PTH::CriticalGuard cg(_access); // - begin critical section
  int tag = (int)wc;
  com_context* ctxt = wait_data( tag);
  deallocate_cc( ctxt );
}



// -----------------------------------------------------------------------------
/** */
void BcastOne2All::async_bcast(
    void*&                     buffer, 
    size_t&                    count,
    const ACOM::Format*        fmt,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_src, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  ++_nbcalled;
  if (index_self != index_src) 
  { 
    /* should receive a message */
    PTH::CriticalGuard cg(_access); // - begin critical section
    postrecv_data( tag, false, fmt, &buffer, &count, cb_func, cb_arg1, cb_arg2);
    return;
  } 

  /* Form a message */
  ACOM::WriteAndSignal<BcastOne2All> ws (_net, this, 0, sz_ranks-1);
  postsend_data( ws.get_request(), tag, fmt, buffer, count );

  /* I initiate the send */
  for (int i=0; i<sz_ranks; ++i) 
  {
    if (i == index_src) continue;
    /* get a channel */
    ACOM::Network::OutChannel* channel = _net->get_default_route( ranks[i] );
    /* Send it */
    ws.post(channel);
  }
  if (cb_func !=0)
    (*cb_func)(cb_arg1, cb_arg2);
}


}
