// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcast_pipesrw.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of bcast using pipeline algorithm
// $Id: ics_bcast_pipesrw.cpp,v 1.4 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_bcast_pipeline1.h>
#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif

namespace ICS {

// -----------------------------------------------------------------------------
void BcastPipeline1::send_iov ( ACOM::Network::OutChannel* ch, int& serial, int tag, 
                                ACOM::IOVectEntry* iov, size_t iovsz )
{
#if 0
std::cout << "IN send_iov #iov=" << iovsz << std::endl;
#endif
  /* sz remains to send */
  size_t io_tosend = iovsz;

  /* sz of bytes in the current bloc of iovector */
  size_t sz_send  = 0;

  int    cur_src  = 0;
  char*  data_src = (char*)iov[cur_src].data;
  size_t len_src  = iov[cur_src].len;

  ACOM::IOVectEntry* iov_dest = (ACOM::IOVectEntry*)::alloca( sizeof(ACOM::IOVectEntry)*iovsz );
  int    cur_dest  = 0;

  while (io_tosend >0) 
  {
    /* size of the message using the current iov[cur_src] 
    */
    size_t szmsg = (sz_send+len_src > _szbloc ? _szbloc - sz_send : len_src);
    sz_send   += szmsg;  
    iov_dest[cur_dest].data = data_src;
    iov_dest[cur_dest].len  = szmsg;
    ++cur_dest;
#if 0
std::cout << "Size iov=" << szmsg << std::endl;
#endif
    
    /* If sz_send >= _szbloc or it will be the last packet, then send it */
    if (((sz_send >= _szbloc) || ((io_tosend ==1) && (len_src == szmsg))) )// && (cur_dest >0)) 
    {
      /* Form the message */
      ACOM::WriteAndSignal<BcastPipeline1> ws (_net, this, 0);
      ws->pack(&serial, sizeof(int));
      ws->pack(&tag, sizeof(int));
      for (int i=0; i < cur_dest; ++i )
        ws->put( iov_dest[i].data, iov_dest[i].len );

      /* Send it */
      ws.post(ch);
      
#if 0
std::cout << "Send, serial=" << serial << ", tag=" << tag 
          << ", #iov=" << cur_dest << " iovdlen=[";
for (int i=0; i < cur_dest; ++i )
  std::cout << iov_dest[i].len << ' ';
std::cout << ']' << std::endl;  
for (int i=0; i < cur_dest; ++i )
{  
  std::cout << "iov[" << i << "]= [";
  for (int k=0; k < iov_dest[i].len; ++k )
    std::cout << (int)((char*)iov_dest[i].data)[k] << ' ';
  std::cout << ']' << std::endl;  
}
#endif
      /* reset counters and usage of src_dest */
      cur_dest = 0;
      sz_send  = 0;
      ++serial;
    }

    /* incr next message */
    len_src -= szmsg;
    if (len_src == 0) 
    {
      --io_tosend;
      if (io_tosend ==0)
        break;
      ++cur_src; 
      len_src  = iov[cur_src].len;
      data_src = (char*)iov[cur_src].data;
    } else {
      data_src += szmsg;
    }
  }
#if 0
std::cout << "OUT send_iov #iov=" << iovsz << std::endl;
#endif
}
                           

// -----------------------------------------------------------------------------
void BcastPipeline1::postrecv_iov ( int& serial, int tag, 
                                    ACOM::IOVectEntry* iov, size_t iovsz )
{
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

  if (iovsz == -1U) 
  {
    /* search if receive already occurs */
    Key key (tag, serial);
    curr = _postrecv.find( key );
    ctxt = 0;
    if (curr == 0) 
    {
      ctxt = allocate_cc();
      ctxt->fmt   = &ACOM::FormatDef::Byte;
      ctxt->count_post = (size_t)-1U;
      ctxt->data_post  = 0;
      ctxt->count_recv = (size_t)-1;
      curr = _postrecv.insert( key, ctxt);
    }
    else {
      ctxt = curr->second;
      if ((curr->first.tag != tag) || (curr->first.serial != serial))
        std::cout << "POST::bad find !!!!!!!!!!!!" << std::endl;
      /* Special case where system allocate data_post: swap _recv with _post */
      ctxt->data_post = (void*)ctxt->data_recv;
      ctxt->data_recv = 0;
    }
    ++serial;
    return;
  }

#if 0
std::cout << "IN postrecv_iov #iov=" << iovsz << std::endl;
#endif
  
  /* sz of iovector remains to recv */
  size_t io_torecv = iovsz;

  /* sz of data that will be received in the current packet */
  size_t sz_recv   = 0;

  int    cur_src   = 0;
  char*  data_src  = (char*)iov[cur_src].data;
  size_t len_src  = iov[cur_src].len;

  ACOM::IOVectEntry* iov_dest = (ACOM::IOVectEntry*)::alloca( sizeof(ACOM::IOVectEntry)*iovsz );
  int    cur_dest  = 0;

  while (io_torecv >0) 
  {
    /* size of the message using the current iov[pos] 
    */
    size_t szmsg = (sz_recv+len_src > _szbloc ? _szbloc-sz_recv : len_src);
    sz_recv   += szmsg;  
    iov_dest[cur_dest].data = data_src;
    iov_dest[cur_dest].len  = szmsg;
    ++cur_dest;
#if 0
std::cout << "Size iov=" << szmsg << std::endl;
#endif

    /* If sz_recv >= _szbloc or it will be the last packet, then recv it */
    if ((sz_recv >= _szbloc) || ((io_torecv ==1) && (len_src == szmsg))) 
    {
      /* search if receive already occurs */
      Key key (tag, serial);
      curr = _postrecv.find( key );
      ctxt = 0;
      if (curr == 0) 
      {
        ctxt = allocate_cc();
        ctxt->fmt   = &ACOM::FormatDef::Byte;
        ctxt->count_post = (size_t)cur_dest;
        curr = _postrecv.insert( key, ctxt);
      }
      else {
        ctxt = curr->second;
        if ((curr->first.tag != tag) || (curr->first.serial != serial))
          std::cout << "POST::bad find !!!!!!!!!!!!" << std::endl;
        if (ctxt->count_recv != (size_t)cur_dest)
          std::cout << "POST::bad split of data in recv, with data that have been received" << std::endl;
      }
      ACOM::IOVectEntry* iov_post =(ACOM::IOVectEntry*)_net->allocate_data( sizeof(ACOM::IOVectEntry)*cur_dest );
      ctxt->count_post = (size_t)cur_dest;
      ctxt->data_post  = iov_post;
      for (int i=0; i < cur_dest; ++i ) {
        iov_post[i] = iov_dest[i];
      }
#if 0
std::cout << "Recv, serial=" << serial << ", tag=" << tag
          << ", #iov=" << cur_dest << " iovdlen=[";
for (int i=0; i < cur_dest; ++i )
  std::cout << iov_dest[i].len << ' ';
std::cout << ']' << std::endl;  
#endif
      
      /* reset counters and usage of src_dest */
      cur_dest = 0;
      sz_recv  = 0;
      ++serial;
    }

    /* incr next message */
    len_src -= szmsg;
    if (len_src == 0) 
    {
      --io_torecv;
      if (io_torecv ==0)
        break;
      ++cur_src; 
      len_src  = iov[cur_src].len;
      data_src = (char*)iov[cur_src].data;
    } else {
      data_src += szmsg;
    }
  }
#if 0
std::cout << "OUT postrecv_iov #iov=" << iovsz << std::endl;
#endif
}

                    
// -----------------------------------------------------------------------------
BcastPipeline1::com_context* BcastPipeline1::wait_iov ( int serial, int tag )
{
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

  /* find tag in _postrecv */
  Key key (tag, serial);
  curr = _postrecv.find( key );
  if (curr == 0) {
    std::cout << "Cannot find ctxt with key(" << serial << ',' << tag << ")" << std::endl;
    throw ACOM::InvalidArgument();
  }
  ctxt = curr->second;
  ctxt->is_wait = true;
  while (ctxt->wait_recv != ctxt->wait_post) {
    _cond.wait(_access);
  }
  _postrecv.erase(curr);

  ACOM::IOVectEntry* iov_recv = (ACOM::IOVectEntry*)ctxt->data_recv;
  ACOM::IOVectEntry* iov_post = (ACOM::IOVectEntry*)ctxt->data_post;
  size_t iovsz = ctxt->count_recv;
  
  /* if iov_recv !=0 and iovsz >0, then a message arrived before post and should
     be recopied 
  */
  if ((iov_recv !=0) && (iov_recv != iov_post))
  {
    for (size_t i=0; i< iovsz; ++i)
    {
      if (iov_recv[i].len != iov_post[i].len) {
        std::cout << "Mismatch between post/recv, may be it is due to splitting ??" << std::endl;
        throw;
      }
      memcpy(iov_post[i].data, iov_recv[i].data, iov_recv[i].len);
      _net->deallocate_data( iov_recv[i].data );
    }
    _net->deallocate_data( iov_recv );
  }

  return ctxt;
}


// -----------------------------------------------------------------------------
void* BcastPipeline1::get_iovect(
  ACOM::Network::Receipt* r,
  ACOM::Network::Header* header, 
  ACOM::IOVectEntry* iov, 
  ACOM::IOVectSize iovsz
)
{
  /* extract immediate data */
  int tag;
  int serial;
  r->unpack(&serial,sizeof(int));
  r->unpack(&tag, sizeof(int));

  /* The key for this packet */
  Key key(tag, serial);  
  
  /* some data */
  com_context* ctxt = 0;
    
  if (serial ==0) 
  { 
    /* first packet: extract fid/count + lengths of iovect for immediate and defered data */
    int fid;
    size_t count;
    r->unpack(&fid,sizeof(int));
    r->unpack(&count,sizeof(size_t));
  }    
  
  /* critical section until the end of this method */
  //PTH::CriticalGuard cg(_access);
  _access.lock();

  /* find tag in _postrecv */
  Trigger_set::iterator curr = _postrecv.find( key );

  ACOM::IOVectEntry* iov_recv =0;
  if (curr == 0) 
  {
    /* Here the message arrived before it was posted
    */
    ctxt  = (com_context*) allocate_cc();
    ctxt->is_wait  = false;
    ctxt->fmt      = &ACOM::FormatDef::Byte;
    ctxt->count_recv = iovsz;
    iov_recv = (ACOM::IOVectEntry*)_net->allocate_data( iovsz );
    for (int i=0; i<iovsz; ++i) {
      iov_recv[i].data = 0; iov_recv[i].len = iov[i].len;
    }
    ctxt->data_recv = iov_recv;
    ctxt->data_post = 0;
    /* insert ctxt into the received table */
    curr = _postrecv.insert(key, ctxt);
    _access.unlock();
  } 
  else 
  {
    /* Here the message arrived after the post to received: reuse data associated
    * to the com_context to received data
    */
    _access.unlock();
    ctxt = curr->second;
    iov_recv = (ACOM::IOVectEntry*)ctxt->data_post;

    /* if ctxt->count == -1, then iov should be allocated by the service */
    if (ctxt->count_recv == -1U) 
    {
      iov_recv = (ACOM::IOVectEntry*)_net->allocate_data( iovsz );
      for (int i=0; i<iovsz; ++i) {
        iov_recv[i].data = 0; iov_recv[i].len = iov[i].len;
      }
      ctxt->data_post = (void**)iov_recv;
      ctxt->data_recv = 0;
      ctxt->count_recv= iovsz;
    } 
  }

  if (ctxt->count_recv != iovsz) 
  {
    std::cout << "Bad size of iovectors for pipeline packet: recv=" 
              << iovsz << ", waiting=" << ctxt->count_recv << std::endl;
  }

  /* allocate entries of the iov_recv[i] and iov[i] */
  for (int i=0; i<iovsz; ++i) 
  {
    if (iov_recv[i].len != iov[i].len) 
    {
      std::cout << "Bad size of iovector entry for pipeline packet: recv=" 
                << iov[i].len << ", waiting=" << iov_recv[i].len << std::endl;
    }
    if (iov_recv[i].data ==0) 
      iov_recv[i].data = iov[i].data = _net->allocate_data( iov_recv[i].len );
    else 
      iov[i].data = iov_recv[i].data;
  }
  
  return ctxt; 
}


}
