// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcastnet_alpha.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_bcastnet_alpha.cpp,v 1.12 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#include <ics_bcastnet_alpha.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
BcastAlphaNetwork::BcastAlphaNetwork()
  : BcastNetwork()
{
  /* Change properties */
  set_property("name",     "BCASTNETALPHA");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("");
}



// --------------------------------------------------------------------------
/** Destructor
*/
BcastAlphaNetwork::~BcastAlphaNetwork()
{
}




// ---------------------------------------------------------------------------
int BcastAlphaNetwork::initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
  throw (ACOM::NetworkError)
{
  set_name("balpha_" + netInit->get_name() );
  return Network::initialize( argc, argv, netInit );
}  



// ---------------------------------------------------------------------------
BcastNetwork::OutChannel* BcastAlphaNetwork::open_channel
  ( int tag, int count_peer, LocalNodeNo* peers, int index_self )
      throw(ACOM::ComFailure)
{
  BcastAlphaNetwork::OutChannel* ch = new BcastAlphaNetwork::OutChannel;
  ch->init( this, tag, count_peer, peers, index_self );
  return ch;
}      



// ---------------------------------------------------------------------------
BcastAlphaNetwork::OutChannel::OutChannel()
 : BcastNetwork::OutChannel(), _alpha(0.5)
{ 
} 



// ---------------------------------------------------------------------------
void BcastAlphaNetwork::process
  ( ACOM::Network::OutChannel* c, ACOM::Network::CallContext* call_cc, ACOM::Network::Request* req )
      throw(ACOM::ComFailure)
{
  ACOM_ASSERT_DEBUG( c->get_net() == this, "bad assertion");

  BcastAlphaNetwork::OutChannel* ch = (BcastAlphaNetwork::OutChannel*)c;
  BcastAlphaNetwork::Header* eh = 0;

  /* what ever it is a post or a redirection : need to
     compute alpha tree from source at index_src or _index_self if it is a post
  */
  int myrank_fromsrc = 0;
  int myrank = 0;

  if (call_cc !=0) {
    /* it a redirection */
    eh = (BcastAlphaNetwork::Header*)call_cc->ext_header;
    myrank_fromsrc = ch->_index_self - eh->index_src;
    if (myrank_fromsrc <0) myrank_fromsrc += ch->_count_peers;
    myrank = myrank_fromsrc;
  }
  int offset = ch->_count_peers;
  int parent = 0;
  int nOthers;
  int threshold;
  
  while (myrank != 0) 
  {
    nOthers = (int)round(ch->_alpha * offset);
    threshold = offset - nOthers;
    if(myrank == threshold) {
      offset = nOthers;
      myrank -= threshold;
    } else {
      if(myrank < threshold)
        offset = threshold;
      else {
        offset = nOthers;
        parent += threshold;
        myrank -= threshold;
      }
    }
  }

  /* count the number of recv to wait */
  int nn = offset;
  int nbpeer = 0;
  while (nn > 1) {
    nn -= (int)round(ch->_alpha*nn);
    ++nbpeer;
  }

  if (call_cc ==0) { 
    /* it a the post : set the extended header for alpha network */
    BcastAlphaNetwork::Header* eh = (BcastAlphaNetwork::Header*)allocate_data( sizeof(BcastAlphaNetwork::Header) );
    req->set_extended_header( eh, sizeof(BcastAlphaNetwork::Header));
    eh->saveid    = req->_header.id;
    eh->tag       = ch->_tag;
    eh->netid     = get_id();
    eh->index_src = ch->_index_self;
  
    /* replace service id by internal service */
    req->_header.id = _service->get_id();
    
    /* Add nbpeer-2 use of the request : 
      1 is already be counted by the user, and send it nbpeer times 
    */
    req->_used_ct += nbpeer -1;
    
    /* I'm the source, then I begin to send */
    int n = offset;
    while(n > 1) {
      n -= (int)round(ch->_alpha*n);
      /* send messag to ranks[ (index_self + n) % sz] */
      int dest = ch->_peers[ (ch->_index_self + n) % ch->_count_peers];
      ACOM::Network::OutChannel* channel = get_subnet()->get_default_route( dest );
      /* Send it */
      channel->post( req );

    }
    /* reset default service, if used multiple times */
    req->_header.id = eh->saveid;
    return;
  }

  /* Else it is a redirection */
  ACOM::Network::Header* h = call_cc->header;
  ACOM_ASSERT_DEBUG( eh !=0, "bad assertion");

  /* The user service : get_iovect has been called in RedirecService */
  ACOM::Service* userserv = ACOM::Service::Manager.resolv( eh->saveid );

  if (nbpeer >0) 
  {
    /* call back to clear call_cc */
    call_cc->acquire( nbpeer );
    RedirectServiceCallBack*  mycbk = 
      new (call_cc->allocate(sizeof(RedirectServiceCallBack))) 
          RedirectServiceCallBack( 0, call_cc );

    /* message to be resend */
    ACOM::Message m ( get_subnet(), (ACOM::Network::ProtocolNo)h->protocol, _service, mycbk, nbpeer );
    
    /* recopy extended header */
    BcastAlphaNetwork::Header* enewh = (BcastAlphaNetwork::Header*)call_cc->allocate(sizeof(BcastAlphaNetwork::Header));
    m.get_request()->set_extended_header( enewh, sizeof(BcastAlphaNetwork::Header));
    *enewh = *eh;

    /* recopy immediate data from receipt */
    ACOM::IOVectSeq* iovs = call_cc->receipt.get_packetbyte_vectseq();
    for( ACOM::IOVectSeq::IteratorEntry ie= iovs->beginEntry(); ie != iovs->endEntry(); ++ie)
      m->pack( ie->data, ie->len );

    /* recopy iovector from  iov, iov_sz */
    for (int i=0; i< call_cc->iov_sz; ++i)
      m->put( &call_cc->iov[i] );
          
    /* re-post the messsage */
    int n = offset;
    while(n > 1) {
      n -= (int)round(ch->_alpha*n);
      /* send messag to ranks[ (index_self+index_src + n) % sz] */
      int dest = ch->_peers[ (eh->index_src + myrank_fromsrc + n) % ch->_count_peers];
//std::cout << net->rank() << "::Send to :" << dest << std::endl;
      ACOM::Network::OutChannel* channel = get_subnet()->get_default_route( dest );

      /* Send it */
      m.post( channel );
    }
  }

  /* call user method */
  h->id = eh->saveid;
  userserv->server_call( call_cc );
  call_cc->release();
}

} // namespace

