// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_bcastnet_chain.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_bcastnet_chain.cpp,v 1.10 2003/07/28 20:21:19 thierry Exp $
//
// ==========================================================================
#include <ics_bcastnet_chain.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
BcastChainNetwork::BcastChainNetwork()
  : BcastNetwork()
{
  /* Change properties */
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("");
}


// --------------------------------------------------------------------------
/** Destructor
*/
BcastChainNetwork::~BcastChainNetwork()
{
}



// ---------------------------------------------------------------------------
int BcastChainNetwork::initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
  throw (ACOM::NetworkError)
{
  set_name("bchain_" + netInit->get_name() );
  return Network::initialize( argc, argv, netInit );
}  



// ---------------------------------------------------------------------------
BcastNetwork::OutChannel* BcastChainNetwork::open_channel
  ( int tag, int count_peer, LocalNodeNo* peers, int index_self )
      throw(ACOM::ComFailure)
{
  BcastChainNetwork::OutChannel* ch = new BcastChainNetwork::OutChannel;
  ch->init( this, tag, count_peer, peers, index_self );
  return ch;
}      



// ---------------------------------------------------------------------------
void BcastChainNetwork::process
  ( ACOM::Network::OutChannel* c, ACOM::Network::CallContext* call_cc, ACOM::Network::Request* req )
      throw(ACOM::ComFailure)
{
//std::cout << rank() << "::IN " << __PRETTY_FUNCTION__ << std::endl;
  ACOM_ASSERT_DEBUG( c->get_net() == this, "bad assertion");

  BcastChainNetwork::OutChannel* ch = (BcastChainNetwork::OutChannel*)c;
  
  /* what ever it is a post or a redirection : need next node */
  int index_next =  (ch->_index_self + 1) % ch->_count_peers;
  
  /* If call_cc ==0, then it is asynchrone post of the request */
  if (call_cc ==0) {
    /* extended header for flat network */
    BcastChainNetwork::Header* eh = (BcastChainNetwork::Header*)allocate_data( sizeof(BcastChainNetwork::Header) );
    req->set_extended_header( eh, sizeof(BcastChainNetwork::Header));
    eh->saveid    = req->_header.id;
    eh->tag       = ch->_tag;
    eh->netid     = get_id();
    eh->index_src = ch->_index_self;
    /* replace service id by internal service */
    req->_header.id = _service->get_id();
/*
    std::cout << "Post: saveid=" << eh->saveid << ", redirec service=" << req->_header.id 
             << ", verif = " << _service->get_id() << std::endl;
*/
    
    //std::cout << _peers[_index_self] << ":: initiate send message to node=" << _peers[index_next] << std::endl;
    ACOM::Network::OutChannel* channel = get_subnet()->get_default_route(ch->_peers[index_next]);
    channel->post( req );
    
    /* reset default service, if used multiple times */
    req->_header.id = eh->saveid;
//std::cout << rank() << "::OUT " << __PRETTY_FUNCTION__ << std::endl;
    return;
  }

  /* Else it is a redirection */
  ACOM_ASSERT_DEBUG( call_cc !=0, "bad assertion");
  ACOM_ASSERT_DEBUG( call_cc->_refcount >0, "bad assertion");

  ACOM::Network::Header* h = call_cc->header;
  BcastChainNetwork::Header* eh = (BcastChainNetwork::Header*)call_cc->ext_header;
  ACOM_ASSERT_DEBUG( eh !=0, "bad assertion");
  //std::cout << "Redirect: saveid=" << eh->saveid << ", redirec service=" << h->id << std::endl;

  /* The user service */
  ACOM::Service* userserv = ACOM::Service::Manager.resolv( eh->saveid );

  if (index_next != eh->index_src) {
  
    /* call back to clear call_cc */
//    std::cout << "repost/";
    call_cc->acquire();
    ACOM_ASSERT_DEBUG( call_cc->_refcount >1, "bad assertion");
    RedirectServiceCallBack*  mycbk = 
      new (call_cc->allocate(sizeof(RedirectServiceCallBack))) 
          RedirectServiceCallBack( 0, call_cc );

    /* message to be resend */
    ACOM::Message m ( get_subnet(), (ACOM::Network::ProtocolNo)h->protocol, _service, mycbk, 1 );

    /* recopy extended header */
    BcastChainNetwork::Header* enewh = (BcastChainNetwork::Header*)call_cc->allocate(sizeof(BcastChainNetwork::Header));
    m.get_request()->set_extended_header( enewh, sizeof(BcastChainNetwork::Header));
    *enewh = *eh;

    /* recopy immediate data from receipt */
    ACOM::IOVectSeq* iovs = call_cc->receipt.get_packetbyte_vectseq();
    for( ACOM::IOVectSeq::IteratorEntry ie= iovs->beginEntry(); ie != iovs->endEntry(); ++ie)
      m->pack( ie->data, ie->len );

    /* recopy iovector from  iov, iov_sz */
    for (int i=0; i< call_cc->iov_sz; ++i)
      m->put( &call_cc->iov[i] );
          
    /* re-post message */
    //std::cout << ch->_peers[ch->_index_self] << "::Begin to resend message to : " << ch->_peers[index_next] << std::endl;
    m.post( get_subnet()->get_default_route( ch->_peers[index_next] ) );
    //std::cout << ch->_peers[ch->_index_self] << "::Message resend ok to : " << ch->_peers[index_next] << std::endl;
  }

  /* call user method */
  h->id = eh->saveid;
  userserv->server_call( call_cc );
//  std::cout << "endcall/";
  ACOM_ASSERT_DEBUG( call_cc->_refcount >0, "bad assertion");
  call_cc->release();

//std::cout << rank() << "::OUT " << __PRETTY_FUNCTION__ << std::endl;
}


}
