// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_dns.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#ifndef _ICS_DNS_H_
#define _ICS_DNS_H_

#include<ics_conf.h>
#include<pth>
#include<acom>
#include<string>
#include<map>

namespace ICS {

/* fwd declaration */
class DNSRequest;


/** DNS service.
 DNS is a ICS service build on any kind of primary Network.
 DNS provides a set of function to associate value (a string) to a
 name (a string) in a namecontext (a string). All remote operations
 are done using a Inuktitut service on top of the primary Network.
 @warning The only requirements is that such network permits to 
 redo an active message in a service call method.
*/
class DNS {
public:
public:
  /** Constructor 
  */
  DNS();

  /** Destructor 
  */
  ~DNS();
  
  /** Exceptions thrown by DNS service
  */
  //@{
    /// Base class for error 
  struct DnsError : public ACOM::Runtime {};
    /// Bad name : name doesnot exist in DNS
  struct BadName : public DnsError {};
    /// Already bind : the name is already bind
  struct AlreadyBind : public DnsError {};
  //@}

  /** Initialisation of the DNS service.
      DNS service should be instanciate on all nodes !
      @param net a Primary network
      @exception ACOM::InvalidArgument if net not a primary network
  */
  void initialize( ACOM::Network* net )
    throw (ACOM::InvalidArgument);

  /** Destroy the local instance of DNS service
  */
  void destroy();
  
  /** Bind : associate a value to name in a naming context
      @param namectxt the naming context
      @param name the name
      @param value the associated value
      @exception AlreadyBind if name in naming context already exists
      @exception ComFailure if dns cannot be contacted
  */
  void bind( const std::string& namectxt, const std::string& name, const std::string& value )
    throw (AlreadyBind, ACOM::ComFailure);

  /** rebind : associate a value to name in a naming context, rebind if already exist
      @param namectxt the naming context
      @param name the name
      @param value the associated value
      @exception ComFailure if dns cannot be contacted
  */
  void rebind( const std::string& namectxt, const std::string& name, const std::string& value )
    throw (ACOM::ComFailure);

  /** Resolve : the the value associated to a name in a naming context
      @param namectxt the naming context
      @param name the name
      @retval the value the associated value
      @exception BadName the name in naming context does not exist
      @exception ComFailure if dns cannot be contacted
  */
  std::string resolve( const std::string& namectxt, const std::string& name )
    throw (BadName, ACOM::ComFailure);

  /** Unbind : suppress entry name in a naming context
      @param namectxt the naming context
      @param name the name
      @exception BadName the name in naming context does not exist
      @exception ComFailure if dns cannot be contacted
  */
  void unbind( const std::string& namectxt, const std::string& name )
    throw (BadName, ACOM::ComFailure);

protected:  
  typedef std::map<std::string, std::string> NamingContext;
  typedef std::map<std::string, NamingContext> AllContext;
  PTH::Mutex     _access;
  PTH::Condition _waitreply;
  ACOM::Network* _net;
  DNSRequest*    _service;
  AllContext     _world;
  
  friend class DNSRequest; 
};

}

#endif
