// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_flatbarrier.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// Implementation of barrier as a service
// $Id: ics_flatbarrier.cpp,v 1.2 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_flatbarrier.h>
#include <iostream>

namespace ICS {


// -----------------------------------------------------------------------------
/** implementation: use a flat tree : key % randcount is the root
 * phase1: root wait all childs, child send to root it arrives
 * phase2: root bcast to all childs to pass through the barrier.
 */
void FlatBarrier::barrier( int key )
{
  /* I will receive data */
  PTH::CriticalGuard cg(_access); // - begin critical section

  /* search if receive already occurs */
  bar_com_context declared;
  bar_com_context* used =0;

  long count;
  int rank = _net->rank();
  
  /* if myself== 0, then wait for all other node in network */
  if (rank ==0)
    count = _net->node_count()-1;
  else 
    count = 1;
    
  /* find a message already received */
  Trigger_set::iterator curr = _postrecv.find(key);
  if (curr == 0) 
  {
#if 0
    std::cout << "Post to recv ack" << std::endl;
#endif
    used = &declared;
    curr = _postrecv.insert(key, used);
  }
  else {
    used = curr->second;
  }

  /* accumulate count with already received signal */
  used->count = count;

  if (rank != 0) 
  { /* send to master (node ==0) */
    ACOM::Network::OutChannel* channel = _net->get_default_route( 0 );
    ACOM::ActiveMessage<BarrierService> am (_net, this, 0);
    am->pack(&key, sizeof(key)); 
    am.post(channel);
  }

  /* what ever am I (slave or master): wait recv message received */
  while (used->recv != used->count) 
  {
    used->is_wait = true;
    _cond.wait(_access); 
  }

  if (rank == 0) 
  {
    /* reply to all other node except master (==0) */
    for (size_t i=1; i< _net->node_count(); ++i)
    {
      ACOM::Network::OutChannel* channel = _net->get_default_route( i );
      ACOM::ActiveMessage<BarrierService> am (_net, this, 0);
      am->pack(&key, sizeof(key)); 
      am.post(channel);
    }
  }
  
  _postrecv.erase(curr);
  if (used != &declared) delete used;
} 



} // ICS namespace

