// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_net_threads.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_net_threads.cpp,v 1.1 2003/07/28 07:45:49 thierry Exp $
//
// ==========================================================================
#include <ics_network.h>

namespace ICS {

// --------------------------------------------------------------------------
std::vector<ThreadRedirec*> ThreadRedirec::allthreads;  // list of all threads for redirection
QueueWork*                  ThreadRedirec::queue;       // list of cc to redirect


// -----------------------------------------------------------------------------
BasicQueueWork::BasicQueueWork()
{
  _cond.init();
  _access.init();
  _nwaiters = 0;
  _wakeup = false;
  _size = 0;
}



// -----------------------------------------------------------------------------
BasicQueueWork::~BasicQueueWork()
{
  clear();
  _cond.destroy();
  _access.destroy();
}



// -----------------------------------------------------------------------------
Network::Work* BasicQueueWork::pop( )
{
  PTH::CriticalGuard critical(_access);
  /* return if wakeup flag is true */
  if (_wakeup) return 0;
  
  /* wait queue not empty or wakeup flag pass to true */
  while (_queue.empty() && !_wakeup)
  {
    ++_nwaiters;
    _cond.wait(_access);
    --_nwaiters;
  }
  /* if wakeup is true and I'm last waiters then reset 
     wakup flag and return 0;
  */
  if (_wakeup) {
    if (_nwaiters ==0)
    _wakeup = false;
    _cond.broadcast();
    return 0;
  }
  
  /* work to return, if queue is not empty and nwaiters >0 then
     signal a waiters
  */
  Network::Work* retval = _queue.pull_front();
  --_size;
  if (!_queue.empty() || (_nwaiters >0)) 
    _cond.signal();
  ACOM_ASSERT_DEBUG( (retval->cc ==0) || (retval->cc->_refcount >0), "bad assertion");
#if 0
  std::cout << "poped work:" << (long)retval << ", req=" << (long)retval->req
            << ", cc=" << (long)retval->cc;
  if (retval->cc !=0) 
    std::cout << ", rc=" << retval->cc->_refcount << std::endl;
  else
    std::cout << std::endl;
  std::cout << std::flush;
#endif
  return retval;
}



// -----------------------------------------------------------------------------
void BasicQueueWork::push( Network::Work* new_work )
{
  bool flag = false;
  { PTH::CriticalGuard critical(_access);
    ACOM_ASSERT_DEBUG( (new_work->cc ==0) || (new_work->cc->_refcount >0), "bad assertion");
    if (new_work->cc !=0) {
//      std::cout << "push/";
      new_work->cc->acquire();
      ACOM_ASSERT_DEBUG( (new_work->cc ==0) || (new_work->cc->_refcount >1), "bad assertion");
    }
    flag = _queue.empty() && (_nwaiters >0);
#if 0
    std::cout << "pushed work:" << (long)new_work << ", req=" << (long)new_work->req
              << ", cc=" << (long)new_work->cc;
    if (new_work->cc !=0) 
      std::cout << ", rc=" << new_work->cc->_refcount << std::endl;
    else
      std::cout << std::endl;
    std::cout << std::flush;
#endif
    _queue.push_back( new_work );
    ++_size;
    /* WARN: it could introduce deadlock */
    if (_size > 128) {
      /* wait size <= 128 */
      ++_nwaiters;
      while (_size > 128)
        _cond.wait(_access);
      --_nwaiters;
    }
  }
  if (flag) _cond.signal();
}



// -----------------------------------------------------------------------------
void BasicQueueWork::wake_up()
{
  PTH::CriticalGuard critical(_access);
  _wakeup = true;
  _cond.broadcast();
  while (_wakeup) {
    _cond.wait( _access );
  }
}



// -----------------------------------------------------------------------------
void BasicQueueWork::clear()
{
  wake_up();
  PTH::CriticalGuard critical(_access);
  _queue.clear();
}



// -----------------------------------------------------------------------------
QueueWork::QueueWork()
 : _size(0), _ntopush(0), _all_queues()
{ 
  _access.init();
}



// -----------------------------------------------------------------------------
QueueWork::~QueueWork()
{ 
  clear(); 
  _access.destroy();
}



// -----------------------------------------------------------------------------
ThreadRedirec::ThreadRedirec( int index )
 : _stop(false), _kill(false), _bq_index(index)
{
  _access.init(); 
  _condstop.init();
}



// -----------------------------------------------------------------------------
ThreadRedirec::~ThreadRedirec( )
{
  _access.destroy(); 
  _condstop.destroy();
}



// -----------------------------------------------------------------------------
void ThreadRedirec::stop()
{
  PTH::CriticalGuard critical(_access);
  _stop = true;
  ThreadRedirec::queue->wake_up( _bq_index );
}



// -----------------------------------------------------------------------------
void ThreadRedirec::kill()
{
  PTH::CriticalGuard critical(_access);
  _kill = true;
  ThreadRedirec::queue->wake_up( _bq_index );
  while (_kill) {
    _condstop.wait(_access);
  }
}



// -----------------------------------------------------------------------------
void ThreadRedirec::restart()
{
  PTH::CriticalGuard critical(_access);
  _stop = false;
  _condstop.signal();
}



// -----------------------------------------------------------------------------
void* ThreadRedirec::run()
{
  /* this is a daemon's like thread, infinite loop except if it stop */
  bool cestfini = false;
  while (!cestfini)
  {
    Network::Work* work =0;

    while (work ==0) {
      /* get a cc to redirect */
      work = ThreadRedirec::queue->pop(_bq_index);
      
      /* stop thread ? */
      if (work ==0) {
        PTH::CriticalGuard critical(_access);
        if (_kill) {
          _kill = false;
          _condstop.broadcast();
          return 0;
        }
        if (_stop) {
          _condstop.broadcast();
          while (_stop) 
            _condstop.wait(_access);
        }
      }
    }
    
    /* process the request */
    ACOM_ASSERT_DEBUG( ((work->net !=0) && (work->subnet ==0)) || ((work->net ==0) && (work->subnet !=0)), "bad assertion");
    ACOM_ASSERT_DEBUG( (work->cc ==0) || (work->cc->_refcount >0), "bad assertion");
    if (work->net !=0) {
#if 0
      std::cout << "process work:" << (long)work << ", cc=" << (long)work->cc;
      if (work->cc !=0) 
        std::cout << ", rc=" << work->cc->_refcount << std::endl;
      else
        std::cout << std::endl;
      std::cout << std::flush;
#endif
      work->net->process( work->ch, work->cc, work->req );
    } else {
      ACOM_ASSERT_DEBUG( (work->cc ==0), "bad assertion" );
      work->subnet->post( work->req, work->ch );
    }
    
    /* deallocate work */
    ThreadRedirec::queue->deallocate( work );
  }
  return 0;
}


 
} // namespace ICS

