// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_network.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all bcast algorithms
// Please read node in .cpp file.
// $Id: ics_network.cpp,v 1.8 2003/07/28 20:21:19 thierry Exp $
//
// ==========================================================================
#include <ics_network.h>

namespace ICS {


// --------------------------------------------------------------------------
/** Constructor
*/
Network::Network()
  : ACOM::Network(), _subnet(0), _service(0), _allchannels()
{
  _access.init();
  _cond.init();
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("");
  /* */
  _service = new RedirectService;
  _service->initialize( this );
  ACOM::Service::Manager.bind( ACOM::ServiceManager::ICS_LNS, ServiceNumber::ICS_REDIRECT_SERVICE+_nid, _service );
}



// --------------------------------------------------------------------------
/** Destructor
*/
Network::~Network()
{
  _access.destroy();
  _cond.destroy();
}



// ---------------------------------------------------------------------------
int Network::initialize( int* argc, char** argv[] )
  throw (ACOM::NetworkError)
{ 
  ACOM::throwit( ACOM::NetworkError("Cannot be initialized without a network") ); 
  return 0;
}



// ---------------------------------------------------------------------------
int Network::initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
  throw (ACOM::NetworkError)
{
  _subnet = netInit;
  _node_rank =  _subnet->rank();
  _node_count = _subnet->node_count();
  if (get_name() == "")
    set_name("ICS_" + _subnet->get_name() );
  ACOM::Acom::bind_network( get_name(), this );
  return 0;
}  



// ---------------------------------------------------------------------------
int Network::commit () throw (ACOM::NetworkError)
{
  return 0;
}  



// ---------------------------------------------------------------------------
int Network::terminate () throw (ACOM::NetworkError)
{
  ACOM::Acom::unbind_network( get_name() );
  return 0;
}  



// ---------------------------------------------------------------------------
ACOM::Network::Request* Network::allocate_request () throw (ACOM::BadAlloc)
{
  ACOM::Network::Request* req = _subnet->allocate_request();
//  std::cout << "[ICS]:: allocate request : " << (long)req << std::endl;
  return req;
}  



// ---------------------------------------------------------------------------
void Network::deallocate_request( ACOM::Network::Request* req ) throw (ACOM::InvalidArgument)
{
//  std::cout << "[ICS]:: deallocate request : " << (long)req << std::endl;
  _subnet->deallocate_request( req );
}  



// ---------------------------------------------------------------------------
void* Network::allocate_data( size_t sz )
  throw (ACOM::BadAlloc)
{ return _subnet->allocate_data( sz ); }



// ---------------------------------------------------------------------------
void Network::deallocate_data(  void* data )
  throw (ACOM::InvalidArgument)
{ _subnet->deallocate_data( data ); }



// -----------------------------------------------------------------------------
Network::CallContext* Network::allocate_callcontext(size_t sz) throw (ACOM::BadAlloc)
{ return _subnet->allocate_callcontext(sz); }



// -----------------------------------------------------------------------------
void Network::deallocate_callcontext(CallContext* p) throw (ACOM::InvalidArgument)
{ _subnet->deallocate_callcontext(p); }



// --------------------------------------------------------------------------
ACOM::Network::CallContext* Network::resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) 
  throw (ACOM::BadAlloc)
{ return _subnet->resize_callcontext( cc, sz ); }



// --------------------------------------------------------------------------
ACOM::Network::CallContext* Network::acquire_callcontext() throw (ACOM::BadAlloc)
{  
  ACOM::Network::CallContext* cc = _subnet->acquire_callcontext( ); 
//  std::cout << "allocate cc=" << (long)cc << std::endl;
  return cc;
}



// --------------------------------------------------------------------------
void Network::release_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument)
{ 
//  std::cout << "deallocate cc=" << (long)cc << std::endl;
  _subnet->release_callcontext( cc ); 
}


// ---------------------------------------------------------------------------
void Network::close_channel( ACOM::Network::OutChannel* c )
    throw(ACOM::ComFailure)
{
  if (c->get_net() == get_subnet()) 
  {
    _subnet->close_channel( c );
  }
  else if (c->get_net() == this) {
    OutChannel* ch = (OutChannel*)c;
    ch->destroy();
  }
  else {
    ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );
  }
}    



// ---------------------------------------------------------------------------
Network::OutChannel* Network::get_channel( uint32_t tag )
{
  PTH::CriticalGuard critical( _access );
  LHashMap<uint32_t, OutChannel*>::iterator curr = Network::_allchannels.find( tag );
  if (curr ==0) return 0;
  return curr->second;
}



// ---------------------------------------------------------------------------
Network::Work* QueueWork::allocate( )
{
  Network::Work* w = 0;
  PTH::CriticalGuard critical(_access);
  w = _free_work.pop();
  if (w ==0)
    w = new Network::Work;
  w->net =0;
  w->ch  =0;
  w->cc  =0;
  w->req =0;
//  std::cout << "alloc work:" << (long)w << std::endl;
  return w;
}



// ---------------------------------------------------------------------------
void QueueWork::deallocate( Network::Work* w )
{
  w->net =0;
  w->ch  =0;
  w->cc  =0;
  w->req =0;
  PTH::CriticalGuard critical(_access);
//  std::cout << "free work:" << (long)w << std::endl;
  _free_work.push( w );
}



// ---------------------------------------------------------------------------
void Network::post( ACOM::Network::Request* req, ACOM::Network::OutChannel* ch )
  throw (ACOM::ComFailure)
{
  if (req ==0) 
    ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );

  if ((ch->get_net() != get_subnet()) && (ch->get_net() != this))
    ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );

  Network::Work* work = ThreadRedirec::queue->allocate();
  work->subnet = (ch->get_net() == this ? 0 : get_subnet());
  work->net    = (ch->get_net() == this ? this : 0);
  work->ch     = ch;
  work->cc     = 0;
  work->req    = req;
//  std::cout << "[ICS]:: Network::post: req=" << (long)req << std::endl;
//  std::cout << "[ICS]:: Network::post: work req=" << (long)work->req << std::endl;
  ThreadRedirec::queue->push( work );
}



// ---------------------------------------------------------------------------
void Network::OutChannel::post( ACOM::Network::Request* req )
  throw (ACOM::ComFailure)
{
  get_net()->post( req, this );
}



// ---------------------------------------------------------------------------
Network::OutChannel::OutChannel( )
 : _tag(0)
{
}



// ---------------------------------------------------------------------------
Network::OutChannel::~OutChannel()
{
}



// ---------------------------------------------------------------------------
void Network::OutChannel::init
  ( Network* net, uint32_t tag )
{
  ACOM::Network::OutChannel::init(net, 0);
  _tag = tag;
  
  /* register it to the network */
  PTH::CriticalGuard critical( net->_access );
  LHashMap<uint32_t, OutChannel*>::iterator curr = net->_allchannels.find( _tag );
  if (curr !=0) {
    ACOM::throwit( ACOM::ComFailure( "channel's tag already exist", 
                                     ACOM::ComFailure::BAD_ADDRESS) );
  }
  net->_allchannels.insert( _tag, this );
}



// -----------------------------------------------------------------------------
void Network::OutChannel::destroy( )
{
  Network* net = (Network*)get_net();
  PTH::CriticalGuard critical( net->_access );
  LHashMap<uint32_t, OutChannel*>::iterator curr = net->_allchannels.find( _tag );
  if (curr ==0) {
    ACOM::throwit( ACOM::ComFailure( "channel's tag do not exist",
                                     ACOM::ComFailure::BAD_ADDRESS) );
  }
  net->_allchannels.erase( curr );
}



// -----------------------------------------------------------------------------
void RedirectServiceCallBack::done( ACOM::Network::Request* r )
{
  if (_req !=0) {
    if ((_req->_ext_header !=0) && (_req->_size_eh > 0))
      _req->get_net()->deallocate_data( _req->_ext_header );
    _req->notify( );
  }
  if (_call_cc !=0) {
//    std::cout << "callback-post/";
    _call_cc->release();
  }
  // allocated on call_context delete this;
}



// -----------------------------------------------------------------------------
void RedirectService::initialize( Network* net )
{
  _attachnet = net;
}



// -----------------------------------------------------------------------------
void RedirectService::server_call( ACOM::Network::CallContext* call_cc ) 
  throw (ACOM::ComFailure)
{
//std::cout << _attachnet->get_rank() << ":: IN " << __PRETTY_FUNCTION__ << std::endl;
  Network::Header* xheader = (Network::Header*)call_cc->ext_header;
  ACOM_ASSERT_DEBUG( xheader !=0, "bad assertion");

  ACOM_ASSERT_DEBUG( call_cc->_refcount >0, "bad assertion");
  Network::Work* work = ThreadRedirec::queue->allocate();
  work->net    = _attachnet;
  work->subnet = 0;
  work->ch     = _attachnet->get_channel( xheader->tag );
  work->cc     = call_cc;
  work->req    = 0;
  ACOM_ASSERT_DEBUG( call_cc->_refcount >0, "bad assertion");
  ThreadRedirec::queue->push( work );
//std::cout << _attachnet->get_rank() << ":: OUT " << __PRETTY_FUNCTION__ << std::endl;
}



// -----------------------------------------------------------------------------
void RedirectService::server_get_iovect(
  ACOM::Network::CallContext* call_cc
) throw (ACOM::ComFailure)
{
//std::cout << _attachnet->get_rank() << ":: IN " << __PRETTY_FUNCTION__ << std::endl;
  Network::Header* xheader = (Network::Header*)call_cc->ext_header;
  ACOM_ASSERT_DEBUG( xheader !=0, "bad assertion");
  ACOM::Service* userserv = ACOM::Service::Manager.resolv( xheader->saveid );
  ACOM_ASSERT_DEBUG( call_cc->_refcount >0, "bad assertion");
  userserv->server_get_iovect( call_cc );
//std::cout << _attachnet->get_rank() << ":: OUT " << __PRETTY_FUNCTION__ << std::endl;
}


} // namespace ICS

