// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_pipeline_network.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all pipeline algorithms
// Please read node in .cpp file.
// $Id: ics_pipeline_network.cpp,v 1.6 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_pipeline_network.h>

namespace ICS {
// --------------------------------------------------------------------------
/** Constructor
*/
PipelineNetwork::PipelineNetwork()
  : Network()
{
  set_property("name",     "PIPELINENET");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("PIPELINENET");
}

// --------------------------------------------------------------------------
/** Destructor
*/
PipelineNetwork::~PipelineNetwork()
{
}


// ---------------------------------------------------------------------------
int PipelineNetwork::initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
  throw (ACOM::NetworkError)
{
  Network::initialize(argc, argv, netInit);

  _service = new PipelineRedirectService;
  _service->initialize( this );
  ACOM::Service::Manager.bind( ACOM::ServiceManager::ICS_LNS, ServiceNumber::ICS_REDIRECT_PIPELINE, _service );  
  return 0;
}  
  

// ---------------------------------------------------------------------------
ACOM::Network::OutChannel* PipelineNetwork::get_default_route( LocalNodeNo peer )
  throw(ACOM::NetworkError)
{
  PTH::CriticalGuard critical( _access );
  LHashMap<uint32_t, Network::OutChannel*>::iterator curr = _allchannels.find( peer );
  if (curr ==0) {
    ACOM::throwit( ACOM::NetworkError( "channel do not exist" ) );
  }
  return curr->second;
}  

// ---------------------------------------------------------------------------
PipelineNetwork::OutChannel* PipelineNetwork::open_channel( uint32_t tag, LocalNodeNo peer, ACOM::Service* packet_service )
       throw(ACOM::ComFailure)
{
  return 0;
}

// ---------------------------------------------------------------------------
PipelineNetwork::OutChannel::OutChannel( )
 : Network::OutChannel(), _serial(0), _size(4096), _packet_service(0)
{
}


// ---------------------------------------------------------------------------
PipelineNetwork::OutChannel::~OutChannel()
{
}


// ---------------------------------------------------------------------------
void PipelineNetwork::OutChannel::init
  ( PipelineNetwork* net, uint32_t tag, LocalNodeNo peer, ACOM::Service* cbk )
{
  _peer = peer;
  _packet_service = cbk;
  _serial         = 0;

  Network::OutChannel::init(net, tag);
}


// ---------------------------------------------------------------------------
ACOM::Network::Request* PipelineNetwork::OutChannel::allocate_request( ) throw (ACOM::BadAlloc)
{
  ACOM::Network::Request* req = _net->allocate_request();
  return req;
}


// ---------------------------------------------------------------------------
void PipelineNetwork::OutChannel::post_init( Request* r )
{
}


// ---------------------------------------------------------------------------
struct ICSPipelineCallBack : public ACOM::Callback {
  ICSPipelineCallBack( volatile long* cd, PipelineNetwork* net ) : _cd(cd), _net(net) {}
  void done( ACOM::Network::Request* )
  {
    _net->_access.lock();
    --*_cd;
    _net->_access.unlock();
    _net->_cond.broadcast();
    std::cout << _net->rank() << "::Callback called :" << std::endl;
  }
private:
  volatile long*   _cd;
  PipelineNetwork* _net;
};


// ---------------------------------------------------------------------------
/* use header of request sended as following :
   - tag    : name of the channel
   - idxsrc : serial number (>0 : immediate, <0 : deferred)
   - saveid : id of the user service
   - no     : serial number of the each packet sended, -1 if it is the last
*/
void PipelineNetwork::OutChannel::post( ACOM::Network::Request* r ) throw (ACOM::ComFailure)
{
#if 0
  PipelineNetwork* net = (PipelineNetwork*)r->get_net();

  int serial_msg; 
  int serial_packet = 0; 
  { PTH::CriticalGuard critical( net->_access ); serial_msg = ++_serial; }
  
  /* decompose immediate data in packet of size at most _size */
  ACOM::Network::OutChannel* channel = net->_subnet->get_default_route( _peer );
  volatile long posted = 0;
  ICSPipelineCallBack mycbk( &posted, net );

  ACOM::WSMessage msg ( net->_subnet, net->_service,  &mycbk, 1 );
  msg.get_request()->_header.tag    = _tag;          /* the name of this pipeline channel */
  msg.get_request()->_header.idxsrc = serial_msg;    /* immediate data */
  msg.get_request()->_header.saveid = r->_header.id;
  msg.get_request()->_header.no     = serial_packet;

  size_t size_packed =0;
  
  /* send immediate data */
  for( ACOM::IOVectSeq::IteratorEntry ie= r->_datastream.beginEntry();
        ie != r->_datastream.endEntry();
        ++ie)
  {
    void* data = ie->data;
    size_t size_topack = ie->len;
    
    while (size_topack >0)
    {
      if (size_packed + size_topack < _size) {
        msg->put( data, size_topack );
        size_packed += size_topack;
        size_topack -= size_topack;
      } else {
        /* fill last packet */
        size_t size_thispacket = _size - size_packed;
        msg->put( data, size_thispacket );
        { PTH::CriticalGuard critical( net->_access ); ++posted; }
        channel->post( msg );

        /* realloc new message */
        msg = ACOM::WSMessage ( net->_subnet, net->_service,  &mycbk, 1 );
        msg.get_request()->_header.tag    = serial_msg;      /* serial number of this post */
        msg.get_request()->_header.idxsrc = 1;               /* immediate data */
        msg.get_request()->_header.no     = ++serial_packet;
        msg.get_request()->_header.saveid = r->_header.id;
        
        data = ((char*)data) + size_thispacket;
        size_packed += size_thispacket;
        size_topack -= size_thispacket;
      }
    } // while
  }
  
  /* post last packet for immediate message */
  if (r->_iovstream.beginEntry() == r->_iovstream.endEntry())
    msg.get_request()->_header.no  = (uint32_t)-1;             /* it is the last */

  { PTH::CriticalGuard critical( net->_access ); ++posted;  }
  channel->post( msg );
  
  /* wait all previous post */
  { PTH::CriticalGuard waitcritical( &net->_access );
    while (posted != 0) 
      net->_cond.wait(net->_access);
  }
  if (r->_iovstream.beginEntry() == r->_iovstream.endEntry())
  {
    r->notify();
    return;
  }

  /* realloc new message for defered data */
  msg = ACOM::WSMessage ( net->_subnet, net->_service,  &mycbk, 1 );
  msg.get_request()->_header.tag    = serial_msg;      /* serial number of this post */
  msg.get_request()->_header.idxsrc = 2;               /* deferred data */
  msg.get_request()->_header.no     = ++serial_packet;
  msg.get_request()->_header.saveid = r->_header.id;

  // - for the defered data part:
  for( ACOM::IOVectSeq::IteratorEntry ie = r->_iovstream.beginEntry();
        ie != r->_iovstream.endEntry();
        ++ie)
  {
    void* data = ie->data;
    size_t size_topack = ie->len;
    
    while (size_topack >0)
    {
      if (size_packed + size_topack < _size) {
        msg->put( data, size_topack );
        size_packed += size_topack;
        size_topack -= size_topack;
      } else {
        /* fill last packet */
        size_t size_thispacket = _size - size_packed;
        msg->put( data, size_thispacket );
        { PTH::CriticalGuard critical( net->_access ); ++posted; }
        channel->post( msg );

        /* realloc new message */
        msg = ACOM::WSMessage ( net->_subnet, net->_service,  &mycbk, 1 );
        msg.get_request()->_header.tag    = serial_msg;      /* serial number of this post */
        msg.get_request()->_header.idxsrc = 2;               /* deferred data */
        msg.get_request()->_header.no     = ++serial_packet;
        msg.get_request()->_header.saveid = r->_header.id;
        
        data = ((char*)data) + size_thispacket;
        size_packed += size_thispacket;
        size_topack -= size_thispacket;
      }
    } // while
  }
  
  /* post last packet for immediate message */
  msg.get_request()->_header.no = (uint32_t)-1;             /* it is the last */

  { PTH::CriticalGuard critical( net->_access ); ++posted;  }
  channel->post( msg );
  
  /* wait all previous post */
  { PTH::CriticalGuard waitcritical( &net->_access );
    while (posted != 0) 
      net->_cond.wait(net->_access);
  }
  r->notify();
#endif
}



// -----------------------------------------------------------------------------
/* use header of request sended as following :
   - tag    : serial number of this post
   - idxsrc : 1 for immediate data, 2 for deferred data
   - saveid : id of the user service
   - no     : serial number of the each packet sended, -1 if it is the last
*/
void PipelineRedirectService::call ( 
      ACOM::ComFailure::Code e,      // if any 
      ACOM::Network::Receipt* r,      
      ACOM::Network::Header* h,
      ACOM::IOVectEntry* iov,
      ACOM::IOVectSize   iov_sz,
      void* cc
) throw (ACOM::ComFailure)
{
#if 0
  ACOM::Service* userserv = ACOM::Service::Manager.resolv( h->saveid );
  return userserv->call( e, r, h, iov, iov_sz, cc );
#endif
}


// -----------------------------------------------------------------------------
void* PipelineRedirectService::get_iovect(
  ACOM::Network::Receipt* r,
  ACOM::Network::Header* h, 
  ACOM::IOVectEntry* iov, ACOM::IOVectSize iov_sz
)
{
#if 0
  ACOM::Service* userserv = ACOM::Service::Manager.resolv( h->saveid );
  return userserv->get_iovect( r, h, iov, iov_sz );
#endif
  return 0;
}

 
}

