// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_pipeline_network.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all pipeline algorithms
// Please read node in .cpp file.
// $Id: ics_pipeline_network.h,v 1.6 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_PIPELINE_NETWORK_H_
#define _INUK_PIPELINE_NETWORK_H_
#include <ics_network.h>


namespace ICS {

/* fwd declaration */
class PipelineNetwork;


// -----------------------------------------------------------------------------
/** PipelineRedirectService : service registered to redirect 
 * call to a user service while permited to add call to internal
 * function. 
 */
class PipelineRedirectService : public RedirectService {
public:

protected:
  /* appele apres que le message soit arrive 
   */
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,      
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void* cc
  ) throw (ACOM::ComFailure);

  /* decodage de l'address pour stocker le message
   */
  void* get_iovect(
    ACOM::Network::Receipt* r,
    ACOM::Network::Header* header, 
    ACOM::IOVectEntry* vector, ACOM::IOVectSize count
  );
};



// -----------------------------------------------------------------------------
/** PipelineNetwork with pipeline capabilities over channel.
 * Such network relies on an existing network to handle point 2 point communication.
 * Allocation of request and channel is specific to such network in order
 * to add extra information as immediate data. 
 */
class PipelineNetwork: public Network {
public:
  // --------------------------------------------------------------------------
  /** Constructor
  */
  PipelineNetwork();

  /** Destructor
  */
  ~PipelineNetwork();

  /** Initialize a network 
  */
  int initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
    throw (ACOM::NetworkError);
    
  // ---------------------------------------------------------------------------
  /** extended classes for broadcast purpose.
  */
  class OutChannel: public Network::OutChannel {
  protected:
    /* Post should be redefined in derived class */
    //void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);

    /** Constructor
    */
    OutChannel( );

    /** destructor
    */
    ~OutChannel();
    
  public:
    /** see PipelineNetwork::open_channel
    */
    virtual void init( PipelineNetwork* net, uint32_t tag, LocalNodeNo peer, ACOM::Service* cbk );

    void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);

  protected:
    /* Allocate a request : add information (immediate data) needed 
       by the algorithm/channel
    */
    ACOM::Network::Request* allocate_request( ) throw (ACOM::BadAlloc);

    /// 
    void post_init( Request* );

  public: /* to all derived service of forwarding */
    uint16_t       _serial;          /* serial number of message */
    uint16_t       _size;            /* size of the packet */
    ACOM::Service* _packet_service;  /* service call at each recv of packet */
  };


  // ---------------------------------------------------------------------------
protected:
  ACOM::Network::OutChannel* get_default_route( LocalNodeNo peer )
    throw(ACOM::NetworkError);

public:
  // ---------------------------------------------------------------------------
  /* extended methods for broadcast purpose.
  */
  /** Open a pipeline channel between two node.
      @param packet_service(IN) the service that is called when a new packet arrived
      \warning ajouter les aliasing d'adresse de noeud
  */
  virtual OutChannel* open_channel( uint32_t tag, LocalNodeNo peer, ACOM::Service* packet_service )
       throw(ACOM::ComFailure);

private:
  friend class PipelineRedirectService;
  friend class ICSPipelineCallBack;
  friend class OutChannel;
};
 


/*
 * inline definitions
 */
 
 
}

#endif
