// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// $Id: ics_reduce.cpp,v 1.3 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_reduce.h>
#include <iostream>

namespace ICS {


// -----------------------------------------------------------------------------
/* */
void ReduceService::notify_cc( Trigger_set::iterator curr )
{
  com_context* ctxt = curr->second;
  if (ctxt->wait_post != ctxt->wait_recv)
    return;

  /* delete temporary data_recv */
  if (ctxt->data_recv !=0) {
    ctxt->fmt->destroy( ctxt->data_recv, ctxt->count_recv );
    _sizefree += ctxt->fmt->get_size()* ctxt->count_recv;
    ctxt->data_recv  = 0;
    ctxt->count_recv = 0;
  }

  /* call the call back function */
  if (ctxt->cb_func !=0) 
    (*ctxt->cb_func)(ctxt->cb_arg1, ctxt->cb_arg2);

  /* erase data if needed */
  if (!ctxt->is_join) 
  {
    deallocate_cc( ctxt );
    _postrecv.erase( curr );
  }
}



// -----------------------------------------------------------------------------
/* Send a io vector by packets such that the whole size of data
   of each packets is no more than szbloc.
   If szbloc == -1, then the whole io vector at once time.
*/
void ReduceService::postsend_data
  ( ACOM::Network::Request* req, int tag, 
    const ACOM::Format* fmt, void* data, size_t count,
    const CumulOperatorFormat* op
  )
{
  int fid  = fmt->get_id();
  int opid = op->get_id();
  req->pack(&tag, sizeof(int));
  req->pack(&fid, sizeof(int));
  req->pack(&opid, sizeof(int));
  req->pack(&count, sizeof(size_t));

  if (count >0) {
    ACOM::OStream msg ( req->get_datastream(), req->get_iovstream() );
    fmt->write(msg, data, count);
    
#if 0
    int* d = (int*)data;
    std::cout << "Send d=[ ";
    for (size_t i=0; i<count; ++i) {
      std::cout << d[i] << ' ';
    }
    std::cout << "] " << std::endl;
#endif
  }
}


// -----------------------------------------------------------------------------
/* Post recv of a iov splited by bloc of size at most szbloc bytes
*/
void ReduceService::postrecv_data
  ( int tag, bool isjoinable, size_t wait_torecv,
    const ACOM::Format* fmt, void** data, size_t* count,
    const CumulOperatorFormat* op,
    void (*cbf)(void*, void*), void* cba1, void* cba2
  )
{
  if ((count ==0) || (data ==0)) {
    ACOM::throwit( ACOM::InvalidArgument() );
  }
  
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

  /* search if receive already occurs */
  Key key (tag);
  curr = _postrecv.find( key );

  if (curr == 0) 
  {
#if 0
  std::cout << "new cc with tag=" << tag << std::endl;
#endif
    ctxt             = allocate_cc();
    curr             = _postrecv.insert( key, ctxt);
    ctxt->fmt        = fmt;
    ctxt->count_post = (size_t)count;
    ctxt->data_post  = (void*)data;
  }
  else 
  {
    /* ctxt already exist: one or serveral messages have been arrived before.
       Result of partial accumulation was done in data_post which is a
       pointer to a temporary, make accumulation and replace data_post
    */
    ctxt             = curr->second;
#if 0
  std::cout << "already exists= cc with tag=" << curr->first.tag << std::endl;
#endif
    if (*data ==0) 
    {
      /* (data,count) get the temporary, no reduction ! */
      *data  = ctxt->data_post;
      *count = ctxt->count_post;
    } 
    else 
    {
      if (*count != ctxt->count_post) {
#if 0
        std::cout << "Recv data with size :\nctxt->count_post=" << ctxt->count_post
                  << ", request size=" << *count << "\nctxt->count_recv=" << ctxt->count_recv
                  << std::endl;
#endif
        ACOM::throwit( ACOM::Length() );
      }
      if (ctxt->count_post ==1)
        ctxt->op->cumul( *data, ctxt->fmt, ctxt->data_post );
      else 
        ctxt->op->cumul( *data, ctxt->fmt, ctxt->data_post, ctxt->count_post );
      
      /* supress the temporary: the recv buff with be destroy only on  root site 
         at the end of reduction
      */
      ctxt->fmt->destroy( ctxt->data_post, ctxt->count_post );
      _sizefree += ctxt->fmt->get_size() * ctxt->count_post;

      /* Now count_post and data_post are reference to pointer on their data */
      ctxt->data_post  = (void*)data;
      ctxt->count_post = (size_t)count;
    }
    ++_nbmatchpost;
  }
  
  /* common set up */
  ctxt->is_join    = isjoinable;
  ctxt->cb_arg1    = cba1;
  ctxt->cb_arg2    = cba2;
  ctxt->cb_func    = cbf;
  ctxt->wait_post  = wait_torecv;

  /* notify here if not joinable */
  if ((ctxt->wait_recv == ctxt->wait_post) && (!ctxt->is_join))
    notify_cc( curr );
}




// -----------------------------------------------------------------------------
ReduceService::com_context* ReduceService::wait_data( int tag )
{
  Trigger_set::iterator curr =  0;
  com_context* ctxt = 0;

  /* find tag in _postrecv */
  Key key (tag);
  curr = _postrecv.find( key );
  if (curr == 0) 
  {
    std::cerr << "Cannot find ctxt with key(" << tag << ")" << std::endl;
    throwit( ACOM::InvalidArgument() );
  }
  ctxt = curr->second;
  
  /* infinite loop */
  ctxt->is_wait = true;
  while (ctxt->wait_recv != ctxt->wait_post) {
    _cond.wait(_access);
  }

  /* notify the com_context */
  if (ctxt->is_join)
    notify_cc( curr );

  /* supress entry in hash table */
#if 0
  std::cout << "delete cc with tag=" << curr->first.tag << std::endl;
#endif
  _postrecv.erase( curr );

  return ctxt;
}


// -----------------------------------------------------------------------------
void ReduceService::call ( 
     ACOM::ComFailure::Code e,      
     ACOM::Network::Receipt* r,
     ACOM::Network::Header* header,
     ACOM::IOVectEntry* iov,
     ACOM::IOVectSize   iovsz,
     void* cc
) throw (ACOM::ComFailure)
{
  if (e != ACOM::ComFailure::OK) {
    std::cerr << "Error in msg='" //<< ACOM::ComFailure::what(e) 
              << "', #iov" << iovsz  << ", len[0]" << (iovsz > 0 ? iov[0].len : 0)
              << std::endl;
    ACOM::throwit( ACOM::ComFailure(e) ) ;
  }

  Trigger_set::iterator curr = (Trigger_set::iterator)cc;
  com_context* ctxt = curr->second;

  /* if count_post/data_post are not yet set, then set them to the value of 
     count_recv and data_recv. Else do reduction.
  */
  if (ctxt->wait_post ==0) 
  {
    if (ctxt->data_post ==0) {
      ctxt->data_post  = ctxt->data_recv;
      ctxt->count_post = ctxt->count_recv;
      ctxt->data_recv  = 0;  /* force reallocation of buffer for next message, if any */
      ctxt->count_recv = 0;
#if 0
std::cout << "In ::call: first recv, no post" << std::endl;
#endif
    } else {
#if 0
std::cout << "In ::call: > #1 recv, no post" << std::endl;
#endif
      if (ctxt->count_recv ==1)
        ctxt->op->cumul( ctxt->data_post, ctxt->fmt, ctxt->data_recv );
      else 
        ctxt->op->cumul( ctxt->data_post, ctxt->fmt, ctxt->data_recv, ctxt->count_recv );
      }
  } else {
#if 0
std::cout << "In ::call: post" << std::endl;
#endif
    if (ctxt->count_recv ==1)
      ctxt->op->cumul( *(void**)ctxt->data_post, ctxt->fmt, ctxt->data_recv );
    else 
      ctxt->op->cumul( *(void**)ctxt->data_post, ctxt->fmt, ctxt->data_recv, ctxt->count_recv );
  }
  
  ++ctxt->wait_recv;
  if (ctxt->wait_recv == ctxt->wait_post) 
  {
    if (ctxt->is_wait)
      _cond.signal();
    if (!ctxt->is_join)
      notify_cc( curr );
  }

  /* lock in get_iovect */
  _access.unlock(); 
}


// -----------------------------------------------------------------------------
/* decoding the address for the message and put it into temporary storage
   ctxt->data_recv.
 */
void* ReduceService::get_iovect(
  ACOM::Network::Receipt* r,
  ACOM::Network::Header* header, 
  ACOM::IOVectEntry* iov, 
  ACOM::IOVectSize iovsz
)
{
  /* extract immediate data */
  int tag;
  int fid;
  int opid;
  size_t count;
  r->unpack(&tag, sizeof(int));
  r->unpack(&fid,sizeof(int));
  r->unpack(&opid,sizeof(int));
  r->unpack(&count,sizeof(size_t));

#if 0
  std::cout << "recv tag  =" << tag << std::endl;
  std::cout << "recv fid  =" << fid << std::endl;
  std::cout << "recv opid =" << opid << std::endl;
  std::cout << "recv count=" << count << std::endl;
#endif

  Key key(tag);
  /* some data */
  com_context* ctxt = 0;
  const ACOM::Format* fmt =0;
  void* data =0;
 
  /* unlock in call */
  _access.lock(); 

  /* find tag in _postrecv */
  Trigger_set::iterator curr = _postrecv.find( key );

  /* common set up */
  fmt = ACOM::Format::get_format( fid );

  if (curr == 0) 
  {
#if 0
    std::cout << "Recv first" << std::endl;
#endif
    /* Here the message arrived before the post to was called
     */
    ctxt  = (com_context*) allocate_cc();

    /* allocate temporary for receiving message */
    if (count >0) {
      data = ctxt->data_recv = fmt->create(count);
      _sizealloc += fmt->get_size() * count;
    }
    ctxt->count_recv = count;
    curr = _postrecv.insert(key, ctxt);
  } 
  else 
  {

    /* Here the message arrived after the post to received reallocated
       the temporary storage if needed
    */
    ++_nbmatchrecv;
    ctxt = curr->second;

#if 0
    std::cout << "Recv after, ctxt->count_recv=" << ctxt->count_recv << std::endl;
    std::cout << "Recv after, ctxt->data_recv =" << ctxt->data_recv << std::endl;
    std::cout << "Recv after, ctxt->count_post=" << ctxt->count_post << std::endl;
    std::cout << "Recv after, ctxt->data_post =" << ctxt->data_post << std::endl;
#endif

    /* if buffer has no enough space, reallocated it */
    if (ctxt->count_recv != count) 
    {
      if (ctxt->data_recv !=0) {
        fmt->destroy( ctxt->data_recv, ctxt->count_recv);
        _sizefree += fmt->get_size()* ctxt->count_recv;
      }
      data = ctxt->data_recv = fmt->create(count);
      _sizealloc += fmt->get_size() * count;
      ctxt->count_recv = count;
    } else {
      data = ctxt->data_recv;
    }
  }
  
  /* common set up */
  ctxt->fmt = fmt;
  ctxt->op  = CumulOperatorFormat::get_cumul_format( opid );

#if 0
  std::cout << "Recv before unpack" << std::endl;
  std::cout << "Recv, ctxt->count_recv=" << ctxt->count_recv << std::endl;
  std::cout << "Recv, ctxt->data_recv =" << ctxt->data_recv << std::endl;
  std::cout << "Recv, ctxt->count_post=" << ctxt->count_post << std::endl;
  std::cout << "Recv, ctxt->data_post =" << ctxt->data_post << std::endl;
  std::cout << "Recv, ctxt->op id =" << ctxt->op->get_id() << std::endl;
  std::cout << "Recv, *ctxt->count_post=" << *((size_t*)ctxt->count_post) << std::endl;
  std::cout << "Recv, *ctxt->data_post =" << *((void**)ctxt->data_post) << std::endl;
  std::cout << "Data post=[ ";
  for (size_t i=0; i<count; ++i) {
    std::cout << (*((int**)ctxt->data_post))[i] << ' ';
  }
  std::cout << "] " << std::endl;
#endif

  /* read the data in the temporary storage data_recv */
  if (count >0) 
  {
    /* Unpack the data and get iovect description
    */
    ACOM::IStream fin ( r->get_datastream(), iov, iovsz );
    fmt->read(fin, data, count);
  }
  return curr;
}



// -----------------------------------------------------------------------------
/* */
void ReduceService::reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag
  )
{
  handler wc = post_reduce( 
                              recv_data, 
                              send_data, 
                              count,
                              fmt,
                              op,
                              sz_ranks, 
                              ranks, 
                              index_self, 
                              index_dest, 
                              tag
  );
  wait_reduce( wc );
}


}
