// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// $Id: ics_reduce.h,v 1.2 2003/07/15 17:22:44 thierry Exp $
//
// ==========================================================================
#ifndef _INUK_REDUCE_H_
#define _INUK_REDUCE_H_
#include <acom>
#include <ics_base.h>
#include <ics_redops.h>

namespace ICS {

/* 
 * Reduce base service for inuktitut.
 * All process in ranks are involved into the collective communication.
 * Data are reduced using the reduction operator op into the variable
 * pointed by recv_data on processus on site ranks[index_dest].
 *
 */
class ReduceService: public BaseService {
public:
  typedef void* handler;
  
  /** */
  virtual handler post_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag
  ) = 0;

  /** */
  virtual void wait_reduce(
    handler wc
  ) = 0;

  /** */
  virtual void async_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  ) = 0;

  /** short cut for post_reduce/wait_reduce */
  void reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag
  );

protected:
 
  /** notify a com_context
   */
  void notify_cc( Trigger_set::iterator curr );

  /* Form a request
  */
  void postsend_data( ACOM::Network::Request* req, int tag, 
                      const ACOM::Format* fmt, void* data, size_t count,
                      const CumulOperatorFormat* op
                    );

  /* post recv for an io vector of bytes by splitting it into packet of size at most sz_bloc 
  */
  void postrecv_data( int tag, bool is_join, size_t wait_torecv,
                      const ACOM::Format* fmt, void** data, size_t* count,
                      const CumulOperatorFormat* op,
                      void (*cb_func)(void*, void*), void* cb_arg1, void* cb_arg2
                  );

  /* wait for a packet for a particular (tag,serial)
  */
  com_context* wait_data( int tag );

  /* appele apres que le message soit arrive 
   */
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,      
       ACOM::Network::Header* header,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iovsz,
       void* cc
  ) throw (ACOM::ComFailure);

  /* decodage de l'address pour stocker le message
   */
  void* get_iovect(
    ACOM::Network::Receipt* r,
    ACOM::Network::Header* header, 
    ACOM::IOVectEntry* vector, ACOM::IOVectSize count
  );

protected:
};


/*
 * inline definitions
 */
 
}

#endif
