// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce_all_2_one.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of flat reduce
// $Id: ics_reduce_all_2_one.cpp,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_reduce_all_2_one.h>
#include <iostream>

namespace ICS {

// -----------------------------------------------------------------------------
/** */
ReduceAll2One::handler ReduceAll2One::post_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag
)
{
  ++_nbcalled;
  if (sz_ranks <=1) return 0;

  if (index_self != index_dest) 
  { 
#if 0
  std::cout << " data to send = ["; 
  for (int i=0; i < count; ++i)
    std::cout << ((int*)send_data)[i] << ' ';
  std::cout << "]" << std::endl;
#endif
    /* Form a message */
    ACOM::WriteAndSignal<ReduceAll2One> ws (_net, this, 0);
    postsend_data( ws.get_request(), tag, fmt, send_data, count, op );
    /* get a channel */
    ACOM::Network::OutChannel* channel = _net->get_default_route( ranks[index_dest] );
    /* Send it */
    ws.post(channel);
    return 0;
  } 

  /* should receive a message */
  PTH::CriticalGuard cg(_access); // - begin critical section
#if 0
  std::cout << " data to reduce = ["; 
  for (int i=0; i < count; ++i)
    std::cout << ((int*)send_data)[i] << ' ';
  std::cout << "]" << std::endl;
  std::cout << " buffer to recv = ["; 
  for (int i=0; i < count; ++i)
    std::cout << ((int*)recv_data)[i] << ' ';
  std::cout << "]" << std::endl;
  std::cout << "tag=" << tag << std::endl;
#endif
  postrecv_data( tag, true, sz_ranks-1, fmt, &recv_data, &count, op, 0, 0, 0);
  op->cumul( recv_data, fmt, send_data, count);
#if 0
  std::cout << " buffer to recv after local reduction = ["; 
  for (int i=0; i < count; ++i)
    std::cout << ((int*)recv_data)[i] << ' ';
  std::cout << "]" << std::endl;
#endif
  return (handler*)tag;
}



// -----------------------------------------------------------------------------
/** */
void ReduceAll2One::wait_reduce(
  handler wc
)
{
  if (wc ==0) return;
  int tag = (int)wc;
#if 0   
  std::cout << " begin wait, with tag =" << tag << std::endl; 
#endif
  PTH::CriticalGuard cg(_access); // - begin critical section
  com_context* ctxt = wait_data( tag );
#if 0
  std::cout << " buffer to recv after reduction = ["; 
  for (int i=0; i < *(size_t*)ctxt->count_post; ++i)
    std::cout << (*(int**)ctxt->data_post)[i] << ' ';
  std::cout << "]" << std::endl;
#endif
  deallocate_cc( ctxt );
}



// -----------------------------------------------------------------------------
/** */
void ReduceAll2One::async_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return;
  
  if (index_self != index_dest) 
  { 
    /* Form a message */
    ACOM::WriteAndSignal<ReduceAll2One> ws (_net, this, 0);
    postsend_data( ws.get_request(), tag, fmt, send_data, count, op );
    /* get a channel */
    ACOM::Network::OutChannel* channel = _net->get_default_route( ranks[index_dest] );
    /* Send it */
    ws.post(channel);

    /* call callback: end of local operation */
    if (cb_func !=0)
      (*cb_func)(cb_arg1, cb_arg2);

    return;
  } 

  /* should receive a message */
  PTH::CriticalGuard cg(_access); // - begin critical section
  postrecv_data( tag, false, sz_ranks-1, fmt, &recv_data, &count, op,  cb_func, cb_arg1, cb_arg2);
  /* Accumulate self data into recv */
  op->cumul( recv_data, fmt, send_data, count);  
}


}
