// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce_alphatree.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of reduce using binomial algorithm
// $Id: ics_reduce_alphatree.cpp,v 1.4 2003/07/18 11:26:00 olobry Exp $
//
// ==========================================================================
#include <ics_conf.h>
#include <ics_reduce_alphatree.h>
#include <math.h>

namespace ICS {

/* Reduction du message (buffer, count), en utilisant l'identificateur tag,
  a tous les processus de rank contenu dans ranks[0] ... ranks[sz-1]. 
  Le processus qui appelle cette fonction a le rank ranks[index_self]. 
  Le processus qui possede la donnee a le rank ranks[index_dest]. 
  Tous les processus qui participe a cette diffusion doivent appeler 
  cette fonction. buffer est alloue sur tous les processus, initialement 
  seul le processus de rank ranks[index_dest] possede une donnee initialisee. 
*/
// -----------------------------------------------------------------------------
/** */
struct RAlphaTree_continuation {
  ReduceAlphaTree*           serv;
  void*                      send_data;
  void**                     recv_data;
  size_t                     count;
  const ACOM::Format*        fmt;
  const CumulOperatorFormat* op;
  int                        sz_ranks;
  int*                       ranks;
  int                        index_self;
  int                        index_dest;
  int                        tag;
  void                       (*cb_func)(void*, void*);
  void*                      cb_arg1;
  void*                      cb_arg2;
  
  /* compute alpha parameters (next data field of the structure) */
  void comp_alpha_param ();
  
  /* internal to alpha tree : */
  int _parent;      /* where to send received data */
  int _offset;      /* offset in index for the sites I'm waiting for */
  int _nbrecv;      /* number of received item */
};


/* */
void RAlphaTree_continuation::comp_alpha_param()
{
  int monRang = index_self;
  _offset = sz_ranks;
  _parent = 0;
  int nOthers;
  int threshold;

  // cached copy of alpha
  double alpha = serv->get_alpha();
  
  while (monRang != 0) 
  {
    nOthers = (int)round(alpha * _offset);
    threshold = _offset - nOthers;
    if(monRang == threshold) {
      _offset = nOthers;
      monRang -= threshold;
    } else {
      if(monRang < threshold)
        _offset = threshold;
      else {
        _offset = nOthers;
        _parent += threshold;
        monRang -= threshold;
      }
    }
  }

  /* count the number of recv to wait */
#if 0
  std::cout << "Rank self= " << ranks[index_self] 
            << std::endl;
  std::cout << "Rank dest= " << ranks[index_dest] 
            << std::endl;
#endif
  int nn = _offset;
  _nbrecv = 0;
  while (nn > 1) {
    nn -= (int)round(alpha*nn);
#if 0
    std::cout << "Recv from : " << ranks[index_self + nn] << std::endl;
#endif
    ++_nbrecv;
  }
#if 0
  std::cout << "ReSend to : " << ranks[_parent] 
            << "------------------------"
            << std::endl;
#endif
}


// -----------------------------------------------------------------------------
/** */
void ReduceAlphaTree::continuation( void* a1, void* a2 )
{
  /* compute the parition of the set of node */
  RAlphaTree_continuation* cc = (RAlphaTree_continuation*)a1;

#if 0
std::cout << "IN :" << __PRETTY_FUNCTION__ << std::endl;
#endif

  /* If I'am the index_dest: do nothing */
  if (cc->index_self != cc->index_dest) 
  {  
    /* Form the message */
    ACOM::WriteAndSignal<ReduceAlphaTree> ws (cc->serv->_net, cc->serv, 0);
    cc->serv->postsend_data( ws.get_request(), cc->tag, cc->fmt, cc->send_data, cc->count, cc->op );
  
    /* Get the route */
    int dest = cc->ranks[ cc->_parent ];
    ACOM::Network::OutChannel* channel = cc->serv->_net->get_default_route( dest );
  
    /* Send it */
    ws.post(channel);
#if 0
  std::cout << " data to Send = ["; 
  for (int i=0; i < cc->count; ++i)
    std::cout << ((int*)cc->send_data)[i] << ' ';
  std::cout << "]" << std::endl;
    std::cout << "Send for reduction tag=" << cc->tag 
              << ", node=" << dest << std::endl;
#endif
    if (cc->count >0) {
      //cc->fmt->destroy( cc->send_data, cc->count );
      cc->send_data =0;
      cc->count =0;
    }
  }
  else 
  {
#if 0
  std::cout << "(Root) data to Recv = ["; 
  for (int i=0; i < cc->count; ++i)
    std::cout << ((int*)cc->recv_data)[i] << ' ';
  std::cout << "]" << std::endl;
#endif
  }
  
  if (cc->cb_func !=0)
    (*cc->cb_func)( cc->cb_arg1, cc->cb_arg2 );
  delete cc;

#if 0
std::cout << "OUT :" << __PRETTY_FUNCTION__ << std::endl;
#endif
}
 


// -----------------------------------------------------------------------------
/** */
ReduceAlphaTree::handler ReduceAlphaTree::post_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag
  )
{
  ++_nbcalled;
  if (sz_ranks <=1) return 0;
  
  RAlphaTree_continuation* cc = new RAlphaTree_continuation;
  cc->serv       = this;
  cc->sz_ranks   = sz_ranks;
  cc->ranks      = ranks;
  cc->index_self = index_self;
  cc->index_dest = index_dest;
  cc->tag        = tag;
  cc->fmt        = fmt;
  cc->op         = op;
  cc->recv_data  = 0;
  cc->send_data  = send_data;
  cc->count      = count;
  cc->cb_func    = 0;
  cc->cb_arg1    = 0;
  cc->cb_arg2    = 0;

  cc->comp_alpha_param();

  if (cc->_nbrecv >0) 
  {
    /* I need to recv before re-send to cc->_parent, recv on my data to send !
        TG: WARN !!!! Do not use send_data: assumed to be const
    */
    /* Allocate a temporary store in cc->send_data, if I'm not the master
     */
    if (index_self == index_dest) {
      cc->recv_data = &recv_data;
      op->cumul( recv_data, fmt, send_data, count);  
#if 0
std::cout << " (Root) Initial data = ["; 
for (int i=0; i < cc->count; ++i)
  std::cout << ((int*)cc->recv_data)[i] << ' ';
std::cout << "]" << std::endl;
#endif
    } else {
      if (count >0) {
        cc->send_data = fmt->create( count );
        fmt->copy( cc->send_data, send_data, count);
      }
      cc->recv_data = &cc->send_data;
#if 0
std::cout << " (Node) Initial data = ["; 
for (int i=0; i < cc->count; ++i)
  std::cout << ((int*)cc->send_data)[i] << ' ';
std::cout << "]" << std::endl;
#endif
    }
    
    /* begin critical section */
    PTH::CriticalGuard cg(_access); 

    /* post recv: continuation will do remainder of the algorithm */
    postrecv_data( tag, true, cc->_nbrecv, fmt, cc->recv_data, &cc->count, op, &ReduceAlphaTree::continuation, cc, 0);
#if 0
    std::cout << "Post reduce recv for tag=" << tag << std::endl;
#endif
    return cc;
  }
  
  /* Here nothing to do : only post a message to my parent 
  */
  ReduceAlphaTree::continuation(cc, 0);
  return 0;
}



// -----------------------------------------------------------------------------
/** */
void ReduceAlphaTree::wait_reduce(
  handler wc
)
{
  if (wc ==0) return; // not the destination site
  PTH::CriticalGuard cg(_access); // - begin critical section
  RAlphaTree_continuation* cc = (RAlphaTree_continuation*)wc;
  com_context* ctxt = wait_data( cc->tag);
  deallocate_cc( ctxt );
}



// -----------------------------------------------------------------------------
/** */
void ReduceAlphaTree::async_reduce(
    void*&                     recv_data, 
    void*                      send_data, 
    size_t                     count,
    const ACOM::Format*        fmt,
    const CumulOperatorFormat* op,
    int                        sz_ranks, 
    int*                       ranks, 
    int                        index_self, 
    int                        index_dest, 
    int                        tag,
    void                       (*cb_func)(void*, void*),
    void*                      cb_arg1,
    void*                      cb_arg2
  )
{
}


}
