// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reduce_network.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// Please read node in .cpp file.
// $Id: ics_reduce_network.cpp,v 1.7 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#include <ics_reduce_network.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
ReduceNetwork::ReduceNetwork()
  : Network()
{
  set_property("name",     "REDUCENET");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("REDUCENET");
}

// --------------------------------------------------------------------------
/** Destructor
*/
ReduceNetwork::~ReduceNetwork()
{
}



// ---------------------------------------------------------------------------
int ReduceNetwork::initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
  throw (ACOM::NetworkError)
{
  Network::initialize(argc, argv, netInit);
  return 0;
}  


  
// ---------------------------------------------------------------------------
ReduceNetwork::Rendezvous* ReduceNetwork::allocate_rendezvous()
{
  return new ReduceNetwork::Rendezvous;
}



// ---------------------------------------------------------------------------
void ReduceNetwork::deallocate_rendezvous(ReduceNetwork::Rendezvous* p)
{
  delete p;
}



// ---------------------------------------------------------------------------
ReduceNetwork::Rendezvous* ReduceNetwork::get_rendezvous(uint32_t tag, uint32_t serial)
{
  Rendezvous* rdv =0;
  Key key (tag,serial);
  _access.lock();
  LHashMap<Key, Rendezvous*>::iterator curr = _allrdv.find( key );
  if (curr ==0) {
    rdv= allocate_rendezvous();
    rdv->key = key;
    _allrdv.insert(key, rdv);
  }
  else rdv = curr->second;
  return rdv;
}


// ---------------------------------------------------------------------------
void ReduceNetwork::erase_rendezvous( ReduceNetwork::Rendezvous* rdv )
{
  LHashMap<Key, Rendezvous*>::iterator curr = _allrdv.find( rdv->key );
  ACOM_ASSERT( curr !=0, "bad assertion" );
  _allrdv.erase( curr );
}


// ---------------------------------------------------------------------------
ReduceNetwork::OutChannel::OutChannel( )
 : Network::OutChannel(), _op(0), _serial(0), _count_peers(0), _peers(0), _index_self(0), _index_dest(0)
{
}


// ---------------------------------------------------------------------------
ReduceNetwork::OutChannel::~OutChannel()
{
}


// ---------------------------------------------------------------------------
void ReduceNetwork::OutChannel::destroy()
{
}



// ---------------------------------------------------------------------------
void ReduceNetwork::OutChannel::init
  ( ReduceNetwork* net, int tag, OpReduceService* op,
    int count_peers, LocalNodeNo* peers, int index_self, int index_dest )
{
  _op         = op;
  _count_peers= count_peers;
  _peers      = peers;
  _index_self = index_self;
  _index_dest = index_dest;
  _serial     = 0;  
  Network::OutChannel::init(net, tag);
}



}

