// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reducenet_alpha.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// Please read node in .cpp file.
// $Id: ics_reducenet_alpha.cpp,v 1.1 2003/07/28 07:45:49 thierry Exp $
//
// ==========================================================================
#include <ics_reducenet_alpha.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
ReduceAlphaNetwork::ReduceAlphaNetwork()
  : ReduceNetwork()
{
  /* Change properties for this network */
  set_property("name",     "REDUCENETALPHA");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("REDUCENETALPHA");
}

// --------------------------------------------------------------------------
/** Destructor
*/
ReduceAlphaNetwork::~ReduceAlphaNetwork()
{
}


// ---------------------------------------------------------------------------
ReduceAlphaNetwork::OutChannel::OutChannel()
 : ReduceNetwork::OutChannel(), _alpha(0.5)
{ 
} 


// ---------------------------------------------------------------------------
ReduceNetwork::OutChannel* ReduceAlphaNetwork::open_channel
  ( int tag, OpReduceService* op,
    int count_peer, LocalNodeNo* peers, int index_self, int index_dest )
      throw(ACOM::ComFailure)
{
  ReduceAlphaNetwork::OutChannel* ch = new ReduceAlphaNetwork::OutChannel;
  ch->init( this, tag, op, count_peer, peers, index_self, index_dest );
  return ch;
}


// ---------------------------------------------------------------------------
void ReduceAlphaNetwork::OutChannel::post( ACOM::Network::Request* r ) throw (ACOM::ComFailure)
{
  ReduceAlphaNetwork* net = (ReduceAlphaNetwork*)r->get_net();

  /* extended header for alpha reduce network */
/*  ReduceAlphaNetwork::Header* eh = 
    (ReduceAlphaNetwork::Header*)r->allocate_extended_header( sizeof(ReduceAlphaNetwork::Header) );
*/
  eh->saveid     = r->_header.id;
  eh->tag        = _tag;
  eh->netid      = net->get_id();
  eh->opid       = _op->get_id();
  eh->index_dest = _index_dest;
  eh->key        = ++_serial;

  /* replace service id by internal service */
  r->_header.id     = net->_service->get_id();

  /* I'm the source, then I begin to send */
  int index_next =  (_index_self + 1) % _count_peers;

  ACOM::Network::OutChannel* channel = net->get_subnet()->get_default_route(_peers[index_next]);
  channel->post( r );
}


// -----------------------------------------------------------------------------
void ReduceAlphaNetwork::OutChannel::process( ACOM::Network::CallContext* call_cc )
{
  ACOM::Network::Header* h = call_cc->header;
  ReduceAlphaNetwork::Header* eh = (ReduceAlphaNetwork::Header*)call_cc->ext_header;
  ACOM_ASSERT( eh !=0, "bad assertion");

  /* Get the associated Bcast network */
  ReduceAlphaNetwork* net = (ReduceAlphaNetwork*)get_net();

  OpReduceService* opserv = (OpReduceService*)ACOM::Service::Manager.resolv( eh->opid );
  
  /* get rendez vous, mutex on net remains locked  */
  Rendezvous* rdv = net->get_rendezvous(eh->tag, eh->key);
  ++rdv->rcv_msg;
  if (rdv->wait_msg ==0) rdv->wait_msg = 1;
    
  /* if rendezvous suppress entry */
  bool rendezvouscompte = rdv->is_ok();
  if (rendezvouscompte) {
    net->erase_rendezvous( rdv );
  }

  /* do reduction */
  if (rdv->call_cc ==0) {
    rdv->call_cc = call_cc;
    opserv->server_reduce( rdv->call_cc, 0 );
  } else {
    opserv->server_reduce( rdv->call_cc, call_cc );
    call_cc->release();
  }
  
  /* unlock network mutex */
  net->mutex().unlock();
  
  if (rendezvouscompte && (_index_self != eh->index_dest) ) 
  {
    /* forward message to next */
    int index_next =  (_index_self + 1) % _count_peers;
    net->deallocate_rendezvous( rdv );

    /* call back to clear rdv->call_cc */
    RedirectServiceCallBack*  mycbk = new RedirectServiceCallBack( net, rdv->call_cc );

    /* message to be resend */
    ACOM::Message m ( net, (ACOM::Network::ProtocolNo)h->protocol, net->_service, mycbk, 1 );

    /* recopie extended header */
    ReduceAlphaNetwork::Header* enewh = 
      (ReduceAlphaNetwork::Header*)m.get_request()->allocate_extended_header( sizeof(ReduceAlphaNetwork::Header) );
    *enewh = *eh;

    /* recopy immediate data from receipt */
    ACOM::IOVectSeq* iovs = rdv->call_cc->receipt.get_packetbyte_vectseq();
    for( ACOM::IOVectSeq::IteratorEntry ie= iovs->beginEntry(); ie != iovs->endEntry(); ++ie)
      m->pack( ie->data, ie->len );

    /* recopy iovector from  iov, iov_sz */
    for (int i=0; i< rdv->call_cc->iov_sz; ++i)
      m->put( &rdv->call_cc->iov[i] );
          
    /* re-post message */
    m.post( net->get_default_route( _peers[index_next] ) );
  }
  else if (rendezvouscompte && (_index_self == eh->index_dest) ) 
  {
    /* Call user service */
    ACOM::Service* userserv = ACOM::Service::Manager.resolv( eh->saveid );
    userserv->server_call( rdv->call_cc );
    rdv->call_cc->release();
    net->deallocate_rendezvous( rdv );
  }
}
 
} // namespace

