// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reducenet_chain.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// Please read node in .cpp file.
// $Id: ics_reducenet_chain.cpp,v 1.1 2003/07/28 07:45:49 thierry Exp $
//
// ==========================================================================
#include <ics_reducenet_chain.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
ReduceChainNetwork::ReduceChainNetwork()
  : ReduceNetwork()
{
  /* Change properties for this network */
  set_property("name",     "REDUCENETCHAIN");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("REDUCENETCHAIN");
}

// --------------------------------------------------------------------------
/** Destructor
*/
ReduceChainNetwork::~ReduceChainNetwork()
{
}


// ---------------------------------------------------------------------------
ReduceNetwork::OutChannel* ReduceChainNetwork::open_channel
  ( int tag, OpReduceService* op,
    int count_peer, LocalNodeNo* peers, int index_self, int index_dest )
      throw(ACOM::ComFailure)
{
  ReduceChainNetwork::OutChannel* ch = new ReduceChainNetwork::OutChannel;
  ch->init( this, tag, op, count_peer, peers, index_self, index_dest );
  return ch;
}


// ---------------------------------------------------------------------------
void ReduceChainNetwork::OutChannel::post( ACOM::Network::Request* r ) throw (ACOM::ComFailure)
{
  ACOM_ASSERT( _index_self != _index_dest, "bad usage of post on dest node");

  ReduceChainNetwork* net = (ReduceChainNetwork*)r->get_net();

  if (_index_self == ((_index_dest + 1) % _count_peers))
  {
    /* extended header for flat reduce network */
    ReduceChainNetwork::Header* eh =  0;
/*
      (ReduceChainNetwork::Header*)r->allocate_extended_header( sizeof(ReduceChainNetwork::Header) );
*/
    eh->saveid     = r->_header.id;
    eh->tag        = _tag;
    eh->netid      = net->get_id();
    eh->opid       = _op->get_id();
    eh->index_dest = _index_dest;
    eh->key        = ++_serial;
  
    /* replace service id by internal service */
    r->_header.id  = net->_service->get_id();
  
    int index_next =  (_index_self + 1) % _count_peers;
    ACOM::Network::OutChannel* channel = net->get_subnet()->get_default_route(_peers[index_next]);
    channel->post( r );
    return;
  }
  /* here I'm a node that should wait rendez-vous before do the post to next node
     - build a dummy call_cc that stores same data as local request
     - call process method to test rendez-vous
  */
  ACOM::Network::CallContext* call_cc = net->get_subnet()->acquire_callcontext();
  call_cc->network                     = net->get_subnet();
  call_cc->header = (ACOM::Network::Header*)call_cc->allocate( sizeof(ACOM::Network::Header) );
  call_cc->header->src          = _peers[_index_self];
  call_cc->header->dst          = _peers[_index_dest];
  call_cc->header->size_h       = r->_datastream.getFillSize();
  call_cc->header->size_d       = r->_iovstream.getByteSize();
  call_cc->header->iovsz_d      = r->_iovstream.getBlockSize();
  call_cc->header->no           = 0;
  call_cc->error_no             = ACOM::ComFailure::OK;

  ReduceChainNetwork::Header* eh =
    (ReduceChainNetwork::Header*)call_cc->allocate( sizeof(ReduceChainNetwork::Header) );
  eh->saveid     = r->_header.id;
  eh->tag        = _tag;
  eh->netid      = net->get_id();
  eh->opid       = _op->get_id();
  eh->index_dest = _index_dest;
  eh->key        = ++_serial;
  call_cc->ext_header = eh;

  call_cc->iov_sz = call_cc->header->iovsz_d;
  call_cc->iov = (ACOM::IOVectEntry*)call_cc->allocate(call_cc->iov_sz);
  call_cc->_im_data_seq.init( 1, r->_iovstream._iovA);

  /* immediate data */
  if (call_cc->header->size_h >0)
    for( ACOM::IOVectSeq::IteratorEntry ie= r->_datastream.beginEntry();
          ie != r->_datastream.endEntry();
          ++ie)
    {
      call_cc->_im_data_seq.put( ie->data, ie->len );
    }
  
  /* deferred data */
  if (call_cc->header->iovsz_d >0)
  {
    int i=0;
	  for( ACOM::IOVectSeq::IteratorEntry ie = r->_iovstream.beginEntry();
	  	   ie != r->_iovstream.endEntry();
         ++ie, ++i)
    {
      call_cc->iov[i].data = ie->data;
      call_cc->iov[i].len = ie->len;
    }
  }
  call_cc->receipt.init( &call_cc->_im_data_seq, call_cc->iov, call_cc->iov_sz, net->get_subnet());          

  process( call_cc, r );
}


// -----------------------------------------------------------------------------
/* If req !=0 then it is the call from post : req and call_cc shared the same data
*/
void ReduceChainNetwork::OutChannel::process( ACOM::Network::CallContext* call_cc, ACOM::Network::Request* req )
{
  ACOM_ASSERT( _index_self != ((_index_dest + 1) % _count_peers), "bad assertion");

  ACOM::Network::Header* h = call_cc->header;
  ReduceChainNetwork::Header* eh = (ReduceChainNetwork::Header*)call_cc->ext_header;
  ACOM_ASSERT( eh !=0, "bad assertion");

  /* Get the associated Bcast network */
  ReduceChainNetwork* net = (ReduceChainNetwork*)get_net();

  OpReduceService* opserv = (OpReduceService*)ACOM::Service::Manager.resolv( eh->opid );
  
  /* get rendez vous, mutex on net remains locked  */
  Rendezvous* rdv = net->get_rendezvous(eh->tag, eh->key);
  ++rdv->rcv_msg;
  if (rdv->wait_msg ==0) {
    if (_index_self == eh->index_dest) 
      rdv->wait_msg = 1;
    else 
      rdv->wait_msg = 2; /* incomming + local post ! */
  }
    
  /* if rendezvous suppress entry */
  bool rendezvouscompte = rdv->is_ok();
  if (rendezvouscompte) {
    net->erase_rendezvous( rdv );
  }

  /* do reduction */
  if (rdv->call_cc ==0) {
    rdv->call_cc = call_cc;
    opserv->server_reduce( rdv->call_cc, 0 );
  } else {
    opserv->server_reduce( rdv->call_cc, call_cc );
    if (req !=0) {
      /* called by local post do soft reset before release call_cc */
      call_cc->header = 0;
      call_cc->ext_header = 0;
      call_cc->iov_sz = 0;
      call_cc->iov   = 0;
    }
    call_cc->release();
  }
  
  /* unlock network mutex */
  net->mutex().unlock();
  
  if (rendezvouscompte && (_index_self != eh->index_dest) ) 
  {
    /* forward message to next */
    int index_next =  (_index_self + 1) % _count_peers;
    net->deallocate_rendezvous( rdv );

    /* call back to clear rdv->call_cc */
    RedirectServiceCallBack*  mycbk = new RedirectServiceCallBack( req, rdv->call_cc );

    /* message to be resend */
    ACOM::Message m ( net, (ACOM::Network::ProtocolNo)h->protocol, net->_service, mycbk, 1 );

    /* recopie extended header */
    ReduceChainNetwork::Header* enewh = 0;
/*
      (ReduceChainNetwork::Header*)m.get_request()->allocate_extended_header( sizeof(ReduceChainNetwork::Header) );
*/
    *enewh = *eh;

    /* recopy immediate data from receipt */
    ACOM::IOVectSeq* iovs = rdv->call_cc->receipt.get_packetbyte_vectseq();
    for( ACOM::IOVectSeq::IteratorEntry ie= iovs->beginEntry(); ie != iovs->endEntry(); ++ie)
      m->pack( ie->data, ie->len );

    /* recopy iovector from  iov, iov_sz */
    for (int i=0; i< rdv->call_cc->iov_sz; ++i)
      m->put( &rdv->call_cc->iov[i] );
          
    /* re-post message */
    m.post( net->get_default_route( _peers[index_next] ) );
  }
  else if (rendezvouscompte && (_index_self == eh->index_dest) ) 
  {
    /* Call user service */
    ACOM::Service* userserv = ACOM::Service::Manager.resolv( eh->saveid );
    userserv->server_call( rdv->call_cc );
    rdv->call_cc->release();
    net->deallocate_rendezvous( rdv );
  }
}
 
} // namespace

