// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/ICS/src/ics_reducenet_flat.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier, L. Pigeon
// Implementation of virtual base class for all reduce algorithms
// Please read node in .cpp file.
// $Id: ics_reducenet_flat.cpp,v 1.4 2003/07/28 08:05:41 thierry Exp $
//
// ==========================================================================
#include <ics_reducenet_flat.h>

namespace ICS {

// --------------------------------------------------------------------------
/** Constructor
*/
ReduceFlatNetwork::ReduceFlatNetwork()
  : ReduceNetwork()
{
  /* Change properties for this network */
  set_property("name",     "REDUCENETFLAT");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("REDUCENETFLAT");
}

// --------------------------------------------------------------------------
/** Destructor
*/
ReduceFlatNetwork::~ReduceFlatNetwork()
{
}


// ---------------------------------------------------------------------------
ReduceNetwork::OutChannel* ReduceFlatNetwork::open_channel
  ( int tag, OpReduceService* op,
    int count_peer, LocalNodeNo* peers, int index_self, int index_dest )
      throw(ACOM::ComFailure)
{
  ReduceFlatNetwork::OutChannel* ch = new ReduceFlatNetwork::OutChannel;
  ch->init( this, tag, op, count_peer, peers, index_self, index_dest );
  return ch;
}


// ---------------------------------------------------------------------------
void ReduceFlatNetwork::OutChannel::post( ACOM::Network::Request* r ) throw (ACOM::ComFailure)
{
  ReduceFlatNetwork* net = (ReduceFlatNetwork*)r->get_net();

  /* extended header for flat reduce network */
  ReduceFlatNetwork::Header* eh =  0;
/*
    (ReduceFlatNetwork::Header*)r->allocate_extended_header( sizeof(ReduceFlatNetwork::Header) );
*/
  eh->saveid     = r->_header.id;
  eh->tag        = _tag;
  eh->netid      = net->get_id();
  eh->opid       = _op->get_id();
  eh->index_dest = _index_dest;
  eh->key        = ++_serial;

  /* replace service id by internal service */
  r->_header.id     = net->_service->get_id();

  ACOM::Network::OutChannel* channel = net->get_subnet()->get_default_route(_peers[_index_dest]);
  channel->post( r );
}


// -----------------------------------------------------------------------------
void ReduceFlatNetwork::OutChannel::process( ACOM::Network::CallContext* call_cc, ACOM::Network::Request* r )
{
  ACOM::Network::Header* h = call_cc->header;
  ReduceFlatNetwork::Header* eh = (ReduceFlatNetwork::Header*)call_cc->ext_header;
  ACOM_ASSERT( eh !=0, "bad assertion");

  /* Get the associated Bcast network */
  ReduceFlatNetwork* net = (ReduceFlatNetwork*)get_net();

  OpReduceService* opserv = (OpReduceService*)ACOM::Service::Manager.resolv( eh->opid );
  
  /* get rendez vous, mutex on net remains locked  */
  Rendezvous* rdv = net->get_rendezvous(eh->tag, eh->key);
  ++rdv->rcv_msg;
  if (rdv->wait_msg ==0) rdv->wait_msg = _count_peers-1;
    
  /* if rendezvous suppress entry */
  bool rendezvouscompte = rdv->is_ok();
  if (rendezvouscompte) {
    net->erase_rendezvous( rdv );
  }

  /* do reduction */
  if (rdv->call_cc ==0) {
    rdv->call_cc = call_cc;
    opserv->server_reduce( rdv->call_cc, 0 );
  } else {
    opserv->server_reduce( rdv->call_cc, call_cc );
    call_cc->release();
  }
  
  /* unlock network mutex */
  net->mutex().unlock();
  
  if (rendezvouscompte) {
    ACOM::Service* userserv = ACOM::Service::Manager.resolv( eh->saveid );
    userserv->server_call( rdv->call_cc );
    rdv->call_cc->release();
    net->deallocate_rendezvous( rdv );
  }
}
 
} // namespace

