//
// $Id: hello_bcast.cpp,v 1.1 2003/07/28 08:15:55 thierry Exp $
//
// Bcast algorithm for hello
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

static ACOM::Network* net = 0;
static ICS::BcastNetwork* bcastnet = 0;
static ICS::BcastNetwork::OutChannel* bcastchannel =0;
static PTH::Mutex     lock;
static PTH::Condition cond;


/* Callback call when all requests are sent
*/
class Mycallback : public ACOM::Callback {
public:
  /* */
  Mycallback() : is_called(false) {}

  /* release call cc */
  void done( ACOM::Network::Request* )
  { 
    std::cout << "Call back executed" << std::endl;
    { PTH::CriticalGuard critical(lock);
      is_called = true;
    } 
    cond.signal();
  }

public:
  bool is_called;
};




/* Service called to print the message 
*/
class HelloService: public ACOM::ServiceAM {
public:
  HelloService() : is_called(false) {}
  void call ( 
	     ACOM::ComFailure::Code e,      // if any 
	     acom::Network::Receipt* receipt,
	     ACOM::Network::Header* h,
	     ACOM::IOVectEntry* iov,
	     ACOM::IOVectSize   iov_sz,
	     void* ctx
	     ) throw (ACOM::ComFailure)
  {
    if (e == ACOM::ComFailure::OK) 
    {
      char toto[8];
      receipt->unpack(toto,8);
      std::cout << net->rank() << ":: Hello: Im data = "<< toto << std::endl;
      std::cout << net->rank() << ":: Hello: De data = "<< (char*)iov[0].data << std::endl;
    }
    else 
      std::cout <<  net->rank() << ":: Error" << std::endl;
   
     { PTH::CriticalGuard critical(lock);
      is_called = true;
    } 
    cond.signal();
  }
public:
  bool is_called;
};

static HelloService hello_service;



/* 
 *
 */
int main (int argc, char* argv[])
{
  try {
    ACOM::Acom::initialize( &argc, &argv );

    ACOM::Service::Manager.bind (1, &hello_service );
    lock.init();
    cond.init();
    
    ACOM::Acom::commit( );

    /* Get the network:
    */
    try {
      net = ACOM::Acom::get_default_network();
      if (net ==0)
        net = ACOM::Acom::resolv_network( "CORNET" );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
      throw e;
    }

    /* build Bcast network */
    //bcastnet = new ICS::BcastFlatNetwork;
    //bcastnet = new ICS::BcastChainNetwork;
    bcastnet = new ICS::BcastAlphaNetwork;
    bcastnet->initialize( &argc, &argv, net );
    net->barrier();

    /* open one BcastChannel from the bcastnet */
    int rank = net->rank();
    int nodecount = net->node_count();
    ICS::uint16_t* ranks = new ICS::uint16_t[nodecount];
    for (int i=0; i<nodecount; ++i)
      ranks[i] = i;
    bcastchannel = bcastnet->open_channel(8, nodecount, ranks, rank );

    net->barrier();
    sleep(3);
    
    if (rank ==0) {
      /* Msg sent from 0 to all other node */
      char* toto="coucou\n";

      /* call back called when all data have been sent */
      Mycallback mycbk;

      /* send message: call back to wait send occurs */
      ACOM::ActiveMessage<HelloService> am (bcastnet, &hello_service, &mycbk);
      am->put(toto, strlen(toto)+1);
      am->pack(toto, strlen(toto)+1);
      am.post(bcastchannel);
      std::cout << "Post done" << std::endl;
      
      PTH::CriticalGuard critical(lock);
      while (!mycbk.is_called)
        cond.wait(lock);
    }
    else {
      PTH::CriticalGuard critical(lock);
      while (!hello_service.is_called)
        cond.wait(lock);
    }
    net->barrier();

    /* close and destroy every things */
    bcastnet->close_channel( bcastchannel );
    delete [] ranks;
    lock.destroy();
    cond.destroy();
    sleep(3);
    ACOM::Acom::terminate();
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    return -1;
  }
}
