//
// $Id: jacobi.cpp,v 1.3 2003/07/09 13:29:17 olobry Exp $
//
// Bcast algorithm
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>
#include <math.h> // on IRIX

static ACOM::Network* net = 0;
ICS::BcastOne2All*   bcastone2all_service =0;
ICS::BcastAlphaTree* bcastalpha_service =0;
ICS::BcastChain*     bcastchain_service =0;
ICS::BcastAlphaChain* bcastalphachain_service =0;
ICS::BcastCompose*   bcastcompose_service =0;
ICS::AckService*     ack_service =0;
ICS::BarrierService* barrier_service =0;

/* pointer to the concret service for bcast */
ICS::BcastService*   thebcast =0;


struct Domain {
  //
  Domain( int el, int il )
   : ext_len(el), int_len(il)
  {
    ext_data = new double[el];
    for (int k=0; k<ext_len; ++k)
      ext_data[k] = sin( drand48());
    int_data = new double[il];

    for (int k=0; k<int_len; ++k)
      int_data[k] = sin( drand48());
  }

  //
  ~Domain()
  {
    delete [] ext_data;
    delete [] int_data;
  }

  void update_fontiere()
  {
    for (int i=0; i<10000; ++i)
      for (int k=0; k<ext_len; ++k)
        ext_data[k] *= sin( drand48());
  }

  void update_internal()
  {
    for (int i=0; i<10000; ++i)
      for (int k=0; k<int_len; ++k)
        int_data[k] = sin( drand48());
  }

  size_t  ext_len;
  double* ext_data;
  long  int_len;
  double* int_data;

};

struct Argument {
  Argument( int s, int nodecount, int nbv, int myrank, int ext_len, int int_len )
   : step(s), dom(ext_len, int_len)
  {
    nbv /= 2;
    nbvoisins = 2*nbv+1;
    group = new int*[nbvoisins];
    index = new int[nbvoisins];
    tag   = new int[nbvoisins];

    /* groupe voisins de mr: mr-nbv,..,mr+nbv: pour envoi/recep */
    for (int v =-nbv; v<= nbv; ++v) {
      int rankv = (v+myrank) % nodecount;
      if (rankv <0) rankv += nodecount;
      group[nbv+v] = new int[nbvoisins];
      tag[nbv+v]   = rankv;
      index[nbv+v] = (myrank-rankv);
      if (index[nbv+v] <0) index[nbv+v] += nodecount;

      /* remplissage des voisins */
      for (int i=0; i<nbvoisins; ++i)
        group[nbv+v][i] = (i+rankv)%nodecount;
    }


#if 1
    for (int i=0; i<nbvoisins; ++i)
    {
      std::cout << myrank << ":: Group=" << i << ", tag["   << i << "]= " << tag[i];
      std::cout << ", index[" << i << "]= " << index[i];
      std::cout << ", group[" << i << "]=[ ";
      for (int k=0; k<nbvoisins; ++k)
       std::cout << group[i][k] << ' ';
      std::cout << "]" << std::endl;
    }
#endif
    mutex.init();
    cond.init();
  }

  ~Argument() {
    mutex.destroy();
    cond.destroy();
  }

  int   step;
  int   nbvoisins;
  int*  index;
  int** group;
  int*  tag;
  Domain dom;
  PTH::Mutex mutex;
  PTH::Condition cond;
};


/* 
 * Thread worker: 
 * @param args struct task
 *
 */
void signal_func(void* pc, void* pcnt )
{
  PTH::Condition* cond = (PTH::Condition*)pc;
  long* count = (long*)pcnt;
  --*count;
//std::cout << net->rank() << ":: In signal ! count=" << *count << std::endl;
  if (*count ==0) cond->broadcast();
}


/* 
 * Thread worker: 
 * @param args struct task
 *
 */
void* worker(void* arg)
{
  long count;
  int nodecount = net->node_count();
  Argument* cc = (Argument*)arg;

  //std::cout << "Wait start init" << std::endl;
  cc->mutex.lock();
  cc->mutex.unlock();
  //std::cout << "Start init ok " << std::endl;

  for (int i=0; i<cc->step; ++i)
  {
  std::cout << "iter=" << i << std::endl;
    /* Update fontiere */ 
    cc->dom.update_fontiere( );     

    count = cc->nbvoisins;
    /* Envoi/Reception de la fontiere aux voisins */ 
    void* data = cc->dom.ext_data;
    for (int v =0; v < cc->nbvoisins; ++v)
      thebcast->async_bcast(
          data,
          cc->dom.ext_len,
          &ACOM::FormatDef::Double,
          cc->nbvoisins,
          cc->group[v],
          cc->index[v],
          0,               /* src indexs in group[v], always 0 */
          i*nodecount + cc->tag[v],      /* tag */
          &signal_func,
          (void*)&cc->cond, (void*)&count
        );
 
    /* Update internal */ 
    cc->dom.update_internal( );     

    /* Wait send */
    cc->mutex.lock();
    while (count >0)
      cc->cond.wait( cc->mutex );
    cc->mutex.unlock();
  }

  /* synchro with the caller */
  cc->step = 0;
  cc->cond.signal();
  return 0;
}



/* 
 *
 */
int doit (size_t nbv, int iter, int szdom)
{
  int myrank    =  net->rank();
  int nodecount =  net->node_count();
  
  Argument* arg = new Argument( iter, nodecount, nbv, myrank, szdom, szdom*szdom  );

  /* create and synchronize with thread */
  arg->mutex.lock();
  PTH::ThreadId thid;
  PTH::Thread::create( &thid, &worker, (void*)arg);

  struct timeval time1,time2;
  double t1,t2;

  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  net->barrier();
  arg->cond.wait(arg->mutex);

  /* wait end */
  while (arg->step >0)
    arg->cond.wait(arg->mutex);

  net->barrier();
  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  arg->mutex.unlock();

  
  double time = (t2-t1) / (double)iter;

  if (myrank ==0) {
    std::cout 
//              << std::setw(15) << std::setfill(' ') << ((double)szdomain)
//              << std::setw(15) << std::setfill(' ') << ((double)nbv)
              << std::setw(20) << std::setfill(' ') << time * 1e-6 / iter
              << std::endl;
  }
  delete arg;
}



int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );

    bcastone2all_service    = new ICS::BcastOne2All; 
    bcastchain_service      = new ICS::BcastChain; 
    bcastalphachain_service = new ICS::BcastAlphaChain; 
    bcastalpha_service      = new ICS::BcastAlphaTree; 
    bcastcompose_service    = new ICS::BcastCompose; 
    ack_service             = new ICS::AckService; 
    barrier_service         = new ICS::FlatBarrier; 
    ACOM::Service::Manager.bind (2, ack_service );
    ACOM::Service::Manager.bind (3, barrier_service );
    ACOM::Service::Manager.bind (4, bcastone2all_service );
    ACOM::Service::Manager.bind (5, bcastchain_service );
    ACOM::Service::Manager.bind (6, bcastalpha_service );
    ACOM::Service::Manager.bind (7, bcastalphachain_service );
    ACOM::Service::Manager.bind (8, bcastcompose_service );

    //AC::System::attach_debugger( );
    AC::Acom::commit( );

    if (argc >6) { 
      std::cout << "Usage: " << argv[0] << " <#iter> <#voisins> <size domain> <method> <network name>" << std::endl;
      std::cout << "      - method = one2all | chain | alpha | alphachain" << std::endl;    
      std::cout << "      - method = chain-chain | chain-alpha | alpha-alpha | alpha-chain " << std::endl;    
      AC::Acom::const_iterator icurr = AC::Acom::begin_network();
      AC::Acom::const_iterator iend  = AC::Acom::end_network();
      std::cout << "      - network name = ";
      while (icurr != iend) {
        std::cout << (*icurr)->get_name() << " | ";
        ++icurr;
      }
      std::cout << std::endl;
      for (int i=0; i<argc; ++i)
        std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
      if (argc >6) 
        return 1;
    }
    
    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( argv[5] );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }

    /* Initialize the service objects 
    */
    ack_service->initialize( net );
    barrier_service->initialize( net );
    bcastone2all_service->initialize( net );
    bcastchain_service->initialize( net );
    bcastalphachain_service->initialize( net );
    bcastalpha_service->initialize( net );


    bcastalpha_service->set_alpha( 0.5 );

    bcastalphachain_service->attach_services( bcastalpha_service, bcastchain_service );
    bcastalphachain_service->set_chainlength( 8 );
    bcastcompose_service->set_groupsize( 8 );

    int iter = atoi(argv[1]);
    int nbvoisins = atoi(argv[2]);
    int len  = atoi(argv[3]);

    if ( std::string(argv[4]) == "one2all" ) {
      thebcast = bcastone2all_service;
      std::cout << "using one2all algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain" ) {
      thebcast = bcastchain_service;
      std::cout << "using chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alphachain" ) {
      thebcast = bcastalphachain_service;
      std::cout << "using alphachain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha" ) {
      thebcast = bcastalpha_service;
      std::cout << "using alpha algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha-alpha" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastalpha_service);
      thebcast = bcastcompose_service;
      std::cout << "using alpha-alpha algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha-chain" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastchain_service);
      thebcast = bcastcompose_service;
      std::cout << "using alpha-chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain-chain" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastchain_service);
      thebcast = bcastcompose_service;
      std::cout << "using chain-chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain-alpha" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastalpha_service);
      thebcast = bcastcompose_service;
      std::cout << "using chain-alpha algorithm" << std::endl;
    } else {
      thebcast = bcastone2all_service;
      std::cout << "using one2all algorithm" << std::endl;
    }

    net->barrier();
    barrier_service->barrier(123);

    int rank = net->rank();
    int nodecount = net->node_count();

#if 0
    std::cout << rank << " :: My_rank is=" << rank 
              << " over #nodes=" << nodecount << std::endl << std::flush;
#endif

    if (rank ==0) {
      std::cout << "# iter=" << iter 
                << ", size_domain=" << len 
                << ", nbvoisins=" << nbvoisins << std::endl;

      std::cout //<< std::setw(15) << std::setfill(' ') << "# )"
                //<< std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::setw(20) << std::setfill(' ') << "# time/iteration (sec)"
                << std::endl;
    }

    /* Call one experiment : */
    doit(nbvoisins, iter, len);

    /* Terminate */
//    std::cout << rank << " :: Ready to Terminate" << std::endl;
    for (int i=0; i<nodecount; ++i)
    {
      barrier_service->barrier( 1 );
      if (i == rank) { 
        std::cout << "Infos node rank=" << i << std::endl;
        bcastone2all_service->print( std::cout << "bcastone2all_service::" ) << std::endl;
        bcastalpha_service->print( std::cout << "bcastalpha_service::" ) << std::endl;
      }
    }
    ACOM::Acom::terminate( );
    delete bcastone2all_service;
    delete bcastchain_service;
    delete bcastalpha_service;
    delete bcastalphachain_service;
    delete bcastcompose_service;
    delete ack_service;
    delete barrier_service;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
