//
// $Id: testbarrier.cpp,v 1.2 2003/06/24 19:21:20 thierry Exp $
//
// Barrier algorithms
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

static ACOM::Network* net = 0;
ICS::AlphaBarrier*   barrieralpha_service =0;
ICS::FlatBarrier*    barrierflat_service =0;
ICS::DefaultBarrier*  barrierdefault_service =0;

/* pointer to the concret service for bcast */
ICS::BarrierService* thebarrier =0;

/* 
 *
 */
int doit (int iter)
{
  int myrank    =  net->rank();
  //int nodecount =  net->node_count();
  
  struct timeval time1,time2;
  double t1,t2;

  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  for (int i=0; i<iter; ++i)
  {
    thebarrier->barrier(i+100);
  }
  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  double time = (t2-t1) / (double)iter;

  if (myrank ==0) {
    std::cout 
              << std::setw(15) << std::setfill(' ') << net->node_count()
              << std::setw(20) << std::setfill(' ') << time/iter
              << std::endl;
  }
}





/* 
 * Usage about test
 */
void Usage(const char *msg)
{
  if (msg !=0)
    std::cout << "****** Error: " << msg << std::endl;
  std::cout << "Usage: testbcast " 
            << " -i <#iter> -m <method> -n <network name> " 
            << std::endl;
  std::cout << "Available method: \n"
            << "      - basic algorithms = alpha | flat | default \n" 
            << std::endl;
  AC::Acom::const_iterator icurr = AC::Acom::begin_network();
  AC::Acom::const_iterator iend  = AC::Acom::end_network();
  std::cout << "Available network name:\n"
            << "      - network name = ";
  while (icurr != iend) {
    std::cout << (*icurr)->get_name() << " | ";
    ++icurr;
  }
  std::cout << std::endl;
  throw;
}




/*
 * Parse option
 */
long   iter=1;       // - number of iteration per timings
char*  methodname=0; // - the name of the method
char*  netname=0;    // - network name
double alpha =0.5;   // - for alphabarrier

void Barrier_parseopt( int argc, char* argv[] )
{
#if 0
for (int i=0; i<argc; ++i)
{
  std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
}
#endif
  int opt;
  while((opt = getopt(argc, argv, "hn:i:m:a:"))>0)
  switch(opt)
  {
    case 'h':
        Usage(0);

    case 'i':
      iter = atoi(optarg);
    break;

    case 'm':
      methodname = optarg;
      if ( std::string(optarg) == "flat" ) {
      //  std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "alpha" ) {
      //  std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "default" ) {
      //  std::cout << optarg << std::endl;
      } else {
        Usage("Bad method");
      }
    break;

    case 'n':
      netname = optarg;
    break;

    case 'a':
      alpha = atof(optarg);
    break;

    default:
      Usage("Bad option");
  }
  if (netname ==0) Usage("Unspecified networkname");
  if (methodname ==0) Usage("Unspecified method name");
}



int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );

    barrieralpha_service    = new ICS::AlphaBarrier; 
    barrierflat_service     = new ICS::FlatBarrier; 
    barrierdefault_service  = new ICS::DefaultBarrier; 
    ACOM::Service::Manager.bind (1, barrieralpha_service );
    ACOM::Service::Manager.bind (2, barrierflat_service );
    ACOM::Service::Manager.bind (3, barrierdefault_service );

    //AC::System::attach_debugger( );
    AC::Acom::commit( );

    /* Parse option
     */
    Barrier_parseopt(argc, argv);

    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( netname );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }


    /* Initialize the service objects 
    */
    barrieralpha_service->initialize( net );
    barrieralpha_service->set_alpha( alpha );
    barrierflat_service->initialize( net );
    barrierdefault_service->initialize( net );

    if ( std::string(methodname) == "alpha" ) {
      thebarrier = barrieralpha_service;
      if (net->rank()==0) std::cout << "# alpha=" << alpha << std::endl;
    } else if ( std::string(methodname) == "default" ) {
      thebarrier = barrierdefault_service;
      if (net->rank()==0) std::cout << "# alpha=" << alpha << std::endl;
    } else if ( std::string(methodname) == "flat" ) {
      thebarrier = barrierflat_service;
      if (net->rank()==0) std::cout << "# flat" << std::endl;
    } else {
      Usage("Bad method");
    }


    net->barrier();

    int rank = net->rank();
    int nodecount = net->node_count();

#if 0
    std::cout << rank << " :: My_rank is=" << rank 
              << " over #nodes=" << nodecount << std::endl << std::flush;
#endif

    if (rank ==0) {
      std::cout << "# " << methodname
                << ", iter=" << iter 
                << std::endl;

      std::cout << std::setw(15) << std::setfill(' ') << "# Nodes"
                << std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::endl;
    }

    /* Call one experiment : */
    doit(iter);

    /* Terminate */
    ACOM::Acom::terminate( );
    delete barrieralpha_service;
    delete barrierflat_service;
    delete barrierdefault_service;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
