//
// $Id: testbcast_async.cpp,v 1.2 2003/06/23 19:52:25 thierry Exp $
//
// Bcast algorithm
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

static ACOM::Network* net = 0;
ICS::BcastOne2All*   bcastone2all_service =0;
ICS::BcastAlphaTree* bcastalpha_service =0;
ICS::BcastChain*     bcastchain_service =0;
ICS::BcastAlphaChain* bcastalphachain_service =0;
ICS::BcastCompose*   bcastcompose_service =0;
ICS::AckService*     ack_service =0;
ICS::BarrierService* barrier_service =0;

/* pointer to the concret service for bcast */
ICS::BcastService*   thebcast =0;

static PTH::CountDown recvok;

void signal_recvok( void *ptc, void* a)
{
  --recvok;
//if (net->rank() !=0) std::cout << "Coundown=" << recvok.getValue() << std::endl;
//if (recvok.getValue() ==0)
//  std::cout << net->rank() << "::RECV ALL MESG FOR TEST, Signal WAIT" << std::endl;
}

/* 
 *
 */
int doit (size_t len, int iter)
{
  int myrank    =  net->rank();
  int nodecount =  net->node_count();
  int ranks[ nodecount ];
  for (int i=0; i<nodecount; ++i)
    ranks[i] =i;
  
  void* data = malloc(10*len);

//std::cout << myrank << "::BEGIN TEST " << len << std::endl;

  struct timeval time1,time2;
  double t1,t2;

  if (myrank ==0) {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = i;
  }
  else {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = 0;
  }

  size_t len0 =0;
  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  for (int i=0; i<iter; ++i)
  {
    //if (len >0) {
      /* call bcast */
      thebcast->async_bcast(
        data,
        len, 
        &ACOM::FormatDef::Byte,
        nodecount,
        ranks,
        myrank,  /* self index in ranks */
        0,       /* src indexs in ranks */
        100+i,        /* tag */
        &signal_recvok, 0, 0
      );
    //}
  }

#ifdef ACK_INUK
    if (myrank ==0) {
      recvok.init( iter );
      for (int i=1; i < nodecount; ++i)
        ack_service->signal( i, len );
      ack_service->wait( len, nodecount -1 );
    } else {
      ack_service->wait( len, 1 );
      recvok.wait();
      recvok.init(iter);
      ack_service->signal( 0, len );
    }
#else
  {//if (myrank !=0) {
    recvok.wait();
    recvok.init(iter);
    //std::cout << myrank << "::End recv message" << std::endl;
    net->barrier();
    //std::cout << myrank << "::END TEST: " << len << std::endl;
  }
#endif

  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  double time = (t2-t1) / (double)iter;

#if 1 /* pour verfier la transimition */
  if (myrank !=0) {
    for (int i=0; i<len; ++i)
      if (((char*)data)[i] != (char)i) {
        std::cout << "Error on receiving data[" << i << "]" << std::endl;
        exit(1);
      }
  }
#endif

  if (myrank ==0) {
    std::cout 
              << std::setw(15) << std::setfill(' ') << ((double)len)/1024.0
              << std::setw(20) << std::setfill(' ') << time
              << std::setw(20) << std::setfill(' ') << ((double)len)/time*((double)1e6)
                / 1024.0/1024.0
              << std::endl;
  }
  free(data);
}





/* 
 * Usage about test
 */
void Usage(const char *msg)
{
  if (msg !=0)
    std::cout << "****** Error: " << msg << std::endl;
  std::cout << "Usage: testbcast " 
            << " -l <size max msg> -s <stride> -i <#iter> -m <method> -n <network name> " 
            << " [-a <alpha value >] [-b <bloc size>] [-g <group size] " 
            << std::endl;
  std::cout << "Defaults values: \n"
            << "* alpha =0.5\n"
            << "* bloc size = 1024 bytes\n"
            << "* group size = 1 \n" 
            << std::endl;
  std::cout << "Available method: \n"
            << "      - basic algorithms = one2all | chain | alpha | alphachain\n" 
            << "      - composed algorithms = chain-chain | chain-alpha | alpha-alpha | alpha-chain\n"
            << std::endl;
  AC::Acom::const_iterator icurr = AC::Acom::begin_network();
  AC::Acom::const_iterator iend  = AC::Acom::end_network();
  std::cout << "Available network name:\n"
            << "      - network name = ";
  while (icurr != iend) {
    std::cout << (*icurr)->get_name() << " | ";
    ++icurr;
  }
  std::cout << std::endl;
  throw;
}




/*
 * Parse option
 */
double alpha =0.5;   // - alpha value for all alpha-based algorithm
long   iter=1;       // - number of iteration per timings
long   maxlength=0;  // - maximal length of message
long   stride=1;      // - stride to increment message size
int    blocsize;     // - size of bloc for pipeline
int    groupsize =1; // - group size for composition of two algorithms
char*  methodname=0; // - the name of the method
char*  netname=0;    // - network name

void Bcast_parseopt( int argc, char* argv[] )
{
#if 0
for (int i=0; i<argc; ++i)
{
  std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
}
#endif
  int opt;
  while((opt = getopt(argc, argv, "hn:l:s:i:m:a:b:g:"))>0)
  switch(opt)
  {
    case 'h':
        Usage(0);

    case 'l':
      maxlength = atoi(optarg);
    break;

    case 's':
      stride = atoi(optarg);
    break;

    case 'i':
      iter = atoi(optarg);
    break;

    case 'm':
      methodname = optarg;
      if ( std::string(optarg) == "one2all" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "chain" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "alphachain" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "alpha" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "alpha-alpha" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "alpha-chain" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "chain-chain" ) {
        std::cout << optarg << std::endl;
      } else if ( std::string(optarg) == "chain-alpha" ) {
        std::cout << optarg << std::endl;
      } else {
        Usage("Bad method");
      }
    break;

    case 'n':
      netname = optarg;
    break;

    case 'a':
      alpha = atof(optarg);
    break;

    case 'b':
      blocsize = atoi(optarg);
    break;

    case 'g':
      groupsize = atoi(optarg);
    break;

    default:
      Usage("Bad option");
  }
  if (netname ==0) Usage("Unspecified networkname");
  if (methodname ==0) Usage("Unspecified method name");
#if 0
      std::cout << "# " << methodname
                << ", SIZE_MAX=" << maxlength 
                << ", stride=" << stride 
                << ", iter=" << iter 
                << ", alpha=" << alpha 
                << ", groupsize=" << alpha 
                << std::endl;
#endif
}



int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );

    bcastone2all_service    = new ICS::BcastOne2All; 
    bcastchain_service      = new ICS::BcastChain; 
    bcastalphachain_service = new ICS::BcastAlphaChain; 
    bcastalpha_service      = new ICS::BcastAlphaTree; 
    bcastcompose_service    = new ICS::BcastCompose; 
    ack_service             = new ICS::AckService; 
    barrier_service         = new ICS::FlatBarrier; 
    ACOM::Service::Manager.bind (2, ack_service );
    ACOM::Service::Manager.bind (3, barrier_service );
    ACOM::Service::Manager.bind (4, bcastone2all_service );
    ACOM::Service::Manager.bind (5, bcastchain_service );
    ACOM::Service::Manager.bind (6, bcastalpha_service );
    ACOM::Service::Manager.bind (7, bcastalphachain_service );
    ACOM::Service::Manager.bind (8, bcastcompose_service );

    //AC::System::attach_debugger( );
    AC::Acom::commit( );

    /* Parse option
     */
    Bcast_parseopt(argc, argv);

    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( netname );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }


    /* Initialize the service objects 
    */
    ack_service->initialize( net );
    barrier_service->initialize( net );
    bcastone2all_service->initialize( net );
    bcastchain_service->initialize( net );
    bcastalphachain_service->initialize( net );
    bcastalpha_service->initialize( net );

    if ( std::string(methodname) == "one2all" ) {
      thebcast = bcastone2all_service;
    } else if ( std::string(methodname) == "chain" ) {
      thebcast = bcastchain_service;
    } else if ( std::string(methodname) == "alphachain" ) {
      thebcast = bcastalphachain_service;
    } else if ( std::string(methodname) == "alpha" ) {
      thebcast = bcastalpha_service;
    } else if ( std::string(methodname) == "alpha-alpha" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastalpha_service);
      thebcast = bcastcompose_service;
    } else if ( std::string(methodname) == "alpha-chain" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastchain_service);
      thebcast = bcastcompose_service;
    } else if ( std::string(methodname) == "chain-chain" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastchain_service);
      thebcast = bcastcompose_service;
    } else if ( std::string(methodname) == "chain-alpha" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastalpha_service);
      thebcast = bcastcompose_service;
    } else {
      Usage("Bad method");
    }


    bcastalpha_service->set_alpha( alpha );
    bcastalphachain_service->attach_services( bcastalpha_service, bcastchain_service );
    bcastalphachain_service->set_chainlength( groupsize );
    bcastcompose_service->set_groupsize( groupsize );

    net->barrier();
    barrier_service->barrier(123);

    int rank = net->rank();
    int nodecount = net->node_count();

#if 0
    std::cout << rank << " :: My_rank is=" << rank 
              << " over #nodes=" << nodecount << std::endl << std::flush;
#endif

//    if (rank ==0) {
      std::cout << "# " << methodname
                << ", SIZE_MAX=" << maxlength 
                << ", stride=" << stride 
                << ", iter=" << iter 
                << ", alpha=" << alpha 
                << ", groupsize=" << alpha 
                << std::endl;

    if (rank ==0) {
      std::cout << std::setw(15) << std::setfill(' ') << "# Msg. Size (KB)"
                << std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::setw(20) << std::setfill(' ') << "BandWidth (MB/s)"
                << std::endl;
    }

    recvok.init( iter );
    int len =0;
    while (len <= maxlength) {
      /* Call one experiment : */
      doit(len,iter);
      len += stride;
    }

    /* Terminate */
//    std::cout << rank << " :: Ready to Terminate" << std::endl;
    for (int i=0; i<nodecount; ++i)
    {
      barrier_service->barrier( 1 );
      if (i == rank) { 
        std::cout << "Infos node rank=" << i << std::endl;
        thebcast->print( std::cout << methodname << "::" ) << std::endl;
      }
    }
    ACOM::Acom::terminate( );
    delete bcastone2all_service;
    delete bcastchain_service;
    delete bcastalpha_service;
    delete bcastalphachain_service;
    delete bcastcompose_service;
    delete ack_service;
    delete barrier_service;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
