//
// $Id: testbcast_charge.cpp,v 1.3 2003/07/09 13:29:03 olobry Exp $
//
// Bcast algorithm
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>
#include <math.h> // on IRIX

static ACOM::Network* net = 0;
ICS::BcastOne2All*   bcastone2all_service =0;
ICS::BcastAlphaTree* bcastalpha_service =0;
ICS::BcastChain*     bcastchain_service =0;
ICS::BcastAlphaChain* bcastalphachain_service =0;
ICS::BcastCompose*   bcastcompose_service =0;
ICS::AckService*     ack_service =0;
ICS::BarrierService* barrier_service =0;

/* pointer to the concret service for bcast */
ICS::BcastService*   thebcast =0;

void* work_load(void* arg)
{
  double f= drand48();
  while (1) {
    f = f*sin(f);
  }
  return 0;
}



/* 
 *
 */
int doit (size_t len, int iter)
{
  int myrank    =  net->rank();
  int nodecount =  net->node_count();
  int ranks[ nodecount ];
  for (int i=0; i<nodecount; ++i)
    ranks[i] =i;
  
  void* data = malloc(10*len);

  struct timeval time1,time2;
  double t1,t2;

  if (myrank ==0) {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = i;
  }
  else {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = 0;
  }

  size_t len0 =0;
  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  for (int i=0; i<iter; ++i)
  {
    if (len >0) {
      /* call bcast */
      thebcast->bcast(
        data,
        len, 
        &ACOM::FormatDef::Byte,
        nodecount,
        ranks,
        myrank,  /* self index in ranks */
        0,       /* src indexs in ranks */
        i + iter * len       /* tag */
      );
    }
  }
#if ACK_INUK
    if (myrank ==0) {
      ack_service->wait( i, nodecount -1 );
      bcastalpha_service->bcast(
        data,
        len0, 
        &ACOM::FormatDef::Byte,
        nodecount,
        ranks,
        myrank,  /* self index in ranks */
        0,       /* src indexs in ranks */
        1        /* tag */
      );
    } else {
      ack_service->signal( 0, i );
      bcastalpha_service->bcast(
        data,
        len0, 
        &ACOM::FormatDef::Byte,
        nodecount,
        ranks,
        myrank,  /* self index in ranks */
        0,       /* src indexs in ranks */
        1        /* tag */
      );
    }
#else
    net->barrier();
#endif

  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  double time = (t2-t1) / (double)iter;

#if 1 /* pour verfier la transimition */
  if (myrank !=0) {
    for (int i=0; i<len; ++i)
      if (((char*)data)[i] != (char)i) {
        std::cout << "Error on receiving data[" << i << "]" << std::endl;
        exit(1);
      }
  }
#endif

  if (myrank ==0) {
    std::cout 
              << std::setw(15) << std::setfill(' ') << ((double)len)/1024.0
              << std::setw(20) << std::setfill(' ') << time
              << std::setw(20) << std::setfill(' ') << ((double)len)/time*((double)1e6)
                / 1024.0/1024.0
              << std::endl;
  }
  free(data);
}



int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );

    bcastone2all_service    = new ICS::BcastOne2All; 
    bcastchain_service      = new ICS::BcastChain; 
    bcastalphachain_service = new ICS::BcastAlphaChain; 
    bcastalpha_service      = new ICS::BcastAlphaTree; 
    bcastcompose_service    = new ICS::BcastCompose; 
    ack_service             = new ICS::AckService; 
    barrier_service         = new ICS::FlatBarrier; 
    ACOM::Service::Manager.bind (2, ack_service );
    ACOM::Service::Manager.bind (3, barrier_service );
    ACOM::Service::Manager.bind (4, bcastone2all_service );
    ACOM::Service::Manager.bind (5, bcastchain_service );
    ACOM::Service::Manager.bind (6, bcastalpha_service );
    ACOM::Service::Manager.bind (7, bcastalphachain_service );
    ACOM::Service::Manager.bind (8, bcastcompose_service );

    //AC::System::attach_debugger( );
    AC::Acom::commit( );

    if (argc >7) { 
      std::cout << "Usage: " << argv[0] << " <size max msg> <stride> <#iter> <method> <#worker> <network name>" << std::endl;
      std::cout << "      - method = one2all | chain | alpha | alphachain" << std::endl;    
      std::cout << "      - method = chain-chain | chain-alpha | alpha-alpha | alpha-chain " << std::endl;    
      AC::Acom::const_iterator icurr = AC::Acom::begin_network();
      AC::Acom::const_iterator iend  = AC::Acom::end_network();
      std::cout << "      - network name = ";
      while (icurr != iend) {
        std::cout << (*icurr)->get_name() << " | ";
        ++icurr;
      }
      std::cout << std::endl;
      for (int i=0; i<argc; ++i)
        std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
      if (argc >6) 
        return 1;
    }
    
    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( argv[6] );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }

    /* Initialize the service objects 
    */
    ack_service->initialize( net );
    barrier_service->initialize( net );
    bcastone2all_service->initialize( net );
    bcastchain_service->initialize( net );
    bcastalphachain_service->initialize( net );
    bcastalpha_service->initialize( net );


    bcastalpha_service->set_alpha( 0.5 );

    bcastalphachain_service->attach_services( bcastalpha_service, bcastchain_service );
    bcastalphachain_service->set_chainlength( 8 );
    bcastcompose_service->set_groupsize( 8 );

    int len = 0;
    int SIZE_MAX  = atoi(argv[1]);
    int stride = atoi(argv[2]);
    int iter = atoi(argv[3]);
    int nworkers = atoi(argv[5]);
    if ( std::string(argv[4]) == "one2all" ) {
      thebcast = bcastone2all_service;
      std::cout << "using one2all algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain" ) {
      thebcast = bcastchain_service;
      std::cout << "using chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alphachain" ) {
      thebcast = bcastalphachain_service;
      std::cout << "using alphachain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha" ) {
      thebcast = bcastalpha_service;
      std::cout << "using alpha algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha-alpha" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastalpha_service);
      thebcast = bcastcompose_service;
      std::cout << "using alpha-alpha algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "alpha-chain" ) {
      bcastcompose_service->attach_services( bcastalpha_service, bcastchain_service);
      thebcast = bcastcompose_service;
      std::cout << "using alpha-chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain-chain" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastchain_service);
      thebcast = bcastcompose_service;
      std::cout << "using chain-chain algorithm" << std::endl;
    } else if ( std::string(argv[4]) == "chain-alpha" ) {
      bcastcompose_service->attach_services( bcastchain_service, bcastalpha_service);
      thebcast = bcastcompose_service;
      std::cout << "using chain-alpha algorithm" << std::endl;
    } else {
      thebcast = bcastone2all_service;
      std::cout << "using one2all algorithm" << std::endl;
    }

    for (int i=0; i< nworkers; ++i)
    {
      PTH::ThreadId thid;
      PTH::Thread::create( &thid, &work_load, (void*)i);
    }

    net->barrier();
    barrier_service->barrier(123);

    int rank = net->rank();
    int nodecount = net->node_count();

#if 0
    std::cout << rank << " :: My_rank is=" << rank 
              << " over #nodes=" << nodecount << std::endl << std::flush;
#endif

    if (rank ==0) {
      std::cout << "# SIZE_MAX=" << SIZE_MAX 
                << ", stride=" << stride 
                << ", iter=" << iter << std::endl;

      std::cout << std::setw(15) << std::setfill(' ') << "# Msg. Size (KB)"
                << std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::setw(20) << std::setfill(' ') << "BandWidth (MB/s)"
                << std::endl;
    }

    while (len <= SIZE_MAX) {
      /* Call one experiment : */
      doit(len,iter);
      len += stride;
    }

    /* Terminate */
//    std::cout << rank << " :: Ready to Terminate" << std::endl;
    for (int i=0; i<nodecount; ++i)
    {
      barrier_service->barrier( 1 );
      if (i == rank) { 
        std::cout << "Infos node rank=" << i << std::endl;
        bcastone2all_service->print( std::cout << "bcastone2all_service::" ) << std::endl;
        bcastalpha_service->print( std::cout << "bcastalpha_service::" ) << std::endl;
      }
    }
    ACOM::Acom::terminate( );
    delete bcastone2all_service;
    delete bcastchain_service;
    delete bcastalpha_service;
    delete bcastalphachain_service;
    delete bcastcompose_service;
    delete ack_service;
    delete barrier_service;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
