//
// $Id: testchannelbcast.cpp,v 1.6 2003/07/28 20:21:19 thierry Exp $
//
// Bcast algorithm
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

static ACOM::Network* net = 0;
static ICS::BcastNetwork* bcastnet = 0;
static PTH::Mutex     lock;
static PTH::Condition cond;
static int tag_channel = 3;



/* 
 * Usage about test
 */
void Usage(const char *msg)
{
  if (msg !=0)
    std::cout << "****** Error: " << msg << std::endl;
  std::cout << "Usage: testbcast " 
            << " -l <size max msg> -s <stride> -i <#iter> -p <protocol> -m <method> -n <network name> " 
            << " [-a <alpha value >] [-b <bloc size>] " 
            << std::endl;
  std::cout << "Defaults values: \n"
            << "* protcol = am (ActiveMessage)\n"
            << "* alpha =0.5\n"
            << "* bloc size = 1024 bytes\n"
            << std::endl;
  std::cout << "Available procotols: \n"
            << "      - am | ws | aws\n" 
            << std::endl;
  std::cout << "Available methods: \n"
            << "      - basic algorithms = flat | chain | alpha\n" 
            //<< "      - composed algorithms = chain-chain | chain-alpha | alpha-alpha | alpha-chain\n"
            << std::endl;
  AC::Acom::const_iterator icurr = AC::Acom::begin_network();
  AC::Acom::const_iterator iend  = AC::Acom::end_network();
  std::cout << "Available network name:\n"
            << "      - network name = ";
  while (icurr != iend) {
    std::cout << (*icurr)->get_name() << " | ";
    ++icurr;
  }
  std::cout << std::endl;
  throw;
}



/*
 * Parse options
 */
double alpha =0.5;   // - alpha value for all alpha-based algorithm
long   iter=1;       // - number of iteration per timings
long   maxlength=0;  // - maximal length of message
long   stride=1;      // - stride to increment message size
int    blocsize;     // - size of bloc for pipeline
int    groupsize =1; // - group size for composition of two algorithms
ACOM::Network::ProtocolNo protocol= ACOM::Network::P_AM; // - the protocol
char*  methodname="flat"; // - the name of the method
char*  netname=0;    // - network name

void Bcast_parseopt( int argc, char* argv[] )
{
#if 0
for (int i=0; i<argc; ++i)
{
  std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
}
#endif
#if 0
  optind   =1;
  optreset =1;
#endif

  int opt;
  while((opt = getopt(argc, argv, "hn:l:s:i:m:p:a:b:g:"))>0)
  {
    //std::cout << "Opt=" << (char)opt << std::endl;
    switch(opt)
    {
      case 'h':
          Usage(0);
  
      case 'l':
        maxlength = atoi(optarg);
      break;
  
      case 's':
        stride = atoi(optarg);
      break;
  
      case 'i':
        iter = atoi(optarg);
        std::cout << "Iter = " << iter << std::endl;
      break;
  
      case 'm':
        methodname = optarg;
        if ( std::string(optarg) == "flat" ) {
          std::cout << optarg << std::endl;
        } else if ( std::string(optarg) == "chain" ) {
          std::cout << optarg << std::endl;
        } else if ( std::string(optarg) == "alpha" ) {
          std::cout << optarg << std::endl;
        } else {
          Usage("Bad method");
        }
      break;
  
      case 'p':
        if ( std::string(optarg) == "am" ) {
          protocol = ACOM::Network::P_AM;
        } else if ( std::string(optarg) == "ws" ) {
          protocol = ACOM::Network::P_WS;
        } else if ( std::string(optarg) == "aws" ) {
          protocol = ACOM::Network::P_AWS;
        } else {
          Usage("Bad protocol");
        }
      break;
  
      case 'n':
        netname = optarg;
      break;
  
      case 'a':
        alpha = atof(optarg);
      break;
  
      case 'b':
        blocsize = atoi(optarg);
      break;
  
      case 'g':
        groupsize = atoi(optarg);
      break;
  
      default:
        ;//Usage("Bad option");
    }
  }
  if (netname ==0) {
    if (ACOM::Acom::get_default_network() ==0)
      Usage("Unspecified networkname");
    else netname = (char*)ACOM::Acom::get_default_network()->get_name().c_str();
  }
  if (methodname ==0) Usage("Unspecified method name");
#if 0
      std::cout << "# " << methodname
                << ", SIZE_MAX=" << maxlength 
                << ", stride=" << stride 
                << ", iter=" << iter 
                << ", alpha=" << alpha 
                << ", groupsize=" << alpha 
                << std::endl;
#endif
}


static void* data = 0;
static ICS::AckService    service_ack;


/* Callback call when all requests are sent
*/
class Mycallback : public ACOM::Callback {
public:
  /* */
  Mycallback() : is_called(false) {}

  /* release call cc */
  void done( ACOM::Network::Request* )
  { 
    std::cout << "Call back executed" << std::endl;
    { PTH::CriticalGuard critical(lock);
      is_called = true;
    } 
    cond.signal();
  }

public:
  bool is_called;
};



/* 
 *
 */
class ServiceData: public ACOM::Service {
public:
  ServiceData() : _iter(0) {}
  void call (
       AC::ComFailure::Code e,      // if any 
       acom::Network::Receipt* receipt,
       AC::Network::Header* h,
       AC::IOVectEntry* iov,
       AC::IOVectSize   iov_sz,
       void* ctx
       ) throw (AC::ComFailure)
  {
    if (e != AC::ComFailure::OK)
    {
       std::cout << "Error !!!" << e << std::endl;
       throw;
    }
    --_iter;
    if (_iter ==0) {
      //std::cout << "Send signal !!!" << std::endl;
      service_ack.signal( 0, 10 );
      _iter = iter;
    }
  }

  void* get_iovect(
    ACOM::Network::Receipt* r,
    ACOM::Network::Header* header,
    ACOM::IOVectEntry* vector, ACOM::IOVectSize count
  ) 
  {
    vector[0].data = data;
  }
  PTH::AtomicCounter _iter;
};

static ServiceData service_data;




/* 
 *
 */
int doit (size_t len, int iter, bool display)
{
  int rank    =  net->rank();
  int nodecount =  net->node_count();
  
  data = malloc(10*len);

  struct timeval time1,time2;
  double t1,t2;

  if (rank ==0) {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = i;
  }
  else {
    for (int i=0; i<len; ++i)
      ((char*)data)[i] = 0;
  }

  ICS::Network::OutChannel* bcastchannel = bcastnet->get_channel(tag_channel);

  /* Form the message on node 0 */
  ACOM::Message* msg =0;
  Mycallback mycbk;
  if (rank ==0) {
    msg = new ACOM::Message( bcastnet, protocol, &service_data, &mycbk, iter);
    (*msg)->pack( &iter, sizeof(int) ); 
    (*msg)->put( data, len ); 
  }
  service_data._iter = iter;
  net->barrier();
  
  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  if (rank ==0) {
    for (int i=0; i<iter; ++i)
    {
      msg->post( bcastchannel ); 
    }
    std::cout << "post done" << std::endl;
    service_ack.wait( 10, nodecount -1 );
  }
  net->barrier();
  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  double time = (t2-t1) / (double)iter;

  delete msg; 

#if 0 /* pour verfier la transimition */
  if (rank !=0) {
    for (int i=0; i<len; ++i)
      if (((char*)data)[i] != (char)i) {
        std::cout << "Error on receiving data[" << i << "]" << std::endl;
        exit(1);
      }
  }
#endif

  if (display && (rank ==0)) {
    std::cout 
              << std::setw(15) << std::setfill(' ') << ((double)len)/1024.0
              << std::setw(20) << std::setfill(' ') << time
              << std::setw(20) << std::setfill(' ') << ((double)len)/time*((double)1e6)
                / 1024.0/1024.0
              << std::endl;
  }
  free(data);
  net->barrier();
}







int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );

    /* Parse option
     */
    Bcast_parseopt(argc, argv);

    lock.init();
    cond.init();

    ACOM::Service::Manager.bind (1, &service_data );
    ACOM::Service::Manager.bind (2, &service_ack );

    //AC::System::attach_debugger( );
    AC::Acom::commit( );

    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( netname );
      std::cout << "Use network = " << net->get_name() << std::endl;
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }

    if ( std::string(methodname) == "flat" ) {
      bcastnet = new ICS::BcastFlatNetwork;
    } else if ( std::string(methodname) == "chain" ) {
      bcastnet = new ICS::BcastChainNetwork;
    } else if ( std::string(methodname) == "alpha" ) {
      bcastnet = new ICS::BcastAlphaNetwork;
    } else {
      Usage("Bad method");
    }

    net->barrier();
    sleep(1);
    net->barrier();

    service_ack.initialize(net);
    bcastnet->initialize(&argc, &argv, net);

    int rank    =  net->rank();
    int nodecount =  net->node_count();
    ICS::uint16_t ranks[ nodecount ];
    for (int i=0; i<nodecount; ++i)
      ranks[i] =i;
    bcastnet->open_channel(tag_channel, nodecount, ranks, rank );

    //net->barrier();

#if 1
    for (int i=0; i<nodecount; ++i)
    {
      if (rank ==i) {
        std::cout << rank << " :: my pid is=" << getpid() 
                  << " over #nodes=" << nodecount << std::endl << std::flush;
      }
      net->barrier();
    }
#endif

    if (rank ==0) {
      std::cout << "# " << methodname
                << ", SIZE_MAX=" << maxlength 
                << ", stride=" << stride 
                << ", iter=" << iter 
                << ", alpha=" << alpha 
                << ", groupsize=" << alpha 
                << std::endl;

      std::cout << std::setw(15) << std::setfill(' ') << "# Msg. Size (KB)"
                << std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::setw(20) << std::setfill(' ') << "BandWidth (MB/s)"
                << std::endl;
    }

    /* to open channel */
    doit(8,10, false);

    int len =0;
    while (len <= maxlength) {
      /* Call one experiment : */
      doit(len,iter, true);
      len += stride;
    }

    /* Terminate */
    lock.destroy();
    cond.destroy();
    
    bcastnet->close_channel( bcastnet->get_channel(tag_channel) );
    net->barrier();
    ACOM::Acom::terminate( );
    delete bcastnet;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
