//
// $Id: testreduce.cpp,v 1.2 2003/06/23 19:52:25 thierry Exp $
//
// Reduce algorithm
//
#include <pth>
#include <acom>
#include <ics>
#include <unistd.h>
#include <sys/time.h>
#include <iomanip>

ACOM::Network* net = 0;
ICS::AckService*     ack_service =0;
ICS::BarrierService* barrier_service =0;
ICS::ReduceService*   thereduce =0;
ICS::ReduceAll2One*   reduce1_service =0;
ICS::ReduceAlphaTree* reduce2_service =0;


/* 
*
*/
int doit (size_t len, size_t iter)
{
  int myrank    =  net->rank();
  int nodecount =  net->node_count();
  int* ranks = new int[nodecount];
  for (int i=0; i<nodecount; ++i)
    ranks[i] =i;
  
  int* send_data = (int*)ACOM::FormatDef::Int.create( len );
  int* recv_data = 0;

  struct timeval time1,time2;
  double t1,t2;

  if (myrank ==0) {
    recv_data = (int*)ACOM::FormatDef::Int.create( len );
    for (int i=0; i<len; ++i)
      recv_data[i] = 0;
  }
  for (int i=0; i<len; ++i)
    send_data[i] = i;

  gettimeofday(&time1,NULL);
  t1= ((double)time1.tv_sec)*((double)1e6) + (double)time1.tv_usec;
  for (int i=0; i<iter; ++i)
  {
    /* call reduce */
    thereduce->reduce (
      (void*&)recv_data,
      (void*&)send_data,
      len,
      &ACOM::FormatDef::Int,
      &ICS::OpAdd<int>::theformat,
      //&ICS::VectorOp<int,ICS::UnitaryOpAdd<int> >::theformat,
      nodecount,
      ranks,
      myrank,               /* self index in ranks */
      0,                    /* src indexs in ranks */
      1+i                   /* tag */
    );
  }
#if 0
  if (myrank ==0) {
    ack_service->wait( len, nodecount -1 );
  } else {
    ack_service->signal( 0, len );
  }
#else  
  barrier_service->barrier( 25 );
#endif

  gettimeofday(&time2,NULL);
  t2= ((double)time2.tv_sec)*((double)1e6) + (double)time2.tv_usec;
  double time = (t2-t1) / (double)iter;

#if 1 /* pour verfier la transimition */
#if 0
  std::cout << myrank << "::Data send = ";
  for (int i=0; i<len; ++i)
    std::cout << (int)send_data[i] << ' ';
  std::cout << std::endl;
#endif
  if (myrank ==0) 
  {
    std::cout << myrank << "::Data recv = ";
    for (int i=0; i<len; ++i)
      std::cout << (int)recv_data[i] << ' ';
    std::cout << std::endl;
  }
#endif

  if (myrank ==0) {
    len *= sizeof(int);
    std::cout 
            << std::setw(15) << std::setfill(' ') << ((double)len)/1024.0
            << std::setw(20) << std::setfill(' ') << time
            << std::setw(20) << std::setfill(' ') << ((double)len)/time*((double)1e6)
              / 1024.0/1024.0
            << std::endl;
  }
  delete []ranks;
  ACOM::FormatDef::Int.destroy( send_data, len );
  if (myrank ==0) {
     ACOM::FormatDef::Int.destroy( recv_data, len );
  }
}



int main (int argc, char* argv[])
{
  try {
    AC::Acom::initialize( &argc, &argv );
    //AC::System::attach_debugger( );

    reduce1_service       = new ICS::ReduceAll2One; 
    reduce2_service       = new ICS::ReduceAlphaTree; 
    ack_service           = new ICS::AckService; 
    barrier_service       = new ICS::FlatBarrier; 

    ACOM::Service::Manager.bind (2, ack_service );
    ACOM::Service::Manager.bind (3, barrier_service );
    ACOM::Service::Manager.bind (4, reduce1_service );
    ACOM::Service::Manager.bind (5, reduce2_service );

    AC::Acom::commit( );

    if (argc >6) { 
      std::cout << "Usage: " << argv[0] << " <size max msg> <stride> <#iter> <method> <network name>" << std::endl;
      std::cout << "      - method = one2all | pipeline1 | alpha" << std::endl;    
      AC::Acom::const_iterator icurr = AC::Acom::begin_network();
      AC::Acom::const_iterator iend  = AC::Acom::end_network();
      std::cout << "      - network name = ";
      while (icurr != iend) {
        std::cout << (*icurr)->get_name() << " | ";
        ++icurr;
      }
      std::cout << std::endl;
      for (int i=0; i<argc; ++i)
        std::cout << "argv[" << i << "]=" << argv[i] << std::endl;
      if (argc >6) 
        return 1;
    }
    
    /* Get the network:
    */
    try {
      net = ACOM::Acom::resolv_network( argv[5] );
    } catch ( const ACOM::InvalidArgument& e ) {
      std::cout << "Error, bad network name !" << std::endl;
    }

    /* Initialize the service objects 
    */
    /* Initialize the service objects */
    ack_service->initialize( net );
    barrier_service->initialize( net );
    reduce1_service->initialize( net );
    reduce2_service->initialize( net );
    reduce2_service->set_alpha( 0.5 );
    
    int len = 0;
    int SIZE_MAX  = atoi(argv[1]);
    int stride = atoi(argv[2]);
    int iter = atoi(argv[3]);
    if ( std::string(argv[4]) == "one2all" ) {
      thereduce = reduce1_service;
      std::cout << "using one2all algorithm" << std::endl;
    } 
#if 0
    else if ( std::string(argv[4]) == "pipeline1" ) {
      thereduce = 0;
      std::cout << "using pipeline1 algorithm" << std::endl;
    } 
#endif
    else if ( std::string(argv[4]) == "alpha" ) {
      thereduce = reduce2_service;
      std::cout << "using alpha algorithm" << std::endl;
    } 
    else {
      thereduce = reduce1_service;
      std::cout << "using one2all algorithm" << std::endl;
    }

    net->barrier();

    int rank = net->rank();
    int nodecount = net->node_count();

#if 0
    std::cout << rank << " :: My_rank is=" << rank 
              << " over #nodes=" << nodecount << std::endl << std::flush;
#endif

    if (rank ==0) {
      std::cout << "# SIZE_MAX=" << SIZE_MAX 
                << ", stride=" << stride 
                << ", iter=" << iter << std::endl;
      std::cout << std::setw(15) << std::setfill(' ') << "# Msg. Size (KB)"
                << std::setw(20) << std::setfill(' ') << "One-way(usecs)"
                << std::setw(20) << std::setfill(' ') << "BandWidth (MB/s)"
                << std::endl;
    }

    while (len <= SIZE_MAX) {
      /* Call one experiment : */
      doit(len,iter);
      len += stride;
    }

    /* Terminate */
    for (int i=0; i<nodecount; ++i)
    {
      barrier_service->barrier( 1 );
      if (i == rank) { 
        std::cout << "Infos node rank=" << i << std::endl;
        thereduce->print( std::cout << "Reduce service::" ) << std::endl;
      }
    }

    ACOM::Acom::terminate( );
    delete reduce1_service;
    delete reduce2_service;
    delete ack_service;
    delete barrier_service;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
