#include <acom>
#include <string>

namespace AC = acom;

void init(int argc, char * argv[]) {

  //TG: WHAT IS THAT ? int _argc = 1; 
  int _argc = argc; 
  char * _arg0 =  argv[0];
  char ** _argv = &_arg0;

  //AC::Acom::initialize(&_argc, &_argv);
  AC::Acom::initialize(&argc, &argv);
  AC::Acom::commit();
}
  

void terminate(int exit_value) {
  AC::Acom::terminate();
  exit(exit_value);
}
  


void list_nets(const std::string& kind) {

  AC::Acom::const_iterator iend  = AC::Acom::end_network();
  AC::Acom::const_iterator icurr = AC::Acom::begin_network();

  for (icurr  = AC::Acom::begin_network(); 
       icurr != iend; icurr++)
    if  ( (kind == "all") || ((*icurr)->get_property(kind) == "true") )
      std::cout << (*icurr)->get_property("name") << " "; 
  std::cout << std::endl;
}

void net_info(char * net_name, char * property) {

  AC::Network * net;
  try {
    net = AC::Acom::resolv_network(net_name);
  } catch (const AC::InvalidArgument& E) {
    std::cerr << "Unknown network '" << net_name << "'" << std::endl;
    terminate(2);
  }


  AC::Properties prop = net->get_properties();
  if (property != NULL) {
    std::string s = prop[property];
    std::cout << prop[property] << std::endl;
  } else {
    AC::Properties::constIterator ipcurr, ipend;
    ipend = prop.end();
    for (ipcurr = prop.begin(); ipcurr != ipend; ipcurr++)
      std::cout << ipcurr->first << ": " << ipcurr->second << std::endl;
  }
}

typedef enum {
  no_action = 0,
  list_all,
  list_primaries,
  list_secondaries,
  show_net_info
} action_t;


char * prog_name;

void usage(int exit_value) {
  std::cout << "Usage: " << prog_name << std::endl;
  std::cout << "          -a: list all networks" << std::endl;
  std::cout << "          -p: list primary networks" << std::endl;
  std::cout << "          -s: list secondary networks" << std::endl;
  std::cout << "          -i net_name [-P property]: show network properties" << std::endl;
  terminate(exit_value);
}


int main(int argc, char * argv[]) {

  int opt;
  prog_name = argv[0];
  action_t action =  no_action; 
  char * net_name;
  char * property = NULL;
  
  init(argc, argv);

  while((opt = getopt(argc, argv, "hapsi:P:")) > 0)
    switch(opt) {
      
    case 'h':
      usage(0);
      break;
      
    case 'a':
      action = list_all;
      break;
      
    case 'p':
      action = list_primaries;
      break;
      
    case 's':
      action = list_secondaries;
      break;
      
    case 'i':
      action = show_net_info;
      net_name = optarg;
      break;
      
    case 'P':
      property = optarg;
      break;
      
    default:
      usage(2);
	break;
    }
  
  switch(action) {
  case list_all:
    list_nets("all");
    break;
    
  case list_primaries:
    list_nets("primary");
    break;
    
  case list_secondaries:
    list_nets("secondary");
    break;
    
  case show_net_info:
    net_info(net_name, property);
      break;
      
  case no_action:
  default:
    usage(2);
  }
  terminate(0);
}
