#! /bin/sh


DN=`dirname $0`

INUKNETINFO=$DN/inuknetinfo


#---------------------------------------------------------------------
isdigit() {
#---------------------------------------------------------------------
# Tests whether *entire string* is numerical.
# In other words, tests for integer variable.
# $1 : value to test

    [ "$#" -eq 1 ] || return 1
#    echo isdigit $1
    case $1 in
	*[!0-9]*|"") return 1;;
	*) return 0;;
    esac
}


#---------------------------------------------------------------------
check_digit() {
#---------------------------------------------------------------------
# check if a variable is a digit
# $1: name of variable

    isdigit "${!1}"
    if [ $? -ne 0 ]; then
	echo "$1 (=${!1}) is not of type integer" >&2
	[ -n "$2" ] && exit $2
	return 1
    fi
    return 0
}



usage() {
	echo "Usage: $0 -p primary_net [options] -- program [program_args]"
	echo "    options"
	echo "        -f hostfile"
	echo "        -m hostname"
	echo "        -n nb_nodes"
	echo "        -s secondary_network"
	exit $1
}

[ $# -eq 0 ] && usage 2

get_hosts() {
	if [ ! -f $1 ]
	then
		echo "File $1 does not exists" 
		exit -1
	fi
	hosts="$hosts `cat $1`" 
	if [ $? -ne 0 ]
	then
		echo "Cannot read file $1"
		exit -1
	fi
}

check_primary_net() {

	known_primary_nets=`${INUKNETINFO} -p`

	if [ xx`${INUKNETINFO} -i $1 -P primary` != xxtrue ] 
	then
		echo "Invalid primary network '$1'"
		echo "Available networks are: $known_primary_nets"
		exit 2
	fi
}

check_secondary_net() {

	known_secondary_nets=`${INUKNETINFO} -s`

	if [ xx` ${INUKNETINFO} -i $1 -P secondary` != xxtrue ] 
	then
		echo "Invalid secondary network '$1'"
		echo "Available networks are: $known_secondary_nets"
		exit 2
	fi
}




while getopts hp:f:m:n:s: opt
do
    case "$opt" in
		p)	primary_net="$OPTARG"; check_primary_net $primary_net;; 
		f)  hostfile="$OPTARG"; get_hosts "$hostfile";;
		m)  hosts="$hosts $OPTARG";;
		h)  usage 0;;
		n)  nb_nodes="$OPTARG"; { check_digit nb_nodes || usage; };;
		s)  check_secondary_net $OPTARG; inuk_args="$inuk_args -inuknet $OPTARG";;
		*)  usage 2;;
    esac
done

shift $(($OPTIND - 1))

[ -z "$primary_net" ] && usage 2


[ $# -lt 1 ] && usage 2


program_name="$1"
shift
program_args="$@"
program="$program_name $inuk_args -- $program_args"

case "$primary_net" in

	CORNET)
		CORNET_ARG="inukrun.cornet"
		[ -n "$hostfile" ] && CORNET_ARG="$CORNET_ARG -f $hostfile"
		CORNET_ARG="$CORNET_ARG -- $program"
		echo "$0: executing $CORNET_ARG"
		$CORNET_ARG
		;;

	TAKTUK)
		echo "Not implemented yet"
		;;
esac



