// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/NetworkCore/doc/network_doc.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat &
// ==========================================================================
/*---------------------------------------------------------------------
------------------------ BEGIN   MANUAL -------------------------------
*----------------------------------------------------------------------
*
* There is a text part which introduce concepts anf how to use them.
*  Text refers to the code (class etc...) or modules. Modules contains only 
* object or methods defining the API for Inuktitut0  thread package
*
------------------------------------------------------------------------
-----------------------------------------------------------------------*/
/**
\htmlonly <IMG SRC="logo_a0.gif"  HEIGHT=200 WIDTH=600 ALIGN=CENTER>
\endhtmlonly 
\author J. Briat 
\date   June 2002, 19
\mainpage  Inuktitut0 Network & protocol  package

This is a short introduction to Inuktitut0  Network package. The following sections present the objective of this package and highlight key points and tricks in using  threads, synchronization etc.. 


The programmer may learn   the Inuktitut0 network package through  sections describing how to initiate  a distributed program, to  create, write and call remote objects etc.. These sections refer to the Inuktitut0 network API. The API is composed  of modules grouping the needed operators and their description. After that the reader may learn about the implementation through browsing classes , files etc...

In the documentation, we use the fully qualified names. When programming, you may use short names depending of C++ contextual rules (using directives, lexical scoping ..). The naming rule used use \c ??? as prefix for all classes which are closed to network & protocol behaviour.

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section Sclient  The client side
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
Remote invocation are based  on active message concept i.e. one sided communication. There is three kind of active messages. Each differs in the manner of receiving data at remote node. From the client point of view, they perform the same. All data to sent must be packed in network buffers (immediate data) or placed in network memory.

\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SCH Communication path
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

An active message is sent on a channel  (ACOM_NET_NAMESPACE::Network::OutChannel) of a given network (ACOM_NET_NAMESPACE::Network). A channel to a peer node may be opened ACOM_NET_NAMESPACE::Network::open_channel(). 


\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SAM Active Message
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

 The three kind of active message are :
	- P_AM  This a pure active message. Received data are stacked and a  function is called to process data.   This function MUST be non blocking to preserve network reactivity. On return received data are discarded.
	- P_WS  This is the Write and Signal case.  The called function must provide to the network the buffers to store received data.
	- P_AWS  This the Alloc, Write and Signal case. The network must allocate memory to the received data. The application must free this memory when no more used.

The ACOM_NET_NAMESPACE::ActiveMessage, ACOM_NET_NAMESPACE::WriteAndSignal, ACOM_NET_NAMESPACE::AllocWriteAndSignal classes are wrapper classes to the generic request (ACOM_NET_NAMESPACE::Network::Request). It is to aid programmer in building the communication request for a given network or channel.

\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SPost Sending an active message
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

 From the client point of view, all three kinds of active message perform the same. Once the communication request created on a given channel, 
	- it is possible to pack immediate value to send (ACOM_NET_NAMESPACE::Network::Request::pack()or to register blocks to be sent deferred. (ACOM_NET_NAMESPACE::Network::Request::put(). The size of immediate data or blocks is only limited by the amount of network memory in  which data are packed or in which blocks are placed. 
	- getting  the network channel ( ACOM_NET_NAMESPACE::Network::OutChannel) to use  may be delayed until sending time.  
	- then execute the send request through posting it to a channel ACOM_NET_NAMESPACE::OutChannel::post()
	- when terminated , a call back is done ( ACOM_NET_NAMESPACE::Callback). The blocks of deferred data may be reused because they are completely sent. 
\code
{ActiveMessage<AckService> am (&net, &ack_service, &cbk); // request handle
am->pack(&n, sizeof(int));
....
am->put( (void*)msg, std::strlen(msg)+1);
.....
am(channel); // post to a channel
}
\endcode

A request may be reused for different posts of it concurrently or in sequence. To avoid automatic destruction when the first post completes, user must give the the number of following posts he wants. At completion of these posts, the callback is called and request( and resources used)  discarded . This is done after the last callback.

\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SDfer  Immediate Data / Deferred data
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

Immediate data are sent as soon as possible after the request post. In general, these data are small & sparsed data or data describing   large data blocks of differnt sizes. These data are sent on a block basis as deferred data and received symetrically with the same block number and layout in an iovector defined par the network ( AM & AWS) or user(WS). 

\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SMEM  Network memory
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

	Reveiving or sending data from memory is possible if this memory is known  by the network to be accesed by DMA devices. In general, the blocks must be pinned in physical memory and their layout depend of the DMA device constrainsts.

 The Network offers a buffer management facility ( ACOM_NET_NAMESPACE::Network::allocate_data() ACOM_NET_NAMESPACE::Network::deallocate_data(). It is also possible to request to set/reset part of the memory in DMA mode (ACOM_NET_NAMESPACE::Network::map_zone(),ACOM_NET_NAMESPACE::Network::map_zone()) . The memory region really put in this mode may be bigger than  the requested region depending of the system page size.

\warning No control of placement of data is done when a post is done. In particular, there no control of multiple mapping of the same region or overlapping region. If part of the data is not in  DMAable memory, result of a send or receive is impredictible.

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section Sserver  The server side
\htmlonly <hr WIDTH="100%"> \endhtmlonly 

Arrival of data and network polling must be done in implementation derived from ACOM_NET_NAMESPACE::Network.
On reception of an active message, the network poller decode the communication header ACOM_NET_NAMESPACE::Network::Header to
get protocol information, immediate data size and blocks count  & block sizes. Then it begin by allocating place for immediate data, receives them. The next action depend of the active message protocol :
	- P_AM  The network agent  allocates place for immediate data and deferred data and received all them. Then it get ACOM_NET_NAMESPACE::ServiceManager::resolv() and call the service procedure . Service::Call(). On return, he deallocate all memory ressources used. Memory management is LIFO.
	- P_AWS  This is the Alloc Write and Signal case.  The agent behaviour is the same. Immediate data management may be put on stack and free on return. But defered data must allocated on heap and free by application ( ACOM_NET_NAMESPACE::Network::deallocate_data())
	- P_WS  This  Write and Signal case. The network agent receives immediate data and manages them as previously. But it needs to request from the service level the blocks adresses. It request this through a call to Service::get_iovect(). It is possible to received deferred data in the provuded buffer. 



\warning THIS THE ACTUAL SOLUTION. Another way is to call the service and received data when the service issue a  ACOM_NET_NAMESPACE::Network::Receipt::sync() indicating it that the iovect for them is defined. In this case, there is no need to split a service in 2 functions.

\htmlonly <hr WIDTH="50%"> \endhtmlonly 
\subsection SCall  What to do in a service code
\htmlonly <hr WIDTH="50%"> \endhtmlonly 
The service code receives arguments :
	-  error code
	- a communication handle ACOM_NET_NAMESPACE::Network::Receipt  from wich he can get immefiate data (ACOM_NET_NAMESPACE::Network::Receipt::unpack()
	- an iovector pointer completely defined (P_AM, P_AWS) or only filled with block sizes for P_WS. In this case, the service code may be split in two functions. The first one  must provide the adresses of the corresponding received buffers. So the second is called with fully defined iovector and may use deferred data. 


\code
class AckService: public ACOM_NET_NAMESPACE::::Service {
public:
  void call ( 
       AC::ComFailure::Code e,      // if any  
			 AC::Network::Receipt* receipt,
       AC::Network::Header* h,
       AC::IOVectEntry* iov,
       AC::IOVectSize   iov_sz
  ) throw (AC::ComFailure)
  { int i; float f; 
	 n;char  * cmsg; 
    if (e == AC::ComFailure::OK){ 
		 
			receipt->unpack(&i, sizeof(int));
			receipt->unpack(&f, sizeof(float));
			....
			for(int i=0;i<iov_sz;i++) // compute with deferred data
			 (SomeBigType)* x= ((SomeBigType)*)(iov[i].data)); 
				......
			}
    }
  };
\endcode
  \warning these buffers must be in a network DMA zone.        


\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section SNET  Network  Construction & connection
\htmlonly <hr WIDTH="100%"> \endhtmlonly 

From a local node point of view, to build a ACOM_NET_NAMESPACE::Network object, it is needed to give :
	- its transport level network address.  ACOM_NET_NAMESPACE::Network::bind_node(). 
  - its rank in the network  ACOM_NET_NAMESPACE::Network::set_rank(). 
	- its immediate neighbour nodes. These neighbours may be physical neighbours or topological neighbours. A neighbour node may be add or delete from neighbourhood :  ACOM_NET_NAMESPACE::Network::add_node() or ACOM_NET_NAMESPACE::Network::del_node(). When adding, It is necessary to set the network rank    for this new node. To add successfully a node, it is needed to give its network address and/or its network name. Double definitions are forbidden.
	- the local node can open a communication channel with a remote node by opening a channel with it ACOM_NET_NAMESPACE::Network::open_channel(). This channel may be used to send active messages. It is closed by ACOM_NET_NAMESPACE::Network::close_channel()


The network must build a default routing table to respond to the  previously seen ACOM_NET_NAMESPACE::Network::get_default_route() request. A channel may be inserted in the default routes by 
ACOM_NET_NAMESPACE::Network::set_default_route(). 


All nodes are registered. All default routes are registered. All opened channels to the same node are registered. Const iterator are provided to consult known nodes, default routes or opened channels to a node.( see ACOM_NET_NAMESPACE::Network)



\warning PROBLEM We must find an appropriate  class to encapsulate the network address ( Myrinet, IP/port, http or corba addressing) to define more precisely ACOM_NET_NAMESPACE::Network::add_node() and ACOM_NET_NAMESPACE::Network::bind_node().

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section SGLOB Global communication
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
 All networks must implement :
	- a broadcast mechanism to post to all neighbours or all known nodes the same active messages by ACOM_NET_NAMESPACE::Network::neighbour_cast() and ACOM_NET_NAMESPACE::Network::broadcast(). This broadcast mechanism is only processor consistent.
	- a barrier synchronization ACOM_NET_NAMESPACE::Network::barrier() 



\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section SINT  Internals & on going Work
\htmlonly <hr WIDTH="100%"> \endhtmlonly 

The user may pack data ( client side) in a request or  unpack data from a receipt  (server side). He put only iovector entries in a request (client side).  Getting iovector entries from a request is for network receiver or poller. A service may get a regular iovector when called. Header is packed/unpacked by network.

The key structures of  the new network interface are
	-   class ACOM_NET_NAMESPACE::BytePacket to build byte sequence
	-   classes ACOM_NET_NAMESPACE::IOVectSeq   to build unlimited stream of ACOM_NET_NAMESPACE::IOVectEntry. Iterators are provided for iterating through the ACOM_NET_NAMESPACE::IOVectCell or the ACOM_NET_NAMESPACE::IOVectEntry

For example, it is possible to iterate through an IOVectSeq by
\code
{// init the iterator from the IOVectSeq iovS
IOVectSeq::IteratorEntry ient= iovS.beginEntry();
for( i=0;; i++){ void * c; int sz;
	if(ient.eos()){
 		break;
	}
// get the next iovect entry
ient.get( &c,&sz);
......
}
\endcode
Another way is 
\code:
{// bild the iterator from the IOVectSeq iovS
for(IOVectSeq::IteratorEntry ie= iovS.beginEntry(); ie!=iovS.endEntry(); ++ie){
			void * c; int sz;
			c= ie->data ; sz= ie->len ;
			........
		 }
}

}
\endcode
It is possiible to iterate through the IOVectCell to use network iovector facilities if any or to iterate on the enties like in the following code

\code:
for(iIOVectSeq::IteratorCellc= iovS.beginCell(); ic!=iovS.endCell(); ++ic){
	for(int k=0; k<ic->_filled ; k++){
		...= ic->_iov[k].data;
		...= ic->_iov[k].len); 
		 }
}
\endcode
	-   classes ACOM_NET_NAMESPACE::ByteSeq to build unlimited stream of  byte for send operations. The basic idea behind this class is that sequence of a priori unbounded size are built as a sequence of fixed size chunks which need to be allocated in network dependant memory region. So the virtual flip() flop() methods may be redefined to be adapted to network properties. A default implementation is provided which used a constant size memory allocator. It is only possible to build an ACOM_NET_NAMESPACE::ByteSeq by packing data sequentially. When built, this sequence mays be read by multiple readers through iterators ACOM_NET_NAMESPACE::ByteSeqReader

\code:
{// Building a ByteSeqReader from a ByteSeq
 ByteSeqReader bR;
  bool err= bR.init(&bS);
	int j;
	while(!bR.eos()){int j;
	bR.unpack((void*)&j,sizeof(int));
	......
}
 }
\endcode

All ACOM_NET_NAMESPACE::ByteSeq or ACOM_NET_NAMESPACE::IOVectSeq sequences are incrementally build ( pack() / put()) by users. They are reread through iterators at the IOVectEntry level  (ACOM_NET_NAMESPACE::IOVectSeq::IteratorEntry) or IOVectCell level(ACOM_NET_NAMESPACE::IOVectSeq::IteratorCell) by network internal
emitters. Network internal receivers build the immediate data sequence as an ACOM_NET_NAMESPACE::IOVectSeq.   When built, these sequence is  reread to be unpacked through ACOM_NET_NAMESPACE::ByteSeqReader iterators.

Deferred data blocks are registered in a contiguous iovector.

\warning ACOM_NET_NAMESPACE::Map and ACOM_NET_NAMESPACE::IStream or ACOM_NET_NAMESPACE::OStream must be rewrite to use this facilities (TO DO !!)
	


\remarks  All the protocol and network information to be send as a communication header may be deferred until posting the request. It's the post which put an end of stream in the immediate or deferred sequences. The network header  is composed of   a fixed size structure terminated by the number of blocks of deferred data. This fixed part is followed by the sizes of the deferred blocks. Request creation needs only protocol & servive number. It encapsulate a byte sequence (ByteSeq) for immediate value,  a sequence of iovector entries for deferred values and  an header structure. The header is packed at posting time when all is known. 


\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section SLoop Intra Node Network
\htmlonly <hr WIDTH="100%"> \endhtmlonly 

Inside a node, threads may use the Inuktitut communication API to sens active messages. The ACOM_NET_NAMESPACE::LoopBack network is provided for this. As a derived class of the root class ACOM_NET_NAMESPACE::Network, it may be also an axample how to specialize the internal classes ACOM_NET_NAMESPACE::Network::OutChannel, ACOM_NET_NAMESPACE::Network::Request.., how to used memory allocators or how to iterate in ACOM_NET_NAMESPACE::IOVectSeq or ACOM_NET_NAMESPACE::ByteSeq. The implementation is very naive. On the receiver part, the immediate data iterator of the ACOM_NET_NAMESPACE::Network::Receipt is built by copying   the  ACOM_NET_NAMESPACE::ByteSeq built on the sender part. The deferred data sequence is scanned and all the blocks are copied in fresh allocated blocks.  All temporary  data needed on the receiver part ( mainly the iovector to give to the service call) are pushed on stack. 

The use of thread safe allocators garantees the safeness of this design.

\see the code for details...

*/ // DON'T REMOVE 
/*----------------------------- END OF MANUAL ----------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------*/
