// ==========================================================================
// $Source: 
// Copyright(c)'2001 by Apache Team
// acom_network.cpp : 
// Authors: Jacques Briat, Thierry Gautier
// ==========================================================================


#include "acom_loopback.h"
#include "acom_service.h"
#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif

namespace ACOM_NET_NAMESPACE {

void LoopBack::simulateReceive( LoopRequest* req, size_t*  iovsz )
{
  int i;Service* s;
  LoopReceipt rcpt;
  void * dd;
  IOVectEntry * iov;
  /* copy the sent header 
  */
  rcpt._header=req->_header;
  /*build the iovector of deferred data on the stack 
  */
  //std::cout << "PUSH IOV ON STACK SZ="<<(sizeof(IOVectEntry)*(rcpt._header.iovsz_d))  << ")\n";
  iov=(IOVectEntry*) alloca(sizeof(IOVectEntry)*(rcpt._header.iovsz_d));
  /* FIRST SOLUTION ::init the receipt structure :
      use the request immediate sequence to build the receipt byteSeqReader iterator
  
  rcpt.init(&req->_datastream,iov,rcpt._header.iovsz_d,this);
  */
  /* SECOND  SOLUTION ::init the receipt structure :
      copy the request immediate sequence in an IOVectSeq; then use it to build the receipt byteSeqReader iterator
  */
  IOVectSeq iovs;
  iovs.init(_nb_entries,&_iovA_RCV);
  for(IOVectSeq::IteratorEntry id= req->_datastream.beginEntry();
  id!=req->_datastream.endEntry(); ++id){
      dd= _bA_RCV.alloc();
      // block copy 
      memcpy( dd,id->data,id->len);
      //int v= *((int*)dd);std::cout << "AFTER IMMEDIATE MEMCOPY ("<<dd<<","<< id->data<<","<<id->len<<", value=="<< v << ")\n";
      iovs.put(dd, id->len);
    }
  iovs.putEos();
  rcpt.init(&iovs,iov,rcpt._header.iovsz_d,this);
  //std::cout << "AFTER ALL COPIES "<< "\n";
  /* END OF THE 2   SOLUTIONs */
  /* init iterator on defered iovec seq.
      it points on the first iovect entry of thz stream
  */ 
  IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
  void* ctxt =0;
  // protocol switch
  switch(rcpt._header.protocol){
  case P_AM :// AM  Allocate + copy defered data  + call + free 
    //  intialize  the iovector ( block & size)
    for( i=0; i<rcpt._header.iovsz_d; i++){
      iov[i].len=iovsz[i]; 
      iov[i].data=allocate_data(iovsz[i]);
      // block copy 
      memcpy( iov[i].data,ie->data,iovsz[i]);
    //std::cout << "AFTER deferred MEMCOPY ("<<iov[i].data<<","<< ie->data<<","<<iovsz[i]<< ")\n";
      ++ie; // next entry
    }
    // call service
    s=Service::Manager.resolv(rcpt._header.id);
    //std::cout << "CALL "<< "\n";
    s->call(ComFailure::OK, &rcpt,&rcpt._header,iov,rcpt._header.iovsz_d,ctxt);
  // std::cout << "AFTER CALL "<< "\n";
    /// free blocks
    for( i=0; i<rcpt._header.iovsz_d; i++){
      //std::cout << "FREE deferred BLOCK ("<<iov[i].data<< ")\n";
      deallocate_data( iov[i].data);
    }
  
    break;
  case P_AWS://AWS  Allocate + copy defered data  + call ( not free)
    //  intialize  the iovector ( block & size)
    for( i=0; i<rcpt._header.iovsz_d; i++){
      iov[i].len=iovsz[i]; iov[i].data=allocate_data(iovsz[i]);
      memcpy( iov[i].data,ie->data,iovsz[i]);
      ++ie;
    }
    // call service
    s= Service::Manager.resolv(rcpt._header.id);
    s->call(ComFailure::OK,&rcpt,&rcpt._header,iov,rcpt._header.iovsz_d,ctxt);
    // blocks must be freed later by applicationthrough deallocate_data
    break;
  case  P_WS :
    //  intialize  the iovector ( size  ONLY)
    for( i=0; i<rcpt._header.iovsz_d; i++){
      iov[i].len=iovsz[i]; iov[i].data=0;
    }
    s= Service::Manager.resolv(rcpt._header.id);
    // call get iovect . Service MUST complete the IOvector
    ctxt = s->get_iovect(&rcpt,&rcpt._header,iov,rcpt._header.iovsz_d);
    for( i=0; i<rcpt._header.iovsz_d; i++){
      if( iov[i].data) memcpy( iov[i].data,ie->data,iovsz[i]);
      ++ie;
    }
    // call service
    s->call(ComFailure::OK,&rcpt,&rcpt._header,iov,rcpt._header.iovsz_d,ctxt);
    break;
  default : break;
  }
  // end of reception
  /* FIRST SOLUTION ::init the receipt structure :
      use the request immediate sequence to build the receipt byteSeqReader iterator
  
  NOTHING TO DO
  */
  /* SECOND  SOLUTION ::init the receipt structure :
      copy the request immediate sequence in an IOVectSeq; then use it to build the receipt byteSeqReader iterator
  
  THE IOVECT SEQ MUST BE FREED
  */
  for(IOVectSeq::IteratorEntry id= iovs.beginEntry();
  id!=iovs.endEntry(); ++id){
      //	std::cout << "FREE immediate BLOCK ("<<id->data<< ")\n";
      _bA_RCV.free(id->data);
    }
  //std::cout << "DESTROY IOV STEAM  \n";
  iovs.destroy();
  /* END OF THE 2   SOLUTIONs */
  rcpt.finish();
  //std::cout << "END CALL "<< "\n";
}

void LoopBack::simulatePost( LoopRequest* req ){ int i;
   size_t* iovsz ; // vector of size of deferred data blocks
	// close the sequences 
	req->_datastream.packEos();
	req->_iovstream.putEos();
	// complete the fixed part of header 
	Header*      h=&req->_header;
	 h-> src=_node_rank;
 	// dest is given by user
 	// time stamp 
 	h->no=_comClock; ++_comClock;
 	h->size_h = req->_datastream.getFillSize();
 	h->size_d  = req->_iovstream.getByteSize();
	// number of blocks of defered data
	h->iovsz_d = req->_iovstream.getBlockSize();
	/* we need to pack the vector containing the deferred block sizes
	alloc it on stack 
	*/
	iovsz=(size_t*) alloca(sizeof(size_t)*(h->iovsz_d));
  // iterate on the iovect sequence to get block size
  i=0;
  for(IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry(); 		
		  ie!=req->_iovstream.endEntry();i++, ++ie){
			//std::cout << "POST DEFERRED ADR "<< ie->data <<" SZ "<<ie->len  << "\n";
			iovsz[i]=ie->len; 
		 }
	// call the receiver part
	//std::cout << "BEFORE RCV "<< "\n";
	simulateReceive(req, iovsz);
	//std::cout << "AFTER RCV "<< "\n";
	// notify request completion
	req->notify();
 	//std::cout << "END NOTIFY "<< "\n";
}
 
} // - namespace ACOM_...

