// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/NetworkCore/src/acom_stack_pool.inl,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: T. Gautier
// ==========================================================================

inline StackPool::StackPool()
{
  _stacks.clear();
  _free_stacks.clear();
}

  /// StackPool
inline void StackPool::init
( size_t stack_size, int stack_capacity, int max_limit )
{
  _stack_size = stack_size;
  _stack_limit = max_limit;
  _stacks.resize( stack_capacity );
  _free_stacks.clear();
  for (int i=0; i<stack_capacity; ++i)
  {
    _stacks[i] = new char[_stack_size];
    _free_stacks.push_back(_stacks[i]);
  }
  _cond.init();
}

  /// ~StackPool
inline void StackPool::destroy()
{
  if (_free_stacks.size() != _stacks.size())
    throw Runtime("[StackPool::~StackPool] some stacks are not released");
  for (int i=0; i< _stacks.size(); ++i)
  {
    delete [] (char*)_stacks[i];
  } 
  _stacks.clear();
  _free_stacks.clear();
}

inline StackPool::~StackPool( )
{
  destroy();
}

  /// return the count of free stack 
inline int StackPool::count_stacks() const
{
  return _stacks.size();
}

  /// acquire a new stack
inline void StackPool::acquire( void*& st, size_t& sz, AT::Mutex& m_release )
{
  int stack_id =0;
  sz = _stack_size;
  while (_free_stacks.empty()) 
  {
    stack_id = _stacks.size();
    if ((_stack_limit !=-1) && (stack_id < _stack_limit)) {
      _stacks.resize( stack_id+1 );
      st = _stacks[stack_id] = new char[_stack_size]; 
      return;
    }
    else {
      ++_sig;
      _cond.wait( m_release );
    }
  }
  st = _free_stacks.front(); 
  _free_stacks.pop_front(); 
  return;
}

  /// acquire a new stack
inline void StackPool::release( void*& st )
{
  _free_stacks.push_front( st );
  st = 0;
  if (_sig >0) {
    --_sig;
    _cond.signal();
  }
}
