//
// $Id: perfbandwidth.cpp,v 1.8 2003/07/28 16:21:12 olobry Exp $
//
// generic bandwidth performance test
//


#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include <sys/time.h>
#include "testgetnet.h"
#include "testutils.h"

namespace AC = acom;



std::string dec;
 
#ifdef VERBOSE_ENABLE
#define DO_VERBOSE 1
#else
#define DO_VERBOSE 0
#endif

// take care when using this macro withinf if.. else.. statements
#define OUT(o) if ((par_verbose >= (o)) && DO_VERBOSE) std::cout << "     [#" << pack_num << "] " << dec << getpid() << " " << rank << "  :" << function << " "
#define ERR std::cerr << "ERROR[#" << pack_num << "] " << dec << getpid() << " " << rank << "  :" << function << " " 


static PTH::CountDown isterm;
static PTH::Mutex mutex;
static PTH::Condition cond;

int rank = -1;
int pack_num;

int par_count = 1;
char * s_net_name, * p_net_name = 0;
char * progname;
int par_verbose = 0;
int par_szpacket = 1;
bool par_simpleoutput = false;
int par_minpacket = 0;
int par_maxpacket = 0;
int par_incpacket = 0;
int par_mulpacket = 0;
bool par_burstmode = false;
int par_period = 0;

void terminate(int exit_value) {
  AC::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: " << progname << " [OPTIONS]" << std::endl;
  std::cout << "OPTIONS:" << std::endl;
  std::cout << " -n net: choose net for secondary network"  << std::endl;
  std::cout << " -c count: number of packets to send" << std::endl;
  std::cout << " -t test_mode: burst or pingpong, default [pingpong] "  << std::endl;
  std::cout << "      * burst mode:"  << std::endl;
  std::cout << "         node #0 sends count packets then waits till node #1 acks the last one"  << std::endl;
  std::cout << "         node #1 only acknowledges last packet from node #0"  << std::endl;
  std::cout << "      * pingpong mode:"  << std::endl;
  std::cout << "         repeat count times: "  << std::endl;
  std::cout << "          node #0 sends one 'ping' packet to node #1 and waits for a 'pong' packet" << std::endl;
  std::cout << "          node #1 waits for a 'ping' packet from node #0 then sends a 'pong' packet"  << std::endl; 
  std::cout << " mono test use: "  << std::endl; 
  std::cout << " -s size: send packet of size bytes" << std::endl;
  std::cout << " multi test use:  "  << std::endl; 
  std::cout << " -m min_size: min packet size in bytes" << std::endl;
  std::cout << " -M max_size: max packet size in bytes" << std::endl;
  std::cout << " -i inc_size: increment packet size in bytes" << std::endl;
  std::cout << " -x mul: multiply factor for packet size" << std::endl;
  std::cout << " -o: simple output format" << std::endl;
  std::cout << " -T period: show intermediate results every period packets" << std::endl;

#ifdef VERBOSE_ENABLE
  std::cout << " -v verb_level: verbose mode" << std::endl;
#endif
  terminate(exit_val);
}


void parse_args(int argc, char* argv[]) {
  int opt;
  progname = argv[0];

  bool multimode, incsz_set, mulsz_set, minsz_set, maxsz_set, sz_set;
  multimode = incsz_set = mulsz_set = minsz_set = maxsz_set = sz_set= false;

  while((opt = getopt(argc, argv, "hn:c:v:s:m:M:i:x:ot:T:"))>0)
    switch(opt)
      {
      case 'n': // network name
        s_net_name = optarg;
        break;

#ifdef VERBOSE_ENABLE
      case 'v': // verbose
        par_verbose= atoi(optarg);
        break;
#endif

      case 'h': // help
        usage(0);
        break;

      case 's': // packet size
        par_szpacket = atoi(optarg);
        sz_set = true;
        break;

       case 'm': // min packet size
        par_minpacket = atoi(optarg);
        minsz_set = true;
        multimode = true;
        break;

      case 'M': // max packet size
        par_maxpacket = atoi(optarg);
        maxsz_set = true;
        multimode = true;
        break;

      case 'i': // increment for packet size
        par_incpacket = atoi(optarg);
        incsz_set = true;
        multimode = true;
        break;
        
      case 'x': // mul foactor for packet size
        par_mulpacket = atoi(optarg);
        mulsz_set = true;
        multimode = true;
        break;

     case 'o': // simple output format
       par_simpleoutput = true;
       break;

     case 't': // test mode
       if (!strcmp(optarg, "burst"))
         par_burstmode = true;
       else if (strcmp(optarg, "pingpong")) {
         std::cerr << "valid test mode are: burst and pingpong" << std::endl;
         usage(2);
       }
       break;

     case 'c': // nb times the test should be run
        par_count = atoi(optarg);
        break;

     case 'T': // period for intermediate results
        par_period = atoi(optarg);
        break;

      default:
        usage(2);
      }

  if (multimode) {
    if (sz_set) std::cerr << "Ignoring -s option with this mode" << std::endl;
    if (!minsz_set) { std::cerr << "missing -m option"  << std::endl; usage(2); }
    if (!maxsz_set) { std::cerr << "missing -M option"  << std::endl; usage(2); }
    if (!incsz_set && !mulsz_set) { 
      std::cerr << "missing -i or -x option"  << std::endl; usage(2); 
    }
    if (incsz_set && mulsz_set) { 
      std::cerr << "-i and -x are exclusive options"  << std::endl; usage(2); 
    }    
    if (par_maxpacket < par_minpacket) {
      std::cerr << "min packet size should not be greater than max packet size"  
                << std::endl; usage(2); 
    } 
  } else {
    par_minpacket = par_szpacket;
    par_maxpacket = par_szpacket;
    par_incpacket = 1;
  }
}



#define function "GET_IOVECT"

class AckService: public AC::ServiceAM {

public:

  // This function is called once the immediate data have been received
  // Deferred data may not have been totally (or even partially) received
  // Its role is give a mapping between received data and local buffers:
  // For each element of the IO vector, it must give a reference to memory 
  // location where the correspondig data have to be copied
  // Note: this function is not called atomatically, so it must be re-entrant
  void* get_iovect(
          AC::Network::Receipt* receipt,
          AC::Network::Header* header,
          AC::IOVectEntry* vector,
          AC::IOVectSize count
          )
  {
    OUT(10) << "called"  << count << std::endl;
    return 0;
  }    
  

#undef function
#define function "SERVICE"

  // This function is called once all deferred data has been copied
  // at the right location
  // Note: this function is not called atomatically, so it must be re-entrant
  void call ( 
         AC::ComFailure::Code e,      // if any 
         acom::Network::Receipt* receipt,
         AC::Network::Header* h,
         AC::IOVectEntry* iov,
         AC::IOVectSize   iov_sz,
         void* ctx
         ) throw (AC::ComFailure)
  {
    if (rank == 1) {
      OUT(10) << "PING" << std::endl;
    }
    else {
      OUT(10) << "PONG" << std::endl;
    }
    //    sleep(1);
    if (par_burstmode)
      --isterm;
    else
      {
        PTH::CriticalGuard g(mutex);
        cond.broadcast();
      }
  } // AckService::call
  
  
}; // class AckService

static AckService ack_service;


#undef function
#define function "DOIT"

int doit (int argc, char* argv[])
{
  int dest_node;
  int cpt, ret;
  size_t sz, packetsz;
  struct timeval starttime, pingpongtime, t2, diff, t3;



  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  mutex.init();
  cond.init();

  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  AC::Service::Manager.bind (2, &ack_service);

  // initialize the NetworCore library
  AC::Acom::initialize(&argc, &argv);
  AC::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()

  // get networks

  AC::Network* p_net;
  AC::Network* s_net;

  parse_args(argc, argv);

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
    terminate(1);


  OUT(10) << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT(10) << "Using " << s_net->get_name() << " as secondary network" << std::endl;
  
  //  AC::Network* network = s_net;
  AC::Network* network = s_net;


  //   p_net->initialize(&argc, &argv);
  //   s_net->initialize(&argc, &argv, p_net);

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  rank = network->rank();
  for (cpt = 0; cpt < rank; cpt++) dec=dec+"   ";

  if (network->node_count() != 2) {
    std::cerr << "this test must be run with exactly 2 nodes" << std::endl;
    terminate(2);
  }

  // Let's go

  dest_node = (rank == 0) ? 1 : 0;
  AC::Network::OutChannel* channel = network->get_default_route(dest_node);
  OUT(10) << "Got a channel to node " << dest_node << std::endl;

  if (par_simpleoutput && (rank == 0)) {
    std::cout << "    #Packets     Size (B)     BW (B/s)    BW (KB/s)    BW (MB/s) ";
    if (!par_burstmode) std::cout << "Latency (ms)" ;
    std::cout << std::endl;
  }

  for (packetsz = par_minpacket; packetsz <= par_maxpacket; ) {
    if (!par_burstmode) { // ping pong mode
      AC::ActiveMessage<AckService> am (network, &ack_service, 0, (int) par_count);

      char * packet = new char[packetsz];
      if (packetsz) {
        ret=am->put(packet, packetsz);
        if (!ret)
          ERR << "could not put the requested " << packetsz << " bytes" << std::endl;
      }


      PTH::CriticalGuard g(mutex);
      // make sure both nodes are in critical section before starting sending
      // messages
      network->barrier();

      if (rank == 0) { 
        gettimeofday(&starttime, NULL); 
        pingpongtime = starttime; 
      }

      for (pack_num = 1; pack_num <= par_count; pack_num++) {

        // node 1 first wait for msg from node 0 (PING)
        if (rank == 1) {
          OUT(5) << "WAITING:" <<  std::endl;
          cond.wait(mutex);
          OUT(5) << "WAKED UP:" << std::endl;
        } else {
          OUT(10) << "############################" << std::endl;
          OUT(1) << "PACKET #" << pack_num << std::endl;
          OUT(10) << "############################" << std::endl;
        }
        
        // send message through this channel
        channel->post(am);

        // node 0 wait for msg from node 1 (PONG)
        if (rank == 0) {
          OUT(5) << "WAITING:" << std::endl;
          cond.wait(mutex);
          OUT(5) << "WAKED UP:" << std::endl;
        }
        if (rank == 0) {
          gettimeofday(&t2, NULL);
          if (par_period && ((pack_num % par_period) == 0)) {
            timersub(&t2, &pingpongtime, &t3);
            std::cout << "[" << packetsz << "|" << pack_num <<"] avg last " << 
              par_period << " pingpong(s) (ms): " << 
              milliseconds(&t3) / par_period << std::endl;
            gettimeofday(&pingpongtime, NULL);
          }
        }
      }

      delete [] packet;

    } else { // burst mode
      // node 0 send all packets without waiting then wait the ack from node 1
      if (rank == 0) {
        AC::ActiveMessage<AckService> am (network, &ack_service, 0, par_count);

        char * packet = new char[packetsz];
        if (packetsz) {
          ret=am->put(packet, packetsz);
          if (!ret)
            ERR << "could not put the requested " << packetsz << " bytes" 
                << std::endl;
        }
    
        isterm.init(1);
        network->barrier();
        gettimeofday(&starttime, NULL);

        for (pack_num = 1; pack_num <= par_count; pack_num++) {
          OUT(10) << "############################" << std::endl;
          OUT(1) << "PACKET #" << pack_num << std::endl;
          OUT(10) << "############################" << std::endl;
          channel->post(am);
        }

        OUT(5) << "WAITING:" << isterm.getValue() << std::endl;
        isterm.wait();
        OUT(5) << "WAKED UP:" << isterm.getValue() << std::endl;
        delete [] packet;

      } else if (rank == 1) {
        AC::ActiveMessage<AckService> am (network, &ack_service, 0, 1);
        // node 1 wait all packets from node 0 then send it an ack
        isterm.init(par_count);
        network->barrier();
        isterm.wait();
        channel->post(am);
      }
    }

    if (rank == 0) {
      gettimeofday(&t2, NULL);
      timersub(&t2, &starttime, &diff);
      if (!par_simpleoutput)
        printf("Packet size: %d Bytes \n", packetsz);
      if (!par_simpleoutput)
        printf("Total elapsed time: %d.%06ds\n", diff.tv_sec, diff.tv_usec);
      timerdiv(&diff, par_count, &diff);
      if (!par_simpleoutput)
        printf("time per ping-pong: %d.%06ds\n", diff.tv_sec, diff.tv_usec);
      float bd = (float) packetsz /
        ((float) diff.tv_sec + (float) diff.tv_usec / 1000000);
      if (!par_burstmode) bd *= 2;
      if (!par_simpleoutput)
        printf("         bandwidth: %.2f B/s, %.2f KB/s, %.2f MB/s\n", 
               bd, bd / 1024, bd / (1024 * 1024));
      else {
        printf("%12d %12d %12.2f %12.2f %12.2f", par_count,
               packetsz, bd, bd / 1024, bd / (1024 * 1024));
        if (!par_burstmode) printf(" %12.3f", milliseconds(&diff) / 2 );
	printf("\n");
      }
      fflush(stdout);
    }

    if (par_incpacket)
      packetsz += par_incpacket;
    else if (par_mulpacket)
      packetsz *= par_mulpacket;
  }
    
  OUT(10) << "Terminated"<< std::endl;
  mutex.destroy();
  cond.destroy();
  terminate(0);
}
  
  


int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}
