//
// $Id: perfbarrier.cpp,v 1.5 2003/07/28 16:21:12 olobry Exp $
//
// generic network barrier test
//


#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include <sys/time.h>
#include "testgetnet.h"
#include "testutils.h"

namespace AC = acom;


char * protocols[AC::Network::P_FREE] = {
  "VOID",
  "ACTIVE MESSAGE",
  "WRITE & SIGNAL",
  "ALLOCATE, WRITE & SIGNAL"
};



std::string dec;
 
#ifdef VERBOSE_ENABLE
#define DO_VERBOSE 1
#else
#define DO_VERBOSE 0
#endif

#define OUT(o) if ((par_verbose >= (o)) && DO_VERBOSE) std::cout << "     [#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " "
#define ERR std::cerr << "ERROR[#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " " 



int rank = -1;
int test_num;

int par_count = 1;
char * s_net_name, * p_net_name = 0;
char * progname;
int par_verbose = 0;

void terminate(int exit_value) {
  AC::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: " << progname << " [OPTIONS]" << std::endl;
  std::cout << "OPTIONS:" << std::endl;
  std::cout << " -n net: choose net for secondary network"  << std::endl;
  std::cout << " -c count: run the test count times" << std::endl;
#ifdef VERBOSE_ENABLE
  std::cout << " -v: verbose mode" << std::endl;
#endif
  terminate(exit_val);
}


void parse_args(int argc, char* argv[]) {
  int opt;
  progname = argv[0];

  while((opt = getopt(argc, argv, "hn:c:v:"))>0)
    switch(opt)
      {
      case 'n': // network name
        s_net_name = optarg;
        break;

#ifdef VERBOSE_ENABLE
      case 'v': // verbose
        par_verbose= atoi(optarg);
        break;
#endif

      case 'h': // help
        usage(0);
        break;

      case 'c': // nb times the test should be run
        par_count = atoi(optarg);
        break;

      default:
        usage(2);
      }
}





#undef function
#define function "DOIT"

int doit (int argc, char* argv[])
{
  int dest_node;
  int cpt;
  size_t sz;
  struct timeval t1, t2, diff;

  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  
  // initialize the NetworCore library
  AC::Acom::initialize(&argc, &argv);
  AC::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()


  parse_args(argc, argv);

  // get networks

  AC::Network* p_net;
  AC::Network* s_net;

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
    terminate(1);


  OUT(10) << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT(10) << "Using " << s_net->get_name() << " as secondary network" << std::endl;
  
  //  AC::Network* network = s_net;
  AC::Network* network = s_net;


  //   p_net->initialize(&argc, &argv);
   //   s_net->initialize(&argc, &argv, p_net);

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  rank = network->rank();
  for (cpt = 0; cpt < rank; cpt++) dec=dec+"   ";


  // Let's go
  OUT(10) << "There is " << network->node_count() << 
    " nodes in the network"<<std::endl; 

  // with some networks like TCPNET the first barrier may be quite longer that the 
  // following ones because connections may have to be established
  // so just do one for nothing
  network->barrier();

  if (rank == 0) gettimeofday(&t1, NULL);
  for (test_num = 0; test_num < par_count; test_num++) {
    OUT(10) << "############################" << std::endl;
    OUT(1) << "TEST #" << test_num << std::endl;
    OUT(10) << "############################" << std::endl;

    network->barrier();
  }
  
  if (rank == 0) {
    gettimeofday(&t2, NULL);
    timersub(&t2, &t1, &diff);
    std::cout << "Number of packets: " << par_count << std::endl;
    printf("Total elapsed time: %d.%06ds\n", diff.tv_sec, diff.tv_usec);
    printf("  time per barrier: %.6fms\n", milliseconds(&diff) / par_count);
    fflush(stdout); 
  }

  OUT(10) << "Terminated"<< std::endl;
  terminate(0);
}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}

