//
// $Id: perflatency.cpp,v 1.7 2003/07/28 16:21:12 olobry Exp $
//
// generic network barrier test
//


#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include <sys/time.h>
#include "testgetnet.h"
#include <sys/time.h>

#include "testutils.h"

  
std::string dec;
 
#ifdef VERBOSE_ENABLE
#define DO_VERBOSE 1
#else
#define DO_VERBOSE 0
#endif

// take care when using this macro within if.. else.. statements
#define OUT(o) if ((par_verbose >= (o)) && DO_VERBOSE) std::cout << "     [#" << pack_num << "] " << dec << getpid() << " " << rank << "  :" << function << " "
#define ERR std::cerr << "ERROR[#" << pack_num << "] " << dec << getpid() << " " << rank << "  :" << function << " " 


//static PTH::CountDown isterm;
static PTH::Mutex mutex;
static PTH::Condition cond;

int rank = -1;
int pack_num;

int par_count = 1;
char * s_net_name, * p_net_name = 0;
char * progname;
int par_verbose = 0;
int par_period = 0;

void terminate(int exit_value) {
  ACOM::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: " << progname << " [OPTIONS]" << std::endl;
  std::cout << "OPTIONS:" << std::endl;
  std::cout << " -n net: choose net for secondary network"  << std::endl;
  std::cout << " -c count: run the test count times" << std::endl;
  std::cout << " -T period: show intermediate results every period packets" <<std::endl;
#ifdef VERBOSE_ENABLE
  std::cout << " -v verb_level: verbose mode" << std::endl;
#endif
  terminate(exit_val);
}


void parse_args(int argc, char* argv[]) {
  int opt;
  progname = argv[0];

  while((opt = getopt(argc, argv, "hn:c:v:T:"))>0)
    switch(opt)
      {
      case 'n': // network name
        s_net_name = optarg;
        break;

#ifdef VERBOSE_ENABLE
      case 'v': // verbose
        par_verbose= atoi(optarg);
        break;
#endif

      case 'h': // help
        usage(0);
        break;

      case 'c': // nb times the test should be run
        par_count = atoi(optarg);
        break;

     case 'T': // period for intermediate results
        par_period = atoi(optarg);
        break;
      default:
        usage(2);
      }
  if (!par_period) par_period = par_count + 1;
  
  if (getenv("INUK_NET") !=0)
    s_net_name = getenv("INUK_NET");
}



class AckService: public ACOM::ServiceAM {

public:

#undef function
#define function "SERVICE"

  // This function is called once all deferred data has been copied
  // at the right location
  // Note: this function is not called atomatically, so it must be re-entrant
  void call ( 
         ACOM::ComFailure::Code e,      // if any 
         acom::Network::Receipt* receipt,
         ACOM::Network::Header* h,
         ACOM::IOVectEntry* iov,
         ACOM::IOVectSize   iov_sz,
         void* ctx
         ) throw (ACOM::ComFailure)
  {
    if (rank == 1) {
      OUT(10) << "PING" << std::endl;
    }
    else {
      OUT(10) << "PONG" << std::endl;
    }
    {
      PTH::CriticalGuard g(mutex);
      cond.broadcast();
    }
  } // AckService::call
  
  
}; // class AckService

static AckService ack_service;


#undef function
#define function "DOIT"

int doit (int argc, char* argv[])
{
  int dest_node;
  int cpt;
  size_t sz;
  struct timeval t1, t2, diff, pingpongtime;

  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  mutex.init();
  cond.init();
  
  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  ACOM::Service::Manager.bind (2, &ack_service);

  // initialize the NetworCore library
  ACOM::Acom::initialize(&argc, &argv);
  ACOM::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()


  parse_args(argc, argv);

  // get networks

  ACOM::Network* p_net;
  ACOM::Network* s_net;

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
    terminate(1);


  OUT(10) << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT(10) << "Using " << s_net->get_name() << " as secondary network" << std::endl;
  
  //  ACOM::Network* network = s_net;
  ACOM::Network* network = s_net;
  network->barrier();

  //   p_net->initialize(&argc, &argv);
  //   s_net->initialize(&argc, &argv, p_net);

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  rank = network->rank();
  for (cpt = 0; cpt < rank; cpt++) dec=dec+"   ";


  if (network->node_count() != 2) {
    std::cerr << "this test must be run with exactly 2 nodes" << std::endl;
    terminate(2);
  }

  ACOM::ActiveMessage<AckService> am (network, &ack_service, 0, (int) par_count);
  
  dest_node = (rank == 0) ? 1 : 0;
  ACOM::Network::OutChannel* channel = network->get_default_route(dest_node);
  OUT(10) << "Got a channel to node " << dest_node << std::endl;

  PTH::CriticalGuard g(mutex);
  // make sure both nodes are in critical section before starting sending
  // messages
  network->barrier();
  if (rank == 0) {
    gettimeofday(&t1, NULL);
    pingpongtime = t1;
  }


  for (pack_num = 1; pack_num <= par_count; pack_num++) {
    if (rank == 1) {
      cond.wait(mutex);
    } else {
      OUT(10) << "############################" << std::endl;
      OUT(1) << "TEST #" << pack_num << std::endl;
      OUT(10) << "############################" << std::endl;
    }

    // send message through this channel
    channel->post(am);

    if (rank == 0) {
      cond.wait(mutex);
      if ((pack_num % par_period) == 0) {
        gettimeofday(&t2, NULL);
        timersub(&t2, &pingpongtime, &t2);
        std::cout << "[" << pack_num << "] avg last " << 
          par_period << " latencies (ms): " << 
          milliseconds(&t2) / (par_period * 2) << std::endl;
        gettimeofday(&pingpongtime, NULL);
      }
    }
  }

  if (rank == 0) {
    gettimeofday(&t2, NULL);
    timersub(&t2, &t1, &diff);
    std::cout << "# of Packets " << par_count << std::endl;
    printf("Total elapsed time: %d.%06ds\n", diff.tv_sec, diff.tv_usec);
    //    timerdiv(&diff, par_count*2, &diff);
    printf("           latency: %.6fms\n", milliseconds(&diff) / (par_count * 2));
  }

  OUT(10) << "Terminated"<< std::endl;
  //throw on MacOs X: mutex.destroy();
  //throw on MacOs X: cond.destroy();
  network->barrier();
  terminate(0);
}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}

