//
// $Id: testall2all.cpp,v 1.3 2003/07/28 08:06:12 thierry Exp $
//
// test active message
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>

static int rank = 0;
static volatile int count_ping = 0;



class PingService: public ACOM::ServiceAM {
public:
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void* cc
  ) throw (ACOM::ComFailure)
  {
    std::cout << rank << " :: PING: " << h->src << " ->";
    if (e == ACOM::ComFailure::OK) 
       std::cout << "msg='" << (char*)iov[0].data << "'" << std::endl; 
    else 
       std::cout << "Error" << std::endl;
    ++count_ping;
  }
};

static PingService ping;


int doit (int argc, char* argv[])
{
  ACOM::Acom::initialize( &argc, &argv );
  ACOM::Service::Manager.bind (2, &ping);
  ACOM::Acom::commit( );

  /* Get the network:
  */
  ACOM::Network* net = 0;
  ACOM::Acom::const_iterator icurr = ACOM::Acom::begin_network();
  ACOM::Acom::const_iterator iend  = ACOM::Acom::end_network();
  if (icurr == iend) {
     std::cout << "No network attached to the process, exit" << std::endl;
     return 0;
  }
  else {
    if (getenv("INUK_NET") != 0) {
      try {
        net = ACOM::Acom::resolv_network( getenv("INUK_NET") );
      } catch (ACOM::InvalidArgument& e) {
        std::cerr << "Network '" << getenv("INUK_NET") << "' does not exist" << std::endl;
        throw e;
      }
    }
    else 
      net = *icurr;
    std::cout << "Using network: " << net->get_name() << std::endl;
  }
  
  int my_netrank = rank = net->rank();
  std::cout << my_netrank << "::Before Barrier 1 ok" << std::endl;
  net->barrier();
  int netcount = net->node_count();
  std::cout << my_netrank << "::After barrier 1 ok" << std::endl;
  

  std::cout << "Got a channel to 1: my_rank is=" << my_netrank 
            << " over #nodes=" << netcount << std::endl;

  int ITER = 10000;
  for (int k=0; k < ITER; ++k) {
    for (int dest=0; dest < netcount; ++dest)
    {
      if (dest == rank) continue;
  
      /* Get a channel:
      */
      ACOM::Network::OutChannel* channel = net->get_default_route(dest);
  
      //std::cout << rank << " :: Make ping channel to " << dest << std::endl;
      char MSG[256];
      sprintf(MSG,"C'est un ping from:%i , #serial:%i",my_netrank, k);
      ACOM::ActiveMessage<PingService> am (net, &ping, 0);
      am->put(MSG, strlen(MSG)+1);
      am.post(channel);
    }
  }
  while (count_ping < ITER*(netcount-1)) {
    usleep( 1000 );
  }
  std::cout << my_netrank << "::Before Barrier 2 ok" << std::endl;
  net->barrier();
  std::cout << my_netrank << "::After Barrier 2 ok" << std::endl;

  std::cout << my_netrank << " :: Ready to Terminate" << std::endl;
  ACOM::Acom::terminate( );
}

int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (ACOM::RuntimeError& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
    ACOM::Acom::terminate( );
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    ACOM::Acom::terminate( );
  }
}
