//
// $Id: testam2.cpp,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// test active message: a call to one service from rank 0
// with callback
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>

namespace AC = acom;

static int rank = -1;
static volatile bool isterm = false;
static AC::Network* net = 0;

class MyCallback : public AC::Callback {
public:
  void done( AC::Network::Request* req )
  {
    static int count =1;
    ++count;
    std::cout << rank << ":: Callback called:" << count 
              << ", node_count=" << net->node_count() << std::endl;
    if (count >= net->node_count()) {
      isterm = true;
    }
  }
};
static MyCallback cbk;

class PingService: public AC::Service {
public:
  void call ( 
       AC::ComFailure::Code e,      // if any 
       AC::Network::Header* h,
       AC::IOVectEntry* iov,
       AC::IOVectSize   iov_sz
  ) throw (AC::ComFailure)
  {
    std::cout << ":: PING: ";
    if (e == AC::ComFailure::OK) 
       std::cout << "msg='" << (char*)iov[0].data << "'" << std::endl; 
    else 
       std::cout << "Error" << std::endl;
    isterm = true;
  }
};

static PingService ping;

int doit (int argc, char* argv[])
{
  AC::System::initialize( &argc, &argv );
  AC::Service::Manager.bind (2, &ping);
  AC::System::initcommit( );

  /* Get the network:
  */
  AC::System::const_iterator icurr = AC::System::begin_network();
  AC::System::const_iterator iend  = AC::System::end_network();
  if (icurr == iend) {
     std::cout << "No network attached to the process, exit" << std::endl;
     return 0;
  }
  else
    std::cout << "Using network: " << (*icurr)->get_name() << std::endl;
  net = *icurr;
  rank = net->rank();

  std::cout << "Got a channel to 1: rank is=" << rank << std::endl;
  if (rank ==0) 
  {
    for (int n=1; n < net->node_count(); ++n)
    {
      /* Get a channel:
      */
      AC::OutChannel* channel = net->get_local_route(n);

      std::cout << "Make ping channel " << std::endl;
      char* MSG = "C'est un ping" ;

      AC::ActiveMessage<PingService> am (channel, &ping, &cbk, 1);
      am->io_v[0] = AC::IOVectEntry(MSG, strlen(MSG)+1);
      am->post();
    }
  }
  // - wait calling of PING or Callback object
  while (!isterm) {
    usleep( 100 );
  }
  std::cout << rank << ":: Ready to Terminate" << std::endl;
  AC::System::terminate( );
  std::cout << rank << ":: End Terminate" << std::endl;
}

int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (AC::Exception& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
