//
// $Id: testam3.cpp,v 1.4 2003/07/28 07:59:34 thierry Exp $
//
// test active message: latency
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

using namespace std;


#define SECPERUSEC 1000000

static int SIZE_MAX = 4096; // - in bytes 
static int ITER_PING;

static int rank = 0;
static volatile int count_ping = 0;
static volatile int count_pong = 0;

static ACOM::Network* net = 0;

class PongService: public ACOM::ServiceAM {
public:
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void *cc
  ) throw (ACOM::ComFailure)
  {
    count_pong =1;
  }
};
static PongService pong;


class PingService: public ACOM::ServiceAM {
public:
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void *cc
  ) throw (ACOM::ComFailure)
  {
    ++count_ping;
    if (e != ACOM::ComFailure::OK) {
      cerr << "Error in msg='" << e 
                << "', #iov" << iov_sz  << ", len[0]" << iov[0].len
                << endl;
    }
    if (count_ping == ITER_PING) {
      //cout << rank << " :: End Phase: " << count_ping << endl;
      ACOM::Network::OutChannel* channel = net->get_default_route(0);
      ACOM::ActiveMessage<PongService> am (net, &pong, 0);
      am.post(channel);
      count_ping =0;
    }
  }
};
static PingService ping;


int doit (int len )
{
  //cout << rank << " :: DOIT == rank is=" << rank << endl;
//  cout << rank << " :: Test using ITER_PING=" << ITER_PING 
//       << ", SIZE_BUFF=" << len << endl;

  net->barrier();
  if (rank ==0) 
  {
    struct timeval time1,time2;
    double t1,t2;

    /* Get a channel, form the message
    */
    char* msg = new char[len];
    ACOM::Network::OutChannel* channel = net->get_default_route(1);
    ACOM::ActiveMessage<PingService> am (net, &ping, 0, ITER_PING);
    am->put(msg, len);

    gettimeofday(&time1,NULL);
    t1= ((double)time1.tv_sec)*((double)SECPERUSEC) + (double)time1.tv_usec;

    for (int k=0; k < ITER_PING; ++k) {
      am.post(channel);
    }
    while (count_pong ==0) usleep( 10 );

    gettimeofday(&time2,NULL);
    t2= ((double)time2.tv_sec)*((double)SECPERUSEC) + (double)time2.tv_usec;

    delete [] msg;

    double time = (t2-t1) / ITER_PING;
    cout << setw(15) << setfill(' ') << ((double)len)/1024.0
         << setw(20) << setfill(' ') << time
         << setw(20) << setfill(' ') << ((double)len)/time*((double)SECPERUSEC)
                                         / 1024.0/1024.0
         << endl;
    /*
    std::cout << "Count_pong =" << count_pong << std::endl;
    std::cout << "Time  us =" << (t2-t1) / ITER_PING << std::endl;
    std::cout << "MBytes/s =" << (ITER_PING*SIZE_MAX*1.0)/(t2-t1)/1024.0/1024.0 
              << std::endl;
    */
    count_pong =0;
    /* Send ACK to server for next loop */
    channel = net->get_default_route(1);
    ACOM::ActiveMessage<PongService> am2 (net, &pong, 0);
    am2.post(channel);
  }
  else {
    while (count_pong ==0) usleep( 10 );
    count_pong =0;
    //std::cout << rank << "::[DoIt] Recv ACK Next loop" << std::endl;
  }
}

int main (int argc, char* argv[])
{
  try {
    ACOM::Acom::initialize( &argc, &argv );
    ACOM::Service::Manager.bind (2, &ping);
    ACOM::Service::Manager.bind (3, &pong);
    ACOM::Acom::commit( );
    //ACOM::Acom::attach_debugger( );

    /* Get the default network:
    */
    net = ACOM::Acom::get_default_network();
    rank = net->rank();
    int nodecount = net->node_count();
    std::cout << "My rank=" << rank << std::endl;
    std::cout << "My nodecount=" << nodecount << std::endl;
    net->barrier();
    rank = net->rank();
    nodecount = net->node_count();

    /* Get the argument
    */
    if (rank ==0) {
      cout << rank << " :: My_rank is=" << rank 
           << " over #nodes=" << nodecount << endl;
    }
    cout << argv[1] << endl;
    cout << argv[2] << endl;
    cout << argv[3] << endl;

    SIZE_MAX  = atoi(argv[1]);
    int stride = atoi(argv[2]);
    ITER_PING = atoi(argv[3]);

    for (int i=0; i<2; ++i)
    {
      int len = 0;
      if (rank ==0) {
        cout << "# Experience:" << i 
             << "\n# SIZE MAX= " << SIZE_MAX
                        << ", STRIDE=" << stride 
                        << ", ITER=" << ITER_PING 
                        << std::endl;
        cout << setw(15) << setfill(' ') << "# Msg. Size (KB)"
            << setw(20) << setfill(' ') << "One-way(usecs)"
            << setw(20) << setfill(' ') << "BandWidth (MB/s)"
            << endl;
      }
  
      while (len <= SIZE_MAX) {
        /* Call one experiment : */
        doit(len);
        len += stride;
      }
    }

    /* Terminate */
    std::cout << rank << " :: Ready to Terminate" << std::endl;
    ACOM::Acom::terminate( );

  } catch (ACOM::Runtime& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
    ACOM::Acom::terminate( );
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    ACOM::Acom::terminate( );
  }
  return 0;
}
