//
// $Id: testaws1.cpp,v 1.2 2003/07/15 17:32:22 thierry Exp $ 
//
// test write and signal of multiple IOvect
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>

namespace AC = acom;

// - Message sended
char* MSG0 = "C'est un ping";
char* MSG1 = "Le chocolat c'est bon !";
char* MSG2 = "Il fait beau aujourd'hui, c'est ben vrai, fait beau!";
int MSG0_len = strlen(MSG0)+1;
int MSG1_len = strlen(MSG1)+1;
int MSG2_len = strlen(MSG2)+1;

static volatile bool isterm = false;
static int rank =-1;

class PingService: public AC::Service
{
public:
  PingService()
  {
    _myiovect[0] = AC::IOVectEntry(_mydata, MSG0_len );
    _myiovect[1] = AC::IOVectEntry(_mydata+MSG0_len, MSG1_len );
    _myiovect[2] = AC::IOVectEntry(_mydata+MSG0_len+MSG1_len, MSG2_len );
  }

  void call ( 
       AC::ComFailure::Code e,      // if any 
       AC::Network::Header* h,
       AC::IOVectEntry* iov,
       AC::IOVectSize   iov_sz
  ) throw (AC::ComFailure)
  {
    std::cout << ":: PING: ";

    std::cout << "msg[0]=@" << iov[0].data << std::endl; 
    std::cout << "msg[1]=@" << iov[1].data << std::endl; 
    std::cout << "msg[2]=@" << iov[2].data << std::endl; 

    if (e == AC::ComFailure::OK) {

       std::cout << "msg[0]='" << (char*)iov[0].data << "'" << std::endl; 
       std::cout << "msg[1]='" << (char*)iov[1].data << "'" << std::endl; 
       std::cout << "msg[2]='" << (char*)iov[2].data << "'" << std::endl; 
    }
    else 
       std::cout << "Error" << std::endl;
    isterm = true;
  }

  void get_iovect(
    AC::Network::Header* header, AC::IOVectEntry* vector, AC::IOVectSize count
  )
  {
    count  =3;
    vector = _myiovect;
  }

  AC::IOVectEntry _myiovect[3];
  char _mydata[256];
};

static PingService ping;

int doit (int argc, char* argv[])
{
  AC::System::initialize( &argc, &argv );
  AC::Service::Manager.bind (2, &ping);
  AC::System::initcommit( );

  /* Get the network:
  */
  AC::System::const_iterator icurr = AC::System::begin_network();
  AC::System::const_iterator iend  = AC::System::end_network();
  if (icurr == iend) {
     std::cout << "No network attached to the process, exit" << std::endl;
     return 0;
  }
  else
    std::cout << "Using network: " << (*icurr)->get_name() << std::endl;
  AC::Network* net = *icurr;
  rank = net->rank();


  if (rank ==0) 
  {
    for (int n=1; n < net->node_count(); ++n)
    {
      try {
        /* Get a channel:
        */
        AC::OutChannel* channel = net->get_local_route(n);
        std::cout << "Got a channel to 1: rank is=" << rank << std::endl;

        std::cout << "Make ping channel " << std::endl;
        AC::AllocWriteAndSignal<PingService> am (channel, &ping, 0, 3);
        am->io_v[0] = AC::IOVectEntry(MSG0, strlen(MSG0)+1);
        am->io_v[1] = AC::IOVectEntry(MSG1, strlen(MSG1)+1);
        am->io_v[2] = AC::IOVectEntry(MSG2, strlen(MSG2)+1);
        am->post();
      } catch (...) {
        std::cerr << " Error for sending to node: " << n << std::endl;
      }
    }
  }
  else {
    while (!isterm) {
      usleep( 1000 );
    }
    // - wait calling of PING
  }
  std::cout << "Ready to Terminate" << std::endl;
  AC::System::terminate( );
}

int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (AC::Exception& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
