//
// $Id: testaws3.cpp,v 1.1 2003/07/15 17:16:47 thierry Exp $
//
// test active message: latency
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <sys/time.h>
#include <iostream>
#include <iomanip>

using namespace std;


#define SECPERUSEC 1000000

static int SIZE_MAX = 4096; // - in bytes 
static int ITER_PING;

static int rank = 0;
static volatile int count_ping = 0;
static volatile int count_pong = 0;

static ACOM::Network* net = 0;

class PongService: public ACOM::ServiceAM {
public:
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void *cc
  ) throw (ACOM::ComFailure)
  {
    count_pong =1;
  }
};
static PongService pong;


class PingService: public ACOM::ServiceAWS {
public:
  void call ( 
       ACOM::ComFailure::Code e,      // if any 
       ACOM::Network::Receipt* r,
       ACOM::Network::Header* h,
       ACOM::IOVectEntry* iov,
       ACOM::IOVectSize   iov_sz,
       void *cc
  ) throw (ACOM::ComFailure)
  {
    ++count_ping;
    if (e != ACOM::ComFailure::OK) {
      cerr << "Error in msg='" << e 
                << "', #iov" << iov_sz  << ", len[0]" << iov[0].len
                << endl;
    }
    if (count_ping == ITER_PING) {
      //cout << rank << " :: End Phase: " << count_ping << endl;
      ACOM::Network::OutChannel* channel = net->get_default_route(0);
      ACOM::ActiveMessage<PongService> am (net, &pong, 0);
      am.post(channel);
			count_ping =0;
    }
    free(iov[0].data);
  }
};
static PingService ping;


int doit (int len )
{
  //cout << rank << " :: DOIT == rank is=" << rank << endl;
//	cout << rank << " :: Test using ITER_PING=" << ITER_PING 
//       << ", SIZE_BUFF=" << len << endl;

  net->barrier();
  if (rank ==0) 
  {
    struct timeval time1,time2;
    double t1,t2;

    /* Get a channel, form the message
    */
    char* msg = new char[len];
    ACOM::Network::OutChannel* channel = net->get_default_route(1);
    ACOM::AllocWriteAndSignal<PingService> am (net, &ping, 0, ITER_PING);
    am->put(msg, len);

    gettimeofday(&time1,NULL);
    t1= ((double)time1.tv_sec)*((double)SECPERUSEC) + (double)time1.tv_usec;

    for (int k=0; k < ITER_PING; ++k) {
      am.post(channel);
    }
    while (count_pong ==0) usleep( 10 );

    gettimeofday(&time2,NULL);
    t2= ((double)time2.tv_sec)*((double)SECPERUSEC) + (double)time2.tv_usec;

		delete [] msg;

		double time = (t2-t1) / ITER_PING;
		cout << setw(15) << setfill(' ') << ((double)len)/1024.0
		     << setw(20) << setfill(' ') << time
		     << setw(20) << setfill(' ') << ((double)len)/time*((double)SECPERUSEC)
												                 / 1024.0/1024.0
				 << endl;
		/*
    std::cout << "Count_pong =" << count_pong << std::endl;
    std::cout << "Time  us =" << (t2-t1) / ITER_PING << std::endl;
    std::cout << "MBytes/s =" << (ITER_PING*SIZE_MAX*1.0)/(t2-t1)/1024.0/1024.0 
              << std::endl;
		*/
		count_pong =0;
		/* Send ACK to server for next loop */
    channel = net->get_default_route(1);
    ACOM::ActiveMessage<PongService> am2 (net, &pong, 0);
    am2.post(channel);
  }
  else {
    while (count_pong ==0) usleep( 10 );
		count_pong =0;
    //std::cout << rank << "::[DoIt] Recv ACK Next loop" << std::endl;
  }
}

int main (int argc, char* argv[])
{
  try {
    ACOM::Acom::initialize( &argc, &argv );
    ACOM::Service::Manager.bind (2, &ping);
    ACOM::Service::Manager.bind (3, &pong);
    ACOM::Acom::commit( );
    //ACOM::Acom::attach_debugger( );

    /* Get the network:
    */
    if (getenv("INUK_NET") ==0) 
    {
      ACOM::Acom::const_iterator icurr = ACOM::Acom::begin_network();
      ACOM::Acom::const_iterator iend  = ACOM::Acom::end_network();
      if (icurr == iend) {
        cout << "No network attached to the process, exit" << endl;
        return 0;
      }
      net = *icurr;
    } else {
      net = ACOM::Acom::resolv_network(getenv("INUK_NET"));
    }
		/*
    else 
      cout << "Using network: " << (*icurr)->get_name() << endl;
		*/
    net->barrier();
    rank = net->rank();
    int netcount = net->node_count();

    /* Get the argument
    */
		if (rank ==0) {
      cout << rank << " :: My_rank is=" << rank 
           << " over #nodes=" << netcount << endl;
    }
		cout << argv[1] << endl;
		cout << argv[2] << endl;
		cout << argv[3] << endl;

		int len = 0;
		SIZE_MAX  = atoi(argv[1]);
		int stride = atoi(argv[2]);
		ITER_PING = atoi(argv[3]);
		if (rank ==0) {
		  cout << "# SIZE MAX= " << SIZE_MAX
                       << ", STRIDE=" << stride 
                       << ", ITER=" << ITER_PING 
                       << std::endl;
		  cout << setw(15) << setfill(' ') << "# Msg. Size (KB)"
		  		 << setw(20) << setfill(' ') << "One-way(usecs)"
		  		 << setw(20) << setfill(' ') << "BandWidth (MB/s)"
		  		 << endl;
		}

		while (len <= SIZE_MAX) {
		  /* Call one experiment : */
      doit(len);
			len += stride;
    }

		/* Terminate */
    std::cout << rank << " :: Ready to Terminate" << std::endl;
    ACOM::Acom::terminate( );

  } catch (ACOM::Runtime& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
    ACOM::Acom::terminate( );
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    ACOM::Acom::terminate( );
  }
  return 0;
}
