//
// $Id: testbarrier.cpp,v 1.7 2003/07/28 16:21:12 olobry Exp $
//
// generic network barrier test
//


#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include "testgetnet.h"
#include "testutils.h"

namespace AC = acom;


char * protocols[AC::Network::P_FREE] = {
  "VOID",
  "ACTIVE MESSAGE",
  "WRITE & SIGNAL",
  "ALLOCATE, WRITE & SIGNAL"
};



std::string dec;
 
#ifndef VERBOSE_DISABLE
#define DO_VERBOSE 1
#else
#define DO_VERBOSE 0
#endif

#define OUT(o) if (DO_VERBOSE && (par_verbose >= (o)) && ((par_verb_node == -1) || (par_verb_node == rank)) ) std::cout << "     [#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " "
#define ERR std::cerr << "ERROR[#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " " 


static PTH::CountDown isterm;
static int volatile test_var;


int rank = -1;
int test_num;

int par_count = 1;
int par_period = 0;
char * par_prot;
char * s_net_name, * p_net_name = 0;
char * progname;
ACOM::Network::ProtocolNo par_prot_no = AC::Network::P_AM;
int par_verbose = 0;
int par_verb_node = -1;

void terminate(int exit_value) {
  AC::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: " << progname << " [OPTIONS]" << std::endl;
  std::cout << "OPTIONS:" << std::endl;
  std::cout << " -n net: choose net for secondary network"  << std::endl;
  std::cout << " -p protocol: use protocol when sending messages [am]" << std::endl;
  std::cout << " -c count: run the test count times [1]" << std::endl;
  std::cout << " -T period: verify every period barriers [count/10] " << std::endl;
#ifndef VERBOSE_DISABLE
  std::cout << " -v verb_level: verbose mode" << std::endl;
  std::cout << " -N node : verbose only node node" << std::endl;
#endif
  terminate(exit_val);
}


void parse_args(int argc, char* argv[]) {
  int opt;
  progname = argv[0];

  while((opt = getopt(argc, argv, "hn:c:p:v:N:T:"))>0)
    switch(opt)
      {
      case 'n': // network name
        s_net_name = optarg;
        break;

#ifndef VERBOSE_DISABLE
      case 'v': // verbose level
        par_verbose= atoi(optarg);
        break;
      case 'N': // verbose node
        par_verb_node= atoi(optarg);
        break;
#endif

      case 'h': // help
        usage(0);
        break;

      case 'c': // nb times the test should be run
        par_count = atoi(optarg);
        break;

      case 'T': // check period
        par_period = atoi(optarg);
        break;

      case 'p': // protocol type
        par_prot = optarg;
        if (!strcmp(par_prot, "am"))
          par_prot_no = AC::Network::P_AM; 
        else if (!strcmp(par_prot, "ws"))
          par_prot_no = AC::Network::P_WS;
        else if (!strcmp(par_prot, "aws"))
          par_prot_no = AC::Network::P_AWS;
        else {
          std::cerr << "Bad protocol type. Possible values are: am, ws, aws" 
                    << std::endl;
          terminate(2);
        }
        break;

      default:
        usage(2);
      }
  // by default check 10 % of barriers
  if (!par_period)
    par_period = par_count / 10;
  if (!par_period)
    par_period = 1;


}

  PTH::Mutex mutex;



class AckService: public AC::Service {

public:

#undef function
#define function "GET_IOVECT"


  // This function is called once the immediate data have been received
  // Deferred data may not have been totally (or even partially) received
  // Its role is give a mapping between received data and local buffers:
  // For each element of the IO vector, it must give a reference to memory 
  // location where the correspondig data have to be copied
  // Note: this function is not called atomatically, so it must be re-entrant
  void* get_iovect(
          AC::Network::Receipt* receipt,
          AC::Network::Header* header,
          AC::IOVectEntry* vector,
          AC::IOVectSize count
          )
  {
    OUT(10) << "called"  << count << std::endl;
    return 0;
  }    
  

#undef function
#define function "SERVICE"

  // This function is called once all deferred data has been copied
  // at the right location
  // Note: this function is not called atomatically, so it must be re-entrant
  void call ( 
         AC::ComFailure::Code e,      // if any 
         acom::Network::Receipt* receipt,
         AC::Network::Header* h,
         AC::IOVectEntry* iov,
         AC::IOVectSize   iov_sz,
         void* ctx
         ) throw (AC::ComFailure)
  {
    int sender;
    int value;

    OUT(10) << "Enter: protocol type: " << protocols[h->protocol] << std::endl;
    
    // check communication errors
    if (e != AC::ComFailure::OK) {
      ERR << "-- Communication errors detected" << std::endl;
      return;
    }
    
    // check that this function is not called too many times
    if (isterm.getValue() <= 0) {
      ERR << "called too many times !!!" << std::endl;
      ERR << "isterm.getValue() = " << isterm.getValue() << std::endl;
      return;
    }

    // check that the sent value equals the expected one
    receipt->unpack(&value, sizeof(value));
    if (value != test_var) {
      ERR << "incorrect received value:" << value << std::endl;
      ERR << "expected value          :" << test_var << std::endl;
    }
    // one more message received
    if (isterm.getValue() == 0)
      ERR << "countdown value is zero !!!" << std::endl;

    --isterm;
    
    if (isterm.getValue() == 0)
      OUT(2) << "should wakeup" << std::endl;

    OUT(10) << "Exit" << std::endl;    
  } // AckService::call
  
  
}; // class AckService

static AckService ack_service;




#undef function
#define function "DOIT"

int doit (int argc, char* argv[])
{
  int dest_node;
  int cpt;
  size_t sz;

  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  
  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  AC::Service::Manager.bind (2, &ack_service);

  // initialize the NetworCore library
  AC::Acom::initialize(&argc, &argv);
  AC::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()


  parse_args(argc, argv);

  // get networks

  AC::Network* p_net;
  AC::Network* s_net;

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
    terminate(1);


  OUT(10) << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT(10) << "Using " << s_net->get_name() << " as secondary network" << std::endl;
  
  AC::Network* network = s_net;


  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  rank = network->rank();
  for (cpt = 0; cpt < rank; cpt++) dec=dec+"   ";


  // Let's go
  OUT(10) << "Start sending messages to other nodes"<<std::endl; 
  OUT(10) << "There is " << network->node_count() << 
    " nodes in the network"<<std::endl; 


  for (test_num = 1; test_num <= par_count; test_num++) {
    OUT(10) << "############################" << std::endl;
    OUT(1) << "TEST #" << test_num << std::endl;
    OUT(10) << "############################" << std::endl;

    if ((test_num % par_period) != 0) {
      network->barrier();
      continue;
    }
    isterm.init(network->node_count()-1); 
    test_var++;
    // synchronize with other nodes to make sure that everybody 
    // has incremented the variable test_var and initialized 
    // the countdown object
    network->barrier();

    // create one message per possible protocol, only once will be used
    // note that the same message is constructed only once and sent many
    // times (actually network->node_count() - 1 times)
    AC::Message msg(network, par_prot_no, &ack_service, 
	0, (int) network->node_count() - 1);

    //Pack value
    sz = msg->pack((int*) &test_var, sizeof(test_var)); 

    if (sz != sizeof(test_var))
      ERR << "packed only " << sz << " bytes out of " 
          << sizeof(test_var) << std::endl;
    
      // Now, the same message is sent to all other nodes
      for (dest_node = 0; dest_node < network->node_count(); ++dest_node)
        {
          if (dest_node != rank) // do not send to myself
            {
              // get a channel to node n through our network
              AC::Network::OutChannel* channel = network->get_default_route(dest_node);
              OUT(10) << "Got a channel to node " 
                  << dest_node << std::endl;
              
              msg.post(channel);
              OUT(10) << "Message sent to node " << dest_node << std::endl;
            }
        }
      
      // wait until all messages have been locally received from other nodes
      // (this coutdown object is decremented in the service)
      OUT(2) << "isterm wait" << std::endl;
      isterm.wait();
      OUT(2) << "isterm waked up" << std::endl;
      OUT(10) << "received all messages from other nodes" << std::endl;
  }

  AC::Acom::terminate();
  OUT(10) << "Terminated"<< std::endl;

}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}

