//
// $Id: testfmt2.cpp,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// test format de definition and stream for format.
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>

namespace AC = acom;


class MyOStream : public AC::OStream {
public:
	MyOStream( void* buff_IA, int sz_IA, AC::IOVectEntry* buff_DA, int sz_DA)
   : AC::OStream(buff_IA,sz_IA, buff_DA,sz_DA)
  { }
  bool overflow( Mode m, size_t len ) { 
    std::cout << "Overflow of stream buffer: want to add len=" << len << endl;
    switch (m) {
    case Map::IA:
      std::cout << "IA:: _IA_pos=" << _IA_pos 
                << ", _IA_len=" << _IA_len << endl;
      break;
    case Map::DA:
      std::cout << "DA:: _DA_pos=" << _DA_pos 
                << ", _DA_iovsz=" << _DA_iovsz << endl;
      break;
    };
  }
};

/* SMALL */
struct SMALL {};
AC::OStream& operator<< (AC::OStream& os, const SMALL& a )
{
  int i=3;
  os << i;
  os.write( AC::Map::DA, new char[10], 10);
}

AC::IStream& operator>> (AC::IStream& os, SMALL& a )
{
  int i;
  os >> i;
  void* data = new char[10];
  os.read( AC::Map::DA, data, 10 );
}

/* BIG_IA */
struct BIG_IA {};

AC::OStream& operator<< (AC::OStream& os, const BIG_IA& a )
{
  int val=3;
  for (int i=0; i<8; ++i)
    os << val;
}

AC::IStream& operator>> (AC::IStream& os, BIG_IA& a )
{
  int val;
  for (int i=0; i<8; ++i)
    os >> val;
}

/* BIG_DA */
struct BIG_DA {};

char tmp[256];

AC::OStream& operator<< (AC::OStream& os, const BIG_DA& a )
{
  for (int i=0; i<4; ++i)
    os.write( AC::Map::DA, tmp, 256 );
}

AC::IStream& operator>> (AC::IStream& os, BIG_DA& a )
{
  for (int i=0; i<4; ++i)
    os.read( AC::Map::DA, tmp, 256 );
}



/* DOIT */
int doit (int argc, char* argv[])
{
  try {
    std::cout <<  "-- Test small" << std::endl;
    SMALL small;

    char* IAiov = new char[8]; 
    std::vector<AC::IOVectEntry> DAiov(2);
  
    MyOStream msg( IAiov, 8, &DAiov[0], 2); 
    msg << small;

  } catch (AC::NoMoreResource& ex) {
    std::cout << "Catch [NoMoreResource] exception: " << ex.what() << std::endl;
  } catch (AC::Exception& ex) {
    std::cout << "Catch exception: " << ex.what() << std::endl;
  } catch (...) {
    std::cout << "Unknown exception" << std::endl;
  }
  std::cout <<  "-- END Test small" << std::endl;

  try {
    std::cout <<  "\n-- Test big IA" << std::endl;
    BIG_IA big;

    char* IAiov = new char[8]; 
    std::vector<AC::IOVectEntry> DAiov(2);
  
    MyOStream msg( IAiov, 8, &DAiov[0], 2); 
    msg << big;

  } catch (AC::NoMoreResource& ex) {
    std::cout << "Catch [NoMoreResource] exception: " << ex.what() << std::endl;
  } catch (AC::Exception& ex) {
    std::cout << "Catch exception: " << ex.what() << std::endl;
  } catch (...) {
    std::cout << "Unknown exception" << std::endl;
  }
  std::cout <<  "-- END Test big IA" << std::endl;

  try {
    std::cout <<  "\n-- Test big DA" << std::endl;
    BIG_DA big;

    char* IAiov = new char[8];
    std::vector<AC::IOVectEntry> DAiov(2);

    MyOStream msg( IAiov, 8, &DAiov[0], 2);
    msg << big;

  } catch (AC::NoMoreResource& ex) {
    std::cout << "Catch [NoMoreResource] exception: " << ex.what() << std::endl;  } catch (AC::Exception& ex) {
    std::cout << "Catch exception: " << ex.what() << std::endl;
  } catch (...) {
    std::cout << "Unknown exception" << std::endl;
  }
  std::cout <<  "-- END Test big DA" << std::endl;

}

int main (int argc, char* argv[])
{
  try {
    AC::System::initialize( &argc, &argv );
    std::cout << "[MAIN] after initialize" << std::endl;

    AC::System::initcommit( );
    std::cout << "[MAIN] after initcommit" << std::endl;

    int ret = doit(argc, argv);

    AC::System::terminate( );
    return ret;

  } catch (AC::NoMoreResource& ex) {
    std::cerr << "Catch [NoMoreResource] exception: " << ex.what() << std::endl;
  } catch (AC::Exception& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  } catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
