//
// $Id: testsndrcv.cpp,v 1.7 2003/07/28 16:21:12 olobry Exp $
//
// generic network test
//

#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include "testgetnet.h"
#include "testutils.h"


char * protocols[ACOM::Network::P_FREE] = {
  "VOID",
  "ACTIVE MESSAGE",
  "WRITE & SIGNAL",
  "ALLOCATE, WRITE & SIGNAL"
};



std::string dec;

#ifdef VERBOSE_ENABLE
#define DO_VERBOSE 1
#else
#define DO_VERBOSE 0
#endif

#define OUT(o) if ((par_verbose >= (o)) && DO_VERBOSE) std::cout << "     [#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " "
#define ERR std::cerr << "ERROR[#" << test_num << "] " << dec << getpid() << " " << rank << "  :" << function << " " 



int rank = -1;
static PTH::CountDown isterm;
int test_num = -1;

int par_nton = 1;
int par_nbimm = 0;
int par_szimm = 1;
int par_nbdef = 0;
int par_szdef = 1;
int par_count = 1;
int par_sync = 0;
int par_barrier = 0;
char * par_prot;
char * s_net_name, * p_net_name = 0;
char * progname;
ACOM::Network::ProtocolNo par_prot_no = ACOM::Network::P_AM;
int par_verbose;

void terminate(int exit_value) {
  ACOM::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: " << progname << " [OPTIONS]" << std::endl;
  std::cout << "OPTIONS:" << std::endl;
  std::cout << " -P net: choose net for primary network"  << std::endl;
  std::cout << " -n net: choose net for secondary network"  << std::endl;
  std::cout << " -p protocol: use protocol when sending messages" << std::endl;
  std::cout << " -d nb_def: send nb_def deferred data in each msg"<< std::endl;
  std::cout << " -D sz_def: send deferred data of size sz_def"  << std::endl; 
  std::cout << " -i nb_imm: send nb_imm immediate data in each msg"<<std::endl;
  std::cout << " -I sz_imm: send immediate data of size sz_imm"<< std::endl; 
  std::cout << " -x: use one to n-1 mode (default is n to n-1)" << std::endl;
  std::cout << " -c count: run the test count times" << std::endl;
  std::cout << " -s: execute calls to service function in mutex" << std::endl;
  std::cout << " -b: synchronize each sending" << std::endl;
#ifdef VERBOSE_ENABLE
  std::cout << " -v verbose_level: verbose mode" << std::endl;
#endif
  terminate(exit_val);
}


void parse_args(int argc, char* argv[]) {
  int opt;
  progname = argv[0];

  while((opt = getopt(argc, argv, "hn:d:D:i:I:xc:sbp:P:v:"))>0)
    switch(opt)
      {
      case 'n': // secondary network name
        s_net_name = optarg; 
        break;

      case 'P': // primary network name
        p_net_name = optarg;
        break;

      case 'h': // help
        usage(0);
        break;

#ifdef VERBOSE_ENABLE
      case 'v': // verbose
        par_verbose = atoi(optarg);
        break;
#endif

      case 'd': // number of deferred data
        par_nbdef = atoi(optarg);
        break;

      case 'D': // size in bytes of deferred data
        par_szdef = atoi(optarg);
        if (par_szdef <= 0) {
          std::cerr << "-D: size should be greater than 0" << std::endl;
          terminate(2);
        }
        break;

      case 'i': // number of immediate data
        par_nbimm = atoi(optarg);
        break;

      case 'I': // size of immediate data
        par_szimm = atoi(optarg);
        if (par_szimm <= 0) {
          std::cerr << "-I: size should be greater than 0"<< std::endl;
          terminate(2);
        }          
        break;

      case 'x': // one to n mode (default is n to n-1)
        par_nton = 0;
        break;

      case 's': // execute service calls in mutex
        par_sync = 1;
        break;

      case 'b': // synchronize each sending
        par_barrier = 1;
        break;

      case 'c': // nb times the test should be run
        par_count = atoi(optarg);
        break;

      case 'p': // protocol type
        par_prot = optarg;
        if (!strcmp(par_prot, "am"))
          par_prot_no = ACOM::Network::P_AM; 
        else if (!strcmp(par_prot, "ws"))
          par_prot_no = ACOM::Network::P_WS;
        else if (!strcmp(par_prot, "aws"))
          par_prot_no = ACOM::Network::P_AWS;
        else {
          std::cerr << "Bad protocol type. Possible values are: am, ws, aws" 
                    << std::endl;
          terminate(2);
        }
        break;

      default:
        usage(2);
      }
}





#define function "GET_IOVECT"

class AckService: public ACOM::Service {

  PTH::Mutex mutex;

public:

  // This function is called once the immediate data have been received
  // Deferred data may not have been totally (or even partially) received
  // Its role is give a mapping between received data and local buffers:
  // For each element of the IO vector, it must give a reference to memory 
  // location where the correspondig data have to be copied
  // Note: this function is not called atomatically, so it must be re-entrant
  void* get_iovect(
          ACOM::Network::Receipt* receipt,
          ACOM::Network::Header* header,
          ACOM::IOVectEntry* vector,
          ACOM::IOVectSize count
          )
  {
    if (par_sync)
      mutex.lock();
    OUT(10) << "IO vector size: "  << count << std::endl;
    if (count != par_nbdef) {
      ERR << "iov_sz ==" << count  << " != " << par_nbdef << std::endl;
    }

    for (int cpt = 0; cpt < count; cpt++) {
      if (vector[cpt].len != par_szdef) {
        ERR << "size IO vector:" << count << std::endl;
        ERR << "expected      :" << par_szdef << std::endl;
      }
      vector[cpt].data= new char[vector[cpt].len];
    }

    if (par_sync)
      mutex.unlock();
    return 0;
  }    
  

#undef function
#define function "SERVICE"

  // This function is called once all deferred data has been copied
  // at the right location
  // Note: this function is not called atomatically, so it must be re-entrant
  void call ( 
         ACOM::ComFailure::Code e,      // if any 
         acom::Network::Receipt* receipt,
         ACOM::Network::Header* h,
         ACOM::IOVectEntry* iov,
         ACOM::IOVectSize   iov_sz,
         void* ctx
         ) throw (ACOM::ComFailure)
  {
    size_t sz;
    char exp_immval = 0;
    int nb_unpacked = 0;
    int sender;

    if (par_sync)
      mutex.lock();
    
    OUT(10) << "Enter: protocol type: " << protocols[h->protocol] << std::endl;
    
    // check communication errors
    if (e != ACOM::ComFailure::OK) {
      ERR << "-- Communication errors detected" << std::endl;
      if (par_sync)
        mutex.unlock();
      return;
    }
    
    // check that this function is not called too many times
    if (isterm.getValue() <= 0) {
      ERR << "called too many times !!!" << std::endl;
      ERR << "isterm.getValue() = " << isterm.getValue() << std::endl;
      if (par_sync)
        mutex.unlock();
      return;
    }
    
    char * immdata = new char[par_szimm];
    
    sender = h->src;
    //while (receipt->unpack(&immdata, sizeof(immdata))) 
    // unpack until end of stream
    while (!receipt->eos()) {
      
      sz = receipt->unpack(immdata, par_szimm);
      nb_unpacked++;

      // check if when are not unpacking data never packed
      if (nb_unpacked > par_nbimm) {
        ERR << "Unpacked to much data :" << nb_unpacked 
            << "/" << par_nbimm << std::endl;
        break;
      }
      
      // check returned size
      if (sz < par_szimm)
        ERR << "unpacked only " << sz << 
          " bytes out of " << par_szimm << std::endl;
      
      
      // check unpacked value
      for (int cpt = 0; cpt < par_szimm; cpt++) {
        //         OUT(10) << "IMMEDIATE[" << (int) exp_immval << "," << cpt 
        //             << "] FROM " << sender << ": " << (int) immdata[cpt] << std::endl;
        if (immdata[cpt] != exp_immval) {
          ERR << "immediate data expected: " << (int) exp_immval << std::endl;
          ERR << "immediate data received: " << (int) immdata[cpt]<< std::endl;
        }
      }
      exp_immval++;
    }
    // check if when have unpacked the right number of time
    if (nb_unpacked < par_nbimm) {
      ERR << "Unpacked too few data :" << nb_unpacked
          << "/" << par_nbimm << std::endl;
    }
 
  
    delete [] immdata; 
    OUT(10) << "ALL IMMEDIATE UNPACKED" << std::endl;
    
    // check deferred data
    if (iov_sz != par_nbdef) {
      ERR << "iov_sz != " << par_nbdef <<"  ==" << iov_sz  << std::endl;
    }
    else {
      for (int cpt = 0; cpt < iov_sz; cpt++) {
        char* def_data = (char*) iov[cpt].data;
        for (int i=0; i<iov[cpt].len; i++) {
          //            OUT(10) << "DEFERRED[" << (int) cpt << "," << i
          //                << "] FROM " << sender << ": " << (int) def_data[i] <<std::endl;
          if (def_data[i] != (char) cpt) {
            ERR << "deferred data expected: " << cpt << std::endl;
            ERR << "deferred data received: " << (int) def_data[i] 
                << std::endl;
          }
        }
        if (h->protocol != ACOM::Network::P_AM) {
          //OUT(10) << "delete data" << std::endl;
          delete [] def_data;
        }
      }
    }
  
    // one more message received
    --isterm;
    OUT(10) << "Exit" << std::endl;
    if (par_sync)
      mutex.unlock();
    
  } // AckService::call
  
  
}; // class AckService

static AckService ack_service;




class MyCallback: public ACOM::Callback {
  
public:


#undef function
#define function "CALLBACK"

  // This function is called when a message has been completely sent
  // the number of times specified when the object was created
  void done(ACOM::Network::Request * req) {

    OUT(10) << "All messages completely sent" << std::endl;
  }
  
};


  
#undef function
#define function "DOIT"

int doit (int argc, char* argv[])
{
  volatile int nb_sent = 0;
  volatile int dest_node;
  int cpt, ret;
  size_t sz;


  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  
  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  ACOM::Service::Manager.bind (2, &ack_service);

  // initialize the NetworkCore library
  ACOM::Acom::initialize(&argc, &argv);
  ACOM::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()


  parse_args(argc, argv);

  // allocate memory for buffers
  char * immdata = new char[par_szimm];
  char ** defdata = (char **) malloc(par_nbdef * sizeof (char*));
  for (cpt = 0; cpt < par_nbdef; cpt++)
    defdata[cpt] = new char[par_szdef];
  
  // get networks

  ACOM::Network* p_net;
  ACOM::Network* s_net;

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
    terminate(1);


  OUT(10) << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT(10) << "Using " << s_net->get_name() << " as secondary network" << std::endl;


  //  ACOM::Network* network = s_net;
  ACOM::Network* network = s_net;


  //   p_net->initialize(&argc, &argv);
   //   s_net->initialize(&argc, &argv, p_net);

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  rank = network->rank();
  for (cpt = 0; cpt < rank; cpt++) dec=dec+"   ";


  // Let's go
  OUT(10) << "Start sending messages to other nodes"<<std::endl; 
  OUT(10) << "There is " << network->node_count() << 
    " nodes in the network"<<std::endl; 

  for (test_num = 0; test_num < par_count; test_num++) {
    OUT(10) << "############################" << std::endl;
    OUT(1) << "TEST #" << test_num << std::endl;
    OUT(10) << "############################" << std::endl;

    if (par_nton) 
      isterm.init(network->node_count()-1); 
    else if (rank == 0)
      isterm.init(0);
    else
      isterm.init(1);

    // synchronize with other nodes to make sure that everybody 
    // has initialized the countdown object and the rank
    OUT(10) << "ENTER BARRIER" << std::endl;
    network->barrier();
    OUT(10) << "EXIT BARRIER" << std::endl;
    
    if (par_nton || rank == 0) {


      // create one message per possible protocol, only once will be used
      // note that the same message is constructed only once and sent many
      // times (actually network->node_count() - 1 times)
      // The call back object of our callback class, will be called 
      MyCallback cb;
      ACOM::Message mesg(network, par_prot_no, &ack_service, 
	&cb, (int) network->node_count() - 1);

      //Pack immediate data
      for (cpt = 0; cpt < par_nbimm; cpt++) {
        memset(immdata, cpt, par_szimm);
        sz=mesg->pack((int *) immdata, par_szimm);
        if (sz != par_szimm)
          ERR << "packed only " << sz << " bytes out of " 
              << par_szimm << std::endl;
      }

      // Put deferred data
      for (cpt = 0; cpt < par_nbdef; cpt++) {
        memset((char *) defdata[cpt], cpt, par_szdef);
        ret=mesg->put(defdata[cpt], par_szdef);
        if (!ret)
          ERR << "could not put the requested " 
              << par_szdef << " bytes" << std::endl;
      }
      

  
      // Now, the same message is sent to all other nodes
      for (dest_node = 0; dest_node < network->node_count(); ++dest_node)
        {
          if (dest_node != rank) // do not send to myself
            {
              // get a channel to node n through our network
              ACOM::Network::OutChannel* channel = network->get_default_route(dest_node);
              OUT(10) << "Got a channel to node " 
                  << dest_node << std::endl;
              
              // send message through this channel
              nb_sent++;
              channel->post( mesg ); 
              OUT(10) << "Message sent to node " 
                  << dest_node << std::endl;

              if (par_nton && par_barrier)
                network->barrier();
            }
        }

    }

    // wait until all messages have been locally received from other nodes
    // (this coutdown object is decremented in the service)
    isterm.wait();
    OUT(10) << "received all messages from other nodes" << std::endl;
  }

  test_num = -1; // to have clean output 
  for (cpt = 0; cpt < par_nbdef; cpt++)
    delete [] defdata[cpt];

  delete [] immdata;


  OUT(10) << "Terminated"<< std::endl;
  terminate(0);

}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}
