//
// $Id: testws.cpp,v 1.4 2003/07/15 17:33:09 thierry Exp $
//
// test active message: a call to one service from rank 0 with nul msg
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <pth>
#include <string.h>
#include "testgetnet.h"


std::string function;

#define OUT std::cout << "" << rank << "  :" << function << " "



namespace AC = acom;

int rank = -1;
static char buffer[256];
static PTH::CountDown isterm;


class AckService: public AC::Service {

  int nb_sent;
  int nb_completely_sent;
  int sender;
  int dest_node;

public:
  

  // This function is called once the immediate data have been received
  // Deferred data may not have been totally (or even partially) received
  // Its role is to give a mapping between received data and local buffers:
  // For each element of the IO vector, it must give a reference to memory 
  // location where the correspondig data have to be copied
  void* get_iovect(
		  AC::Network::Receipt* receipt,
		  AC::Network::Header* header,
		  AC::IOVectEntry* vector,
		  AC::IOVectSize count
		  )
  {
	function = "GET_IOVECT";
    OUT << "get_iovect called with IO vector of size="  << count << std::endl;
    vector[0].data=&dest_node;
    vector[1].data=&nb_sent;
    vector[2].data=&nb_completely_sent;
    return 0;
  }	
  

  // This function is called once all deferred data has been copied
  // at the right location
  void call ( 
	     AC::ComFailure::Code e,      // if any 
	     acom::Network::Receipt* receipt,
	     AC::Network::Header* h,
	     AC::IOVectEntry* iov,
	     AC::IOVectSize   iov_sz,
	     void* ctx
	     ) throw (AC::ComFailure)
  {

	function="SERVICE";
    OUT << "Service called " << std::endl;
    if (e == AC::ComFailure::OK) {
      OUT << "-- No communication error detected"<<std::endl;

      // Get immediate data in reverse order (LIFO)
      receipt->unpack(&sender, sizeof(sender));
      OUT << "-- called from node " << sender << std::endl;
      OUT << "-- at that time, sender had requested to send " \
		<< nb_sent << " messages" << std::endl;
      OUT << "    and " << nb_completely_sent 
		<< " was completely sent" << std::endl;
      for (int i=0; i<iov_sz; i++)
		OUT << "-- Data pointed by IO vect[" << i << "]: [" 
			<< *((int *) iov[i].data) << "]" << std::endl;
    }
    else 
      OUT << "-- Error" << std::endl;
    --isterm;
  }
};

static AckService ack_service;




class MyCallback: public AC::Callback {
  
public:

  volatile int nb_completely_sent;
  
  // CSTOR
  MyCallback() : nb_completely_sent(0)
  { }
  

  // This function is called each time a message has been completely sent
  // Asociated data are kept until the last call to this function, 
  // and are then deallocated
  void done(AC::Network::Request * req) {
	function="CALLBACK";

    nb_completely_sent++;
    OUT << "CallBack called, " << nb_completely_sent \
	      << " messages has been completely sent" << std::endl;
    if (nb_completely_sent == 0) {
      // after that, the message can no longer be used
      OUT << "Last call..." << std::endl; 
    }
  }
  
};

char * s_net_name, * p_net_name = 0;

void terminate(int exit_value) {
  AC::Acom::terminate();
  exit(exit_value);
}

int usage(int exit_val) {
  std::cout << "Usage: testws -p primary_net -s secondary_net" << std::endl;
  terminate(exit_val);
}

void parse_args(int argc, char* argv[]) {
  int opt;
  optind = 0;
  while((opt = getopt(argc, argv, "hp:s:"))>0)
	switch(opt)
	  {
	  case 'p':
		p_net_name = optarg;
		break;
	  case 's':
		s_net_name = optarg;
		break;

	  case 'h':
		usage(0);
	  default:
		usage(2);
	  }
}



int doit (int argc, char* argv[])
{
  volatile int nb_sent = 0;
  volatile int dest_node;
  function="DOIT";

  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  
  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  AC::Service::Manager.bind (2, &ack_service);

  // initialize the NetworCore library
  AC::Acom::initialize(&argc, &argv);
  AC::Acom::commit();
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()


  parse_args(argc, argv);


  // get networks

  AC::Network* p_net;
  AC::Network* s_net;

  if (get_networks(p_net, p_net_name, s_net, s_net_name))
	terminate(1);


  OUT << "Using " << p_net->get_name() << " as primary network" << std::endl;
  OUT << "Using " << s_net->get_name() << " as secondary network" << std::endl;
  
  //  AC::Network* network = s_net;
  AC::Network* network = s_net;


  //   p_net->initialize(&argc, &argv);
   //   s_net->initialize(&argc, &argv, p_net);

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus one
  isterm.init(network->node_count()-1);
  rank = network->rank();

  // synchronize with other nodes to make sure that everybody has initialized
  // the countdown object and the rank
  s_net->barrier();

  

  // Let's go
  OUT << "Start sending messages to other nodes"<<std::endl; 
  OUT << "There is " << network->node_count() << 
	" nodes in the network"<<std::endl; 

  OUT << "Service has number=" << ack_service.get_id() << std::endl;


  // create an 'write&signal' message to call our service
  // note that the same message is constructed only once and sent many
  // times (actually network->node_count() - 1 times)
  // The call back object of our callback class, will be called 
  MyCallback cb;
  AC::WriteAndSignal<AckService> ws (network, 
  				     &ack_service, 
  				     &cb, 
  				     (int) network->node_count() - 1);
  // Pack immediate data
  ws->pack(&rank, sizeof(rank));

  // Put deferred data
  ws->put((int *) &dest_node, sizeof(dest_node));
  ws->put((int *) &nb_sent, sizeof(nb_sent));
  ws->put((int *) &(cb.nb_completely_sent), sizeof(cb.nb_completely_sent));

  
  // Now, the same message is sent to all other nodes
  for (dest_node = 0; dest_node < network->node_count(); ++dest_node)
    {
      if (dest_node != rank) // do not send to myself
	{
	  // get a channel to node n through our network
	  AC::Network::OutChannel* channel = network->get_default_route(dest_node);
	  OUT << "Got a channel to node " 
		    << dest_node << std::endl;

	  // send message through this channel
	  nb_sent++;
	  ws(channel);
	  OUT << "Message sent to node " 
		    << dest_node << std::endl;
	}
    }

  //  sleep(5);

  // wait until all messages have been locally received from other nodes
  // (this coutdown object is decremented in the service)
  isterm.wait();

  OUT << "All messages from other nodes received" << std::endl;

  AC::Acom::terminate();

}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}
