//
// $Id: testws2.cpp,v 1.2 2003/07/15 17:33:09 thierry Exp $
//
// test write and signal : a call to one service from rank 0, with 
// a service header
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>

namespace AC = acom;

static char* MSG = "C'est un ping" ;
static int rank = -1;
static volatile bool isterm = false;

class PingService: public AC::Service {
public:
  struct Header {
    int  data;
  };
  PingService()
  {
    int sz = strlen(MSG);
    _myiovect = AC::IOVectEntry( new char[sz+1], sz+1);
  }
  ~PingService()
  { 
    delete [] (char*)_myiovect.data; 
  }
  void call ( 
       AC::ComFailure::Code e,      // if any 
       AC::Network::Header* h,
       AC::IOVectEntry* iov,
       AC::IOVectSize   iov_sz
  ) throw (AC::ComFailure)
  {
    Header* h_serv = (Header*)h->get_service_header();
    std::cout << ":: PING: id=" << h_serv->data << ", ";
    if (e == AC::ComFailure::OK) 
       std::cout << "msg='" << (char*)iov[0].data << "'" << std::endl; 
    else 
       std::cout << "Error" << std::endl;
    isterm = true;
  }
  void get_iovect(
    AC::Network::Header* header, AC::IOVectEntry* vector, AC::IOVectSize& count  )
  {
    count  =1;
    vector = &_myiovect;
  }

protected:
  AC::IOVectEntry _myiovect;
};

static PingService ping;

int doit (int argc, char* argv[])
{
  AC::System::initialize( &argc, &argv );
  AC::Service::Manager.bind (2, &ping);
  AC::System::initcommit( );

  /* Get the network:
  */
  AC::System::const_iterator icurr = AC::System::begin_network();
  AC::System::const_iterator iend  = AC::System::end_network();
  if (icurr == iend) {
     std::cout << "No network attached to the process, exit" << std::endl;
     return 0;
  }
  else
    std::cout << "Using network: " << (*icurr)->get_name() << std::endl;
  AC::Network* net = *icurr;
  rank = net->rank();

  std::cout << "Got a channel to 1: rank is=" << rank << std::endl;
  if (rank ==0) 
  {
    for (int n=1; n < net->node_count(); ++n)
    {
      /* Get a channel:
      */
      AC::OutChannel* channel = net->get_local_route(n);

      std::cout << "Make ping channel " << std::endl;

      AC::WriteAndSignal<PingService> am (channel, &ping, 0, 1);
      am->h_serv->data = 123456789;
      am->io_v[0] = AC::IOVectEntry(MSG, strlen(MSG)+1);
      am->post();
    }
  }
  else {
    while (!isterm) {
      usleep( 1000 );
    }
    // - wait calling of PING
  }
  std::cout << "Ready to Terminate" << std::endl;
  AC::System::terminate( );
}

int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (AC::Exception& ex) {
    std::cerr << "Catch exception: " << ex.what() << std::endl;
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
  }
}
