// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/doc/pth_doc.h,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat &
// $Id: 
// ==========================================================================

/*---------------------------------------------------------------------
------------------------ BEGIN   MANUAL -------------------------------
*----------------------------------------------------------------------
*
* There is a text part which introduce concepts anf how to use them.
*  Text refers to the code (class etc...) or modules. Modules contains only 
* object or methods defining the API for Inuktitut0  thread package
*
------------------------------------------------------------------------
-----------------------------------------------------------------------*/
/**
\image html ./logo_a0.gif
\author J. Briat 
\date   March 2002, 19
\mainpage  Inuktitut0  Portable THread  package

This is a short introduction to Inuktitut0  portable thread package. The following sections present the objective of this package and highlight key points and tricks in using  threads, synchronization etc.. 

In the documentation, we use fully qualified names. When programming, you may use short names depending of C++ contextual rules (using directives, lexical scoping ..). 



\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\section SPaim Objectives
\htmlonly <hr WIDTH="100%"> \endhtmlonly 


   This package provides a low-level thread  library  independant of & portable on  low-level API e.g. Posix threads or W32 threads. All functionnalities are grouped in the PTH namespace.

Some specific functions are added to deal with processor capabilities such that atomic counters or system dpendant features. It includes also a  logging facility to log messages. 

\remarks Methods throw exceptions. The exception hierarchy matches the std::exception hierarchy.

This document contains :
	- \ref SPinit : How to start/terminate the thread portability layer  package
	- \ref SPth : The minimal thread API
	- \ref SPsync  : Locks and waiting queues
	- \ref Stsl :   API to register & read private variables with no conflict.
	- \ref Serr     Exception hierarchy
	- \ref SPlib :  like queues, semaphores, events or atomic counters
	- \ref SLog  : how to log or not, on what files etc..

\htmlonly  <hr WIDTH="100%"> \endhtmlonly 
\section  SPinit Initialisation & terminaison
\htmlonly  <hr WIDTH="100%"> \endhtmlonly 


Before using the Inuktitut0 pprtability layer PTH , we need to initialize some static values. So like all Inuktitut0  packages, there is a PTH::Pth class for initializing it, terminating it.

\code
#include <pth>
main( int argc, char* argv[])
{//Inuktitut0 thread package is not usable
  <....>
PTH::Pth::initialize(&argc, &argv);
 // we may initialize application static values

PTH::Pth::commit();
// Inuktitut0 thread package is up

PTH::Pth::terminate();
// Inuktitut0 thread package is no more  usable
..
}
\endcode

The terminate semantics is the Posix one. The concurrent Posix programm terminates when :
	- the main program is finished,
	- all the joinable threads are  finished.
Detached threads are ignored.


\htmlonly  <hr WIDTH="100%"> \endhtmlonly 
\section  SPth  About  Threads
\htmlonly <hr WIDTH="100%"> \endhtmlonly 


 The thread harness is minimal. Following the Posix semantics,  PTH::Th threads are  created detached (PTH::Thread::DETACHED) or joinable (PTH::Thread::JOINABLE). From Posic standard, it only includes operators throwing or not exception for :
	- cancellation disabling PTH::Thread::disableCancel()
 	- yielding control PTH::Thread::yield()
 	- sleeping during a period of time PTH::Thread::sleep()
 	- creating Pthread in detached or joinable mode PTH::Thread::create() PTH::Thread::createnothrow()
	- joining joinable thread PTH::Thread::join() PTH::Thread::joinnothrow() 
 	- getting the current thread id PTH::Thread::getSelf()

 The package provides a minimal wrapper from C++ object to native threads : PTH::ThreadModel. Any classes with a void run() method may be executed as a thread. It may be defined (simplified) and used as follow :
\code
template < class XX > class ThreadModel : public Thread {
static void runIt( XX *s){x->run();};
public :
// Native thread id . 
ThreadId       _threadid;    
// start a thread object
void start() throw(PosixError){create(& _threadid, (void *(*)(void*))runIt , this ); };
// get native  thread id
ThreadId getThreadId(){return  _threadid;} ;
};
// define  a derived thread class
class SomeThreadKind : public ThreadModel<SomeThreadKind>, SomeClass0, SomeClass1,... {
....... // states variables
void run() { // the new run method. May be in inherited classes};
.....
}
//  build and start a derived thread 
SomeThreadKind * th= new SomeThreadKind();
th->start();
\endcode
It inherits of the Thread class operators. It  includes  operators :
	-  to start a thread and encapsulate its id. PTH::ThreadModel::start() PTH::ThreadModel::startnothrow()
	-  to join a thread PTH::ThreadModel::join() PTH::ThreadModel::joinnothrow()
	- to get the Posix thread id PTH::ThreadModel::getThreadId()


\htmlonly  <hr WIDTH="100%"> \endhtmlonly 
\section  SPsync  About  Synchronization
\htmlonly <hr WIDTH="100%"> \endhtmlonly 




\htmlonly  <hr WIDTH="50%"> \endhtmlonly 
\subsection  SPmutex  Mutual exclusion
\htmlonly <hr WIDTH="50%"> \endhtmlonly 



Synchronization is based on Posix locks. Locks must be unlocked by the thread owning it i.e. the thread  previously getting the lock. Before using them , mutexes must be initialized by PTH::Mutex::init().  We only provides basic locks. A lock may be recursively locked by the owner or not. Default lock property is non recursive. See PTH::Mutex.


To assist in locking a critical section and automatically unlock it when leaving explicitly or when raising an exception, we provide a class PTH::CriticalGuard  to build critical C++ blocks with automatic unlocking (Thanks to the C++ automatic call of destructors at lexical block exit.).
\code

// - declaration but not initialization
static Mutex my_mutex;

void dummy1() {
  // here do something

  {CriticalGuard (my_mutex) guard; 
   // I want this block to be executed in exclusive access
     
    ....
    ....
  } // here the mutex is released when "guard" is destroyed ..

  // here do something
}
\endcode

It is possible to use in place the macro CRITICAL().
\code

// - declaration but not initialization
static Mutex my_mutex;

void dummy1() {
  // here do something

  {CRITICAL (my_mutex); 
   // I want this block to be executed in exclusive access
     
    ....
    ....
  } // here the mutex is released when "guard" is destroyed ..

  // here do something
}
\endcode

\htmlonly  <hr WIDTH="50%"> \endhtmlonly 
\subsection  SPcond  Queueing blocked threads
\htmlonly <hr WIDTH="50%"> \endhtmlonly 


 Queueing & blocking threads is based on the monitor's condition concept as provide by the Posix standard : PTH::Condition . When a thread blocks PTH::Condition::wait(), it releases the mutex which protects the shared variables used in the wait conditionnal expression. When issued, a wait enqueues the thread in the queue argument of the wait. The waiting period may be bounded in time. So the thread is awaken when a signal is sent on the condition or when the timeout accurs. A signal on a condition  may wake up only one waiting thread PTH::Condition::signal() or all the waiting threads PTH::Condition::broadcast(). If the condition queue is empty, the signal is lost and had no effect.
\code
// - declaration but not initialization
static Mutex my_mutex;
static Condition my_queue;

void dummy1() {
  // here do something
  {CriticalGuard (my_mutex) guard; 
   // I want this block to be executed in exclusive access
     
    while( ...) .my_queue.wait(&my_mutex);
.
    ....
  } // here the mutex is released when "guard" is destroyed ..
 // here do something
}

void dummy2() {
  // here do something
  {CriticalGuard (my_mutex) guard; 
   // I want this block to be executed in exclusive access
     
    my_queue.signal();
    ....
  } // here the mutex is released when "guard" is destroyed .
  
}
\endcode
As for mutexes, condition must be initialized before used.



\htmlonly  <hr WIDTH="50%"> \endhtmlonly 
\subsection  SPbeh  Monitor behaviour
\htmlonly <hr WIDTH="50%"> \endhtmlonly 

There are many implementation method of the monitor concept. Unfortunately, these methods implies different synchronization behaviours and, if ignored by the programmers, may lead to race condition errors or state invariant violation. Grossly, these  methods  are caracterised by the scheduling rules used to arbitrate the access to mutual exclusive execution of the monitor. 
	- entry requests are issued by threads trying to get initial access to the object or to a method. Let us called this threads the \b Newbies
	- at a notify point, the notifier may yield the monitor to the awaked threads (Hoare semantic) or not. In any case, there is an owner of the monitor and threads ready to continue but waiting for the monitor. Let us called these threads the \b Olders
	
So it is possible to classifiy an implementation of a monitor by the priority rules used to arbitrate between threads trying to get the monitor and the queue strategy used between threads of the same kind.
	- \b  Notify \b Rule (2 cases)  At a notify point, awaked threads have priority on a notifier thread. In this case, the notifier yields the monitor to an awaked thread. Others threads and it are recycled as Olders in this order. In the opposite case, notifier remains owner of the monitor and all awaked threads are recycled.
	- \b  Locking \b Rule (3 cases) : The lock is first granted  to the Newbies or  to the Olders or non deterministically to one of them. 
	- \b Queue \b Rule : At most two queue are needed to sort Newbies and Olders. One queue is needed when no distinction is done between newbies and olders. A classical implementation is a FIFO strategy. Another is to use static priority (real time priority) or dynamic priority ( Unix scheduling). When mixing them, this leads to impredictible behaviours. So we must program under the assumption that any thread execution interleavings may occur.

The Inuktitut implementation is based on the Posix or W32 natives threads. The  \b Queue \b  Rule is a priority based. The \b   Locking  \b  Rule is non deterministic. The \b  Notify \b Rule gives priority to notifiers.

This is the \b Posix \b threads \b semantic.


\htmlonly <hr WIDTH="100%"> \endhtmlonly
\section Stsl  Thread's private  variables
\htmlonly <hr WIDTH="100%">\endhtmlonly

Threads may register and access private variables through a key. Keys are atomically allocated on demand. Once allocated, a key may be used by any threads which share this key value. Keys are shared but registering values under a given key is thread specific. This means that values registered by differents threads under the same key value don't conflict.  The PTH::Key class provides the  functionnalities of Posix threads. This class is a template class which parameter is  the type of the  values accessed through the key.

When creating a key, one may provide a finalyze function to be called with the key value as parameter at the end of the thread.
This is used to signal the end of the thread and/or to garbage the key value.

\htmlonly <hr WIDTH="100%"> \endhtmlonly
\section Serr  Error & Exception hierarchy
\htmlonly <hr WIDTH="100%">\endhtmlonly

The inuktitut exception hierarchy is rooted on std::exception through the PTH::RuntimeError. It encapsulates a system error code and a message( char *) which depend on the error context.  The PTH::RuntimeError::what() return the concatenation of the message and the system error code message. When a  new exception is defined by inheriting the PTH::RuntimeError class, it is needed to redefine the virtual method PTH::RuntimeError::whatmsg() which converts the error code to a string in order to be able to print a readable message.


\htmlonly <hr WIDTH="100%"> \endhtmlonly
\section SPlib Programming facilities
\htmlonly <hr WIDTH="100%">\endhtmlonly


PTH package comes with some useful  programming objects like  list, queues, atomic counters, semaphore  etc..

\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection Stko  Lexical scoped object & action
\htmlonly <hr WIDTH="50%">\endhtmlonly
It is sometime needed to garantee that when some actions are done when entering a block,  the corresponding undoing actions will be done when leaving the block normally or when throwing an exception. This the purpose of the following classes :
	- a local PTH::StAction object is used to define the do and undo actions to be done on entering & leaving a lexical block,
	- a local PTH::StObjAction or  PTH::StPtrAction is used with reference or pointer to object providing a do and undo method to be applied on entering & leaving a lexical block.

For exemple, in the following code,f_at_the_end is never executed if an exception is thrown in do_main_compute.
\code
// -- exception unsafe code
void dummy1_toto() {
  // - do pre computation
  f_at_the_begining();

  // here do that can throw exception
  do_main_compute(); 

  // - do post computation
  f_at_the_end();
}
\endcode
In the following code, f_at_the_end is always executed. If 
an exception is thrown in do_main_compute, f_at_the_end is called when the stackobject is destroyed,

\code
// -- Exception safe code
void dummy1_toto() {
  
  StAction stackobject( f_at_the_begining, f_at_the_end );

  // here do that can throw exception
  do_main_compute(); 
\endcode

\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection Sq  List & queue
\htmlonly <hr WIDTH="50%">\endhtmlonly


 To be put in single linked or double linked list, an object must inherit the property to be PTH::Linkable or PTH::Queable ( from the Portable thread package). 
\code
class X : Public Queable<X>{....};
\endcode

It may put in a PTH::Queue or a PTH::List or a PTH::StackList. For the first class, all kinds of insert/retract operators are usable. For the second, its is possible to insert or retact in front  of list and put at end. For the third, only push and pop operators are allowed. 

\code
class X : public Queable<X>{....};
Queue<X> q;
 
q.push_front( new X(..));
\endcode

A linkable or queueable  object must be in \b atmost \b one \b queue.

These three classes are \b thread \b unsafe. If it is needed to use shared list or queue, all operators on a given list or queue must be  done in mutual exclusion.

\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection Sqa  Cell allocator
\htmlonly <hr WIDTH="50%">\endhtmlonly

It is often needed to do fast allocation of object in concurrent programms. One way to do that is to maintain a cache list of deallocated cells. The size of this list is bounded. Each time a cell is requested, the allocator tries to find a cell in the cache list. If not, it calls the normal allocation mechanism. When a cell is deallocated, the allocator puts the cell in cache list if the list size is less than the trhreshold else the cell is deallocated. 

The PTH::CellAllocator template class implements this method for static  sized object ( size given by sizeof( )).  Such allocator manages objects of only one size. 

For object of 
\code 
 CellAllocator<X> A_of_X; 
  
 X * x=A_for_X.alloc();
  .....
 A_for_X.free(x);  
\endcode  

This  allocator is  thread unsafe. Use the PTH::SyncCellAllocator with threads.

 With object composed of a fixed part and a variable size part ( typically including a variable size array), use the PTH::BlockAllocator or PTH::SyncBlockAllocator. The object size is given at allocator creation time. 

PTH::ByteAllocator and PTH::SyncByteAllocator are for fixed size byte blocks.


\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection SPlibconc Concurrent programming facilities
\htmlonly <hr WIDTH="50%">\endhtmlonly

PTH package comes with some useful concurrent programming objects like atomic counters, semaphore  or events.

For atomic counters PTH::AtomicCounter, there is a machine independant implementation. Nevertheless, the interest of these object is to used native processor instructions to speedup some critical operations.

Class PTH::CountDown provides a count down counter on which to wait until it reacches zero. Semaphore PTH::Semaphore and events are classical programming concepts. PTH::PersistentEvent and PTH::TransientEvent are closed to W32 events.


\htmlonly <hr WIDTH="100%"> \endhtmlonly
\section SLog Log service
\htmlonly <hr WIDTH="100%">\endhtmlonly

 The inuktitut PTH package includes a logging facility : PTH::InukLogger. This facility is initialized at PTH initialization time (see PTH::Pth class)

\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection SLevel  How to log events 
\htmlonly <hr WIDTH="50%">\endhtmlonly 


The logging point must be inserted in the code by writind lines like this :
\code
LOG( "Level::Event::"<< any_value<<..<<etc... ); 
\endcode
where \b level is a user chosen name qualifying a set  of  events related to a piece of code or module etc.. and \b Event is the event occuring.
Changing LOG() to NOLOG() inhibits the recording of the log message. These macros are threadsafe macros. There are thread unsafe macros ( _LOG() and _NOLOG() )


\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection SLogger The log files 
\htmlonly <hr WIDTH="50%">\endhtmlonly

The PTH::InukLogger  manages two streams  to log messages.  Each of these streams must be bound to an IO stream before used.  Any of the 2 log streams may be independantly bound or unbound and enable or disable.   PTH::Logger::bind(), PTH::Logger::unbind(), PTH::Logger::enable(), PTH::Logger::disable() is the subset of streams to bind/unbind, etc..
A log message is output on the enabled streams. At end, the logger must be flushed (PTH::Logger::flush())

\htmlonly <hr WIDTH="100%"> \endhtmlonly
\section STrac Tracer service
\htmlonly <hr WIDTH="100%">\endhtmlonly
The inuktitut PTH package includes a tracing facility in a proper namespace InukTracer. This facility is initialized at PTH initialization time (see PTH::Pth class)

 The basic idea is to:
	
	- have in the portable thread  module all the stuff to build a tracer:the initialization / termination of trace, allocation / deallocation of trace buffers to threads etc...
	-  have in the portable thread  module the Inuktitut tracer definition. The  pth.evt file begins  with TRACER(InukTracer)) from which are built the trace recorder object. This file contains also the event definitions used to trace the portable thread  module module. All the needed classes are generated in the dedicated namespace InukTracer by the stub compiler (see the Tracer module). Events are grouped by level. An event is defined like a procedure definition.
\code
#include "pth_pthreadsync.h"
// define the inuktitut tracer
TRACER(InukTracer)
using namespace PTH;
LEVEL(PTH_LEVEL,1)
EVENT(PTH_YIELD_0,ThreadId pthreadId)
EVENT(PTH_YIELD_1,ThreadId pthreadId)
EVENT(PTH_SLEEP_0,ThreadId pthreadId,long milli,long nano)
EVENT(PTH_SLEEP_1,ThreadId pthreadId,long milli,long nano)
EVENT(PTH_CREATE_1,ThreadId pthreadId)
\endcode
 
	-  put in each modules to trace ( Jthread, Network, etc ..) an .evt file beginning only with USE_TRACER(InukTracer)) followed by level+event declarations. Stubs are generated from theses files with only external reference to basic type ans values ( defined in the tracer stubs).
	- insert in the code trace points in the form of procedure call to trace stubs.
\code
inline  void Thread::yield()throw ()
{
  PTH_YIELD_0(Thread::getSelf());
	sched_yield();
  PTH_YIELD_1(Thread::getSelf());
}
\endcode
Depending on the Inuktitut configuration, the stub compiler  outputs effective stubs or no-op stubs. 

At runtime, tracing of events may be temporarily and dynamically disable through control commands. It is not possible to discard trace points but it is possible to enable or disable the tracing mechanism using the tracer mask, level masks or event masks. 
Setting a mask to true, disable recording of the corresponding set of events or a single event.
\code
using namespace InukTracer;
InukTracer_msk=1; // disable the tracer
...
InukTracer_msk=0; ; // enable the tracer
...
PTH_LEVEL_msk=1; // disable all events of the PTH_LEVEL level
....
PTH_LEVEL_msk=0; // enable all events of the PTH_LEVEL level
PTH_YIELD_0_msk=1;// disable yield event 
......
\endcode


\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection Trrecord Event recorder, buffer and  file management
\htmlonly <hr WIDTH="50%">\endhtmlonly

The tracer building blocks used to day are  :
	-  the PTH::Binder is  used to link a thread to its trace buffer
	- The buffer manager  PTH::SimpleBufferManager is very simple. On demand, it flushes event buffers in a trace in a blocking manner. After that, it returns the flushed buffer as a fresh one. When a new trace is opened, it allocates a buffer from the standard free memory and frees the buffer when the trace is closed. 
	- one buffer class PTH::SimpleBuffer which is used to record events. A generic template operator <<  (see file tr_recorder.h) record by copy basic type and pointer. To copy more complex type, this template must be specialized. To day, thgis is only done for the std::string class.
	- one generic recorder PTH::SimpleRecorder which glues together all the previous things.
	- a physical clock PTH::PhysicalClock providing time on a second/microsecond basis PTH::PhysicalTime ( Unix gettimeofday output).

\htmlonly <hr WIDTH="50%"> \endhtmlonly
\subsection TrInuk The Inuktitut tracer
\htmlonly <hr WIDTH="50%">\endhtmlonly
 To define a tracer, it's only needed to instantiate a recorder InukTracer::recorder and a mask InukTracer::InukTracer_msk to start/stop tracing. 




*/ // DON'T REMOVE 
/*----------------------------- END OF MANUAL ----------------------------
------------------------------------------------------------------------
------------------------------------------------------------------------*/


