// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/src/pth_pthreadsync.cpp,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Authors: J. Briat & T. Gautier 
// $Id: pth_pthreadsync.cpp,v 1.5 2003/07/17 14:11:12 cyril Exp $
//
// ==========================================================================
// the namespace identifier
#include "pth_ns.h"
#include "pth_pthreadsync.h"
#include "pth_system.h" 

#ifdef PTH_USE_TRACER
#include "Pthread_stub_tracer.h"
#endif

namespace PTH
{

pthread_mutexattr_t Mutex::default_attr;
pthread_mutexattr_t Mutex::recursive_attr;


void Mutex::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;
  // default attribute for Mutex
  ASSERT(!(err =pthread_mutexattr_init(&default_attr)), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
  ASSERT(!(err =pthread_mutexattr_init(&recursive_attr)), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
  ASSERT(!(err =pthread_mutexattr_settype(&default_attr,MUTEX_TYPE_NORMAL )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
  ASSERT(!(err =pthread_mutexattr_settype(&recursive_attr, MUTEX_TYPE_RECURSIVE )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
};

pthread_condattr_t  Condition::default_attr;

void Condition::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;
  // default attribute for Condition
  ASSERT(!(err =pthread_condattr_init(&default_attr)), 
      PosixError(SUFX( "[Condition:: Init]: cannot define condition attributes "),err));

#ifdef PTH_USE_PTHREAD_IRIX
  ASSERT(!(err =pthread_condattr_setpshared(&default_attr, COND_TYPE )), 
      PosixError(SUFX( "[Mutex:: Init]: cannot define   mutex attributes "),err));
#endif

  
};



pthread_attr_t Thread::detached_attr;
pthread_attr_t Thread::joinable_attr;

Thread::Kind Thread::DETACHED;
Thread::Kind Thread::JOINABLE;      

void Thread::init_static()throw (PosixError)
{
  int err;
  static bool once=true;if(once)once=false;else return;

#ifdef PTH_USE_PTHREAD_IRIX
  ASSERT(!( err= pthread_setconcurrency(256)),
    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
#endif

  // -- initialize detached attribut for thread creation
  ASSERT(!( err= pthread_attr_init(&detached_attr)),
      PosixError(SUFX( "[Pthread:: Init]: cannot define   pthread attributes ") ,err));
  ASSERT(!( err= pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED)), 
      PosixError(SUFX( "[Pthread:: Init]: cannot define  pthread attributes "), err));


  // -- initialize joinable attribut for thread creation
  ASSERT(!( err= pthread_attr_init(&joinable_attr)),
      PosixError(SUFX( "[Pthread:: Init]: cannot define   pthread attributes ") ,err));
  ASSERT(!( err= pthread_attr_setdetachstate(&joinable_attr, PTHREAD_CREATE_JOINABLE)), 
      PosixError(SUFX( "[Pthread:: Init]: cannot define  pthread attributes "), err));

#ifdef PTH_USE_PTHREAD_IRIX
  // 1MByte / per stack on IRIX
  ASSERT(!( err= pthread_attr_setstacksize(&detached_attr, 2 * 1024*1024)),
    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
//  ASSERT(!( err= pthread_attr_setstacksize(&joinable_attr, 1024*1024)),
//    PosixError(SUFX( "[Pthread:: Init]: cannot set concurrency ") ,err));
#endif

  DETACHED= &detached_attr; JOINABLE= &joinable_attr;


}

int Mutex::initnothrow(bool recursive) throw ()
{
  int err;
#ifdef PTH_USE_TRACER
  MUTEX_INIT_0(&_mutex);
#endif
  if(recursive)
    { err = pthread_mutex_init(&_mutex, &recursive_attr); }
  else
    { err = pthread_mutex_init(&_mutex, &default_attr); }
#ifdef PTH_DEBUG
  if (err ==0) _isinit =true;  
#endif
#ifdef PTH_USE_TRACER
  MUTEX_INIT_1(&_mutex);
#endif
  return err;
}

void Mutex::init(bool recursive) throw (PosixError)
{ 
  int err;
  ASSERT( !(err=initnothrow(recursive)),
  PTH::PosixError(SUFX("[Mutex::init] POSIX error"), err));
}

void Mutex::init_notrace() throw ( )   // Without trace verions of init()
{
  pthread_mutex_init(&_mutex,&default_attr); 
}

int Mutex::destroynothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_DESTROY_0(&_mutex);
#endif
  int err;
  err=pthread_mutex_destroy(&_mutex);
#ifdef PTH_DEBUG
  if (err ==0) _isinit =false;  
#endif
#ifdef PTH_USE_TRACER
  MUTEX_DESTROY_1(&_mutex);
#endif
  return err;
}

void Mutex::destroy() throw (PosixError)
{
  int err;
  ASSERT( !(err=destroynothrow()),
  PTH::PosixError(SUFX("[Mutex::destroy] POSIX error"),err));
}


// -------------------------------------------------------------------------
// Condition
// -------------------------------------------------------------------------
int Condition::initnothrow() throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_INIT_0(&_condition);
#endif
  int err;
  err=pthread_cond_init(&_condition, &default_attr);
#ifdef PTH_USE_TRACER
  CONDITION_INIT_1(&_condition);
#endif  
  return err;
}

void Condition::init() throw (PosixError)
{
  int err;
  ASSERT( !(err=initnothrow()), PosixError (SUFX("[Condition()] POSIX error"), err));
}

int Condition::destroynothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_DESTROY_0(&_condition);
#endif
  int err;
  err=pthread_cond_destroy( &_condition ); 
#ifdef PTH_USE_TRACER
  CONDITION_DESTROY_1(&_condition);
#endif
return err;
}

void Condition::destroy()throw (PosixError)
{
  int err;
  ASSERT( !(err=destroynothrow()),  
  PosixError(SUFX( "[Condition::~()] POSIX error"),err)); 
}


} // -namespace
