// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/src/pth_pthreadsync.inl,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat & T. Gautier
// $Id: pth_pthreadsync.inl,v 1.7 2003/07/17 14:11:12 cyril Exp $
// ==========================================================================


#ifdef PTH_USE_TRACER
#include "Pthread_stub_tracer.h"
#endif

namespace PTH {

// -------------------------------------------------------------------------
// Mutex
// -------------------------------------------------------------------------
// 
// The without trace versions of functions are used by tracer
//
inline int Mutex::locknothrow()throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_0(&_mutex);
#endif
  int ret= pthread_mutex_lock(&_mutex);
  if(ret) std::cout<< " ERROR......................................."<< strerror(ret)<<"\n";
#ifdef PTH_USE_TRACER
  MUTEX_LOCK_1(&_mutex);
#endif
  return ret;
}

inline void Mutex::lock()throw (PosixError) 
{
	int err; 
	ASSERT( !(err= locknothrow()), 
      PTH::PosixError(SUFX("[Mutex::lock] POSIX error"),err));	
}

inline void Mutex::lock_notrace()throw (PosixError)    // Without trace verions of lock()
{
  int err; 
  ASSERT( !(err= pthread_mutex_lock(&_mutex)), 
      PTH::PosixError(SUFX("[Mutex::lock] POSIX error"),err));
}


inline int Mutex::trylocknothrow()throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_0(&_mutex);
#endif
	int ret= pthread_mutex_trylock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_TRYLOCK_1(&_mutex);
#endif
  return ret;
}


inline bool Mutex::trylock()throw (PosixError)
{
	int err;
	err= trylocknothrow();
	if (err==EBUSY) return false;
	if (err==0) return true;
  ASSERT( !err, PTH::PosixError(SUFX("[Mutex::trylock] POSIX error"),err));
  return false;
}


inline int Mutex::unlocknothrow() throw ()
{
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_0(&_mutex);
#endif
	int err; 
	err= pthread_mutex_unlock(&_mutex);
#ifdef PTH_USE_TRACER
  MUTEX_UNLOCK_1(&_mutex);
#endif
	return err;
}


inline void Mutex::unlock() throw (PosixError)
{
	int err;
	ASSERT( !(err= unlocknothrow()), PosixError(SUFX("[Mutex::unlock] POSIX error"),err));	
}


inline void Mutex::unlock_notrace()throw (PosixError) // without trace version of unlock
{
  int err;
  ASSERT( !(err= pthread_mutex_unlock(&_mutex)), PosixError(SUFX("[Mutex::unlock] POSIX error"),err));
}

 


// -------------------------------------------------------------------------
// Condition
// -------------------------------------------------------------------------
inline int Condition::waitnothrow( Mutex& mutex )throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_0(&(mutex._mutex),&_condition,0,0);
#endif
	int ret= pthread_cond_wait( &_condition, &(mutex._mutex)); 
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_1(&(mutex._mutex),&_condition,0,0);
#endif
  return ret;
}

inline void Condition::wait( Mutex& mutex ) throw (PosixError)
{
	int err;
	ASSERT( !(err=waitnothrow(mutex)), PosixError(SUFX( "[Condition::wait] POSIX error"),err)); 
}

inline int Condition::waitnothrow( Mutex& mutex, const struct timespec& time ) throw ()
{ 
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_0(&(mutex._mutex),&_condition,time.tv_sec,time.tv_nsec);
#endif	
  int r = pthread_cond_timedwait( &_condition, &(mutex._mutex),&time);
#ifdef PTH_USE_TRACER
  CONDITION_WAIT_1(&(mutex._mutex),&_condition,time.tv_sec,time.tv_nsec);
#endif
  return r;
}

inline bool Condition::wait( Mutex& mutex, const struct timespec& time ) throw (PosixError)
{
  int r=waitnothrow(mutex,time);
	if (r ==0) return false;
	if (r == ETIMEDOUT) return true;
  ASSERT( !r,PosixError(SUFX(  "Condition::timedwait"), r));
  return false;
}

inline int Condition::waitnothrow( Mutex& mutex, long milli, int nano ) throw ()
{
	struct timespec req;
	struct timeval now;    
	gettimeofday(&now,0);  
	req.tv_sec  =  now.tv_sec+ milli / 1000;
	req.tv_nsec =  (now.tv_usec*1000)+ (1000000 * (milli % 1000)) + nano;
	return waitnothrow( mutex,req);
}

inline bool Condition::wait( Mutex& mutex, long milli, int nano ) throw (PosixError)
{
  int r=waitnothrow(mutex,milli,nano);
	if (r ==0) return false;
	if (r == ETIMEDOUT) return true;
  ASSERT( !r,PosixError(SUFX(  "Condition::timedwait"), r));
  return false;
}

inline int Condition::signalnothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_SIGNAL_0(&_condition);
#endif	
  int ret= pthread_cond_signal(&_condition); 
#ifdef PTH_USE_TRACER
  CONDITION_SIGNAL_1(&_condition);
#endif
  return ret;
}

inline void Condition::signal()throw (PosixError)
{
	int err;
  ASSERT( !(err=signalnothrow()), PosixError (SUFX("[Condition::signal] POSIX error"),err)); 
} 

inline int Condition::broadcastnothrow()throw ()
{
#ifdef PTH_USE_TRACER
  CONDITION_BROADCAST_0(&_condition);
#endif
	int ret= pthread_cond_broadcast(&_condition); 
#ifdef PTH_USE_TRACER
  CONDITION_BROADCAST_1(&_condition);
#endif
  return ret;
}

inline void Condition::broadcast()throw (PosixError)
{
	int err;
	ASSERT( !(err=broadcastnothrow()), PosixError(SUFX("[Condition::broadcast] POSIX error"),err)); 
}



// -------------------------------------------------------------------------
// Cleaner (For Posix only)
// -------------------------------------------------------------------------  
#if defined(PTH_USE_PTHREAD_LINUX)
inline Cleaner::~Cleaner()
{
	_pthread_cleanup_pop(&_clup,0);
}

inline Cleaner::Cleaner( void (*hdl)(void*), void*v)   
{
	_pthread_cleanup_push(&_clup,hdl,v);
}   
#elif defined(PTH_USE_PTHREAD_DARWIN)
/* TG : implemented using /usr/include/pthread.h definition ! */
inline Cleaner::~Cleaner()
{
  pthread_t __self = pthread_self();
  __self->cleanup_stack = __handler.next;
  if (0) (__handler.routine)(__handler.arg);
}

inline Cleaner::Cleaner( void (*hdl)(void*), void* v)   
{
  pthread_t __self = pthread_self();
  __handler.routine = hdl;
  __handler.arg = v;
  __handler.next = __self->cleanup_stack;
  __self->cleanup_stack = &__handler;
}   
#endif

  
// -------------------------------------------------------------------------
// Thread 
// -------------------------------------------------------------------------
/// current thread yield the processor
inline  void Thread::yield()throw ()
{
#ifdef PTH_USE_TRACER
  PTH_YIELD_0(Thread::getSelf());
#endif
  sched_yield();
#ifdef PTH_USE_TRACER
  PTH_YIELD_1(Thread::getSelf());
#endif
}
	

#if defined(PTH_USE_PTHREAD_DARWIN)
extern "C" int nanosleep(struct timespec* req, int );
#endif

inline int Thread::sleep( long milli, long nano) throw () 
{
#ifdef PTH_USE_TRACER
  PTH_SLEEP_0(pthread_self(),milli, nano);
#endif
#if defined(PTH_HAVE_NANOSLEEP)
	struct timespec req;
	req.tv_sec  = milli / 1000;
	req.tv_nsec = 1000000 * (milli % 1000) + nano;
  int ret=nanosleep(&req,0);
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return ret;
#elif defined(PTH_HAVE_MICROSLEEP)
	unsigned int microsec = milli*1000 + nano/1000;
	usleep( microsec ); 
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return 0;
#elif defined(PTH_HAVE_SLEEP)
	unsigned int sec = milli/1000 + nano/1000000000;
	sleep(sec); 
#ifdef PTH_USE_TRACER
  PTH_SLEEP_1(getSelf(),milli, nano);
#endif
  return 0;
#else
#error UNABLE TO DEFINE SLEEP
#endif
}

 
/// current thread disable automatic cancellation

inline  int Thread::disableCancel( )throw()
{
	// disable cancelation 
	return pthread_setcancelstate( PTHREAD_CANCEL_DISABLE ,0);
}

// create a  thread 
inline  int Thread::createnothrow(ThreadId * th, void * (*f)(void *), void * a, Kind k) throw()
{ 
  int ret= pthread_create(th,k, f , a);
#ifdef PTH_USE_TRACER
  PTH_CREATE_1(*th);
#endif
	return ret;
}
   
// create a  thread 
inline  void Thread::create(ThreadId * th, void * (*f)(void *), void * a, Kind k ) throw(PosixError)
{
  int err; 
  ASSERT( (!(err=createnothrow(th, f , a,k))),
	PosixError(SUFX("[Thread::create] unable to create POSIX thread "),err));
}


inline int Thread::joinnothrow(ThreadId  th, void **v) throw()
{
#ifdef PTH_USE_TRACER
  PTH_JOIN_0(th);
#endif
  int ret= pthread_join(th,v);
#ifdef PTH_USE_TRACER
  PTH_JOIN_1(th);
#endif
  return ret;
}

inline void Thread::join(ThreadId  thid, void **v) throw(PosixError)
{
  int err; 
  ASSERT( (!(err=joinnothrow(thid,v))),
	PosixError(SUFX("[Thread::join] unable to join POSIX thread "),err));
}

template < class XX > 
void*  ThreadModel<XX>::runIt( void* arg)
{
#ifdef PTH_USE_TRACER
  InukTracer::openTrace();  
#endif
  //std::cout << "Runit: @obj =" << (void*)s << std::endl;
  XX* s = (XX*)arg;
  void* retval = s->run();
#ifdef PTH_USE_TRACER 
  InukTracer::closeTrace();
#endif	
  return retval;
};  

} // end namespace
