// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/Pthread/src/pth_util.inl,v $
// Copyright(c)'2001 by Apache Team
// see the copyright file
// Author: J. Briat & T. Gautier
// $Id: pth_util.inl,v 1.8 2003/07/28 16:08:05 olobry Exp $
//
//

#ifndef _PTH_UTIL_INL_ 
#define _PTH_UTIL_INL_

#include "pth_system.h" 
#include "pth_util.h" 

// ==========================================================================
namespace PTH
{

// -----
// CountDown
// -----
/**
\brief This class implements interface to Conut down counter
   

*/


    /// Initialize event. Initial value is undefined.
inline  CountDown::CountDown() throw (PosixError):_ct(-1),_mutex(),_cond(){}; 
   /// Initialize event. Default value is 0
inline    void CountDown::init(int i=0) throw ( PosixError )
{
  _mutex.init();_cond.init(); _ct = i ;
}
 
inline void CountDown::destroy() throw ( PosixError )
{
  _mutex.destroy(); _cond.destroy();
}

  
    /// decrement by one the counter. Result is the new value
inline   int CountDown::operator--(){ int x;
         CriticalGuard gd(_mutex);
         if (_ct ==0) return 0;
         if ((x=--_ct)==0){
            _cond.broadcast();
          }
    return x;
    } ;
    /// wait event set
inline     void CountDown::wait() throw (PosixError){
    CriticalGuard gd(_mutex);
      if(_ct ==0)return ;
      { 
        _cond.wait(_mutex);
      }
    };
 



// -------------------------------------------------------------------------
// Persistent Event 
// -------------------------------------------------------------------------
#
inline PersistentEvent::PersistentEvent() throw (PosixError) :_set(false), _mutex(), _cond(){};

inline void PersistentEvent::init() throw ( PosixError ) {_mutex.init();_cond.init(); };

/// set event; wake up all waiting threads ; event remains set 
inline void PersistentEvent::set()throw (PosixError){
      /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
        _set=true; _cond.broadcast();
};    

    /// set event; wake up all waiting threads ; event is reset 

inline    void PersistentEvent::pulse()throw (PosixError){
      /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
        _cond.broadcast();_set=false;
};
   

    /// reset event

inline    void PersistentEvent::reset()throw (PosixError){_set=false;};



inline    void PersistentEvent::wait()throw (PosixError){
/* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
        if (!_set)_cond.wait(_mutex);
};

inline    void PersistentEvent::wait(long milli, int nano)throw (PosixError){
/* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
       if (!_set) _cond.wait(_mutex,milli,nano);
};


// -------------------------------------------------------------------------
// Transient Event 
// -------------------------------------------------------------------------

inline TransientEvent::TransientEvent() throw (PosixError) :_set(false), _ct(0),_mutex(), _cond(){};

inline void TransientEvent::init() throw ( PosixError ) {_mutex.init();_cond.init(); };

/// set event ; reset only when a thread may be awaken
inline void TransientEvent::set()throw (PosixError){
      /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
        if(_ct==0)_set=true;else {_cond.signal();}
};

    
  /// set event; wake one waiting thread if any ; event is allways reset 

inline  void TransientEvent::pulse()throw (PosixError){
      /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
       if(_ct!=0) _cond.signal();
      
};


    /// reset event
inline  void TransientEvent::reset()throw (PosixError){_set=false;};



inline  void TransientEvent::wait()throw (PosixError){
/* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
       if(!_set){  _ct++; _cond.wait(_mutex); _set=false;_ct--;}
};

inline    void TransientEvent::wait(long milli, int nano)throw (PosixError){
/* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex); 
        if(!_set){  _ct++; _cond.wait(_mutex,milli,nano); _set=false;_ct--;}
};




// -------------------------------------------------------------------------
// Sem
// -------------------------------------------------------------------------
inline Semaphore::~Semaphore()throw (PosixError)
{
  /* Get the mutex and release it at the end */
  { CriticalGuard gd(_mutex);
  ASSERT( (_counter>= 0), PosixError(SUFX("[Semaphore::destroy] there are waiting threads"),0));
  }
  _mutex.destroy();
  _cond.destroy();    
}

inline void Semaphore::p()throw( PosixError)
{
  /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);

  /* Wait on the condition if counter  less or equal 0*/
  _counter--;
  if (_counter < 0 )  _cond.wait(_mutex);
}

inline void Semaphore::v()throw (PosixError)
{
  /* Get the mutex and release it at the end */
  CriticalGuard gd(_mutex);
  /* Increment the counter */
  _counter ++;
  /* Signal the condition variable (no broadcast needed here because at most one blocked thread can be released). */
  if( _counter<=0)_cond.signal();
}


} // end namespace
#endif
