//
// -- Example of thread creation using stack allocated object
//


#include <iostream>
#include <fstream>
#include <pth> // - include kernel namespace in global one


extern "C" {
//#include "inukterm.h" // - include kernel namespace in global one
};


//
// -- Global variables
//

static std::ofstream* mylog = 0;
//#define COUT *mylog 
#define COUT std::cout 
static PTH::Mutex m_print;
static PTH::Condition condwait;
static int count_wait = 0;

void* run(void* arg) 
{
  int data = (int) arg;
  COUT << "Running thread" << std::endl;
  for (int i=0; i<100; i++)
  { 
    // -- Example of using a stack mutex: constructor locks m_print
    // and destructor unlocks it at the end of the scope.
    {
      PTH::CriticalGuard critical(m_print);
      COUT << "Thread " << data << ": " << i << std::endl;
    }
    PTH::Thread::sleep(100,0);
    PTH::Thread::yield();
  }
  {
    PTH::CriticalGuard critical(m_print);
    if (--count_wait ==0) 
      condwait.signal();
  }
  return 0;
}



int main(int argc, char** argv)
{
  try 
    {
      //
      // Initialisation
      //
      //MacOSX_initialize(argc, (const char**)argv);
      mylog = new std::ofstream("/tmp/res");
      PTH::Pth::initialize(&argc, &argv);
      PTH::Pth::commit(); 
      m_print.init();
      condwait.init();
      PTH::Thread::yield();
      {
        PTH::CriticalGuard critical(m_print);
        //
        // Create two new threads
        //
        PTH::ThreadId id1, id2;
        PTH::Thread::create( &id1, &run, (void*)1);
        PTH::Thread::create( &id2, &run, (void*)2);
        count_wait = 2;
        
        //
        // Wait them
        //
        condwait.wait( m_print );
      }
      
      //
      // Finalize
      //

//      m_print.destroy();
//      condwait.destroy();
      
      PTH::Pth::terminate();
    }
  
  catch(PTH::PosixError &e)
    {
      COUT << "exception PTH::Pth::commit " << e.what() 
	   << std::endl << std::flush;
      mylog->close();
      return 127;
    }
  
  COUT << "Ending" << std::endl;
  
  mylog->close();
  //MacOSX_terminate();
  return 0;
}
