
//
// -- Example of thread creation using stack allocated object
//

#include <iostream>
#include <fstream>
#include <pth> // - include kernel namespace in global one


//
// -- Global variables
//
static std::ofstream* mylog = 0;
//#define COUT *mylog 
#define COUT std::cout 
static PTH::Mutex m_print;
static PTH::Condition condwait;
static int count_wait = 0;


class MyThread : public PTH::ThreadModel<MyThread> { 
  int _data;
public:
  MyThread( int d ) : _data(d) {}
  virtual int get_derived_data()
  { return 0; }
  void* run() 
  {
    std::cout << "Run, @obj=" << (void*)this << std::endl;
    COUT << "Running thread" << std::endl;
    for (int i=0; i<100; i++)
      { 
        {
          // -- Example of using a stack mutex: constructor locks m_print
          // and destructor unlocks it at the end of the scope.
          {
            PTH::CriticalGuard critical(m_print);
            COUT << "Thread " << _data << ": " << i << ", derived data=" << get_derived_data() << std::endl;
          }
          PTH::Thread::sleep(100,0);
        }
        PTH::Thread::yield();
      }
    {
      PTH::CriticalGuard critical(m_print);
      if (--count_wait ==0) 
        condwait.signal();
    }
    return 0;
  }
};

class MyThreadDerived : public MyThread {
  int _dataderived;
public:
  MyThreadDerived( int d ) : MyThread(d), _dataderived(10) {}
  virtual int get_derived_data()
  { return _dataderived; }
};

int main(int argc, char** argv)
{
  try 
    {
      //
      // Initialisation
      //
      mylog = new std::ofstream("/tmp/res");
      PTH::Pth::initialize(&argc, &argv);
      PTH::Pth::commit(); 
      m_print.init();
      condwait.init();
      PTH::Thread::yield();
      {
        PTH::CriticalGuard critical(m_print);
        //
        // Create two new threads
        //
        MyThread t1(1);
        MyThreadDerived t2(2);
        t1.start();
        t2.start();
        count_wait = 2;
        
        //
        // Wait them
        //
        condwait.wait( m_print );
      }
      
      //
      // Finalize
      //
//      m_print.destroy();
//      condwait.destroy();
      
      PTH::Pth::terminate();
    }
  
  catch(PTH::PosixError &e)
    {
      COUT << "exception PTH::Pth::commit " << e.what() 
     << std::endl << std::flush;
      mylog->close();
      return 127;
    }
  
  COUT << "Ending" << std::endl;
  
  mylog->close();
  return 0;
}
