// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_channel.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: Thierry Gautier
//
//
//
// ==========================================================================

#include <SOCKNET_network.h>
#include <SOCKNET_service.h>
#include <SOCKNET_selector.h>
#include <sstream>

#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif


namespace SOCKNET {


// --------------------------------------------------------------------------
/** send  to a node (AM,AWS,WS protocols)
    @param   req  request (IN)
    @param   ch  channel  (IN)
*/
void SOCKNET_net::post( ACOM::Network::Request* req , ACOM::Network::OutChannel* ch)
      throw (ACOM::ComFailure) 
{
  if (req) 
  {
    if (req->get_net()!=this)
      ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );
    else {
      ch->post(req);
      return;
    }
  }
  ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_REQUEST) );
}


// --------------------------------------------------------------------------
SOCKNET_net::OutChannel::OutChannel( )
 : Connexion(), ACOM::Network::OutChannel(), _urlpeer()
{
}

// --------------------------------------------------------------------------
void SOCKNET_net::OutChannel::initialize( Network* net, int peer )
{
  _access.init();
  init(net, peer);
}


// --------------------------------------------------------------------------
SOCKNET_net::OutChannel::~OutChannel()
{
}


// --------------------------------------------------------------------------
void SOCKNET_net::OutChannel::destroy()
{
  disconnect();
  _access.destroy();
}


// --------------------------------------------------------------------------
//
// InukCorba::OutChannel::post
//
void SOCKNET_net::OutChannel::post( ACOM::Network::Request* req ) 
  throw(ACOM::ComFailure)
{
  ACOM_ASSERT_DEBUG( req !=0, "[SOCKNET_net::OutChannel::post] nul request");
  ACOM_ASSERT_DEBUG( get_net() !=0, "[SOCKNET_net::OutChannel::post] nul network");

//std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
  /* close the sequences */
  if (req->get_state() != ACOM::Network::S_POSTED)
  {
    req->set_state (ACOM::Network::S_POSTED);
    req->_datastream.packEos();
    req->_iovstream.putEos();
  }
  
  /* complete the fixed part of header */
  Header* h = &req->_header;
  h->src    = get_net()->get_rank();
  h->dst    = get_peer();
  h->size_eh= req->_size_eh;
  h->size_h = req->_datastream.getFillSize();
  h->size_d = req->_iovstream.getByteSize();

  /* number of blocks of defered data */
  h->iovsz_d = req->_iovstream.getBlockSize();
  h->no=0;  

#ifdef SOCKNET_TRACE_DEBUG
  std::cout << "Header for output message is=\n"
            << "size_h=" << h->size_h << std::endl
            << "size_eh=" << h->size_eh << std::endl
            << "size_d=" << h->size_d << std::endl
            << "sz_iov=" << h->iovsz_d << std::endl
            << "src="    << h->src << std::endl
            << "dest="   << h->dst << std::endl
            << "protocol="<< (int)h->protocol << std::endl
            << "id="     << h->id << std::endl;
#endif

  /* size of blocs of deferred data */ 
  size_t* iovsz=(size_t*) ::alloca(sizeof(size_t) * h->iovsz_d);
  int i=0;
  for( ACOM::IOVectSeq::IteratorEntry ie = req->_iovstream.beginEntry();
       ie != req->_iovstream.endEntry();
       ++i, ++ie)
  {
    iovsz[i] = ie->len;
  }

  /* send the message */
  int err;

  try {
    PTH::CriticalGuard critical(_access);
    
    /* connect the Channel if needed to be */
    if (!_is_connected) connect();

#if (SOCK_NODELAY_MODE==2)
    _speer.setSockOpt(SOCK::Socket::no_delay, false);
#endif

    /* send the header  */
    err = _speer.writen( h, sizeof(ACOM::Network::Header) );
    if (err <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
  
    /* send the extended header */
    if (h->size_eh >0) {
      err = _speer.writen( req->_ext_header, req->_size_eh );
      if (err <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
    }
  
    /* send the blocs size of deferred data */
    err = _speer.writen( iovsz, sizeof(size_t) * h->iovsz_d );
    if (err <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
    //::free(iovsz);
    iovsz =0;
    /* send immediate data */
    for( ACOM::IOVectSeq::IteratorEntry ie= req->_datastream.beginEntry();
	  	   ie != req->_datastream.endEntry();
         ++ie)
    {
      err = _speer.writen( ie->data, ie->len );
      if (err <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
    }

    // - for the defered data part:
	  for( ACOM::IOVectSeq::IteratorEntry ie = req->_iovstream.beginEntry();
	  	   ie != req->_iovstream.endEntry();
         ++ie)
    {
//      std::cout << "write bloc=" << ie->len << std::endl;
      err = _speer.writen( ie->data, ie->len );
      if (err <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
    }
#if (SOCK_NODELAY_MODE==2)
      _speer.setSockOpt(SOCK::Socket::no_delay, true);
#endif
    req->notify();
  } catch (const ACOM::ComFailure& e ) {
    get_net()->deallocate_request( req );
    disconnect();
    throw e;
  }

  { PTH::CriticalGuard critical(get_net()->_access);
    get_net()->touch_connexion( this );
  }
//std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
  
}



// --------------------------------------------------------------------------
void SOCKNET_net::OutChannel::connect( )
  throw (ACOM::ComFailure)
{
  if (_is_connected) return;
  
  /* get the url of the peer node, if non valid */
  if (_urlpeer == "") {
    /* Set a message to get url of the peers node 
       Warning it was a blocking call !!!
    */
    get_net()->_service_geturl->get_urlpeer( _urlpeer, get_peer() );
  }

  /* parse url: host:port, host : ip=<n1>.<n2>.<n3>.<n4> | host name */
  std::string host;
  int port = 0;
  typedef std::string::size_type csize_type;
  csize_type beg_port = _urlpeer.find(":");
  if (beg_port >0) 
  {
    std::string str_port;
    host = _urlpeer.substr(0, beg_port );
    str_port = _urlpeer.substr(beg_port+1, _urlpeer.size()-beg_port-1 );
#ifdef SOCKNET_TRACE_DEBUG
    std::cout << "host=" << host << std::endl;
    std::cout << "port=" << str_port << std::endl;
#endif
    std::istringstream input(str_port);
    input >> port;
  }
  else {
    ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::BAD_ADDRESS) );
  }

  int err = 0;
  int trycount = 0;
  do {
    err = _speer.connect(host, port );
    if (_speer.getAddress().getLastErrorNo() == ECONNREFUSED)
    {
      PTH::Thread::sleep(50, 0); // 20 ms
      ++trycount;
    }
  } while ((err !=0) && (trycount < 10));
  if (err !=0) {
    ACOM::throwit( ACOM::ComFailure(ACOM::ComFailure::FAIL) );
  }
#if 0 //(SOCK_NODELAY_MODE==1)
  _speer.setSockOpt( SOCK::Socket::no_delay, true );
#endif

  _is_connected = true;  

  /* add new connexion */
  PTH::CriticalGuard critical(get_net()->_access);
  get_net()->add_connexion( this );
}


// --------------------------------------------------------------------------
void SOCKNET_net::OutChannel::disconnect( )
{
  { PTH::CriticalGuard critical(get_net()->_access);
    get_net()->_all_connexions.erase( this );
  }
  if (!_is_connected) return;  
  _speer.close();
  _is_connected = false;
}



// --------------------------------------------------------------------------
SOCKNET_net::InChannel::InChannel( )
 : Connexion(),  _net(0)
{
}


// --------------------------------------------------------------------------
void SOCKNET_net::InChannel::initialize( SOCKNET_net* net )
{
  _access.init();
  _net = net;
  _is_connected = false;
}


// --------------------------------------------------------------------------
SOCKNET_net::InChannel::~InChannel()
{
}


// --------------------------------------------------------------------------
void SOCKNET_net::InChannel::destroy()
{
  disconnect();
  _access.destroy();
}

// --------------------------------------------------------------------------
void SOCKNET_net::InChannel::disconnect()
{
  { PTH::CriticalGuard critical(get_net()->_access);
    get_net()->_all_connexions.erase( this );
  }
  PTH::CriticalGuard critical(_access);
  if (!_is_connected) return;
  _speer.close();
  get_net()->_selector->remove(Selector::SET_READ, this );
  get_net()->release_index( get_index() );
  set_index( -1 );
  _is_connected = false;  
}



/* Per thread buffer for the incomming message depending of their
    protocol.
*/
enum { CLEANUP_BUFFER_SIZE = 4096 };
static char cleanup_buffer[CLEANUP_BUFFER_SIZE];

// --------------------------------------------------------------------------
//
bool SOCKNET_net::InChannel::server_stub( SOCKNET_net::CallContext*& cc )
{
// std::cout << "In : " << __PRETTY_FUNCTION__ << std::endl;
 
  /* get data structure to store the call */
  int nread;

  /* current position in the buffer: */
  ACOM::Network::Header* header = 
    cc->header = (ACOM::Network::Header*)cc->allocate(sizeof(ACOM::Network::Header));
  
  /* Read only the fixed part of the header.
  */
  nread = _speer.readn( cc->header, sizeof(ACOM::Network::Header) );
  if (nread < 0) 
  {
    return true;
  }

  /* recv the extended header */
  if (cc->header->size_eh >0) 
  {
    cc->ext_header = cc->allocate(cc->header->size_eh);
    nread = _speer.readn( cc->ext_header, cc->header->size_eh );
    if (nread <0) ACOM::throwit( ACOM::ComFailure( ACOM::ComFailure::FAIL ) );
  }
  
#ifdef SOCKNET_TRACE_DEBUG
  std::cout << "Header for incomming message is=\n"
            << "size_h=" << cc->header->size_h << std::endl
            << "size_eh=" << cc->header->size_eh << std::endl
            << "size_d=" << cc->header->size_d << std::endl
            << "sz_iov=" << cc->header->iovsz_d << std::endl
            << "src="    << cc->header->src << std::endl
            << "dest="   << cc->header->dst << std::endl
            << "protocol="<< (int)cc->header->protocol << std::endl
            << "sid="     << cc->header->id << std::endl;
#endif
  if (cc->header->protocol == SOCKNET_net::P_SHUTDOWN) 
  {
    return true;
  }

  /* size to read all data : 
    - description of iovector : h->size_d*sizeof(size_t)
    - immediate data: h->size_h
    - If AM protocol read also size_d bytes.
    The stack will be :
      - ACOM::Network::Header
      - description of IOVectEntry: sizeof(ACOM::IOVectEntry)*size_d
      - all IOVectEntry size : sizeof(size_t)*size_d
      - immediate data : size_h
      - deferred data if AM protocol
  */
  size_t sz_desc  = sizeof(ACOM::IOVectEntry)*cc->header->iovsz_d;
  size_t sz_iovsz = cc->header->iovsz_d * sizeof(size_t);
  size_t sz_input = sz_iovsz + cc->header->size_h;
  if (cc->header->protocol == ACOM::Network::P_AM)
    sz_input += cc->header->size_d;
  
  /* reallocate buffer if necessary */
  if ( sz_input + sz_desc > cc->free_size() )
  {
    size_t offset_in_header = cc->get_offset( cc->header );
    size_t offset_in_ext_header = cc->get_offset( cc->ext_header );
    size_t newsz = (2+sz_input+sz_desc)*3/2;
    cc = cc->network->resize_callcontext( cc, newsz );
    cc->header = (ACOM::Network::Header*)cc->get_address( offset_in_header );
    cc->ext_header = (ACOM::Network::Header*)cc->get_address( offset_in_ext_header );
  }
  cc->iov = (ACOM::IOVectEntry*)cc->allocate(sz_desc);
  cc->iov_sz = cc->header->iovsz_d;
  size_t* iovsz = (size_t*)cc->allocate( sz_iovsz );
  void* im_data = cc->allocate( cc->header->size_h );
  
  /* read all data : description of deferred data */
  if (sz_iovsz >0)
  {
    nread = _speer.readn( iovsz, sz_iovsz );
    if (nread < 0) {
      std::cout << __LINE__ << ":" << __FILE__ << ": Cannot read immediate data" << std::endl;
      ACOM::throwit(ACOM::ComFailure( ACOM::ComFailure::FAIL ));
    }
  }
  
  /* read immediate data in one bloc */
  if (cc->header->size_h >0)
  {
    nread = _speer.readn( im_data, cc->header->size_h );
    if (nread < 0) {
      std::cout << __LINE__ << ":" << __FILE__ << ": Cannot read immediate data" << std::endl;
      ACOM::throwit(ACOM::ComFailure( ACOM::ComFailure::FAIL ));
    }
  }
  
  /* build description of data begin in buffer */
  if (cc->header->iovsz_d >0) {
    for (size_t i=0; i<cc->header->iovsz_d; ++i)
    {
      cc->iov[i].len = iovsz[i];
      cc->iov[i].data = 0;
    }
  }
  else cc->iov = 0;
  
  /* get pointer to the service id: */
  ACOM::Service* stub_server = ACOM::Service::Manager.resolv(cc->header->id);
  cc->_im_data_seq.init( _net->_nb_entries, &_net->_iovA);
  if (cc->header->size_h >0) 
    cc->_im_data_seq.put( im_data, cc->header->size_h );
  cc->receipt.init( &cc->_im_data_seq, cc->iov, cc->iov_sz, _net);          
    
  
  /* allocate io vector */
  switch (cc->header->protocol)
  {
    case ACOM::Network::P_AM: {
      /* allocate in the stack the deferred data */
      for (size_t i=0; i < cc->header->iovsz_d; ++i)
        cc->iov[i].data = cc->allocate( cc->iov[i].len );
    } break;
  
    case ACOM::Network::P_WS: {
      stub_server->server_get_iovect( cc );
    } break;
  
    case ACOM::Network::P_AWS: {
      /* allocate in the stack the deferred data */
      for (size_t i=0; i < cc->header->iovsz_d; ++i)
        cc->iov[i].data = _net->allocate_data( cc->iov[i].len );
    } break;
  };
  
  /* read iovector */
  if (cc->header->iovsz_d >0) {
    for (size_t i=0; i<cc->header->iovsz_d; ++i)
    {
//      std::cout << "read bloc len=" << cc->iov[i].len << "@addr=" << cc->iov[i].data << std::endl;
      nread = _speer.readn( cc->iov[i].data, cc->iov[i].len );
//      std::cout << "read bloc=" << cc->iov[i].len << std::endl;
      if (nread < 0) {
        std::cout << __LINE__ << ":" << __FILE__ << ": Cannot read defered data" << std::endl;
        ACOM::throwit(ACOM::ComFailure( ACOM::ComFailure::FAIL ));
      }
    }
  }
    
  /* call user service */
  stub_server->server_call( cc );

// std::cout << "Out : " << __PRETTY_FUNCTION__ << std::endl;
  return false;
}


} // namespace InukCorba

