// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_dispatcher.cpp,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#include <SOCKNET_network.h>
#include <SOCKNET_selector.h>
#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif


namespace SOCKNET {

// --------------------------------------------------------------------------
/* Listen incomming message a process them: args is a socket client
*/
void* SOCKNET_net::dispatcher_1( void * arg )
{
  SOCKNET_net::InChannel* input = (SOCKNET_net::InChannel*)arg;
  SOCKNET_net* net = input->get_net();
  
  bool shutdown = false;
  Network::CallContext* cc = 0;

  try {

    /* infinite loop that read protocol information ... 
    */
    while (!net->_shutdown) 
    {
      cc = net->acquire_callcontext();
      shutdown = input->server_stub( cc );
      if (shutdown) {
        //std::cout << "shutdown due to ... error ?" << std::endl;
        break;
      }
      net->release_callcontext( cc );
    } // while1
  } catch (const ACOM::Runtime& e ) {
    std::cout << "Throw exception ! " << e.what() << std::endl;
  }
  cc = 0;

  input->_speer.close();
  delete input;
#ifdef SOCKNET_TRACE_DEBUG
  std::cout << "OUT " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 0;
}


// ------------------------------------------------------------------------
/* Listen incomming message a process them: args is a SOCKNET_net
*/
void* SOCKNET_net::dispatcher_2( void * arg )
{
  SOCKNET_net* net = (SOCKNET_net*)arg;
  
  bool shutdown = false;
  Network::CallContext* cc = 0;

  try {
    struct timeval timeinselect;
    timeinselect.tv_sec = 0;
    timeinselect.tv_usec = 500000; /* 500 ms */
    
    /* infinite loop that read protocol information ... 
    */
    while (!net->_shutdown) 
    {
      SOCKNET_net::InChannel* input = 0;
      try {
        do {
          input = (SOCKNET_net::InChannel*)net->_selector->acquire( Selector::SET_READ, &net->_access, &timeinselect );
          if (net->_shutdown) 
            break;
        } while (input ==0);
        if (input !=0) {
//std::cout << "Thread id=" << PTH::Thread::getSelf() << " handles new message on socket fd=" << input->get_socket()->get_fd() << std::endl;
          cc = net->acquire_callcontext();
          shutdown = input->server_stub( cc );
          net->_selector->release( input, Selector::SET_READ, &net->_access);
          net->release_callcontext( cc );
        }
      } catch (const ACOM::ComFailure& e ) {
        delete input;
      }
      if (shutdown) 
        break;
    } // while1
  } catch (const ACOM::Runtime& e ) {
    std::cout << "Internal error: POA_Server::dispatcher"
              << "\nerr=" << e.what() << std::endl;
    ACOM::Acom::abort(-1);
  }
  cc = 0;
#ifdef SOCKNET_TRACE_DEBUG
  std::cout << "OUT " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 0;
}



// ------------------------------------------------------------------------
/** Create a Selector descriptor and spawn nthread to listen on it.
    Each new incomming connection are added to the set.
*/
void* SOCKNET_net::scavanger( void * arg )
{
  SOCKNET_net* net = (SOCKNET_net*)arg;
  
  /* create a pool of waiting thread for communication */
  if (net->_nthreads >0) {
//std::cout << "#threads for incomming message =" << net->_nthreads << std::endl;
    for (int i=0; i<net->_nthreads; ++i)
    {
      PTH::ThreadId tid;
      PTH::Thread::create( &tid, &SOCKNET_net::dispatcher_2, (void*)net);
//std::cout << "thread[" << i << "]=" << tid << std::endl;
    }
    //sleep(1);
  }
  
  try {

    SOCKNET_net::InChannel* newchannel = 0;
    while (!net->_stopscavanger) 
    {
      newchannel = 0;
#ifdef SOCKNET_TRACE_DEBUG
      std::cout << "Scavanger ready for accepting connection" << std::endl;
#endif
      newchannel = new SOCKNET_net::InChannel;
      newchannel->initialize(net);
      net->_s_connect.accept( net->_myaddress, newchannel->_speer );
      if (newchannel->_speer.get_fd() <0)
        continue;
  
      //newchannel->_speer.setSockOpt( SOCK::Socket::no_delay, true );
      /* add a new connextion */
      if (net->_nthreads >0)
      {
        newchannel->_is_connected = true;
//        newchannel->get_socket()->setSockOpt(SOCK::Socket::no_delay,true);
        PTH::CriticalGuard critical(net->_access);
//std::cout << "New input socket: " << newchannel->get_socket()->getURL() << std::endl;
        net->add_connexion( newchannel );
        newchannel->set_index( net->acquire_index() );
        net->_selector->insert( Selector::SET_READ, newchannel );
        net->_selector->force_rebuild_sets();
      }
  
#ifdef SOCKNET_TRACE_DEBUG
      std::cout << "Scavanger new connection:" <<  newchannel->_speer << std::endl;
      SOCK::Address addr;  newchannel->_speer.getPeerName( addr );
      std::cout << "Peer=" << addr << std::endl;
#endif
      if (net->_nthreads ==-1) {
        PTH::ThreadId tid;
        PTH::Thread::create( &tid, &SOCKNET_net::dispatcher_1, (void*)newchannel);
      }
    }
  } catch (ACOM::RuntimeError& e ) {
  }
  return 0;
}



}
