// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_network.cpp,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#include <SOCKNET_network.h>
#include <SOCKNET_selector.h>
#include <SOCKNET_service.h>

namespace SOCKNET {

// --------------------------------------------------------------------------
  /** iovect entries in a cell
   */
const int SOCKNET_net::_nb_entries= 16384;

  /** packed block size
   */
const int SOCKNET_net::_block_sz = 65535;

  /** size of cached free blocks
   */
const int SOCKNET_net::_block_threshold= 16384;

  /** size of cached free blocks
   */
const int SOCKNET_net::_iovect_threshold=64;

  /** size of cached free blocks
   */
const int SOCKNET_net::_request_threshold=64;


// --------------------------------------------------------------------------
//
// Constructor
//
SOCKNET_net::SOCKNET_net()
 : ACOM::Network()
{
  set_property("name",     "SOCKNET");
  set_property("primary",  "false");
  set_property("secondary","true");
  set_name("SOCKNET");

  initAllocators();

  _subnet = 0;
  _is_init = false;
  _nb_init = 0;
  _is_initcommit = false;
  _is_terminate = false;

  _nthreads  = 1; //-1; //2;  /* #threads to listen on select, -1 : one per connexion */
  _cachesize = -1;
  _opencx = 0;
  _selector = new Selector;
  _selector->initialize();
  // not need by the current implementation of PTH, cstor initial them
  _access.init();
  _cond.init();
  _all_routes.clear();
  _all_connexions.clear();

  _service_geturl = new ServiceGetUrl;
  _service_reply = new ServiceReply;
  _service_geturl->initialize( this );
  _service_reply->initialize( this );
  ACOM::Service::Manager.bind( ACOM::ServiceManager::SOCKNET_LNS, 1, _service_geturl );
  ACOM::Service::Manager.bind( ACOM::ServiceManager::SOCKNET_LNS, 2, _service_reply );
}


// --------------------------------------------------------------------------
SOCKNET_net::~SOCKNET_net()
{
  clear();
}


// --------------------------------------------------------------------
void SOCKNET_net::clear()
{
//  ACOM::NetworkClient::clear();
  ACOM::NetworkServer::clear();
//  ACOM::NetworkAdmin::clear();
  _is_init = false;
  _nb_init = 0;
  _is_initcommit = false;
  _is_terminate = false;
  _all_routes.clear();
  Connexion* c =0;
  while (!_all_connexions.empty())
  {
    c = _all_connexions.pull_front();
    if (c !=0) c->destroy();
    delete c;
  };
  _selector->destroy();
  delete _selector;
  delete _service_geturl;
  delete _service_reply;
}

// ---------------------------------------------------------------------------
/** Network specific barrier
    \warning  parameters count, time out etc ...( voir avec Cyrille)
*/
void SOCKNET_net::barrier()
  throw (ACOM::ComFailure) 
{
  _subnet->barrier(); 
}


// ---------------------------------------------------------------------------
void SOCKNET_net::initAllocators() 
{
  _comClock=0;
  _reqA.init( _request_threshold );
  _bA.init(_block_sz, _block_threshold  );
  _iovA.init( sizeof(ACOM::IOVectCell)+(_nb_entries*sizeof(ACOM::IOVectEntry)),
              _iovect_threshold );
}


// ---------------------------------------------------------------------------
/** Return the route (channel) to send data to LocalNodeNo peer
    @param   peer(IN) remote node id
*/
ACOM::Network::OutChannel* SOCKNET_net::get_default_route( LocalNodeNo peer )
    throw(ACOM::NetworkError)
{
  
  ACOM::Network::OutChannel* ch = _all_routes[peer];
  if (ch ==0) ch = (ACOM::Network::OutChannel*)open_channel( peer );
  return ch;
}  


// ---------------------------------------------------------------------------
/** Open a channel between local node  to remote peer
    @param   peer(IN) remote node id
    \warning ajouter les aliasing d'adresse de noeud
*/
ACOM::Network::OutChannel* SOCKNET_net::open_channel( LocalNodeNo peer )
      throw(ACOM::ComFailure)
{
  PTH::CriticalGuard critical(_access);
  if (peer > _all_routes.size()) 
  {
    /* reget the node count ... */
    _node_count = _subnet->node_count();
    _all_routes.resize( _node_count );
    if (peer > _node_count) 
      ACOM::throwit( ACOM::ComFailure("no route to node", ACOM::ComFailure::NOROUTE) );
  }

  SOCKNET_net::OutChannel* ch = _all_routes[peer];
  if (ch !=0) return ch;
  ch = _all_routes[peer] = new OutChannel();
  ch->initialize( this, peer );
  return ch;
}


// ---------------------------------------------------------------------------
/** Close a channel between local node  to remote peer
    @param   ch (IN)channel to close
    \warning ajouter les aliasing d'adresse de noeud
*/
void SOCKNET_net::close_channel( ACOM::Network::OutChannel* c )
    throw(ACOM::ComFailure)
{
  SOCKNET_net::OutChannel* ch =(SOCKNET_net::OutChannel*)c;
  int peer = ch->get_peer();
  _all_routes[peer]->destroy();
  PTH::CriticalGuard critical(_access);
  _all_routes[peer] = 0;
  delete ch;
}  


// ---------------------------------------------------------------------------
ACOM::Network::Request* SOCKNET_net::allocate_request () 
    throw (ACOM::BadAlloc) 
{
  SOCKNET_net::Request* req = _reqA.alloc() ;
  if (!req) ACOM::throwit( ACOM::BadAlloc("[[SOCKNET_net::allocate_request::]] fail") );
  if (!req->init( _block_sz, &_bA, _nb_entries, &_iovA, (ACOM::Network*)this ))
  {
    _reqA.free(req);
    ACOM::throwit( ACOM::BadAlloc("[[SOCKNET_net::allocate_request::]] fail") );
  }
//  std::cout << "[SOCKNET]:: allocate request:" << (long)req << std::endl;
  return req;
}


// ---------------------------------------------------------------------------
void SOCKNET_net::deallocate_request( ACOM::Network::Request* rq ) 
    throw (ACOM::InvalidArgument) 
{
  SOCKNET_net::Request* req = (SOCKNET_net::Request*)rq;
//  std::cout << "[SOCKNET]:: dellocate request:" << (long)req << std::endl;
  if (req) 
  {
    if (req->get_net()!= this )
      ACOM::throwit( ACOM::InvalidArgument("[[SOCKNET_net::deallocate_request::]] not aloopback request") );
    else
      _reqA.free(req);
      //delete req; 
    return;
  }
  else 
    ACOM::throwit( ACOM::InvalidArgument("[[SOCKNET_net::deallocate_request::]] null request") );
}

#define CC_ONEBLOC

// --------------------------------------------------------------------------
ACOM::Network::CallContext* SOCKNET_net::allocate_callcontext(size_t sz) throw (ACOM::BadAlloc)
{
  if (sz == (size_t)-1) sz = 2048;
#ifdef CC_ONEBLOC  /* allocate one bloc in which we store CallContext + the stack for call context */
  void* zone = malloc( sizeof(ACOM::Network::CallContext)+ sz + 2048);
  ACOM::Network::CallContext* cc = new (zone) ACOM::Network::CallContext;
  cc->init( this, 0 ); 
  cc->_base = ((char*)zone) + sizeof(ACOM::Network::CallContext);
  cc->_top = cc->_base;
  cc->_stack_size = sz;
#else
  ACOM::Network::CallContext* cc = new ACOM::Network::CallContext;
  cc->init( this, sz ); 
#endif  
  cc->_im_data_seq.init( _nb_entries, &_iovA);
//  std::cout << "alloc cc=" << (long)cc << std::endl;
  return cc;
}



// --------------------------------------------------------------------------
void SOCKNET_net::deallocate_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument)
{
#ifdef CC_ONEBLOC  /* allocate one bloc in which we store CallContext + the stack for call context */
  cc->_stack_size =0;
  cc->destroy( );
  cc->ACOM::Network::CallContext::~CallContext();
  free( cc );
#else
  cc->destroy( );
  delete cc;
#endif
}



// --------------------------------------------------------------------------
ACOM::Network::CallContext* SOCKNET_net::resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) 
  throw (ACOM::BadAlloc)
{
  if (cc->_stack_size > sz) return cc;
  size_t offset = cc->size();
  ACOM::Network::CallContext*  new_cc = allocate_callcontext(sz+128);
  memcpy( new_cc->_base, cc->_base, offset );
  new_cc->_top = ((char*)new_cc->_base) + offset;
  new_cc->error_no = cc->error_no;
  new_cc->cc       = cc->cc;
//  cc->reset();
//  std::cout << "resize cc=" << (long)cc << " new cc=" << (long)new_cc << std::endl;
  deallocate_callcontext( cc );
  return new_cc;
}



// --------------------------------------------------------------------------
ACOM::Network::CallContext* SOCKNET_net::acquire_callcontext() throw (ACOM::BadAlloc)
{
  PTH::CriticalGuard critical( _access);
  return ACOM::Network::acquire_callcontext();
}


// --------------------------------------------------------------------------
void SOCKNET_net::release_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument)
{
  PTH::CriticalGuard critical( _access);
  ACOM::Network::release_callcontext( cc );
}


// --------------------------------------------------------------------------
void SOCKNET_net::add_connexion( Connexion* c )
{
  if ((_cachesize >0) && (_opencx >= _cachesize)) 
  {
retry:
    Connexion* todel = _all_connexions.back();
    while (todel !=0)
    {
      int cpt = 11; 
      while ((--cpt !=0) && !todel->_access.trylock()) ;
      if (cpt ==0) 
        todel = todel->pred();
    }
    if (todel ==0) goto retry;
    del_connexion( todel );
    todel->_access.unlock();
  }
  else ++_opencx;
  _all_connexions.push_front( c );
}


// --------------------------------------------------------------------------
void SOCKNET_net::del_connexion( Connexion* c )
{
  c->disconnect();
  _all_connexions.erase(c);
}


// --------------------------------------------------------------------------
void SOCKNET_net::touch_connexion( Connexion* c )
{
  _all_connexions.erase( c );
  _all_connexions.push_front( c );
}


// --------------------------------------------------------------------------
int SOCKNET_net::acquire_index()
{
  static int counter_index = 0;
  if (_free_index.empty()) {
    return counter_index++;
  }
  int retval = _free_index.front();
  _free_index.pop_front();
}


// --------------------------------------------------------------------------
void SOCKNET_net::release_index(int id)
{
  _free_index.push_front(id);
}


} //End of namespace SOCKNET_NET

extern "C" {

	static SOCKNET::SOCKNET_net* socknet =0;

	// - dynamic library initialization:
	static void factory_initialize() 
	{
		socknet = new SOCKNET::SOCKNET_net();
		ACOM::Acom::bind_network("SOCKNET", socknet );
	}
	extern void SOCKNET_factory_initialize() 
	{
		//dbg.log("","factory init\n");
		factory_initialize();
		//dbg.log("","factory init OK\n");
	}
	
	// - dynamic library terminaison:
	static void factory_destroy() 
	{
		delete socknet;
	}
	extern void SOCKNET_factory_destroy() 
	{
		return factory_destroy();
	}
}

