// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_network.h,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#ifndef __SOCKNET_NET_H
#define __SOCKNET_NET_H

#include <pth>
#include <sock>
#include <acom>

namespace SOCKNET {

/* fwd declaration */
class Selector;

/* fwd declaration */
class ServiceGetUrl;

/* fwd declaration */
class ServiceReply;

/** SOCKET Network implementation over a primary network.
 *  SOCKET is a secondary network that should initialized over an existing
 * initialized network
 */
class SOCKNET_net: public ACOM::Network {
  // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------

public:
  // --------------------------------------------------------------------------
  //
  // Constructor
  //
  SOCKNET_net();
  ~SOCKNET_net();

  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
  int initialize( int* argc, char** argv[] ) 
    throw (ACOM::NetworkError);

  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
  int initialize( int* argc, char** argv[], ACOM::Network* netInit ) 
    throw (ACOM::NetworkError);

  /** Set the number of threads to wait to treat incomming message
      If parameter is -1, then one thread per connexion is created.
      @param nthread (IN)
  */
  void set_poolsize( int nthread );

  /** Set the number of entries in the cache of open connexion.
      A value equals to -1, means no cache limit.
      @param nentries (IN)
  */
  void set_cachesize( int nentries );

  // ---------------------------------------------------------------------------
  /** Second stage of the initialisation process.
      End of the first stage of initialisation: all local initialisations
      are done, communication could occurs after the second stage (after
      the call to initcommit) or during the call of initcommit in an other
      module.
  */
  int commit () throw (ACOM::NetworkError);

  /** Terminate the net...
  */
  int terminate () throw (ACOM::NetworkError);

  /** Clear 
  */
  void clear();

  /** Display usage about SOCKNET, throw error if msg != ""
  */
  void usage( const std::string& msg = "" ) throw (ACOM::Runtime);

public:
  /** Allocate a new request.
  */
  ACOM::Network::Request* allocate_request () throw (ACOM::BadAlloc);

  /** free request descriptor when unused
  */
  void deallocate_request( ACOM::Network::Request* req ) throw (ACOM::InvalidArgument);


  // ---------------------------------------------------------------------------
  /** Allocate  an immediate buffer of sz bytes in network memory
      @param  sz (IN) in bytes
  */
  void* allocate_data( size_t sz )
    throw (ACOM::BadAlloc) 
  { return std::malloc(sz); }


  /** Deallocate a previously allocated immediate buffer  in network memory
      @param  data (IN) byte block to deallocate
  */
  void deallocate_data(  void* data )
    throw (ACOM::InvalidArgument) 
  { std::free(data); }


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Return the route (channel) to send data to LocalNodeNo peer
      @param   peer(IN) remote node id
  */
  ACOM::Network::OutChannel* get_default_route( LocalNodeNo peer )
    throw(ACOM::NetworkError);

  /** Open a channel between local node  to remote peer
      @param   peer(IN) remote node id
      \warning ajouter les aliasing d'adresse de noeud
  */
	ACOM::Network::OutChannel* open_channel( LocalNodeNo peer )
       throw(ACOM::ComFailure);

  /** Close a channel between local node  to remote peer
      @param   ch (IN)channel to close
      \warning ajouter les aliasing d'adresse de noeud
  */
  void close_channel( ACOM::Network::OutChannel* ch )
      throw(ACOM::ComFailure);

  /** send  to a node (AM,AWS,WS protocols)
      @param   req  request (IN)
      @param   ch  channel  (IN)
  */
  void post( ACOM::Network::Request* req, ACOM::Network::OutChannel* ch)
    throw (ACOM::ComFailure);

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */
  void barrier()
    throw (ACOM::ComFailure); 


public:
	/** The definition of a connexion class
  */
  class Connexion : public PTH::Queable<Connexion> {
  protected:
    Connexion() : _index(-1), _is_acquired(false), _is_connected(false)
    {}

    virtual ~Connexion()
    {}

    /** destroy
    */
    virtual void destroy( ) =0;
    
    /** disconnect
    */
    virtual void disconnect( ) =0;
    
    /* Get socket */
    SOCK::SocketClient* get_socket()
    { return &_speer; }

    /* Get index in network */
    int get_index()
    { return _index; }

    /* Set index in network */
    void set_index(int i)
    { _index =i; }
  protected:
    PTH::Mutex         _access;       // - critical region
    SOCK::SocketClient _speer;        // - socket
    int                _index;        // - socket
    bool               _is_acquired;  // true if acquired
    bool               _is_connected; // true if connected
    friend class SOCKNET_net;
    friend class Selector;
  };

	/** The definition of OutChannel class
  */
  class OutChannel: public Connexion, public ACOM::Network::OutChannel {
  public:
    void post( ACOM::Network::Request* r ) throw (ACOM::ComFailure);

  protected:
    /** Constructor
    */
    OutChannel( );

    /** destructor
    */
    ~OutChannel();

    /** initialization
    */
    void initialize( Network* net, int peer );

    /** destroy and disconnect if needed
    */
    void destroy( );

    /** connect to remote peer node
     */
    void connect() throw (ACOM::ComFailure);

    /** disconnect to remote peer node
     */
    void disconnect();

    /** overload get_net
    */
    SOCKNET_net* get_net()
    { return (SOCKNET_net*)_net; }
  protected:
    std::string  _urlpeer;
    friend class SOCKNET_net;
  };


  /** The input channel class
  */
  class InChannel : public Connexion {
  protected:
    /** Constructor
    */
    InChannel( );

    /** destructor
    */
    ~InChannel();
    
    /** Initialize it
    */
    void initialize( SOCKNET_net*  net );

    /** Destroy it 
    */
    void destroy( );

    /** disconnect
    */
    void disconnect( );

    /* Return the associated net 
    */
    SOCKNET_net* get_net()
    { return _net; }
    
    /* Server stub : return true if shutdown
    */
    bool server_stub( CallContext*& cc );
    
  protected:
    SOCKNET_net*       _net;          // - associated network
    friend class SOCKNET_net;
  };


  /** The request class 
  */
  class Request: public ACOM::Network::Request, public PTH::Linkable<Request> {
  protected:
    friend class SOCKNET_net;
  };

protected:

  /** Receipt class
  */
  class Receipt: public ACOM::Network::Receipt {
  protected:
    friend class SOCKNET_net;
  };

  /** Allocate a new call context to treats incomming message
  */
  ACOM::Network::CallContext* allocate_callcontext(size_t sz) throw (ACOM::BadAlloc);

  /** Deallocate a call context
  */
  void deallocate_callcontext(ACOM::Network::CallContext*) throw (ACOM::InvalidArgument);

  /** Reallocate of the stack of the call context
  */
  ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* ctxt, size_t sz ) throw (ACOM::BadAlloc);

  /** Acquire a call context for a new incomming communication
  */
  ACOM::Network::CallContext* acquire_callcontext() throw (ACOM::BadAlloc);
  
  /** Release a call context 
  */
  void release_callcontext(ACOM::Network::CallContext*) throw (ACOM::InvalidArgument);

  /** iovect entries in a cell
   */
  static const int _nb_entries;

  /** packed block size
   */
  static const int _block_sz;

  /** Logical clock for timestamping request
   */
  PTH::AtomicCounter _comClock;

  /** size of cached free blocks
   */
  static const int _block_threshold; 

  /** size of cached free blocks
   */
  static const int _iovect_threshold;

  /** size of cached free blocks
   */
  static const int _request_threshold;

  /** byte cell allocator
   */
  ACOM::ByteCellAllocator _bA;

  /** iovec allocator
   */
  ACOM::IOVectCellAllocator _iovA;

  void initAllocators();

  /** request allocator
   */
  typedef PTH::SyncCellAllocator<Request> RequestAllocator ;
  RequestAllocator _reqA;

  // ------------------------------------------------------------------------
protected:
  bool   _is_init;         // to avoid multiple calls to the associated
  int    _nb_init;         // number of calls
  bool   _is_initcommit;   // methods
  bool   _is_terminate;
  int    _cachesize;       // number of cached active connection
  int    _opencx;          // number of open connexion
  int    _nthreads;        // number of thread that wait for message
  bool   _stopscavanger;   // to stop scavanger thread
  bool   _shutdown;        // to stop all threads !

  ACOM::Network*           _subnet;
  PTH::Mutex               _access; // to lock access to internal data
  PTH::Condition           _cond;   // to wait initial group
  std::vector<OutChannel*> _all_routes;      // - routes to other 
  PTH::Queue<Connexion>    _all_connexions;  // - In and Out connected channel
  std::list<int>           _free_index;      // - list of free index to number the connexion in Selector
  Selector*                _selector;        // - set of all OutChannel connected
  
  /* cache management : _access should be lock */
  void add_connexion( Connexion* c );

  /* cache management : _access should be lock, as weel as lock on connexion */
  void del_connexion( Connexion* c );

  /* touch a connexion : _access should be lock */
  void touch_connexion( Connexion* c );
  
  /* return a new index for connexion : _access should be lock */
  int acquire_index();

  /* release usage of a new index for connexion : _access should be lock */
  void release_index(int id);

private:
  enum AddProtocolNo {
    P_SHUTDOWN = ACOM::Network::P_FREE 
  };
  std::string        _url;        /* my url to connect to me */
  SOCK::Address      _myaddress;  /* my address */
  SOCK::SocketServer _s_connect;  /* socket for connexion */

  /** function called by a thread to wait for a new message 
      arg is the SocketClient to read for new message
  */
  static void* dispatcher_1( void * arg );

  /** function called by a thread to wait for a new message 
      arg is the network to wait on Selector _bigset
  */
  static void* dispatcher_2( void * arg );

  /** function called by a thread to wait for a new connexion 
  */
  static void* scavanger( void * arg );
  
  /** to get a remote url for connexion
  */
  ServiceGetUrl* _service_geturl;
  ServiceReply*  _service_reply;
  
  friend class OutChannel;
  friend class InChannel;
  friend class ServiceGetUrl;
  friend class ServiceReply;
}; // -- end class Network

} // End of namespace

#endif //end of define __SOCKNET_NET_H
