// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_network_init.cpp,v $
// Copyright(c)'2001 by Apache Team
// Author: Thierry Gautier
//
//
//
// ==========================================================================
#include <SOCKNET_network.h>
#include <SOCKNET_selector.h>

#include <string>
#include <fstream>
#include <sstream>
#ifdef PTH_USE_PTHREAD_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif

#include <sys/types.h> // - for getpid
#include <unistd.h>    // - for getpid

//#define SOCKNET_TRACE_CALL

namespace SOCKNET {


// --------------------------------------------------------------------------
//
// Static object that will register to the list of modules
//
// --------------------------------------------------------------------------//
static bool verboseon = false;



// --------------------------------------------------------------------------
void SOCKNET_net::usage( const std::string& msg ) 
  throw (ACOM::Runtime)
{
  if (msg != "")
    std::cerr << "*** Error:" << msg << std::endl;
  std::cerr << "Options for SOCKNET are: "
            << "[-- SOCKNET [-h] [-v] [-poolsize <n threads] [-cachesize <n connections> \n"
            << "Where:\n"
            << "* -h : display this help\n"
            << "* -v : set verbose mode on, default off\n"
            << "* -poolsize <nthreads: set the number of threads to process input messages,\n"
            << "                       -1 means one threads per connections\n"
            << "* -cachesize <n connections: set the maximal number of open connexions\n"
            << "                       -1 means no limit\n"
            << std::endl << std::flush;
  if (msg !="")
   ACOM::throwit( ACOM::Runtime(msg.c_str()) );
}



// --------------------------------------------------------------------------
//
int SOCKNET_net::initialize( int* argc, char** argv[] ) 
    throw (ACOM::NetworkError)
{
  std::cerr << "SOCKNET is a secondary network that need an already connected network"
            << std::endl;
  ACOM::throwit( ACOM::NetworkError() );
}



// --------------------------------------------------------------------------
//
int SOCKNET_net::initialize( int* argc, char** argv[], ACOM::Network* primary_net )
    throw (ACOM::NetworkError)
{
  if (primary_net ==0) {
    std::cerr << "SOCKNET is a secondary network that need an already connected network"
              << std::endl;
    ACOM::throwit( ACOM::NetworkError() );
  }
  
  // to avoid multiple call of this method
  _nb_init++;
  if (_is_init) return 0;
  _is_init = true;
  _subnet = primary_net;

  /* parse my options:
     -h display help
     -poolsize <nthreads> : number of threads to treats request, 
                            -1 means one threads per connexions.
     -cachesize <nconnections> : maximal number of active connexions at the same time.
  */
  for (int i=0; i<*argc; ++i)
  {
    if (std::string((*argv)[i]) == "-h")
      usage();
    else if (std::string((*argv)[i]) == "-poolsize")
    {
      if (*argc <= i+1)
        usage("Usage: -poolsize <nthreads>");
      int nth = atoi((*argv)[i+1]);
      if ((nth <=0) && (nth != -1))
        usage("Invalid argument for -poolsize");
      _nthreads = nth;
      ++i;
    }
    else if (std::string((*argv)[i]) == "-cachesize")
    {
      if (*argc <= i+1)
        usage("Usage: -cachesize <n connexions>");
      int nc = atoi((*argv)[i+1]);
      if ((nc <=0) && (nc != -1))
        usage("Invalid argument for -cachesize");
      _cachesize = nc;
      ++i;
    }
    else if (std::string((*argv)[i]) == "-v")
    {
      verboseon = true;
      std::cout << "[SOCKNET]:: setup verbose on" << std::endl << std::flush;
    }
  }

#ifdef SOCKNET_TRACE_CALL
  std::cout << "In " << __PRETTY_FUNCTION__ << std::endl;
#endif

  PTH::Pth::initialize(argc, argv);
	PTH::Pth::commit();

  _node_count = _subnet->node_count();
  _node_rank  = _subnet->rank();

  return 1;
}


// --------------------------------------------------------------------------
//
// InitCommit
//
//#define SOCKNET_TRACE_DEBUG
int SOCKNET_net::commit( )
    throw (ACOM::NetworkError)
{
	// to avoid multiple call of this method
	if (_is_initcommit) return 0;
	_is_initcommit = true;

  //_subnet->commit();

  // WARN SHOULD BLOC FOR DYNAMIC NET
  //_subnet->barrier();
  PTH::CriticalGuard critical(_access);
  _node_count = _subnet->node_count();
  _node_rank  = _subnet->rank();

  /* initialize socket for incomming connection */
  //_myaddress.setByName("localhost"); 
  //_myaddress.setPort(5000);
  _myaddress = SOCK::Address::getLocalHost();
  //_myaddress = SOCK::Address::getByName( SOCK::Address::getLocalHost().getHostName() );
  int err    = _s_connect.bind( _myaddress );
  if (err <0) {
    perror("SOCKNET_net::initialize, bind:: ");
    ACOM::throwit( ACOM::NetworkError() );
  }
  err        = _s_connect.listen(1);
  _url       = _s_connect.getURL();

  if (verboseon)
    std::cout << _node_rank << "::my url=" << _url << std::endl;

  /* resize the node count */
  _all_routes.resize( _node_count );
  for (int i=0; i<_node_count; ++i)
  {
    _all_routes[i] = 0;
  }

  if (verboseon) {
    std::cout << " My ranks=" << _node_rank << std::endl;
    std::cout << " #nodes  =" << _node_count << std::endl;
  }

  /* SIGSEGV if using Pthread start method: TG a verfier */
  _stopscavanger = false;
  PTH::ThreadId tid;
  PTH::Thread::create( &tid, &SOCKNET_net::scavanger, (void*)this);

#ifdef SOCKNET_TRACE_CALL
	std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif

  return 1;
}



// --------------------------------------------------------------------------
//
// Terminate
//
int SOCKNET_net::terminate( )
    throw (ACOM::NetworkError)
{
#ifdef SOCKNET_TRACE_CALL
  std::cout << "In " << __PRETTY_FUNCTION__ << std::endl;
#endif
  _nb_init--;
  if ((_nb_init > 0) || (_is_terminate)) return 0;
  _is_terminate = true;

  _stopscavanger = true;
  _shutdown = true;
  //_selector->force_rebuild_sets();
  _selector->wakeup_all( &_access );
  
  /* Create a dummy connection to local node to wake up scavanger */
  SOCK::SocketClient swakeup;
  int err = swakeup.connect( _s_connect.getAddress() );
  /* */
  for (int i=0; i< _node_count; ++i)
  {
    if (_all_routes[i] !=0) close_channel( _all_routes[i] );
  }
  swakeup.close();


  try {
    _s_connect.close();
    PTH::Pth::terminate();
  }
  catch (...) {
    std::cerr << "Unknown exception" << std::endl;
    PTH::Pth::terminate();
  }

#ifdef SOCKNET_TRACE_CALL
  std::cout << "Out " << __PRETTY_FUNCTION__ << std::endl;
#endif
  return 1;
}


} // -namespace SOCKNET

