// ==========================================================================
// $Source: /users/huron/gautier/CVS/Inuktitut2/SOCK_net/src/SOCKNET_selector.cpp,v $
// Copyright(c)'2001 by Apache Team
// Authors: Thierry Gautier
//
//
// ==========================================================================
#include <SOCKNET_selector.h>

//#define SOCKNET_SELECTOR_TRACE

namespace SOCKNET {

// ---------------------------------------------------------------------------
void Selector::initialize()
{
  _access_select.init();
  _condcritical.init();
  _force        = false;
  _nselect      = 0;
  _incritical   =false;
  _waitcritical = 0;
  _shouldwakeup = false;
  _fdmax        = 0;
  for (int k=0; k<3; ++k)
  {
    _pos[k] =0;
    _sc[k].clear();
    FD_ZERO(&_set[k]);
    FD_ZERO(&_setselected[k]);
//    _set[k].zero();
//    _setselected[k].zero();
  }
  if (SOCK::SocketClient::sock_pair( &_wakeupW, &_wakeupR ) < 0)
  {
    ACOM::throwit( ACOM::RuntimeError("cannot initialise sock_pair", errno) );
  }
  _wakeupW.setSockOpt(SOCK::Socket::no_delay,true);
  
  /* add wakeupR in set READ */
  FD_SET(_wakeupR.get_fd(), &_set[0]);
  _fdmax = _wakeupR.get_fd();
//  std::cout << "FDMAX at init=" << _fdmax << std::endl;
}


// ---------------------------------------------------------------------------
void Selector::destroy()
{
  _wakeupW.close();
  _wakeupR.close();
  _access_select.destroy();
  _condcritical.destroy();
}
  


// ---------------------------------------------------------------------------
void Selector::wakeup_all( PTH::Mutex* lockset )
{
  PTH::CriticalGuard critical(*lockset);
//  std::cout << "In wakeup_all" << std::endl;
  _shouldwakeup = true;
  while (_waitcritical >0) {
//    std::cout << "In wakeup_all: in acquire=" << _incritical << std::endl;
    int err = _wakeupW.writen(&_force, 1 );
    lockset->unlock();
    _condcritical.broadcast();
    PTH::Thread::yield();
    PTH::Thread::sleep(10, 0); // 10 ms, 0 nano sec
    lockset->lock();
  }
  _shouldwakeup = false;
}


// ---------------------------------------------------------------------------
void Selector::force_rebuild_sets()
{ 
  _force = true;
  int err = _wakeupW.writen(&_force, 1 );
//std::cout << "Force rebuild sets, write 1, err=" << err << std::endl;
}


// ---------------------------------------------------------------------------
void Selector::release( SOCKNET_net::Connexion* c, SetKind v, PTH::Mutex* lockset ) 
  throw (ACOM::InvalidArgument)
{
  ACOM_ASSERT_DEBUG( c !=0, "bad assertion");
  ACOM_ASSERT_DEBUG( lockset !=0, "bad assertion");
  c->_is_acquired = false;
  c->_access.unlock();
}



// ---------------------------------------------------------------------------
static Selector::SetKind MASK[3] = { Selector::SET_READ, Selector::SET_WRITE, Selector::SET_EXCEP };

SOCKNET_net::Connexion* Selector::acquire( SetKind v, PTH::Mutex* lockset, struct timeval* t ) 
  throw (ACOM::ComFailure)
{
  ACOM_ASSERT_DEBUG( lockset !=0, "bad assertion");
  ACOM_ASSERT_DEBUG( ((v>=0) && (v <= SET_ALL)), "bad assertion");

  /* choosen set for select */
  fd_set* set_r;
  fd_set* set_w;
  fd_set* set_e;
  SOCKNET_net::Connexion* retval = 0;
  int loop_search = 2;

  lockset->lock();
  if ( _shouldwakeup ) {
    lockset->unlock();
    return 0;
  }

  if (_incritical) 
  {
    ++_waitcritical;
    while (_incritical) 
    {
#ifdef SOCKNET_SELECTOR_TRACE  
      std::cout << "Thread: " << PTH::Thread::getSelf() << " wait in select, #wait critical=" << _waitcritical << std::endl;
#endif
      _condcritical.wait( *lockset );
      if ( _shouldwakeup ) 
      {
        --_waitcritical;
        lockset->unlock();
        return 0;
      }
    }
    --_waitcritical;
  }
  _incritical = true;

#ifdef SOCKNET_SELECTOR_TRACE  
  std::cout << "Thread: " << PTH::Thread::getSelf() << " enter in select, #wait critical=" << _waitcritical << std::endl;
#endif

redo_selection:
  if (_nselect ==0)
  {
    /* always add socket to wakeup from select */
    set_r = &_setselected[0];
    FD_ZERO(set_r);
    if ((v & SET_READ) !=0) {
#ifdef SOCK_USE_DARWIN
      FD_COPY(&_set[0], set_r);
#else
      *set_r = _set[0];
#endif
    } else {
      /* Copy _wakeupR */
      FD_SET(_wakeupR.get_fd(), set_r);
    }
    
    if ((v & SET_WRITE) !=0) {
      set_w = &_setselected[1];
#ifdef SOCK_USE_DARWIN
      FD_COPY(&_set[1], set_w);
#else
      *set_w = _set[1];
#endif
    } else {
      set_w =0;
    }

    if ((v & SET_EXCEP) !=0) {
      set_e = &_setselected[2];
#ifdef SOCK_USE_DARWIN
      FD_COPY(&_set[2], set_e);
#else
      *set_e = _set[2];
#endif
    } else {
      set_e =0;
    }
    
    /* what ever is the set on which we want to acquire, reset 
       iteration on fd_set 
    */
    _pos[0] = 0;
    _pos[1] = 0;
    _pos[2] = 0;

    /* call select and unlock lockset */
#ifdef SOCKNET_SELECTOR_TRACE  
    std::cout << "Thread: " << PTH::Thread::getSelf() << " do select, #wait critical=" 
              << _waitcritical << ", fdmax=" << _fdmax << std::endl;
#endif
    lockset->unlock();
    _nselect = ::select( 1+_fdmax, set_r, set_w, set_e, 0); //t );
    lockset->lock();
#ifdef SOCKNET_SELECTOR_TRACE  
    std::cout << "Thread: " << PTH::Thread::getSelf() << " enddo select=" << _nselect 
              << ", #wait critical=" << _waitcritical << std::endl;
#endif
    if (_nselect <0)
      ACOM::throwit( ACOM::ComFailure("[Selector:: select]") );
  }
  if (_nselect ==0) {
    goto return_value;
  }


  /* search for the activated socket */
  if (FD_ISSET(_wakeupR.get_fd(), &_setselected[0])) 
  {
    /* mark it unavailable for next select if _nselect >1 */
    FD_CLR(_wakeupR.get_fd(), &_setselected[0]);

#ifdef SOCKNET_SELECTOR_TRACE  
    std::cout << "Force rebuild sets, read 1...";
#endif    
    int err = _wakeupR.readn(&_force, 1 );
#ifdef SOCKNET_SELECTOR_TRACE  
    std::cout << "Ok ?, err=" << err << std::endl << std::flush;
#endif    
    _fdmax = _wakeupR.get_fd();
    for (int k=0; k<3; ++k)
    {
      for (int i=0; i< _sc[k].size(); ++i)
        if (_sc[k][i] ==0) 
          continue;
        else if (_sc[k][i]->get_socket()->get_fd() > _fdmax)
          _fdmax = _sc[k][i]->get_socket()->get_fd();
    }
    _nselect = 0;
    goto redo_selection;
  }
    
redo_search:
  /* Search the connexion in all set */
  for (int k= 0; k < 3; ++k)
  {
    if ( (MASK[k] & v) == 0) 
      continue;
    /* ok flag on the k-th set : _setselected[k] */
    for (size_t i=_pos[k]; i< _sc[k].size(); ++i)
    {
      /* connexion was already acquired */
      if (_sc[k][i] == 0) continue;
      if (_sc[k][i]->_is_acquired) continue;
      if (FD_ISSET(_sc[k][i]->get_socket()->get_fd(), &_setselected[k]))
      {
        --_nselect;

        /* next position to iterates */
        _pos[k] = i+1;

        /* the connexion */
        retval = _sc[k][i];

        /* lock it before return it */
        retval->_access.lock();
        retval->_is_acquired = true;
        goto return_value;
      }
    }
    _pos[k] = 0;
  }

  /* redo search */
  if (--loop_search >0)
    goto redo_search;
  _nselect = 0;
  
return_value:
  _incritical = false;
  bool flag = _waitcritical >0;
#ifdef SOCKNET_SELECTOR_TRACE  
  std::cout << "Thread: " << PTH::Thread::getSelf() << " leave in select, #wait critical=" << _waitcritical << std::endl;
#endif
  lockset->unlock();
  if (flag) _condcritical.broadcast();
  return retval;
}

 
} // End of namespace
