#include<TCP_network.h>
#include<TCP_manager.h>
#include <acom>
#ifdef INUKTITUT_USE_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif


namespace TCP_NET
{

acom::Network::OutChannel* tcp_net::get_default_route( acom::Network::LocalNodeNo peer )
       throw(acom::NetworkError) 
{
  TCP_Channel* ch;

  //std::cout<<"get default route"<<std::endl;
  defroute.lock();
    if(_default_routes.find(peer)!=_default_routes.end())
      ch =(TCP_Channel*) _default_routes[peer];  
    else
      ch =NULL; 
  defroute.unlock();
  if(ch==NULL)
  {
    //std::cout<<"route to "<<peer<<" don't exist create it"<<std::endl;
    try{
      ch=(TCP_Channel*)open_channel(peer);
    }catch(acom::ComFailure e)
    {
      throw(e);
    }
  }
  
  return ch;
}

void tcp_net::set_default_route(  acom::Network::OutChannel* ch  )
       throw(acom::NetworkError) 
{
  if(ch->get_net()!=this)
    throw(acom::NetworkError("Bad Network"));
  defroute.lock();
  _default_routes[ch->get_peer()] =(TCP_Channel*) ch;
  defroute.unlock();
}

void tcp_net::AddNewChannel(TCP_Channel* ch)
{
//  std::cout<<"ADD NEW CHANNEL"<<std::endl;
  bool newhost=false;
  Khosts.lock();
  if(_known_nodes.find(ch->get_peer())==_known_nodes.end())
  {
    acom::Network::NodeDesc* desc = new acom::Network::NodeDesc();
    desc->init(ch->get_peer(),"TCP_NET");
    desc->_opened_channels.push_back(ch);
    _known_nodes[ch->get_peer()]=desc;
    newhost=true;
  }
  Khosts.unlock();
  if(newhost)
    set_default_route(ch);
  infmut.lock();
  _info.all_links.push_back(*ch->_link);
  //TCP_router::RefreshInfo(_info);
  //for(int i=0;i<_info.all_links.size();i++)
  //{
  //  std::cout<<"info: link "<<i<<" to node "<<_info.all_links[i]._peer<<(_info.all_links[i].open() ? " OPEN" : " CLOSED") <<std::endl;
  //}
  TCP_router::LoadInfo(_info);
  infmut.unlock();
  //std::cout<<"Router info reloaded"<<std::endl;
}

acom::Network::OutChannel* tcp_net::open_channel( acom::Network::LocalNodeNo peer )
       throw(acom::ComFailure)
{
  TCP_Channel* ch=NULL;
  InfoTCPnet::link* l=NULL;

  //std::cout<<"open channel to "<<peer<<std::endl;

  if((ch=_tcpm->connectto(peer))==NULL)
  {
    std::cerr<<"connection refused by node "<<peer<<std::endl;
    throw acom::ComFailure("connection to node refused");
  }
  
  //std::cout<<"channel's link to "<<peer<<" openned"<<std::endl;
  return ch;
}

void tcp_net::close_channel( acom::Network::OutChannel* ch )
       throw(acom::ComFailure) 
{
  // remove it of defaults route and close the socket
  if(ch->get_net()!=this)
    throw(acom::NetworkError("Bad Network"));
  defroute.lock();
   if(_default_routes.find(ch->get_peer())!=_default_routes.end())
    _default_routes.erase(ch->get_peer());
  defroute.unlock();
  ((TCP_Channel*)ch)->_link->close();
}

////////////////////////////////////////////////////////////////////////////
//class TCP_channel

void tcp_net::TCP_Channel::post( acom::Network::Request* r ) throw (acom::ComFailure)
{

  ACOM_ASSERT( r !=0, "[TCP::OutChannel::post] nul request");
  ACOM_ASSERT( get_net() !=0, "[TCP::OutChannel::post] nul network");
  TCP_Request* req = (TCP_Request*)r;

  acom::ComFailure::Code e = acom::ComFailure::FAIL;
  int i;
  size_t* iovsz ; // vector of size of deferred data blocks
    
  // close the sequences
    if (req->get_state() != ACOM::Network::S_POSTED)
      {
        req->set_state( ACOM::Network::S_POSTED);
        req->_datastream.packEos();
        req->_iovstream.putEos();
      }
    
  // complete the fixed part of header
  Header*      h=&req->_header;
   h->src=get_net()->rank();
  // dest is given by user
  h->dst=_peer;
  // time stamp
  h->no=0;
  h->size_h = req->_datastream.getFillSize();
  h->size_d  = req->_iovstream.getByteSize();
  // number of blocks of defered data
  h->iovsz_d = req->_iovstream.getBlockSize();
  /* we need to pack the vector containing the deferred block sizes
       alloc it on stack
  */
  //std::cout<<"NODE "<<h->src<<" POST A MESSAGE TO "<<h->dst<<std::endl;
  iovsz=(size_t*) alloca(sizeof(size_t)*(h->iovsz_d));
  i=0;
  for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
    ie!=req->_iovstream.endEntry();i++, ++ie){
      
      iovsz[i]=ie->len;
  }
    
    
    if(_peer!=InfoTCPnet::ALL)
      {
        try {
          //send this request
          _link->lock();

#if (SOCK_NODELAY_MODE==2)
          // do not send data immediately
          _link->_sock->setSockOpt(SOCK::Socket::no_delay, false);
#endif

          //the header:
          if(_link->_sock->writen(h,sizeof(acom::Network::Header))<0)
            throw acom::ComFailure(e);
          
          //the description for defered data
          if(_link->_sock->writen(iovsz,sizeof(size_t)*(h->iovsz_d))<0)
            throw acom::ComFailure(e);
          
          //the imediate data
          for (acom::IOVectSeq::IteratorEntry ie = 
                 req->_datastream.beginEntry();
               ie!=req->_datastream.endEntry(); ++ie)
            {
              if(_link->_sock->writen(ie->data,ie->len)<0)
        throw acom::ComFailure(e);
            }
          
          //the defered data
          for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
              ie!=req->_iovstream.endEntry(); ++ie)
            {
              if(_link->_sock->writen(ie->data,ie->len)<0)
        throw acom::ComFailure(e);
            }
        } catch (const ACOM::ComFailure& e ) {
          _link->unlock();
          std::cerr<<"PB pdt le post"<<std::endl;
          throw e;
        }

#if (SOCK_NODELAY_MODE==2)
        // flush data
        _link->_sock->setSockOpt(SOCK::Socket::no_delay, true);
#endif
    _link->unlock();
      }
    // notify request completion
    req->notify();
}

} //end namespace
