#ifndef __TCP_DNS_H
#define __TCP_DNS_Hi

#include<sock>
#include<acom>
#include<pth>
#include<map>



class TCP_dns
{
	public:
		typedef unsigned short uint16; 
	private:
	class send_addr : public acom::ServiceAM
	{
		public:
		void init(TCP_dns* d)
		{
			_dns =d;
		}
		void call(
					acom::ComFailure::Code error_no,
					acom::Network::Receipt* receipt,
					acom::Network::Header* header,
					acom::IOVectEntry* vector,
					acom::IOVectSize count,
          void* ctxt
					) throw(acom::ComFailure)
		{	
			if(error_no != acom::ComFailure::OK)
				return;
			SOCK::Address addr;
			std::string host;
			int port;
			host= std::string ((char*) vector[0].data,vector[0].len);
			port= *(uint16*)vector[1].data;
			addr.setByName(host);
			addr.setPort(port);
			_dns->setaddr(header->src,addr);
		}
		private:
		TCP_dns* _dns;
	};

	class request_addr : public acom::ServiceAM
	{
		public:
		void init(TCP_dns* d)
		{
			_dns =d;
		}
		void call(
					acom::ComFailure::Code error_no,
					acom::Network::Receipt* receipt,
					acom::Network::Header* header,
					acom::IOVectEntry* vector,
					acom::IOVectSize count,
          void* ctxt
					) throw(acom::ComFailure)
		{	
			std::string host;
			uint16 port;
			if(error_no != acom::ComFailure::OK)
				return;
			//dbg.log("", " DNS REQUEST RECEIVED \n");
			host=_dns->_myaddr.getHostName();
			port=_dns->_myaddr.getPort();
			//dbg.log("", " DNS REQUEST before get default route ...... \n");
			acom::Network::OutChannel* dest= _dns->_net->get_default_route(header->src);
			acom::ActiveMessage<send_addr> am (_dns->_net, &_dns->_sendaddr, 0);
			am->put((void*)host.c_str(),host.size());
			am->put((void*)&port,sizeof(uint16));
			//dbg.log("", " DNS REQUEST send ...... \n");
			am.post(dest);
			//dbg.log("", " DNS REQUEST proceed send OK \n");
		}
		private:
			TCP_dns *_dns;
	};


	public:
		TCP_dns()
		{
			PTH::Pth::initialize();
			PTH::Pth::commit();
			_mapmut.init();
			_waitcond.init();
			dbg.log("", "Registering service DNS \n");
			acom::Service::Manager.bind(acom::ServiceManager::DNS_LNS,1,(acom::Service*)&_sendaddr);
			acom::Service::Manager.bind(acom::ServiceManager::DNS_LNS,2,(acom::Service*)&_reqaddr);
			dbg.log("", "Service DNS registered\n");
		}
		~TCP_dns()
		{
			_mapmut.destroy();
			_waitcond.destroy();
		}

		void initialize(acom::Network* net,SOCK::Address& myaddr)
		{
			_net=net;
			_myaddr=myaddr;
			_reqaddr.init(this);
			_sendaddr.init(this);
		}
		
		SOCK::Address get_addr(acom::Network::LocalNodeNo nid)
		{
			SOCK::Address addr;
			_mapmut.lock();
			bool once = true;
			while(_known_Addr.find(nid)==_known_Addr.end())
			{
				if(once)
				{
					getinfo(nid);
					once=false;
				}
				//dbg.log("", "DNS: before cond_wait \n");
				_waitcond.wait(_mapmut);
				//dbg.log("", "DNS: after cond_wait \n");
			}
			addr=_known_Addr[nid];
			_mapmut.unlock();
			return addr;
		}
	private:
		void getinfo(acom::Network::LocalNodeNo nid)
		{
			acom::Network::OutChannel* dest = _net->get_default_route(nid);
      acom::ActiveMessage<request_addr> am (_net, &_reqaddr, 0);
      am.post(dest);
			//dbg.log("", "DNS message sent \n");
		}
		void setaddr(acom::Network::LocalNodeNo n,SOCK::Address a)
		{
			//dbg.log("", "DNS waiting address \n");
			_mapmut.lock();
			_known_Addr[n]=a;
			_waitcond.broadcast();
			_mapmut.unlock();
			//dbg.log("", "DNS address OK \n");
		}
		std::map<acom::Network::LocalNodeNo, SOCK::Address> _known_Addr;
		PTH::Mutex _mapmut;
		PTH::Condition _waitcond;
		acom::Network* _net;
		SOCK::Address _myaddr;
		request_addr _reqaddr;
		send_addr _sendaddr;
		friend class request_addr;
		friend class send_addr;
};

#endif
