#include<TCP_manager.h>
#include<TCP_network.h>
#include<CoreTCP.h>
#include<iostream>



static void TCP_mgr_entry(TCPmanager * mgr) {
	mgr->run();
}
	
void TCPmanager::initialize(acom::Network* netbase,TCP_NET::tcp_net* net)
{
	_net=netbase;
	_mynet=net;
	_state.resize(_net->node_count());
	/*first bind the socket server*/
	SOCK::Address Myaddr= SOCK::Address::getLocalHost();
	Myaddr.setPort(0);
	
	if(_sockserv.bind(Myaddr)<0)
	{
		std::cerr<<"Inuktitut TCP network can't bind his primary socket"<<std::endl;
//		return;
	}
	if(_sockserv.listen(300)<0)
	{
		std::cerr<<"Inuktitut TCP network can't listen on his primary socket"<<std::endl;
//		return;
	}
#ifdef TCPNET_TRACE_DEBUG
  std::cout << "::my url=" << _sockserv.getURL() << std::endl;
#endif


	hostAddr=_sockserv.getAddress();
	hostIP=hostAddr.getHostName();
	hostPort=hostAddr.getPort();
//	hostAddr.print(std::cout);
//	_sockserv.print(std::cout);
	//initialize the dns object
	_dns.initialize(_net,hostAddr);
	// begin the thread which do the accept sequence
//	std::cout<<"tcpmanager start the connect thread"<<std::endl;
	//_sockserv.print(std::cout);
//	this->start();
	_TCP_mgr_thread.create(&_TCP_mgr_threadId,(void *(*)(void*))TCP_mgr_entry, this, PTH::Thread::JOINABLE);
}

void TCPmanager::terminate()
{
//	_sockserv.close(); // blocks on IRIX64
}

TCP_NET::tcp_net::TCP_Channel* TCPmanager::connectto(const TCP_NET::tcp_net::LocalNodeNo& node)
{
	int result=1;
	TCP_NET::tcp_net::TCP_Channel* ch=NULL;
	InfoTCPnet::link* l=NULL;
	char ack;
	TCP_NET::tcp_net::LocalNodeNo rank;
	//std::cout<<"open a new connection"<<std::endl;
	_state[node].DoOrWait();
	if(_state[node].get_value()==doing)
	{
		//std::cout<<"TRY to open a new channel"<<std::endl;
		//I'm first it's at me to create the link
		//std::cout<<"I'm the first i do it"<<std::endl;
		SOCK::SocketClient * s = new SOCK::SocketClient();
		s->setAddress(_dns.get_addr(node));

		//std::cout<<"i got the remote address"<<std::endl;
		//s->print(std::cout);
		if(s->connect()<0)
		{
			std::cerr<<"can't connect to node "<<node<<std::endl;
			delete(s);
			s=NULL;
			_state[node].SetLink(NULL);
		}
		else
		{
			//std::cout<<"Connect OK :-)"<<std::endl;
			//s->print(std::cout);

			//send my rank
			rank=_net->rank();
			//std::cout<<"i send my rank"<<std::endl;
			if(s->writen((void*)&rank,sizeof(TCP_NET::tcp_net::LocalNodeNo))<0)
			{
				std::cerr<<"can't send my rank to the remote host ABORT"<<std::endl;
				delete s;
				s=NULL;
				_state[node].SetLink(NULL);
			}else
			{

				//std::cout<<"I wait remote ACK"<<std::endl;
				if(s->readn((void*)&ack,1)<0)
				{
                  // 					std::cerr<<"can't receive ACK from remote host ABORT I lost the DUAL wait for remote connection"<<std::endl;
					delete s;
					s=NULL;
					// in this case remote WIN it will create the connection We must wait
					_state[node].DoOrWait();
				}
				else
				{
					//std::cout<<"I read remote ACK"<<std::endl;
					if(s!=NULL)
					{
#if (SOCK_NODELAY_MODE==1)
                      s->setSockOpt(SOCK::Socket::no_delay, true);
#endif
						l=InfoTCPnet::AllocateLink();
						l->init(s,node);
						l->_flag=InfoTCPnet::link::free;
						ch=new TCP_NET::tcp_net::TCP_Channel(l,_mynet);
						//std::cout<<"LOCAL ADD CHANNEL"<<std::endl;
						_mynet->AddNewChannel(ch);
						_state[node].SetLink(ch);
					}
				}
			}
		}
	}
	// set link have been call in every case for the case DUAL lost the thread
	// which accept the connection do it
	ch=_state[node].getchannel();
	return ch;
}


int TCPmanager::close(const TCP_NET::tcp_net::LocalNodeNo& node)
{
	//not yet implement
	return -1;
}

/*
int TCPmanager::connectall()
{
	for(int i=0;i<_net->node_count();i++)
	{
		connectto(i);
	}
}

int TCPmanager::closeall()
{
	//not yet implement
}

*/
void TCPmanager::run()
{	
	//std::cout<<"START the thread which accept connections"<<std::endl;
	//hostAddr.print(std::cout);
	//_sockserv.print(std::cout);
	while(_sockserv.valid())
//	while(true)
	{
		SOCK::SocketClient* _sock = new SOCK::SocketClient();
		//std::cout<<"I wait for accept a connection"<<std::endl;
		_sockserv.accept(*_sock);
		//std::cout<<"I have accepted a connection"<<std::endl;
		process(_sock);
	}
	//std::cout<<"STOP the thread which accept connections"<<std::endl;
}


void TCPmanager::process(SOCK::SocketClient* s)
{
	TCP_NET::tcp_net::LocalNodeNo Id;
	TCP_NET::tcp_net::TCP_Channel* ch=NULL;
	InfoTCPnet::link* l=NULL;
	state st;
		
	//std::cout<<"Process a connection request"<<std::endl;
	// read which node want to be connected to me
	if(s->readn((void*)&Id,sizeof(TCP_NET::tcp_net::LocalNodeNo))<=0)
		return;

	//std::cout<<"Process receive node ID= "<<Id<<std::endl;
	//node Id want to establish a connection with me :-)
	//verify connection not already done
	// If my rank is greater than the remote i win the DUAL if exist
	st=_state[Id].getset_value(doing);
	if((st!=doing && st!=connected) || (st==doing && _net->rank()<Id))
	{

		//std::cout<<"Process connection accepted"<<std::endl;
		//send an ack
#if (SOCK_NODELAY_MODE==1)
      s->setSockOpt(SOCK::Socket::no_delay, true);
#endif
		s->writen((void*)"O",1);
		
		l=InfoTCPnet::AllocateLink();
		l->init(s,Id);
		l->_flag=InfoTCPnet::link::free;
		ch=new TCP_NET::tcp_net::TCP_Channel(l,_mynet);
		//std::cout<<"REMOTE ADD CHANNEL"<<std::endl;
		_mynet->AddNewChannel(new TCP_NET::tcp_net::TCP_Channel(l,_mynet));
		_state[Id].SetLink(ch);
		
	}
	else
	{
		//std::cout<<"Process connection refused"<<std::endl;
		//already done close and delete this new connection I WIN THE FATAL DUAL 
		s->close();
	}
	
}

