// Authors: Mareschal arnaud, Menetrieux Marie-Laure
#include <TCP_network.h>
#include <TCP_manager.h>




namespace TCP_NET{

  // --------------------------------------------------------------------------
  //
  // Constructor
  //
  tcp_net::tcp_net():acom::Network(),TCP_router(this)
  {
    initAllocators();
     _netBase = NULL;
    _dynamic = false;
    _is_init = false;
    _nb_init = 0;
    _is_initcommit = false;
    _is_terminate = false;
    //_is_listen = false;
    set_property("name","TCPNET");
    set_name("TCPNET");
    set_property("primary","false");
    set_property("secondary","true");
    PTH::Pth::initialize();
    PTH::Pth::commit();
    defroute.init();
    infmut.init();
    Khosts.init();
    _tcpm=new TCPmanager();
   
  }

  tcp_net::~tcp_net()
  {
    _is_init = false;
    _nb_init = 0;
    _is_initcommit = false;
    _is_terminate = false;
    defroute.destroy();
    Khosts.destroy();
    infmut.destroy();
    delete _tcpm;
  }

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
  int tcp_net::initialize( int* argc, char** argv[], acom::Network* netInit,bool dyn )
      throw (acom::NetworkError){
      
      // to avoid multiple call of this method
      _nb_init++;
      if   (_is_init) 
        return 0;
      _is_init = true;
      _dynamic = dyn;
      _netBase = netInit;

      _primary_net = netInit;

      // create & init barrier object
      ICS::Ics::initialize(argc, argv);
      ICS::Ics::commit();
      _barrier = new ICS::FlatBarrier;
      ACOM::Service::Manager.bind( ACOM::ServiceManager::TCP_LNS,1, _barrier );
      _barrier->initialize( this );
      //      _barrier->initialize( _primary_net );

      return 1;
   }

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
   int tcp_net::initialize( int* argc, char** argv[], acom::Network* netInit )
     throw (acom::NetworkError){
      return initialize(argc,argv,netInit,true);
   }



 /** Second stage of the initialisation process.
      End of the first stage of initialisation : all local initialisations
      are done, communication could occurs after the second stage (after
      the call to initcommit) or during the call of initcommit in an other
      module.
  */
  int tcp_net::commit ()throw (acom::NetworkError) 
  {
    if(_is_initcommit)
      return 0;
    _is_initcommit = true;
    _node_rank = _netBase->rank();
    _node_count = _netBase->node_count();
    _info.rank = _node_rank;
    _info.count_nodes = _node_count;
    _tcpm->initialize(_netBase,this);
    //connection_manager initialization here!!
    
    TCP_router::LoadInfo(info);
    TCP_router::commit();
    if ((_netBase != NULL)&&(!_dynamic)) {
      //on effectue une initialisation secondaire
      //connection_manager.connectAll();
      
    }
    //demarrage de l'ecoute des channels
    TCP_router::NET_on();
    _primary_net->barrier();
    return 0;
  }

  /** Terminate the net...
  */
  int tcp_net::terminate ()throw (acom::NetworkError) {


    _nb_init--;
    if ((_nb_init > 0)||(_is_terminate))return 0;

    
    _is_terminate = true;
    //on libere l'objet point par grouter
    //_net_barrier->terminate();
    _tcpm->terminate();
    TCP_router::terminate();
    //-> fermeture propre des channels (sockets)
    /*for(int i = 0; i < _node_count; i++){
      if (_default_routes[i] !=NULL){
        close_channel(_default_routes[(LocalNodeNo)i]);
      }
    }
    */
    return 0;
  };

 /** neighbour_cast to nodes (AM,AWS,WS protocols)
      @param   req  request (IN)
      @param   ch  channel  (IN)
  */

  void tcp_net::neighbour_cast( Network::Request* req ,OutChannel* ch)
       throw (acom::ComFailure) 
  {

    if(req){
      if(req->get_net()!=this)
        throw acom::ComFailure(acom::ComFailure::BAD_REQUEST)  ;
      else{
        Header*   h=&req->_header;
        h-> src= _node_rank;
        defroute.lock();
        std::map<LocalNodeNo, OutChannel*>::iterator cur  = _default_routes.begin();
        std::map<LocalNodeNo, OutChannel*>::iterator end  = _default_routes.end();
        while ((*cur).first != (*end).first)
        {
          LocalNodeNo distantNode = (*cur).first;
          if (distantNode != _node_rank)
          {
            //send the data
            h->dst = distantNode;
            OutChannel* distantChannel = (*cur).second;
            if (distantChannel != NULL){
              (*distantChannel).post(req);
            }
          }
          cur++;
        }
        defroute.unlock();
        return;

      }
    }
    else throw acom::ComFailure(acom::ComFailure::BAD_REQUEST)  ;
  };

 /** send  to a node (AM,AWS,WS protocols)
      @param   req  request (IN)
       @param   ch  channel  (IN)
  */
void  tcp_net::post( Network::Request* req ,OutChannel* ch)
       throw (acom::ComFailure) {
  if(req){
    if(req->get_net()!=this)
      throw acom::ComFailure(acom::ComFailure::BAD_REQUEST)  ;
    else{
      ch->post(req);
      return;
    }
  }
  else throw acom::ComFailure(acom::ComFailure::BAD_REQUEST)  ;
};

// ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */
  void tcp_net::barrier() throw (acom::ComFailure) {    
    static int counter = 0;
    // use ics barrier
    ++counter;
    _barrier->barrier(counter); 
    //             _net_barrier->sync();
  };








} //End of namespace TCP

extern "C" {

  static TCP_NET::tcp_net* tcpnet =0;

  // - dynamic library initialization:
  static void factory_initialize() 
  {
    tcpnet = new TCP_NET::tcp_net();
    acom::Acom::bind_network("TCPNET", tcpnet );
  }
  extern void TCPNET_factory_initialize() 
  {
    //dbg.log("","factory init\n");
    factory_initialize();
    //dbg.log("","factory init OK\n");
  }
  
  // - dynamic library terminaison:
  static void factory_destroy() 
  {
    delete tcpnet;
  }
  extern void TCPNET_factory_destroy() 
  {
    return factory_destroy();
  }
}

