// Authors: Mareschal arnaud, Menetrieux Marie-Laure, Cyrille Martin
#ifndef __TCP_NET_H
#define __TCP_NET_H

#include <acom>
#include <sock>
#include <map>
#include <string>
#include <TCP_router.h>
#include <ics> // for barrier


class TCPmanager;

namespace TCP_NET{

class tcp_net : public acom::Network, TCP_router
{
 // ---------------------------------------------------------------------------
  // Network interface
  // ---------------------------------------------------------------------------

public :
// --------------------------------------------------------------------------
  //
  // Constructor
  //
  tcp_net();
  ~tcp_net();




  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
   int initialize( int* argc, char** argv[] )throw (acom::NetworkError)
   {
      throw acom::NetworkError("TCP_NET is a secondary network you must specify a primary network\n");
      return -1;
   }

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
   int initialize( int* argc, char** argv[], acom::Network* netInit )throw (acom::NetworkError);

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** First stage initialisation from commandline arguments.
      All recognized arguments are deleted from the command line,
      except of --help.
  */
   int initialize( int* argc, char** argv[], acom::Network* netInit ,bool dyn)throw (acom::NetworkError);

  /** Second stage of the initialisation process.
      End of the first stage of initialisation : all local initialisations
      are done, communication could occurs after the second stage (after
      the call to initcommit) or during the call of initcommit in an other
      module.
  */
  int commit ()throw (acom::NetworkError) ;

  /** Terminate the net...
  */
  int terminate ()throw (acom::NetworkError) ;

public:

  /** the only channel class
  */
  class TCP_Channel: public acom::Network::OutChannel 
  {
    public :
    /** Constructor
              */
      TCP_Channel(InfoTCPnet::link* l, Network*  net)
      {
        _link = l;
        _net = net;
        _peer = _link->_peer;
      };
      ~TCP_Channel()
      {
        _link->lock();
        _link->_sock->close();
        _link->unlock();
      }

    void post( acom::Network::Request* r ) throw (acom::ComFailure);

    InfoTCPnet::link* _link;
  };


protected:

   /** iovect entries in a cell
   */
  static const int _nb_entries=10; // iovect entries in a cell
  /** packed block size
   */
  static const int _block_sz = 1500; // packed block size

  /** Logical clock for timestamping request
   */
  PTH::AtomicCounter _comClock;

  /** size of cached free blocks
   *     */
  static const int _block_threshold=5; // size of cached free blocks
  /** size of cached free blocks
  *     */
  static const int _iovect_threshold=5; // size of cached free blocks
  /** size of cached free blocks
  *     */
  static const int _request_threshold=5; // size of cached free blocks


  acom::ByteCellAllocator _bA;
  /** iovec allocator
   */
  acom::IOVectCellAllocator _iovA;

  void initAllocators(){
      _comClock=0;
     _reqA.init(_request_threshold);
     _bA.init(_block_sz, _block_threshold  );
     _iovA.init(sizeof(acom::IOVectCell)+(_nb_entries*sizeof(acom::IOVectEntry)) ,_iovect_threshold    );
     };




  class TCP_Request : public acom::Network::Request, public PTH::Linkable<TCP_Request>
    {
      protected :
        bool init(size_t bsz, acom::ByteCellAllocator* bufa,
                          int iovsz, acom::IOVectCellAllocator * iova,
                          Network * manager)
        {
          return acom::Network::Request::init(bsz,bufa,iovsz, iova,manager);
        }

        friend class TCP_Channel;
        friend class tcp_net;
        friend class TCP_router;
    };
  protected:
    class TCP_Receipt : public acom::Network::Receipt
    {
      protected :
      bool init( acom::IOVectSeq * iovs , acom::IOVectEntry * iov, size_t sz,
                  acom::Network* net)
      {return acom::Network::Receipt::init(iovs,iov , sz,net);}
      friend class TCP_Channel;
      friend class tcp_net;
      friend class TCP_router;
    };

  /** request allocator
   */

  typedef PTH::SyncCellAllocator<TCP_Request> RequestAllocator ;
  RequestAllocator _reqA;



  public:



  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

  /** Allocate a new request.
  */
   Request* allocate_request () throw (acom::BadAlloc) {
    TCP_Request* req = _reqA.alloc() ;
          if(!req) throw acom::BadAlloc("[[TCP_net::allocate_request ::]] fail") ;
      if(!req->init( _block_sz,& _bA,_nb_entries ,& _iovA,(Network*)this ))
        {_reqA.free(req);throw acom::BadAlloc("[[TCP_net::allocate_request ::]] fail")  ;}
                                  return req;
  }

  /** free request descriptor when unused
  */
   void deallocate_request( Request* rq ) throw (acom::InvalidArgument) {
    TCP_Request* req = (TCP_Request*)rq;
    if(req){
      if(req->get_net()!=this)
        throw acom::InvalidArgument("[[TCP_net::deallocate_request ::]] not aloopback request")  ;
      else
        _reqA.free (req); return ;
      }
    else throw acom::InvalidArgument("[[TCP_net::deallocate_request ::]] null request")  ;
  };

  /** Cancel a request allocated by allocate_request ( true == cancel done)
      Cancelation of request depend on state of progress
      @param req (IN)
  */
  bool cancel_request( Request* req )
    throw (acom::InvalidArgument)
  { return false; }


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Allocate  an immediate buffer of sz bytes in network memory
      @param  sz (IN) in bytes
  */
   void* allocate_data( size_t sz )
       throw (acom::BadAlloc) { return ::malloc(sz);};

  /** Deallocate a previously allocated immediate buffer  in network memory
      @param  data (IN) byte block to deallocate
  */
   void deallocate_data(  void* data )
       throw (acom::InvalidArgument) { ::free(data);};


  // --------------------------------------------------------------------------
  ACOM::Network::CallContext* allocate_callcontext(size_t sz) throw (ACOM::BadAlloc)
  {
    ACOM::Network::CallContext* cc = new ACOM::Network::CallContext;
    cc->init( this, sz );
    return cc;
  }
  
  void deallocate_callcontext(ACOM::Network::CallContext* cc) throw (ACOM::InvalidArgument)
  {
    cc->destroy( );
    delete cc;
  }

  ACOM::Network::CallContext* resize_callcontext(ACOM::Network::CallContext* cc, size_t sz ) throw (ACOM::BadAlloc)
  {
    if (cc->_stack_size > sz) return cc;
    size_t offset = cc->size();
    ACOM::Network::CallContext*  new_cc = allocate_callcontext(sz+128);
    memcpy( new_cc->_base, cc->_base, offset );
    new_cc->_top = ((char*)new_cc->_base) + offset;
    deallocate_callcontext( cc );
    return new_cc;
  }

  /** plug the current node to a  network address
       @param   nodelabel(IN) network label of remote node
      \warning ajouter les aliasing d'adresse de noeud struct  ( A voir  partir
 IP/port, url corba ..)

  */
   void bind_node( std::string nodelabel   )
       throw(acom::NetworkError) {
       throw acom::NetworkError("[[TCP_net::bind_node ::]]  not yet implmented");

    };



  /** add a new node to the network
    @param  nodename(IN) network identification of remote node
    @return local node number of the new node
  \warning MUST fail when multiple add of same addreses

   \warning ajouter les aliasing d'adresse de noeud   ( A voir  partir fd,
 IP/port, url corba ..)
  */
   void  add_node( std::string nodelabel , LocalNodeNo  peer )
       throw(acom::NetworkError) {
       throw acom::NetworkError("[[TCP_net::add_node ::]]  not yet implmented");


   };


  /** del a previously added node to the network
      @param  peer  local node Number of remote node to disconnect
      \warning ajouter les aliasing d'adresse de noeud     ( A voir  partir
 IP/port, url corba ..)
  */
  void  del_node( LocalNodeNo  peer  )
      throw(acom::NetworkError) {
                throw acom::NetworkError("[[TCP_net::del_node ::]]  not yet implemented");


  } ;



  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Return the route (channel) to send data to LocalNodeNo peer
      @param   peer(IN) remote node id
  */
   OutChannel* get_default_route( LocalNodeNo peer )
       throw(acom::NetworkError) ;

/** Set a channel as a default channel to a destination node
      @param   ch default channel
  */
   void set_default_route(  OutChannel* ch  )
       throw(acom::NetworkError) ;


  /** Open a channel between local node  to remote peer
      @param   peer(IN) remote node id
      \warning ajouter les aliasing d'adresse de noeud
  */
  OutChannel* open_channel( LocalNodeNo peer )
       throw(acom::ComFailure);

  /** Close a channel between local node  to remote peer
      @param   ch (IN)channel to close
      \warning ajouter les aliasing d'adresse de noeud
  */
   void close_channel( OutChannel* ch )
       throw(acom::ComFailure);


  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------

/** broadcast to nodes (AM,AWS,WS protocols)
      @param   req  request (IN)
      @param   ch  channel  (IN)
  */
  void broadcast( Network::Request* req, OutChannel* ch)
        throw (acom::ComFailure){
                neighbour_cast(req,ch);

        };

  /** neighbour_cast to nodes (AM,AWS,WS protocols)
      @param   req  request (IN)
      @param   ch  channel  (IN)
  */
  void neighbour_cast( Network::Request* req ,OutChannel* ch)
       throw (acom::ComFailure) ;

/** send  to a node (AM,AWS,WS protocols)
      @param   req  request (IN)
       @param   ch  channel  (IN)
  */
void  post( Network::Request* req ,OutChannel* ch)
       throw (acom::ComFailure) ;

  // ---------------------------------------------------------------------------
  // ---------------------------------------------------------------------------
  /** Network specific barrier
      \warning  parameters count, time out etc ...( voir avec Cyrille)
  */
 
   void barrier()
       throw (acom::ComFailure) ;

  /** Get the name of the Network.
  */
  //std::string get_name( ) const {return _net_name;};

  /** Set the name of the Network.
  */
  //std::string set_name(std::string name )  {
  //         _net_name=name;
  //};


  protected:

    bool _is_init;        // to avoid multiple calls to the associated
    int _nb_init;         // number of calls
    bool _is_initcommit;  // methods
    bool _is_terminate;
    bool _dynamic;
    InfoTCPnet _info;
    PTH::Mutex infmut;
    Network * _netBase;
    TCPmanager* _tcpm;
    /** synchronized access to _default_routes
    */
    PTH::Mutex defroute;
    PTH::Mutex Khosts;
    void AddNewChannel(TCP_Channel* ch);
    friend class TCPmanager;
    TCP_Channel* getroute(LocalNodeNo node);

    friend class TCP_router;

 private:
    ICS::BarrierService* _barrier;
    ACOM::Network*       _primary_net;


}; // -- end class Network

} // End of namespace

#endif //end of define _ACOM_TCP_H
