// Authors: Mareschal arnaud, Menetrieux Marie-Laure
/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 *
 *
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include <barrier.h>
#include <sys/types.h>

namespace BARRIER_NAMESPACE {




Barrier::Barrier(acom::Network* net_)
{
        net=net_;

}

int Barrier::initialize() throw (acom::NetworkError){

       _barrier_state = 1;

        PTH::Pth::initialize();
	PTH::Pth::commit();

        mutex.initnothrow();
        cond.initnothrow();
        ServiceBarrier = new ServBarrier(this);
        ServiceAckBarrier= new ServAckBarrier(this);

        acom::Service::Manager.bind (acom::ServiceManager::TCP_LNS,2, (acom::Service *) ServiceBarrier);
        acom::Service::Manager.bind (acom::ServiceManager::TCP_LNS,3, (acom::Service *) ServiceAckBarrier);

        _count_barrier = net->node_count();

}

int Barrier::commit()throw (acom::NetworkError){

}

int Barrier::terminate()throw (acom::NetworkError){
        PTH::Pth::terminate();
        try{
                mutex.destroy();
                cond.destroy();
        }catch(...){;}

        delete (ServiceBarrier);
        delete (ServiceAckBarrier);

}
Barrier::~Barrier()
{
;
}


int Barrier::sync()
{

        if (net->rank()!=0)
        {

                 // send an AM to the node 0
                acom::ActiveMessage<ServBarrier> am (net,ServiceBarrier,0);
                acom::Network::OutChannel* tch = net->get_default_route(0);
                am(tch);


                mutex.locknothrow();
                while (_barrier_state != 0)
                        cond.waitnothrow(mutex);
                _barrier_state = 1;
		mutex.unlocknothrow();
        }
        //else wait the AMs Barrier of the others nodes
        else
        {
        	_count_barrier--;

	        if (_count_barrier == 0)
        	{
		   	 sendack();
           	}
		else
		{
               		mutex.locknothrow();
     			while (_barrier_state != 0)
                       		 cond.waitnothrow(mutex);
			//il faut reinitialiser l'etat pour la barriere suivante
			_barrier_state = 1;
     			mutex.unlocknothrow();
		}

        }



}


void Barrier::waitall()
{

        _count_barrier--;
        if (_count_barrier == 0)
        {

                 sendack();
           }

 }


void Barrier::receipt_ack(){
  // getting info from the children

  mutex.locknothrow();
  _barrier_state = 0;
  cond.signalnothrow();
  mutex.unlocknothrow();

}

void Barrier::sendack()
{


        for (int i =1; i< net->node_count(); i++){
                acom::ActiveMessage<ServAckBarrier> am (net,ServiceAckBarrier,0);
                acom::Network::OutChannel* tch = (acom::Network::OutChannel *)net->get_default_route((acom::Network::LocalNodeNo) i);
                am(tch);

          }
          mutex.locknothrow();
          _barrier_state = 0;
          cond.signalnothrow();
          mutex.unlocknothrow();
          _count_barrier = net->node_count();
}


};/*end namespace BARRIER_NAMESPACE */
