#include <taktukconf.h>
#include <taktuknetwork.h>
#include <CoreTCP.h>
#include <acom>
#ifdef INUKTITUT_USE_DARWIN
#include <stdlib.h>
#else
#include <alloca.h>
#endif

  
namespace Taktuk
{
void Network::TaktukLocalChannel::post( acom::Network::Request* r )
    throw(acom::ComFailure)
{
  
    ACOM_ASSERT( r !=0, "[Taktuk::OutChannel::post] nul request");
    ACOM_ASSERT( get_net() !=0, "[Taktuk::OutChannel::post] nul network");
    TaktukRequest* req = (TaktukRequest*)r;

    dbg.log("post","local begin post operation\n");
    InfoTCPnet::link* _link;
    _link=&(_info->all_links[_peer]);

    dbg.log("post","local post a message to local node %d info* = %x\n",_peer,_info);
    dbg.log("post","local build the header\n");
          
    acom::ComFailure::Code e = acom::ComFailure::FAIL;
    int i;
    size_t* iovsz ; // vector of size of deferred data blocks
    // close the sequences 
    dbg.log("post","close the sequence\n");
    // WARN: TG + OL
    if (req->get_state() != ACOM::Network::S_POSTED)
    {
      req->set_state( ACOM::Network::S_POSTED );
      req->_datastream.packEos();
      req->_iovstream.putEos();
    }
    //req->_datastream.packEos();
    //req->_iovstream.putEos();
    // complete the fixed part of header 
    Header*      h=&req->_header;
    h->src=~acom::Network::LocalNodeNo(0);
    dbg.log("post","set src=%d\n", h->src);
    // dest is given by user
    h->dst=InfoTCPnet::YOU;
    dbg.log("post","set dst=%d\n",h->dst);
    // time stamp  for tracing 
    //    h->no=_net->_comClock; 
    //    ++_net->_comClock;
    dbg.log("post","set stamp clock%d\n",h->no);
    h->size_h = req->_datastream.getFillSize();
    dbg.log("post","set size_h=%d\n",h->size_h);
    h->size_d  = req->_iovstream.getByteSize();
    dbg.log("post","set size_d=%d\n",h->size_d);
    // number of blocks of defered data
    h->iovsz_d = req->_iovstream.getBlockSize();
    dbg.log("post","set iovsz_d=%d\n",h->iovsz_d);
    /* we need to pack the vector containing the deferred block sizes
       alloc it on stack 
    */

    dbg.log("post","build header done\n");
    dbg.log("post","post a message to header dst no %d\n",h->dst);
    if(h->iovsz_d>0)
      iovsz=(size_t*) alloca(sizeof(size_t)*(h->iovsz_d));
    // iterate on the iovect sequence to get block size
    i=0;
    for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();                 
      									ie!=req->_iovstream.endEntry();i++, ++ie){
      iovsz[i]=ie->len; 
      dbg.log("post","build defered data description no %d = %d\n",i,ie->len);
    }
    
    //send this request
    _link->lock();
    //the header:
    if(_link->_sock->writen(h,sizeof(acom::Network::Header))!=sizeof(acom::Network::Header))
    {
      dbg.log("post","SEND ERROR\n");
      _link->unlock();
      throw acom::ComFailure(e);
    }
        
    dbg.log("post","header sent correctly\n");
    //the description for defered data
    if(_link->_sock->writen(iovsz,sizeof(size_t)*(h->iovsz_d))!=sizeof(size_t)*(h->iovsz_d))
    {
       dbg.log("post","SEND ERROR\n");
       _link->unlock();
       throw acom::ComFailure(e);
    }
        
    //the imediate data
    for(acom::IOVectSeq::IteratorEntry ie= req->_datastream.beginEntry();
            ie!=req->_datastream.endEntry(); ++ie)
    {
      if(_link->_sock->writen(ie->data,ie->len)!=ie->len)
      {
        dbg.log("post","SEND ERROR\n");
        _link->unlock();
        throw acom::ComFailure(e);
      }
    }
      
    //the defered data
    for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
            ie!=req->_iovstream.endEntry(); ++ie)
    {
      dbg.log("post","send defered data block size = %d\n",ie->len);
      dbg.Xmem(ie->data,ie->len);
      if(_link->_sock->writen(ie->data,ie->len)!=ie->len)
      {
         dbg.log("post","SEND ERROR\n");
         _link->unlock();
         throw acom::ComFailure(e);
    	}
    }
        
		_link->unlock();
       
    dbg.log("post","local post completed notify the completion\n");

    // notify request completion
    req->notify();
    dbg.log("post","local post completed \n");
}
	
void Network::TaktukChannel::post( acom::Network::Request* r )
    throw(acom::ComFailure)
{
#ifdef TAKTUK_USER_TRACER
  TAK_POST_0();
#endif
  
    ACOM_ASSERT( r !=0, "[Taktuk::OutChannel::post] nul request");
    ACOM_ASSERT( get_net() !=0, "[Taktuk::OutChannel::post] nul network");
    TaktukRequest* req = (TaktukRequest*)r;

    dbg.log("post","begin post operation\n");
    InfoTCPnet::link* _link;
    if(_peer!=InfoTCPnet::ALL)
      _link=&(_info->all_links[_info->route_table[_peer]]);

        dbg.log("post","post a message to node %d info* = %x\n",_peer,_info);
        if(_peer!=InfoTCPnet::ALL)
                dbg.log("post","route table said %d\n",_info->route_table[_peer]);
        dbg.log("post","build the header\n");
          
    acom::ComFailure::Code e = acom::ComFailure::FAIL;
    int i;
    size_t* iovsz ; // vector of size of deferred data blocks
    // close the sequences 
        dbg.log("post","close the sequence\n");
    req->_datastream.packEos();
    req->_iovstream.putEos();
    // complete the fixed part of header 
    Header*      h=&req->_header;
    h->src=get_net()->rank();
        dbg.log("post","set src=%d\n", h->src);
    // dest is given by user
        h->dst=_peer;
        dbg.log("post","set dst=%d\n",h->dst);
    // time stamp  for tracing
	//    h->no=_net->_comClock; 
	//    ++_net->_comClock;
        dbg.log("post","set stamp clock%d\n",h->no);
    h->size_h = req->_datastream.getFillSize();
        dbg.log("post","set size_h=%d\n",h->size_h);
    h->size_d  = req->_iovstream.getByteSize();
        dbg.log("post","set size_d=%d\n",h->size_d);
    // number of blocks of defered data
    h->iovsz_d = req->_iovstream.getBlockSize();
        dbg.log("post","set iovsz_d=%d\n",h->iovsz_d);
    /* we need to pack the vector containing the deferred block sizes
       alloc it on stack 
    */

        dbg.log("post","build header done\n");
        dbg.log("post","post a message to header dst no %d\n",h->dst);
        if(h->iovsz_d>0)
      //iovsz=(size_t*) malloc(sizeof(size_t)*(h->iovsz_d));
      iovsz=(size_t*) alloca(sizeof(size_t)*(h->iovsz_d));
    // iterate on the iovect sequence to get block size
    i=0;
    for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();                 
        ie!=req->_iovstream.endEntry();i++, ++ie){
      iovsz[i]=ie->len; 
          dbg.log("post","build defered data description no %d = %d\n",i,ie->len);
    }
 
#ifdef TAKTUK_USE_TRACER
    // dbg.log("tracer","point trace cyril post1 %d %d \n",(unsigned int)h->src , (unsigned int)h->dst);
    TAK_POST_1( (unsigned int)h->src , (unsigned int)h->dst, h->no);
#endif
   
    if(_peer!=InfoTCPnet::ALL)
      {
        //send this request
        _link->lock();
        //the header:
//MODIF
				_link->_sock->setSockOpt(SOCK::Socket::no_delay, true);
        if(_link->_sock->writen(h,sizeof(acom::Network::Header))!= sizeof(acom::Network::Header) )
          {
                  dbg.log("post","SEND ERROR\n");
            _link->unlock();
            throw acom::ComFailure(e);
          }
        
        dbg.log("post","header sent correctly\n");
        //the description for defered data
        if(_link->_sock->writen(iovsz,sizeof(size_t)*(h->iovsz_d)) != sizeof(size_t)*(h->iovsz_d))
          {
                  dbg.log("post","SEND ERROR\n");
            _link->unlock();
            throw acom::ComFailure(e);
          }
        
        //the imediate data
        for(acom::IOVectSeq::IteratorEntry ie= req->_datastream.beginEntry();
            ie!=req->_datastream.endEntry(); ++ie)
          {
            if(_link->_sock->writen(ie->data,ie->len)!= ie->len)
              {
                  dbg.log("post","SEND ERROR\n");
                _link->unlock();
                throw acom::ComFailure(e);
              }
          }
      
        //the defered data
        for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
            ie!=req->_iovstream.endEntry(); ++ie)
          {
                  dbg.log("post","send defered data block size = %d\n",ie->len);
                dbg.Xmem(ie->data,ie->len);
            if(_link->_sock->writen(ie->data,ie->len)!=ie->len)
              {
                  dbg.log("post","SEND ERROR\n");
                _link->unlock();
                throw acom::ComFailure(e);
              }
          }
        
        _link->unlock();
      }
    else
      {//it's for all
          dbg.log("post","post a message to all\n");
        for(int i=0;i<_info->all_links.size();i++)
          {
            _link=&_info->all_links[i];
            
            //send this request
            _link->lock();
            //the header:
//MODIF
						_link->_sock->setSockOpt(SOCK::Socket::no_delay, true);
            if(_link->_sock->writen(h,sizeof(acom::Network::Header))!=sizeof(acom::Network::Header))
              {
                _link->unlock();
                throw acom::ComFailure(e);
              }
            dbg.log("post","post a message to all header OK\n");
            
            //the description for defered data
            if(_link->_sock->writen(iovsz,sizeof(size_t)*(h->iovsz_d))!=sizeof(size_t)*(h->iovsz_d))
              {
                _link->unlock();
                throw acom::ComFailure(e);
              }
            dbg.log("post","post a message to all def data OK\n");
            
            //the imediate data
            for(acom::IOVectSeq::IteratorEntry ie= req->_datastream.beginEntry();
                ie!=req->_datastream.endEntry(); ++ie)
              {
                if(_link->_sock->writen(ie->data,ie->len)!=ie->len)
                  {
                    _link->unlock();
                    throw acom::ComFailure(e);
                  }
              }
            
            dbg.log("post","post a message to all im data OK\n");
            //the defered data
            for(acom::IOVectSeq::IteratorEntry ie= req->_iovstream.beginEntry();
                ie!=req->_iovstream.endEntry(); ++ie)
              {
                if(_link->_sock->writen(ie->data,ie->len)!=ie->len)
                  {
                    _link->unlock();
                    throw acom::ComFailure(e);
                  }
              }
             dbg.log("post","post a message to all defered data OK\n");
            
            _link->unlock();
          }
      }
   
    dbg.log("post","post completed notify the completion\n");

/*
        if(h->iovsz_d>0)
      ::free(iovsz); 
*/  
    // notify request completion
    req->notify();
    dbg.log("post","post completed \n");
}


} //namespace ...
