/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#include <string>
#include <vector>
#include <pwd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <sys/wait.h>
#include <fcntl.h>

#include <acom>

#include <taktuknetwork.h>
#include <topology_analyser.h>
#include <connector.h>
#include <IOredir.h>
#include <spawner.h>
#include <taktukrouter.h>
#include <debuglog.h>
#include <barrier.h>
#include <taktukinstall.h>
#include <sock>


namespace Taktuk
{

	static char* prog_name;
	static void attach_dbg()
	{
		// -- Spawn a process dbx with the program name and the pid of the
		// current process.
		// - fork a child on which the debugger run ?
		pid_t pid = getpid();
		if (pid !=0) {
			char cmd_line[512];
			sprintf(cmd_line, "%s/taktukdbx %s %i\n", TAKTUKBIN,prog_name, getpid() );
			system(cmd_line);
		}
	}





  Network::Network():acom::Network(),taktuk_router(this)
  {
		PTH::Pth::initialize();
		PTH::Pth::commit();
		_barrier=new Barrier();
		Spawner=new spawner();
		_ioredir=NULL;
		initAllocators();
    resuming=false;
		_AMWS=false;
		_IOREDIR=false;
		set_property("name","TAKTUK");
		set_name("TAKTUK");
		set_property("primary","true");
		set_property("secondary","false");
  }
  Network::~Network()
  {
		delete Spawner;
		delete _barrier;
		if(_ioredir!=NULL)
			delete _ioredir;
		PTH::Pth::terminate();
  }

  int Network::initialize(int* argc, char** argv[]) throw (acom::NetworkError)
  {
		prog_name=*argv[0];
#ifdef DEBUGER
		dbg.log("init","TAKTUK launch debugger\n");
		attach_dbg();
		dbg.log("init","TAKTUK debugger launched\n");
#endif		
    return initialize(argc,argv,true,true,true,true);
  }

	int Network::InitIO(SOCK::SocketClient** sc)
	{
		SOCK::SocketClient* devnull=NULL;
		dbg.log("init","Init IO begin\n");	
		if(sc!=NULL)
		{
			SOCK::SocketClient* _in = new SOCK::SocketClient();
			_in->set_fd(0);
			SOCK::SocketClient* _out = new SOCK::SocketClient();
			_out->set_fd(1);
			SOCK::SocketClient* _err = new SOCK::SocketClient();
			_err->set_fd(2);
			
			//redir 012 on sc (this node is not root)
			*sc=_in->dup();
			//create new 012 redirected on /dev/null

			int devnullfd=open("/dev/null",O_RDWR,O_APPEND);
			if(devnullfd<=0)
			{	
				dbg.log("WARNING","InitIO can't open /dev/null\n");
				return 0;
			}
			else
			{
				devnull = new SOCK::SocketClient();
				devnull->set_fd(devnullfd);
			
			
				_in->dup2(*devnull);
				_out->dup2(*devnull);
				_err->dup2(*devnull);

				devnull->close();
			}
			delete _in;
			delete _out;
			delete _err;
			if(devnull!=NULL)
				delete devnull;
		} 
		return 0;
	}

  int Network::initialize(int* argc, char** argv[],bool AMWS,bool IO,bool _cloexec, bool new_arg_specif) 
    throw (acom::NetworkError)
  {
    topology_analyser analyser;


    //perf meter 
    gettimeofday(&time_init,0);

		dbg.log("init","TAKTUK NET INITIALIZE .........................\n");

    if(topo.env_recovery()<0)
		{
			dbg.log("init","NOT A RESUME INITIALIZE ALL................\n");
			resuming=false;
			if(analyser.get(argc,argv,&topo,new_arg_specif)<0)
			{
				dbg.log("PANIC:","init Topo_analyser [FAILED]\n");
				return -1;
			}

			dbg.log("init","Topo_analyser [OK]\n");
            
			if(topo.info.rank!=0)
				std::cout<<CHAR_CTRL<<std::flush;
				
		}
		else
		{
			dbg.log("init","RESUME DETECTED....................\n");
			resuming=true;
			if(topo.info.env_recovery()<0)
				return -1;
		}
    _node_rank=topo.info.rank;
    dbg.setrank(_node_rank);
    _node_count=topo.info.count_nodes;
    
		if(!_cloexec)
      topo.cloexec=0;

		InitNetwork();
    dbg.log("init","network table of channels initialized\n");
		InitIO(_node_rank!=0 ? &topo.info.all_links[0]._sock : NULL);
		if(IO)
		{
			_ioredir=new IOredir();
    	_ioredir->initialize(this);  
		}
    dbg.log("init","IOredir initialize completed\n");
 
    topo.print();
		
		if(IO)
		{
			if(_ioredir->commit(topo.cloexec,topo.verbose)<0)
			{
				dbg.log("WARNING","init redirIO [FAILED]\n");
				return -2;
			}
			else
				dbg.log("init","redirIO [OK]\n");
  	}
		for(int i=0;i<topo.info.all_links.size();i++)
		{
			dbg.log("init","info.all_links[%d]._flag=%d\n",i,topo.info.all_links[i]._flag);
		}


    taktuk_router::initialize();
		taktuk_router::LoadInfo(topo.info);
    taktuk_router::commit();
        
		_AMWS=AMWS;
		_IOREDIR=IO;
		dbg.log("init","init with costum AMWS IO completed \n");
		return 1;
  }





  void Network::exec_taktuk(char *argv[])
  {
    //char inter[512];
    //int size=0;

    //first stop the router and IO_mechanism
    taktuk_router::kill();
		if(_ioredir!=NULL)
		{
    	_ioredir->IO_kill();
    	_ioredir->env_export();
		}
    topo.env_export();
    topo.info.env_export();

    struct passwd* pwd;


    if((pwd=getpwuid(getuid()))== (struct passwd* )NULL)
		{
	 		perror("get user env: ");
 		}
    dbg.log("taktukexec","exec user passwd struct OK\n");

    if (chdir(pwd->pw_dir) < 0) 
		{
			(void) chdir("/");
		}

    dbg.log("taktukexec","exec possible chdir OK\n");

    dbg.log("taktukexec","we execute [%s]\n",argv[0]);

    execvp(argv[0],argv);

  }


  int Network::terminate() throw (acom::NetworkError)
  {
		dbg.log("init","TAKTUK NET TERMINATE..............................\n");
		if(_ioredir!=NULL)
			_ioredir->terminate();

		_barrier->sync();        
		
		taktuk_router::terminate();
		if(_ioredir!=NULL)
			_ioredir->terminatecommit();
    dbg.log("terminate","thread_router shutdown [OK]\n");

    //perf meter
    gettimeofday(&time_term,0);
    if(topo.perf && topo.info.rank==0)
      std::cout<<"Taktuk Time: "<<(double)( time_term.tv_sec - time_init.tv_sec)+0.000001*(double)(time_term.tv_usec - time_init.tv_usec)<<std::endl;;
  
    dbg.log("terminate","TAKTUK TERMINATE ;-) [OK]\n"); 
		return 1;
  }

	//template<class T>
//	void Network::add_connector(const std::string& id)
//	{
		//Spawner->add<T>(id);
//	}
	
	void Network::InitNetwork()
	{
		//if we are not the root
		if(topo.info.rank!=0)
		{
			if(!resuming)
			{
				dbg.log("init"," InitNetwork NOT RESUMING UPDATE info\n");
				topo.info.all_links.resize(1);
				topo.info.all_links[0]._flag=InfoTCPnet::link::critical;
				topo.info.all_links[0]._mut=new PTH::Mutex();
				topo.info.all_links[0]._mut->init();
				topo.info.route_table.resize(1);
				topo.info.route_table[0]=0;
			}
			_default_routes[0]=new TaktukChannel(this,0,&topo.info);

			LocalChannel.push_back(new TaktukLocalChannel(this,0,&topo.info));
	    NodeDesc* desc;
			desc=new NodeDesc();
			desc->_node_rank=0;
			desc->_node_label=topo.my_root;
			desc->_opened_channels.push_back(get_default_route(0));
			_known_nodes[0]=desc;
		}

		_default_routes[-1u]=new TaktukChannel(this,-1u,&topo.info);
    _node_label=topo.HostName;
	}

  void Network::RefreschNetwork()
  {
		//int begin=0;
    _node_rank=topo.info.rank;
    dbg.setrank(_node_rank);
    _node_count=topo.info.count_nodes;
		topo.arity=topo.hostsmap.size();
    dbg.log("init","RefreshNetwork rank %d, node_count %d\n",_node_rank,_node_count);
		_barrier->init(_node_rank,_node_count,this);

		if(topo.info.rank!=0)
			_known_nodes[0]->_node_rank=info.all_links[0]._peer;
    
		for(int i=0; i<topo.info.route_table.size();i++)
		{
			dbg.log("init","RefreshNetwork rank %d RT: route to %d --> %d\n",_node_rank,i,topo.info.route_table[i]);
		}
		
		for(int i=0; i<topo.info.count_nodes;i++)
		{
			if(i!=topo.info.rank)
				_default_routes[i]=new TaktukChannel(this,i,&topo.info);
		}

    dbg.log("init","build net default routes initialized\n");
    dbg.log("init","RefreschNetwork hostsmap.size = %d\n",topo.hostsmap.size());

    NodeDesc* desc;
    for(int i=0; i<topo.hostsmap.size();i++)
		{
			desc=new NodeDesc();
			desc->_node_rank=topo.hostsmap[i]->rank;
			dbg.log("init","build net hnown host %d for rank %d \n",i,topo.hostsmap[i]->rank);
			desc->_node_label=topo.hostsmap[i]->name;
			dbg.log("init","build net hnown host %d for name %s \n",i,topo.hostsmap[i]->name.c_str());
			desc->_opened_channels.push_back(get_default_route(topo.hostsmap[i]->rank));
			_known_nodes[(topo.info.rank==0)?i:i+1]=desc;
			dbg.log("init","build net hnown host %d initialized\n",i);
		}
    dbg.log("init","build net known hosts initialized\n");
		for(acom::Network::iter_known_nodes neighbour = begin_known_nodes();	neighbour!=end_known_nodes(); neighbour++)
		{
			dbg.log("init","neighbour no %d is in the global net node %d\n",neighbour->first,neighbour->second->_node_rank);
		}
  }
  
  void Network::newarg(int* argc, char** argv[])
  {
    topo.set_args(*argc, *argv);
  }

  void Network::newarg(std::string new_arg)
  {
    topo.set_args(new_arg);
  }

  int Network::spawn()
  {
    int err=0;
    //int state=0;
  
    if(!resuming)
		{
			Spawner->initialize(&topo.hostsmap,topo.arity,&topo.info,topo.steal_num,this);
			err=Spawner->connect(topo.reject,topo.local_user,(topo.steal_num!=0));
			Spawner->terminate();
		}
		else
		{
			err=1;
			RefreschNetwork();
		}
		// after this call all channels are OK :-)
    return err;
  }

  int Network::commit() throw (acom::NetworkError)
  {
    static bool commit_done = false;
    if (commit_done)
      return 0;
    commit_done = true;

		dbg.log("init","TAKTUK NET COMMIT..............................\n");
		int res=0;
		NET_on();
    res=spawn();
//    if(_AMWS)
//      NET_on();
    if(_ioredir!=NULL)
      _ioredir->IO_on();

#ifdef TAKTUK_USE_TRACER
    TAK_COMMIT();
#endif 

    return res;
  }

  std::string Network::getsub(::topology::host* _host)
  {
    std::string list;
    char* str;
    char* begin;
    char* end;
  //  int size;
    char* name;

    str=strcpy((char*)malloc(_host->cmd.size()+1),(char*)_host->cmd.c_str());
    str[_host->cmd.size()]=0;

    char* strinter =str;


    while(strinter!=NULL)
		{
			begin = strstr(strinter,"-m");
			if(begin==0)
				break;
			begin+=2;
			while(*begin==' ' && *begin!=0)
				begin++;
			if(begin==0)
				break;
			end=strstr(begin," ");
			name = strncpy((char*)malloc(end-begin+1),begin, (end-begin+1));
			name[end-begin]=0;
			list+=name;
			list+=" ";
			strinter=end;
		}
    return (list);
  }

  void Network::print_rejected()
  {
    for(int i=0;i<topo.reject.size();i++)
      std::cerr<<topo.reject[i]->name<<" "<<getsub(topo.reject[i])<<std::endl;
  }

  void Network::barrier()throw (acom::ComFailure)
  {
    dbg.log("sync","INIT SYNCHRO .....\n");
    _barrier->sync();
    dbg.log("sync",".....SYNCHRO OK\n");
  }


  void Network::usage()
  {
  
    std::cerr<<" [-a Arity] || [-d [ Steal Number ]] [-i] [-v] [-P] [-c Connector] [-l user]";
    std::cerr<<" [-f filename]  [-m hostname || -m host=XXX,connect=XXX,login=XXX... ]...] -- arguments ..."<<std::endl;
    std::cerr<<"\t-a: arity of the spanning tree, by default is the number of nodes"<<std::endl;
    std::cerr<<"\t-d: dynamic spanning tree, by default is the number of nodes, but you can give there a number of nodes to steal"<<std::endl;
    std::cerr<<"\t-l: username on the remote host"<<std::endl;
    std::cerr<<"\t-c: specify the connector you want use to connect remote host (rexec, rsh, other...)"<<std::endl;
    std::cerr<<"\t\tlist of connectors with suid bit on rshp: rshs (sequential), rsha (asynchronous), rsht (threaded)"<<std::endl; 
    std::cerr<<"\t\tlist of connectors user level: rshf, ssh, env"<<std::endl; 
    std::cerr<<"\t\tfor the env connector you must declare the TAKTUK_RSH variable "<<std::endl;
    std::cerr<<"\t-f: filename, file containing a list of host, this option can be used more time and mixed with -m option"<<std::endl; 
    std::cerr<<"\t-i: Dev Option let fd opened by application when the exec command is use.\n\t\t\t WARNING don't use it if you don't know what you do ;-)"<<std::endl; 
    std::cerr<<"\t-m: specify one hostname"<<std::endl;
    std::cerr<<"\t-m ... : this option can be increase with suboptions:"<<std::endl;
    std::cerr<<"\t\t\thost=hostname"<<std::endl;
    std::cerr<<"\t\t\tconnect=connector_name, which connector we must use for this node"<<std::endl;
    std::cerr<<"\t\t\tlogin=log_name, login for this node"<<std::endl;
    std::cerr<<"\t-[ and -]: are delimiter for childrens of one node specified by -m before \n\tthe first -[ occurence"<<std::endl;
    std::cerr<<"\t-v: verbose mode"<<std::endl;
    std::cerr<<"\t-P: profiling mode: execution time"<<std::endl;
    std::cerr<<"#### Release 3.0 ############"<<std::endl;
  }


	// --------------------------------------------------------------------------
  //
  // Route
  //




  Network::TaktukChannel* Network::get_default_route( LocalNodeNo peer ) 
    throw(acom::NetworkError)
  {
    if (_default_routes.find(peer) == _default_routes.end())
      throw acom::ComFailure(acom::ComFailure::NOROUTE);
//      throw acom::NetworkError("PB", 1);
    return (Network::TaktukChannel*)_default_routes[peer];
  }



}//namespace Taktuk

extern "C" {

	static Taktuk::Network* taktuk_net =0;

	// - dynamic library initialization:
	static void factory_initialize() 
	{
		taktuk_net = new Taktuk::Network();
		acom::Acom::bind_network("TAKTUK", taktuk_net );
	}
	extern void TAKTUK_factory_initialize() 
	{
		//dbg.log("","factory init\n");
		factory_initialize();
		//dbg.log("","factory init OK\n");
	}
	
	// - dynamic library terminaison:
	static void factory_destroy() 
	{
		delete taktuk_net;
	}
	extern void TAKTUK_factory_destroy() 
	{
		return factory_destroy();
	}
}
