/*
 * The contents of this file are subject to the Dyade Public License,
 * as defined by the file DYADE_PUBLIC_LICENSE.TXT
 *
 * You may not use this file except in compliance with the License. You may
 * obtain a copy of the License on the Dyade web site (www.dyade.fr).
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
 * the specific terms governing rights and limitations under the License.
 *
 * 
 * 
 *
 * The Initial Developer of the Original Code is Dyade. The Original Code and
 * portions created by Dyade are Copyright Bull and Copyright INRIA.
 * All Rights Reserved.
 */
/* Copyright (c) 1997 by Groupe Bull. Read the COPYRIGHT file. */
/* $Id: */
/* Author: Cyrille.Martin@imag.fr */

/* Cyrille Martin */
/* laboratoire ID-IMAG, BULL*/
/* Projet LIPS action DYADE*/
/* 07/06/2001 */

#ifndef __TOPOLOGY_H
#define __TOPOLOGY_H
#include <CoreTCP.h>
#include <vector>
#include <map>
#include <string>

#include <pth.h>



#ifndef MAX_HOST_NAME
#define MAX_HOST_NAME 256
#endif

class topology_analyser;
class spawner;
class connector;

class topology
{
 public:
  
  InfoTCPnet info;


  //!several option flag
  //! bool for the write console function more or less information to the console
  int verbose;
  int perf;
  int cloexec;
  int steal_num;
  //!...

  // host args for one host

  struct host
  {
    std::string name;
    std::string connector;
    std::string taktukargs;
    std::string userargs;
    std::string login;
    std::string passwd;
    std::string cmd;
    std::string badtxt;
    int fd;
    int rank;
    int size;
    int type; //compute node 0 or passerelle 1
		int laststeal;
		int timeout;
    
    host(std::string n, std::string c,std::string s,std::string u, std::string l,std::string p, int _type, int _timeout){
      name=n;
      connector=c;
      taktukargs=s;
      userargs=u;
      login=l;
      passwd=p;
      type=_type;
      size=1;//this node
			laststeal=1;
			timeout=_timeout;
    }

		host(std::string n, std::string c,std::string s,std::string u, std::string l,std::string p, std::string _cmd,int _size, int _type, int _timeout){
      name=n;
      connector=c;
      taktukargs=s;
      userargs=u;
      login=l;
      passwd=p;
			cmd=_cmd;
      type=_type;
      size=_size;
			laststeal=1;
			timeout=_timeout;
    }
		
    host(){size=1;timeout=0;laststeal=1;}
  };

  //! topological and routing info
  char HostName[MAX_HOST_NAME];

  //!hosts[i] == list of all the sub node of this node (his childs)
  std::vector<host*> hostsmap;
  std::vector<host*> reject;  
  //name of my up node
	std::string my_root;
  
  int arity;
  //! size of this subtree this node include
  //int size_sub_tree;


  //! default connecting info for all nodes if no specific host options are specified
  std::string user; // remote user name
  std::string local_user; // local user name
  std::string passwd; // passwd for the remote user (optionnal)
  std::string connector; // how connect their remote hosts
//  std::string arguments; // user's programme arguments


  //! constructor initialize some topology variable & redirect the standarts I/O to fd_local_X, with i in [0 count_childs] 0 is the node up
  topology();
  //! destructor flush & close fd_net & 0 1 2nothing at the moment
  ~topology();

  /**
     set new user's programme arguments
  */
  void set_args(std::string& args);
  /**
     return the user's programme arguments as string
  */
  std::string get_args();
  /**
     set new user's programme arguments
  */
  void set_args(int argc, char* argv[]);
  /**
     print the topology configuration with log facilities
  */
  void print();
  /**
     print the topology configuration on stdout
  */
  void printstdout();

  /**
     export all necessary data on environement variables
   */
  int env_export();
  /**
     rebuild a valide topology structure with environement variables
  */
  int env_recovery();

  friend class topology_analyser;
  friend class spawner;
  friend class connector;
};

#endif /* __TOPOLOGY_H*/
