//
// $Id: testforward.cpp,v 1.1.1.1 2003/05/26 15:59:28 olobry Exp $
//
// test active message: a call to one service from rank 0 with nul msg
//
#include <acom>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include <taktuknetwork.h>
#include <pth>
#include <string.h>

char *indent[]= {"", "  ", "    ", "      ", "        " };

#define RANK "[" << rank << "] "
//#define OUT std::cout
#define OUT f
#define BUFF_SIZE 50

namespace AC = acom;

int rank = -1;
int next = -1;
int senders[BUFF_SIZE];
int sender;
int nb_nodes = -1;
std::ofstream f;
#define NB_INIT_MESSAGES 10

#define NEXT(n) (((n) == nb_nodes - 1) ? 0 : (n) + 1)
#define PREV(n) (((n) == 0) ? nb_nodes - 1 : (n) - 1)


static char buffer[256];
static PTH::Condition cond;
static PTH::Mutex mutex;
static PTH::Mutex loc_mutex; // to synchronize call between themselves



class AckService: public AC::Service {

public:
  

  // This function is called once all deferred data has been copied
  // at the right location
  void call ( 
	     AC::ComFailure::Code e,      // if any 
	     acom::Network::Receipt* receipt,
	     AC::Network::Header* h,
	     AC::IOVectEntry* iov,
	     AC::IOVectSize   iov_sz
	     ) throw (AC::ComFailure)
  {
    PTH::CriticalGuard guard(mutex); 
    PTH::CriticalGuard loc_guard(loc_mutex); 
    
    //    OUT << RANK << ">>>" << std::endl;
    if (e == AC::ComFailure::OK) {
      // Get sender id
      int i;
      for (i = 0; i < BUFF_SIZE; i++)
	if (senders[i] == -1) break;
      if (i == BUFF_SIZE)
	OUT << "Cannot register message: to many in waiting queue" 
		  << std::endl;
      else {
	senders[i] = *(int *) iov[0].data;
	OUT << RANK << "<- [?]: " << senders[i];
	cond.signal();
	OUT << std::endl;
      }
    }
    else 
      OUT << RANK << "-- Error" << std::endl;

    //OUT << RANK << "<<<" << std::endl;

  } // call

};

static AckService ack_service;




int doit (int argc, char* argv[])
{
  // initialize the Pthread library used through the CountDown class
  PTH::Pth::initialize();
  PTH::Pth::commit();
  char fname[100];
  for (int i = 0; i < BUFF_SIZE; i++) 
    senders[i] = -1;


  // declare our service to the service manager and bind it to
  // service number 2 of the user service familly 
  AC::Service::Manager.bind (2, &ack_service);

  
  // create and initialize a taktuk network, passing it the command 
  // line arguments
  Taktuk::Network* net = new Taktuk::Network();
  net->initialize(&argc,&argv);

  // network initlization barrier
  net->commit(); 
  // at this stage, the application has been launched on every nodes specified
  // in the command line, and has exited from net->commit()
  

  // initliaze the countdown object with the number of nodes that are 
  // supposed to send a message to the local node, that is the number of
  // nodes in the network minus the local one
  rank = net->rank();
  nb_nodes = net->node_count();

  sprintf(fname, "./test_fwd.%d", rank);
  f.open(fname);
  net->barrier();

  // synchronize with other nodes to make sure that everybody has initialized
  // the countdown object and the rank
  net->barrier();

  // Let's go
  OUT << RANK << "START"<<std::endl; 
  OUT << RANK << "OPEN->" << NEXT(rank) << std::endl;
  next = NEXT(rank);
  AC::Network::OutChannel* channel = net->get_default_route(next);

  // send initial messages
  //  AC::ActiveMessage<AckService> am (net, &ack_service, 0, 
  //				    nb_nodes *  NB_INIT_MESSAGES);
  AC::ActiveMessage<AckService> am (net, &ack_service);
  sender = rank;
  am->put(&sender, sizeof(sender));
  for (int i = 0; i < NB_INIT_MESSAGES; i++) {
    OUT << RANK << "-> [" << next << "]: " << rank << std::endl;
    am(channel);
  }


  int nb_forwarded = 0;
  int nb_to_forward = nb_nodes *  NB_INIT_MESSAGES;
  // forward requests
  {
    PTH::CriticalGuard guard(mutex); 
    while (true) 
      {
	OUT << RANK << nb_forwarded << " < " << nb_to_forward  << std::endl;
	// forward only messages not initiated by me
	int i;
	for (i = 0; i < BUFF_SIZE; i++) {
	    
	  if (senders[i] == -1)
	    continue;
	  if (senders[i] != rank) {
	    OUT << RANK << "-> [" << next << "]: " << 
	      senders[i] << std::endl;
	    sender = senders[i];
	    am(channel);
	  } else 
	    OUT << RANK << "FINISH"  << std::endl;
	  nb_forwarded++;
	  senders[i] = -1;
	}
	if (nb_forwarded == nb_to_forward)
	  break;
	OUT << RANK << "WAIT" << std::endl;
	cond.wait(mutex);
	OUT << RANK << "WAKE UP" << std::endl;
	
      }
  }
  OUT << RANK << "EXIT"  << std::endl;

  net->barrier();

  // terminate the network
  net->terminate( );

  // terminate tht Pthread library
  PTH::Pth::terminate();
}




int main (int argc, char* argv[])
{
  try {
    return doit(argc, argv);
  } catch (...){
    std::cerr << "Unknown exception" << std::endl;
  }
}
