#! /bin/bash

EVT_FILE=$1.evt

./tr_gen $1 $2 || exit 1


cat > /tmp/tracereader.cpp <<EOF
#include <iostream>
#include <string>
#include <fstream>
#include "tracereader.h"
#include <stdlib.h>  
#include <sstream>
#include <algorithm>
#include <deque>
#include <functional>
#include <iterator>
#include <list>
#include <map>
#include <memory>
#include <numeric>
#include <set>
#include <stack>
#include <utility>
#include <vector>

using namespace std;

const int NBMAXFILE=300; 


/*  OVERLOADING OF RECORD POINTER 
    for  sorting and marging trace  traces by dates*/
class RefRec
{
public:  
  InukTracer::Record *_rcd; //the pointer of event records 
  RefRec():_rcd(0){};
  RefRec(InukTracer::Record *rcd):_rcd(rcd){};
  ~RefRec(){};
  
  /// Logical  less
  bool operator < (RefRec  rp)
  {    return (_rcd->_d) < (rp._rcd->_d);  }
  /// Logical  equal
  bool operator == (RefRec rp)
  {    return _rcd==rp._rcd;  }
  /// Logical not operator. to check a valid  record?
  bool operator!() const { return _rcd == 0; }
  /// Logical not operator. to check a valid  object?
  operator bool() const { return _rcd!= 0; }
  ///  operator : make transparent access to record
  InukTracer::Record 	* operator->() const { return  _rcd;}
  /// assignment  by a ref to record
  RefRec&operator=(InukTracer::Record *rcd ){_rcd=rcd; return *this;}
};

/****************************************************************
*	PROGRAM PARAMETERS ANALYZER
*****************************************************************/
// program parameters

char * fn[NBMAXFILE];

void AfficheAide()
{
  std::cout<< "***************************************************\n";
  std::cout<< "*  Usage : prog_name  fn                          *\n";
  std::cout<< "*    where fn : trace file  to read               *\n";
  std::cout<< "***************************************************\n";
}

// program parameters analyzer
void scanArg( int argc, char ** argv)
{ 
  if(argc < 2)
    {
      AfficheAide();
      exit(0); 
    }
  
  for(int i=0; i<(argc-1); i++)
    {
      fn[i]=argv[i+1];
    }
}

/********************************************************************
*	Purge trace on event basis (example)
*********************************************************************/

 
//static bool isEvent(RefRec rcd){ return rcd->_ev==3;}

template <int LV,  int EV > bool is_event(const RefRec rcd ){ return rcd->_ev==EV && rcd->_lv==LV;};
template <int LV, int EV > bool  not_event(const RefRec rcd ){ return !(rcd->_ev==EV && rcd->_lv==LV);};

/******************************************************************
 * Event trace list                                               *
 ******************************************************************/

class  Trace: public list<RefRec> 
{
public:
  Trace():list<RefRec>(){};
  ~Trace(){};
  std::ostream&  print(std::ostream&);
}; 

std::ostream&  Trace::print(std::ostream& out)   
{
  Trace::iterator il;
  for (il=this->begin();il!=this->end();il++)
    {
      (*il)->print(out);out<<std::endl;
    }
  return out;
};

/*****************************************************************
* Map of pairs <SThread,Trace of the thread>
******************************************************************/

class TracesByThread :  public map<PTH::ThreadId, Trace *>
{
protected:
  bool flag;
  Trace * tr;PTH::ThreadId th;
public :
  TracesByThread() : map<PTH::ThreadId, Trace *>(),flag(false),tr(0){};
  ~TracesByThread(){};
  std::ostream& print(std::ostream&);
  bool recordEvent(RefRec);
};

std::ostream& TracesByThread::print(std::ostream& out)   
{
  TracesByThread::iterator im;
  Trace* l;
  for (im=this->begin();im!=this->end(); im++){
    out<<"MONO TRACE OF THREAD("<< (*im).first<<")::"<<endl;
    out<<"----------------------------------------------------------\n";
    l=(*im).second;l->print(out);
    out<<"----------------------------------------------------------\n";
  }
  return out;
}

bool TracesByThread::recordEvent(RefRec rcd )
{
  
  if(!flag){ // first thread 
    flag=true; th=rcd->_tid;
    tr= new Trace();
    tr->push_back(rcd);
    (*this)[th]=tr;
    return false;
  }
  if(th==rcd->_tid){
    tr->push_back(rcd); 
    return true ;
  }
  th=rcd->_tid;
  if (this->find(th)!=this->end()) {
    tr=(*this)[th];
    tr->push_back(rcd);
    return true;
  }
  tr= new Trace();
  tr->push_back(rcd);
  (*this)[th]=tr;
  return false;
}

/****************************************************************
*        Handling Traces  :: examples
*****************************************************************/

TracesByThread tracesByThread; // map of private trace of thread
Trace tracesAll; // all events : must be sorted by date

int main(int ac, char** av)
{
  std::cout<< "***************************************************\n";
  std::cout<< "*                                                 *\n";
  std::cout<< "*                READ TRACE FILE                  *\n";
  std::cout<< "*                                                 *\n";
  std::cout<< "***************************************************\n";
  
  scanArg( ac,av );
  
  int evcount=0;
  InukTracer::Reader reader;
  TR::TraceFile fout;
  /* // trace output file 
     fout.open("./TOTO",std::ios::out); if(fout.fail())std::cout << " CATCH FAIL TO OPEN OUTPUT TRACE FILE \n";
  */
 
  // dump map of known event descriptors
  InukTracer::recordMap.dump(cout);

  for(int i=0; i<(ac-1); i++)
    {
      try
	{
	  std::cout << "jouvre fichier " << i << " nom:" << fn[i] << std::endl;
	  reader.open(fn[i]); 
	  RefRec rcd;
	  try
	    {
	      while(1)
		{
		  rcd=reader.readEvent();  
		  evcount++;
		  TR::PhysicalClock::Date d= rcd->getDate();
		  // echoes the record
		  //(rcd->print(cout))<<"\n" ; // print record 
		  rcd->_node=i;
		  tracesAll.push_back(rcd);
		  tracesByThread.recordEvent(rcd);
		  //rcd->write(fout); // copy an event from trace file to  another trace file
		}
	      //fout.close();
	    }
      
	  catch( EndOfTrace &e)
	    { cout << " CATCH END OF TRACE WITH  THE "<< evcount <<" EVENT \n";}  
	  catch( UnknownEvent &e)
	    { cout << " CATCH AN UNKOWN EVENT EXCEPTION AT THE "<< evcount <<" EVENT \n";}
	  catch( UnknownEvent2 &e)
	    { cout << " CATCH AN UNKOWN EVENT EXCEPTION AT THE "<< evcount <<\
		" EVENT -- ev=" << e.e << " lv=" << e.l  << "\n";}
	  catch( BadEvent &e)
	    { cout << " CATCH A BAD EVENT EXCEPTION AT THE "<< evcount <<" EVENT \n";}
	  reader.close();

	  
	  // std::cout << "close file \n";
	}
      catch(FailOnOpen &e)
	{    cout << " CATCH FAIL TO OPEN TRACE FILE \n";} 
    }
  
  /*
    std::cout<< "**************************************************\n";
    std::cout << " PRINT TRACE THREAD BY THREAD--------------------\n";
    std::cout<< "**************************************************\n";
    tracesByThread.print(cout);
  */
      
  // sort traces by date
  tracesAll.sort();
      
  //  eliminate flushBuffer event
  tracesAll.remove_if(is_event<tracer_level_no,flushBuffer_no>); 
  /*
  //  eliminate all events but flushBuffer event
  tracesAll.Trace::remove_if(not_event<tracer_level_no,flushBuffer_no>); 
  */
  // print sorted trace 
  tracesAll.print(cout);

  
  std::cout << " ----------------------------> EXIT READTRACE \n";
}
EOF

g++ /tmp/tracereader.cpp ${1}_stub_reader.cpp -ltracereader -o ${1}_reader





