/*
 * analyser : this program convert XXX.evt files 
 * into XXX_stub_conv.h and XXX_stub_conv.cpp files
 * using the rules defined in XXX.regles
 ****************************************************/
 

#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <string>
#include <fstream>
#include "analyser_yacc.h"


std::ofstream* ph;
std::ofstream* pcpp;

std::string name;
std::string pevt_name;
std::string ph_name;
std::string pcpp_name;


extern "C"
{
  int yyparse(void);
  char *yycopy;
}

  extern FILE* yyin;

int main(int argc, char** argv)
{
  FILE* fin;
  if (argc > 1)
    {
      name += argv[1];
      pevt_name = name;
      pevt_name += ".evt";
      
      fin = fopen(pevt_name.c_str(), "rt");
      if (!fin) 
	{
	  std::cout <<  "Unable to open file :" << argv[1] << ".evt\n";
	  exit(1);
	}
      yyin = fin;
    }
  else  // less than one argument on command line
    {
      std::cerr << "! No prefix specified !\n  Aborting...\n";
      exit(1);
    }

  ph_name = name;
  pcpp_name = name;

  ph_name += "_stub_conv.h";
  pcpp_name += "_stub_conv.cpp";
  
  ph = new std::ofstream (ph_name.c_str(),  std::ios::out | std::ios::binary);
  pcpp = new std::ofstream (pcpp_name.c_str(),  std::ios::out | std::ios::binary);
  
  
  *ph << "/*===================================================\n"
      << "File automatically generated by "<< argv[0] << std::endl 
      << "from tracer definition file " << name <<".evt \n"
      << "===================================================*/\n\n";
  
  *pcpp << "/*===================================================\n"
	<< "File automatically generated by "<< argv[0] << std::endl 
	<< "from tracer definition file " << name <<".evt \n"
	<< "===================================================*/\n\n";
  

  // protection against re-def  
  *ph << "#ifndef _" << name << "_stub_conv_h_ \n" 
      << "#define _" << name << "_stub_conv_h_\n"
      << "#include <tr_reader.h>\n";
  
  
  *pcpp << "#include \"iomanip\"\n";
  *pcpp << "#include \""<< name << "_stub_conv.h\"\n";


  /**************/
  yyparse();
  /**************/


  // close open files
  fclose (fin);
  ph->close();
  pcpp->close();
  
  return 0;
}
