
dnl ================================================================
dnl PROG_M4
dnl ================================================================
AC_DEFUN([PROG_M4], [
    AC_CACHE_CHECK(for "gnu m4",
    M4, [
    which_m4=`whereis m4 | cut -d" " -f2- `
    for m4 in $which_m4; do
        m4version=`echo no version | ${m4} --version`
        tmp=`echo ${m4version} | grep GNU`
        if test -n "$tmp"; then
            M4=${m4}
            AC_SUBST(M4)
            break
        fi
    done
])
])

dnl ================================================================
dnl PROG_DOT
dnl ================================================================
AC_DEFUN([PROG_DOT], [
    AC_CACHE_CHECK(for dot,
    dot, [
    dot=`which dot 2> /dev/null`
    have_dot=yes
    test x$dot == x && dot=none && have_dot=no
    ])
])


dnl ================================================================
dnl CHECK_ROUND
dnl ================================================================

AC_DEFUN([FUNC_ROUND], [
	AC_CACHE_CHECK(for round,
	have_round,
	AC_TRY_RUN([
#include <math.h>
int main()
{
  int i = round(10.7);
  if (i != 11) return 1;
  return 0;
}],
	AC_DEFINE(HAVE_ROUND, 1, Define if round is available)
    have_round=yes,
    AC_TRY_RUN([
#include <tgmath.h>
int main()
{
  int i = round(10.7);
  if (i != 11) return 1;
  return 0;
}],
	AC_DEFINE(HAVE_ROUND, 1, Define if round is defined)
    have_round=yes,
    have_round=no
	)))
])

dnl ================================================================
dnl CHECK_PTHREAD_SCOPE_PROCESS
dnl ================================================================

AC_DEFUN([CHECK_PTHREAD_SCOPE_PROCESS], [
	AC_CACHE_CHECK(for pthread_scope_process,
	pth_pthread_scope_process,
	AC_TRY_RUN([
#include <pthread.h>
#include <errno.h>
int main()
{
  int err;
  pthread_attr_t attr;
  pthread_attr_setscope( &attr, PTHREAD_SCOPE_PROCESS);
  if (err != 0) { return 1; } else return 0;
}],
	pth_pthread_scope_process=yes
	AC_DEFINE(PTH_CANUSE_PTHREAD_SCOPE_PROCESS),
	pth_pthread_scope_process=no
	))
])

dnl ================================================================
dnl CHECK_PTHREAD_ATTR_SETSTACKADDR
dnl ================================================================

AC_DEFUN([CHECK_PTHREAD_ATTR_SETSTACKADDR], [
	AC_CACHE_CHECK(for pthread_attr_setstackaddr,
	pthread_attr_setstackaddr,
	AC_TRY_RUN([
#include <pthread.h>
#include <errno.h>
int main()
{
  int err;
  pthread_attr_t attr;
  void* addr;
  pthread_attr_setstackaddr( &attr, &addr);
  if (err != 0) { return 1; } else return 0;
}],
	pthread_attr_setstackaddr=yes
  	AC_DEFINE(PTH_USE_SET_STACKADDR),
	pthread_attr_setstackaddr=no
	))
])


dnl ================================================================
dnl CHECK_PTHREAD_ATTR_GETSTACKADDR
dnl ================================================================

AC_DEFUN([CHECK_PTHREAD_ATTR_GETSTACKADDR], [
	AC_CACHE_CHECK(for pthread_attr_getstackaddr,
	pthread_attr_getstackaddr,
	AC_TRY_RUN([
#include <pthread.h>
#include <errno.h>
int main()
{
  int err;
  pthread_attr_t attr;
  void* addr;
  pthread_attr_getstackaddr( &attr, &addr);
  if (err != 0) { return 1; } else return 0;
}],
	pthread_attr_getstackaddr=yes
	AC_DEFINE(PTH_USE_GET_STACKADDR),
  	pthread_attr_getstackaddr=no
	))
])

dnl ================================================================
dnl CHECK_PTHREAD_ATTR_SETSTACKSIZE
dnl ================================================================
AC_DEFUN([CHECK_PTHREAD_ATTR_SETSTACKSIZE], [
	AC_CACHE_CHECK(for pthread_attr_setstacksize,
	pthread_attr_setstacksize,
	AC_TRY_RUN([
#include <pthread.h>
#include <errno.h>
int main()
{
  int err;
  pthread_attr_t attr;
  pthread_attr_setstacksize( &attr, 1024);
  if (err != 0) { return 1; } else return 0;
}],
	pthread_attr_setstacksize=yes
  	AC_DEFINE(PTH_USE_SET_STACKSIZE),
	pthread_attr_setstacksize=no
	))
])



dnl ================================================================
dnl CHECK_PTHREAD_ATTR_GETSTACKSIZE
dnl ================================================================
AC_DEFUN([CHECK_PTHREAD_ATTR_GETSTACKSIZE], [
	AC_CACHE_CHECK(for pthread_attr_getstacksize,
	pthread_attr_getstacksize,
	AC_TRY_RUN([
#include <pthread.h>
#include <errno.h>
int main()
{
  int err;
  pthread_attr_t attr;
  size_t size;
  pthread_attr_getstacksize( &attr, &size);
  if (err != 0) { return 1; } else return 0;
}],
	pthread_attr_getstacksize=yes
  	AC_DEFINE(PTH_USE_GET_STACKSIZE),
  	pthread_attr_getstacksize=no
	))
])


dnl ================================================================
dnl CHECK_CORBA
dnl ================================================================


AC_DEFUN([CHECK_CORBA],
	[
	AC_ARG_WITH(corbadir,
  	[   --with-corbadir=<corba_directory>. Recognized CORBA implantations are: omniORB3, omniORB4],
  	[ if test -d "$withval"; 
    then 
		corbadir="$withval"
    else 
      	AC_MSG_ERROR([--with-corbadir=DIR, invalid directory]) 
    fi
  	])

#	AC_SUBST(CORBA_IDLCC)
	AC_SUBST(CORBA_IDLCCFLAGS)

	AC_CHECKING(for a corba implementation)

	if test x$corbadir == x 
	then
		corbadir="/usr"
	fi

	CORBA_ORBNAME=unknown 


	if test x$CORBA_ORBNAME == xunknown 
	then
		CORBA_CHECK_OBE
	fi

	if test x$CORBA_ORBNAME == xunknown 
	then
		CORBA_CHECK_OB
	fi

	if test x$CORBA_ORBNAME == xunknown 
	then
		CORBA_CHECK_OMNIORB3
	fi

	if test x$CORBA_ORBNAME == xunknown 
	then
		CORBA_CHECK_OMNIORB4
	fi

	
	AC_MSG_RESULT(found corba implementation: $CORBA_ORBNAME)
	])


AC_DEFUN([CORBA_CHECK_OBE], [

	###################################
	# OBE
	###################################
	if test -d "$corbadir/include/OBE"
	then
#		CORBA_ORBNAME="OBE"
		AC_MSG_WARN([recognised ORbacus/OBE, but no configuration for it]);
#	    AC_DEFINE(CORBA_USE_OBE, 1, Defined if ORBACUS/OBE is used)
	fi 
])

AC_DEFUN([CORBA_CHECK_OB], [


	###################################
	# OB
	###################################
	if test -d "$corbadir/include/OB"
	then
#		CORBA_ORBNAME="OB"
		AC_MSG_WARN([recognised ORbacus/OB, but no configuration for it]);
#	    AC_DEFINE(CORBA_USE_OB, 1, Defined if ORBACUS/OB is used)
	fi 
])

AC_DEFUN([CORBA_CHECK_OMNIORB3], [

	###################################
	# OMNIORB3
	###################################

	if test -d "$corbadir/include/omniORB3"
  	then
#		AC_MSG_RESULT(found omniORB3)
		CORBA_ORBNAME="omniorb3"
		if test $corbadir != "/usr"
		then
			compile_COR_net_CPPFLAGS="-I$corbadir/include"
		fi
		CORBA_CPPFLAGS="-I$corbadir/include/omniORB3"
		if test -d "$corbadir/lib/i586_linux_2.0_glibc2.1"
		then
			CORBA_LIBS="-L$corbadir/lib/i586_linux_2.0_glibc2.1 $CORBA_LIBS"
		  	AC_PATH_PROG(CORBA_IDLCC, omniidl,, $corbadir/bin/i586_linux_2.0_glibc2.1)
		fi
		if test -d "$corbadir/lib/powerpc_darwin_6.5"
		then
		  	AC_CHECK_PROG(CORBA_IDLCC, omniidl,, $corbadir/bin/powerpc_darwin_6.5/)
		  	CORBA_LIBS="-L$corbadir/lib/powerpc_darwin_6.5 $CORBA_LIBS"
		fi

		AC_PATH_PROG(CORBA_IDLCC, omniidl, , $corbadir/bin) 
		if test x"$CORBA_IDLCC" == x
		then 
			 AC_MSG_ERROR(Cannot find omniidl in $corbadir/bin)
		fi

		CORBA_LIBS="$CORBA_LIBS -lomniDynamic3 -lomniORB3 -ltcpwrapGK -lomnithread"
		CORBA_IDLCCFLAGS="-bcxx -Wba -Wbh=.h -Wbd=_dyn.cpp -Wbs=.cpp"
		AC_DEFINE(CORBA_USE_OMNIORB, 1, Defined if omniorb is used)
		AC_DEFINE(CORBA_USE_OMNIORB3, 1, Defined if omniorb3 is used)
	fi 
])

AC_DEFUN([CORBA_CHECK_OMNIORB4], [


	###################################
	# OMNIORB4
	###################################

	if test -d "$corbadir/include/omniORB4"
	then
		CORBA_ORBNAME="OMNIORB4"
		if test $corbadir != "/usr"
		then
			CORBA_CPPFLAGS="-I$corbadir/include"
		fi
		CORBA_CPPFLAGS="$CORBA_CPPFLAGS -I$corbadir/include/omniORB4"
		AC_PATH_PROG(CORBA_IDLCC, omniidl, , $corbadir/bin) 
		if test x$CORBA_IDLCC == x
		then
			 AC_MSG_ERROR(Cannot find omniidl in $corbadir/bin)
		fi

		CORBA_LIBS="$CORBA_LIBS -L$corbadir/lib -lomniDynamic4 -lomniORB4 -lomnithread"
		CORBA_IDLCCFLAGS="-bcxx -Wba -Wbh=.h -Wbd=_dyn.cpp -Wbs=.cpp"
		AC_DEFINE(CORBA_USE_OMNIORB, 1, Defined if omniorb is used)
		AC_DEFINE(CORBA_USE_OMNIORB4, 1, Defined if omniorb4 is used)
	fi 

])


